/**
  * Created by kdelmas on 07/09/16.
  */
package tool

import java.io.File

import analyzers._
import util.analyzers.analyzerTypes._
import ast.dsl.{ConfigIdent, ConfigurationDeclaration, Transformer}
import interpreter.TraceInterpreter
import monocle.macros.Lenses
import perf.{CutEnumerationPerformancePerf, DesignExplorationPerf, ReliabilityComputationBDDvsSTTPerf}
import scopt._
import ui.UI
import util.io.SystemInfo.Linux
import util.analyzers._
import util.io.{LoadDSLFiles, SystemInfo}

import scala.concurrent.ExecutionContext.Implicits.global
import scala.concurrent.duration._
import scala.language.postfixOps
import scala.reflect.io.Path

/**
  * Main object for command line version of the tool
  */
object MainCLI {

  /**
    * command line parser
    */
  private val parser= new OptionParser[RunConfig](s"java -jar ${new File(MainCLI.getClass.getProtectionDomain.getCodeSource.getLocation.getPath).getName}.jar") {
    head(sbtInfo.Info.name, sbtInfo.Info.version)

    arg[File]("<model files>") unbounded() optional() action { (x, config) =>
      config.copy(systemFiles = config.systemFiles :+ x)
    } validate { x => if (!x.exists) failure(s"$x does not exists") else success } text ("KCR input files to read. " +
      "Tool exits if none specified or files do not exist.")

    help("help") text "prints this usage text"

    opt[String]('c', "configuration") action { (name, config) => config.copy(confName = Some(name)) } text "the configuration identifier to analyse"

    opt[Unit]("compute-r") action { (_, config) => (RunConfig.anaConf composeLens AnalysisConfig.computeR).modify(_ => true)(config)
    } text "asks to compute the system reliability"

    opt[File]("print-phi-dot") action { (output, config) => (RunConfig.exportConf composeLens ExportConfig.printDot).modify(_ => Some(output))(config)
    } text "exports characteristic function as dot graph"

    opt[Unit]("check-r") action { (_, config) => (RunConfig.anaConf composeLens AnalysisConfig.checkR).modify(_ => true)(config)
    } text "checks system reliability according to requirement"

    opt[Unit]("check-monotony") action { (_, config) => (RunConfig.anaConf composeLens AnalysisConfig.checkMonotony).modify(_ => true)(config)
    } text "checks system monotony"

    opt[Unit]("gui") action { (_, config) => config.copy(gui = true)
    } text "run in gui mode"

    opt[Unit]("compute-min-card") action { (_, config) => (RunConfig.anaConf composeLens AnalysisConfig.computeMinCard).modify(_ => true)(config)
    } text "compute minimal cardinality of MCS"

    opt[Unit]("check-min-card") action { (_, config) => (RunConfig.anaConf composeLens AnalysisConfig.checkMinCard).modify(_ => true)(config)
    } text "check minimal cardnality of MCS according to requirement"

    opt[Int]("compute-MCS") action { (k, config) => (RunConfig.anaConf composeLens AnalysisConfig.computeMCSUntil).modify(_ => Some(k))(config)
    } validate { k => if (k >= 0) success else failure(s"time must be greater than 0") } text "enumerates MCS until k where 0 means all cutsize"

    version("version") action { (_, config) => println(s"${sbtInfo.Info.name} ${sbtInfo.Info.version}"); config } text "prints the tool version"

    opt[File]("export-r-py") action { (output, config) => (RunConfig.exportConf composeLens ExportConfig.exportRSage).modify(_ => Some(output))(config)
    } text "export symbolic reliability as SYMPI script"

    opt[File]("export-smv") action { (output, config) => (RunConfig.exportConf composeLens ExportConfig.exportSMV).modify(_ => Some(output))(config)
    } text "export problem as SMV script"

    opt[File]("export-xfta") action { (output, config) => (RunConfig.exportConf composeLens ExportConfig.exportXFTA).modify(_ => Some(output))(config)
    } text "export phi as XFTA tree"

    opt[File]("export-hiphops") action { (output, config) => (RunConfig.exportConf composeLens ExportConfig.exportHipHops).modify(_ => Some(output))(config)
    } text "export problem as HipHops script"

    opt[File]("export-cg") action { (output, config) => (RunConfig.exportConf composeLens ExportConfig.exportCG).modify(_ => Some(output))(config)
    } text "exports phi as circuit graph"

    opt[Unit]('o', "optimise") action { (_, config) => (RunConfig.expConf composeLens ExplorationConfig.explore).modify(_ => true)(config)
    } text "run optimisation method"

    opt[Unit]("run-explore-perf") action { (_, config) => (RunConfig.perfConf composeLens PerfConfig.runDSEPerf).modify(_ => true)(config)
    } text "generate exploration execution time table for bench in bench/exploration/KCR"

    opt[Unit]("run-cut-perf") action { (_, config) => (RunConfig.perfConf composeLens PerfConfig.runCutPerf).modify(_ => true)(config)
    } text "generate cutset enumeration execution time table for bench in bench/cutsetEnumeration/KCR"

    opt[Unit]("run-stmdd-perf") action { (_, config) => (RunConfig.perfConf composeLens PerfConfig.runSTMDDPerf).modify(_ => true)(config)
    } text "generate STMDD vs BDD time table for bench in bench/exploration/KCR"

    opt[Int]("timeout") action { (timeout, config) => (RunConfig.perfConf composeLens PerfConfig.timeout).modify(_ => Some(timeout second))(config)
    } text "timeout for analysis and exploration"

    opt[File]("path-xfta") action { (file, config) => (RunConfig.perfConf composeLens PerfConfig.XFTAPath).modify(_ => Some(Path(file.getAbsolutePath)))(config)
    } text s"path for XFTA (to file ${XFTAInterface.mustContain})"

    opt[File]("path-xsap") action { (file, config) => (RunConfig.perfConf composeLens PerfConfig.XSAPath).modify(_ => Some(Path(file.getAbsolutePath)))(config)
    } text s"path for XSAP (to file ${XSAPInterface.mustContain})"

    opt[File]("path-hiphops") action { (file, config) => (RunConfig.perfConf composeLens PerfConfig.HIPHOPSPath).modify(_ => Some(Path(file.getAbsolutePath)))(config)
    } text s"path for HIPHOPS (to file ${HipHOPSInterface.mustContain})"

    opt[File]("path-wine") action { (file, config) => (RunConfig.perfConf composeLens PerfConfig.WINEPath).modify(_ => Some(Path(file.getAbsolutePath)))(config)
    } text s"path for WINE (to file ${WineInterface.mustContain})"

    opt[File]("path-clasp") action { (file, config) => (RunConfig.perfConf composeLens PerfConfig.CLASPPath).modify(_ => Some(Path(file.getAbsolutePath)))(config)
    } text s"path for CLASP (to file ${ClaspInterface.mustContain})"
  }

  /**
    * Naive measure of executime
    * @param f the function to execute
    * @tparam A the type of the result
    * @return the result and the exeuction time
    */
  private def runtime[A](f: () => A): (A, Double) = {
    val start = System.nanoTime()
    (f(), (System.nanoTime() - start) * 1e-9)
  }

  /**
    * Run a function with a given timeout
    * @param f the function to run
    * @param timeout the timeout
    */
  private def runWithTimeout(f: => Unit, timeout: FiniteDuration): Unit = {
    val task = TimeoutTask(f, timeout, println("[ERROR] timeout"))
    task.get()
  }

  /**
    * Execute the command given by the user
    * @param runConf the requested command
    */
  def cliMode(runConf: RunConfig): Unit = {


    //build a full analyser to perform requested command
    for (confName <- runConf.confName) yield {
      val fullAnalyzer= new FileParsing
        with SystemConsole
        with SMTTranslation
        with MonotonyAnalyser
        with AcceptableCardAnalyzer
        with IterativeCutEnumerator
        with ReliabilityComputer
        with AcceptableReliabilityAnalyzer
        with TraceInterpreter
        with SubstitutionAnalyzerSim
        with STMDDBuilderSim
        with DSEProblemSolverKCRSim
        with SolutionAnalyzer
        with CardinalityComputer
        with SymbolicFailureRateComputer
        with HipHOPSExporter
        with SMVExporter
        with CircuitGraphExporter
        with OpenPSAExporter {
        val confId: String = confName
        val fileNames: List[String] = runConf.systemFiles.map(_.getAbsolutePath)
        val BDDFactory = new BDDFactory with Z3BoolExprToBDDTranslator
        val InstBoolBDDFactory = new InstBoolBDDFactory
      }

      fullAnalyzer.printInfo("[INFO] starting analysis phase")
      for (output <- runConf.exportConf.printDot) yield {
        fullAnalyzer.phiCarac.printDot(output.getAbsolutePath)
        fullAnalyzer.printResult(s"[RESULT] phi BDD exported in DOT file: ${output.getName}")
      }
      for (output <- runConf.exportConf.exportRSage) yield {
        fullAnalyzer.printSymbIniRBDD(output.getAbsolutePath)
        fullAnalyzer.printResult(s"[RESULT] r formula exported in SAGE script: ${output.getName}")
      }
      for (output <- runConf.exportConf.exportCG) yield {
        fullAnalyzer.exportToCG(new File(output.getAbsolutePath))
        fullAnalyzer.printResult(s"[RESULT] phi exported as Circuit Graph in: ${output.getName}")
      }
      for (output <- runConf.exportConf.exportHipHops) yield {
        fullAnalyzer.exportToHipHOPS(new File(output.getAbsolutePath))
        fullAnalyzer.printResult(s"[RESULT] problem exported as Hiphops script in: ${output.getName}")
      }
      for (output <- runConf.exportConf.exportSMV) yield {
        fullAnalyzer.exportToSMV(new File(output.getAbsolutePath))
        fullAnalyzer.printResult(s"[RESULT] problem exported as SMV script in: ${output.getName}")
      }
      for (output <- runConf.exportConf.exportXFTA) yield {
        fullAnalyzer.exportToOpenPSA(new File(output.getAbsolutePath))
        fullAnalyzer.printResult(s"[RESULT] phi exported as XFTA script in: ${output.getName}")
      }
      if (runConf.anaConf.checkMinCard) {
        fullAnalyzer.conf.minCardValue match {
          case Some(_) =>
            val (cardOK, t) = runtime(() => fullAnalyzer.cardOK)
            fullAnalyzer.printResult(s"[RESULT] mincard req satisfied (runtime:$t s): ${cardOK.get}")
          case None => fullAnalyzer.printError(s"[Error] no minimal order requirement defined in configuration ${fullAnalyzer.conf.ident.baseId.name}")
        }
      }
      if(runConf.anaConf.checkMonotony){
        if(fullAnalyzer.isCoherent()){
          fullAnalyzer.printResult(s"[RESULT] system is monotonic")
        }else {
          fullAnalyzer.printResult(s"[RESULT] ${fullAnalyzer.counterExample.get}")
        }
      }
      if (runConf.anaConf.computeMinCard) fullAnalyzer.printResult(s"[RESULT] system mincard is: ${fullAnalyzer.cardinality}")
      for (k <- runConf.anaConf.computeMCSUntil) yield {
        val (cuts, t) = fullAnalyzer.profilerCutEnumeration.runWithProfile(if (k > 0) Some(k) else None)(1)
        if (k == 0) {
          fullAnalyzer.printResult(s"[RESULT] system cutsets (runtime:$t ):")
        } else {
          fullAnalyzer.printResult(s"[RESULT] system cutsets of size lesser or equal than $k (runtime:$t ):")
        }
        if (cuts.isEmpty) {
          fullAnalyzer.printResult("No cut")
        } else {
          cuts.toList.sortBy(cut => cut.size).foreach(cut => fullAnalyzer.printResult(cut.map(evt => evt.toString()).mkString(",")))
        }
      }
      if (runConf.anaConf.computeR) {
        val (unr, t) = runtime(() => fullAnalyzer.reliability)
        fullAnalyzer.printResult(s"[RESULT] system reliability is (runtime:$t s): $unr")
      }
      if (runConf.anaConf.checkR) {
        runConf.expConf.minR orElse {
          fullAnalyzer.conf.reliability.map(s => BigDecimal(s.repr))
        } match {
          case Some(obj: BigDecimal) =>
            val (isSat, t) = runtime(() => fullAnalyzer.reliability >= obj)
            fullAnalyzer.printResult(s"[RESULT] reliability req satisfied (runtime $t s): $isSat")
          case None =>
            fullAnalyzer.printError(s"[Error] no reliability requirement defined in configuration ${fullAnalyzer.conf.ident.baseId.name}")
        }
      }
      fullAnalyzer.printInfo("[INFO] end of analysis phase")
      if (runConf.expConf.explore) {
        fullAnalyzer.printInfo("[INFO] starting exploration phase")
        val (solution, t) = fullAnalyzer.designSpaceExplorationProfilerKCR.runWithProfile()(1)
        solution match {
          case Some(slt) =>
            val filename = s"${runConf.systemFiles.map(f => f.getName).mkString("_").replace(".kcr", "")}_${fullAnalyzer.conf.ident}_slt.kcr"
            Transformer.writeSolutionFile(new File(filename), fullAnalyzer.conf.parent, fullAnalyzer.conf, slt)
            fullAnalyzer.printResult(s"[RESULT]: solution found (runtime $t): $slt")
            fullAnalyzer.printResult(s"[RESULT]: solution written in: $filename")
            fullAnalyzer.printResult(s"[RESULT]: solution reliability: ${Measure[Real](1) - fullAnalyzer.computeUnrWithMemo(slt)}")
            fullAnalyzer.printResult(s"[RESULT]: solution order: ${fullAnalyzer.computeCard(slt)}")
          case None =>
            fullAnalyzer.printResult(s"[RESULT]: no solution found (runtime $t)")
        }
        fullAnalyzer.printInfo("[INFO] end of exploration phase")
      }
    }

  }

  /**
    * Run tool in GUI mode
    * @param runConfig the requested command
    */
  def guiMode(runConfig: RunConfig): Unit = {
    val ui = new UI()
    runConfig.systemFiles.foreach(ui.openTabFromFile)
  }


  /**
    * entry point of the program parsing and processing user requests
    * @param args the input arhs
    */
  def main(args: Array[String]): Unit = {
    for (runConf <- parser.parse(args, RunConfig())) yield {
      for (p <- runConf.perfConf.XFTAPath) yield {
        if (!XFTAInterface.setToolPath(p)) {
          println(s"[WARNING] cannot find ${XFTAInterface.mustContain} in $p")
        }

      }

      for (p <- runConf.perfConf.XSAPath) yield {
        if (!XSAPInterface.setToolPath(p)) {
          println(s"[WARNING] cannot find ${XSAPInterface.mustContain} in $p")
        }
      }

      for (p <- runConf.perfConf.HIPHOPSPath) yield {
        if (!HipHOPSInterface.setToolPath(p)) {
          println(s"[WARNING] cannot find ${HipHOPSInterface.mustContain} in $p")
        }

      }

      for (p <- runConf.perfConf.CLASPPath) yield {
        if (!ClaspInterface.setToolPath(p)) {
          println(s"[WARNING] cannot find ${ClaspInterface.mustContain} in $p")
        }

      }

      for (p <- runConf.perfConf.WINEPath) yield {
        if (!WineInterface.setToolPath(p)) {
          println(s"[WARNING] cannot find ${WineInterface.mustContain} in $p")
        }
      }

      if (runConf.perfConf.runCutPerf) {
        if (SystemInfo.userOs == Linux) {
          CutEnumerationPerformancePerf(runConf.perfConf.timeout)
          //process only one performance test for each program execution
          System.exit(0)
        }
        else
          println("[ERROR] performance command must be used under Linux OS")
      }

      if (runConf.perfConf.runDSEPerf) {
        if (SystemInfo.userOs == Linux) {
          DesignExplorationPerf(runConf.perfConf.timeout)
          //process only one performance test for each program execution
          System.exit(0)
        }
        else
          println("[ERROR] performance command must be used under Linux OS")
      }

      if (runConf.perfConf.runSTMDDPerf) {
        if (SystemInfo.userOs == Linux) {
          ReliabilityComputationBDDvsSTTPerf(runConf.perfConf.timeout)
          //process only one performance test for each program execution
          System.exit(0)
        }
        else
          println("[ERROR] performance command must be used under Linux OS")
      }

      if (runConf.gui)
        guiMode(runConf)
      else if (runConf.systemFiles.nonEmpty) {
        runConf.perfConf.timeout match {
          case Some(t) => runWithTimeout({
            cliMode(runConf)
          }, t)
          case None => cliMode(runConf)
        }
      }
    }
  }
}

sealed trait Config

@Lenses case class RunConfig(
                              confName: Option[String] = None,
                              systemFiles: List[File] = Nil,
                              gui: Boolean = false,
                              perfConf: PerfConfig = PerfConfig(),
                              anaConf: AnalysisConfig = AnalysisConfig(),
                              expConf: ExplorationConfig = ExplorationConfig(),
                              exportConf: ExportConfig = ExportConfig()) extends Config

@Lenses case class PerfConfig(
                               timeout: Option[FiniteDuration] = None,
                               XFTAPath: Option[Path] = None,
                               XSAPath: Option[Path] = None,
                               HIPHOPSPath: Option[Path] = None,
                               WINEPath: Option[Path] = None,
                               CLASPPath: Option[Path] = None,
                               runDSEPerf: Boolean = false,
                               runCutPerf: Boolean = false,
                               runSTMDDPerf: Boolean = false
                             )

@Lenses case class AnalysisConfig(
                                   computeR: Boolean = false,
                                   checkR: Boolean = false,
                                   computeMinCard: Boolean = false,
                                   checkMinCard: Boolean = false,
                                   checkMonotony:Boolean=false,
                                   computeMCSUntil: Option[Int] = None
                                 ) extends Config

@Lenses case class ExportConfig(
                                 exportSMV: Option[File] = None,
                                 exportHipHops: Option[File] = None,
                                 exportXFTA: Option[File] = None,
                                 exportCG: Option[File] = None,
                                 exportSTMDD: Option[File] = None,
                                 printDot: Option[File] = None,
                                 exportRSage: Option[File] = None
                               ) extends Config

@Lenses case class ExplorationConfig(
                                      explore: Boolean = false,
                                      mincard: Option[Int] = None,
                                      minR: Option[BigDecimal] = None,
                                      time: Option[Int] = None) extends Config
