package analyzers

import ast.dsl.{CompInstanceIdent, CompSubstitution, DecRealLit, InstBoolIdent}
import util.analyzers._
import util.analyzers.analyzerTypes.{ProbabilityMapType, Real}

/**
  * trait checking if the reliability requirement is met by the initial architecture
  */
trait AcceptableReliabilityAnalyzer extends AcceptabilityAnalyzer with BDDFactoryField with BDDAnalyzer {
  self: Parsing with SMTTranslation =>

  //compute the BDD of the structure function
  lazy val phiCarac = new Phi(phi) with BDDBuild {
    val BDDFactory: BDDFactory with Z3BoolExprToBDDTranslator = self.BDDFactory
  }

  lazy val reliabilityOk: Boolean = isAcceptableR

  /**
    * Compute the reliability and compare it to the requirement
    *
    * @return true if the reliability is greater or equal to the requirement
    */
  private def isAcceptableR: Boolean = {
    conf.reliability match {
      case Some(x: DecRealLit) =>
        computeIniRBDD() >= Real(x.repr)
      case None => true
    }
  }

  /**
    * Compute the unreliability out of the BDD of the structure function
    *
    * @return
    */
  private def computeIniUnrBDD() = {
    val graph = rootDef.namespace.compInstanceDepGraph

    //collect the initial version of each primary component of the system
    val alternatives = graph.inOrder.collect({ case graph.Node(data: CompInstanceIdent) if data.path != Nil => data }).map(comp => (comp,
      conf.designSpace.find(sub => sub.instance == comp) match {
        case None => throw MissingSubstitution(comp)
        case Some(sub: CompSubstitution) => sub.init
      })
    )
    val time: Real = conf.duration.map(time => Real(time.repr)).getOrElse(1)

    //the probability of occurrence of an event is p(e)=1-exp(-lambda*t)
    val formula = (lambda: Real, time: Real) => Real(1 - math.exp(-(lambda * time).toDouble))

    //associate to each event its probability
    val iniEvtPr = alternatives.flatMap(
      altDef => altDef._2.lambdaValues.map(p => (InstBoolIdent(p._1, CompInstanceIdent(altDef._1.path)), formula(Real(p._2.repr), time)))
    ).toMap
    val unrMap: ProbabilityMapType = iniEvtPr.map(e => (phiCarac.varMap(e._1).`var`(), e._2))
    computeProbaBDD(phiCarac.phiBDD, unrMap)
  }

  def computeIniRBDD(): Real = {
    1 - computeIniUnrBDD()
  }
}

