package analyzers

import util.analyzers.analyzerTypes._
import ast.dsl._
import util.analyzers._

import scala.language.postfixOps

/**
  * Created by kdelmas on 08/04/16.
  */
trait ConfigurationAnalyzer extends Analyser {
  self: Parsing =>

  /**
    * implicit transformation from Symbol to Ordered Symbol
    *
    * @param s Scala Symbol
    * @return ordered symbol
    */
  implicit def toOSymb(s: Symbol): OrderedSymbol = OrderedSymbol(s)

  /**
    * Order over instance identifiers in a system component
    */
  implicit object compOrder extends Ordering[CompInstanceIdent] {

    /**
      * Compare two instance identifiers
      *
      * @param x instance identifier
      * @param y instance identifier
      * @return the comparison of the index position of x and y
      *         in the system instances sorted list [[rootCompInstsInOrder]]
      */
    def compare(x: CompInstanceIdent, y: CompInstanceIdent): Int = {
      rootCompInstsInOrder.indexOf(x).compare(rootCompInstsInOrder.indexOf(y))
    }
  }

  /**
    * Order over flow identifiers
    */
  implicit object compBasicFlow extends Ordering[BasicFlowIdent] {

    /**
      * Comparison by the classic alphanumeric order
      *
      * @param x flow identifier
      * @param y flow identifier
      * @return the comparison over flows names
      */
    def compare(x: BasicFlowIdent, y: BasicFlowIdent): Int = {
      x.baseId.name.compare(y.baseId.name)
    }
  }

  /**
    * Order over constants
    */
  implicit object compFlowConst extends Ordering[FlowConst] {

    /**
      * Comparison by the classique alphanumeric order
      *
      * @param x constant
      * @param y constant
      * @return comparison of the constants names
      */
    def compare(x: FlowConst, y: FlowConst): Int = {
      x.toString.compare(y.toString)
    }
  }

  /**
    * Class encoding of component traces which can be ordered
    *
    * @param value          the value for each input and output flows of the component
    * @param orderFlowIdent the implicit ordering according to flow identifiers
    * @param orderFlowConst the implicit ordering according to constants names
    */
  class OrderedCompTrace(val value: Map[BasicFlowIdent, FlowValue])
                        (implicit orderFlowIdent: Ordering[BasicFlowIdent],
                         orderFlowConst: Ordering[FlowConst]) extends Ordered[OrderedCompTrace] {

    /**
      * valuations of the flows sorted with [[orderFlowConst]] and translated to String
      */
    val valuationInOrder: String =
      value.mapValues(_.toList.sorted.mkString(",")).toList.sortBy(_._1).unzip._2.mkString("|")

    /**
      * Compare to anothe component trace by comparing valuations strging
      *
      * @param that the other component trace
      * @return comparison with the other trace
      */
    def compare(that: OrderedCompTrace): Int = {
      valuationInOrder.compare(that.valuationInOrder)
    }

    override lazy val hashCode: Int = FactoryP.sHashCode(value)

    override def toString: String = value.map(kv => s"${kv._1} -> ${kv._2.mkString("(", ",", ")")}").mkString(",")
  }

  /**
    * Companion object of component traces
    */
  object OrderedCompTrace {

    /**
      * Hash consing cache
      */
    private val makeCache = collection.mutable.HashMap[Map[BasicFlowIdent, FlowValue], OrderedCompTrace]()

    /**
      * create component trace iff not already created
      *
      * @param value          the trace
      * @param orderFlowIdent the ordering over flows
      * @param orderFlowConst the ordering over constants
      * @return component trace
      */
    def apply(value: Map[BasicFlowIdent, FlowValue])
             (implicit orderFlowIdent: Ordering[BasicFlowIdent],
              orderFlowConst: Ordering[FlowConst]): OrderedCompTrace = {
      makeCache.getOrElseUpdate(value, new OrderedCompTrace(value))
    }
  }

  /**
    * Class encoding instance identifiers which can be ordered
    *
    * @param value the instance identifier
    * @param order the order over instance identifiers
    */
  class OrderedCompInstIdent(val value: CompInstanceIdent)
                            (implicit order: Ordering[CompInstanceIdent]) extends Ordered[OrderedCompInstIdent] {
    /**
      * Comparison using [[order]]
      *
      * @param that the other instance identifier
      * @return the comparison to the other instance identifier
      */
    def compare(that: OrderedCompInstIdent): Int = {
      order.compare(value, that.value)
    }

    override lazy val hashCode: Int = FactoryP.sHashCode(value)

    override def toString: String = value.toString

  }

  /**
    * Companion object for ordered instance identifiers
    */
  object OrderedCompInstIdent {

    /**
      * Hash consing cache
      */
    private val makeCache = collection.mutable.HashMap[CompInstanceIdent, OrderedCompInstIdent]()

    /**
      * Create ordered instance identifier iff not already created
      *
      * @param value instance identifier
      * @param order order over instance identifiers
      * @return ordered instance identifier
      */
    def apply(value: CompInstanceIdent)(implicit order: Ordering[CompInstanceIdent]): OrderedCompInstIdent = {
      makeCache.getOrElseUpdate(value, new OrderedCompInstIdent(value))
    }
  }

  /**
    * Class encoding scala Symbol that can be compared
    *
    * @param s scala Symbol
    */
  class OrderedSymbol(val s: Symbol) extends Ordered[OrderedSymbol] {

    def compare(that: OrderedSymbol): Int = s.name.compare(that.s.name)

    override lazy val hashCode: Int = FactoryP.sHashCode(s)

    override def toString: String = s"${s.name}"
  }

  /**
    * Companion object for ordered symbols
    */
  object OrderedSymbol {

    /**
      * Hash consing cache
      */
    private val makeCache = collection.mutable.HashMap[Symbol, OrderedSymbol]()

    /**
      * Create ordered symbol iff not already created
      *
      * @param value instance identifier
      * @return ordered symbol
      */
    def apply(value: Symbol): OrderedSymbol = makeCache.getOrElseUpdate(value, new OrderedSymbol(value))
  }

}
