package analyzers

import ast.dsl.InstBoolIdent
import util.analyzers._

/**
  * Cut enumeration implementing all the possible cut computation techniques
  */
trait CutEnumeratorAll extends CutEnumerator with BDDFactoryField {
  self: Parsing with SMTTranslation =>


  private lazy val cutPhiSMT = new DelegatedParsing
    with SMTTranslation
    with SMTCutEnumerator{
    val delegated: Parsing = self
  }

  private lazy val cutPhiBDD = new DelegatedParsing
    with SMTTranslation
    with CutEnumeratorBDD {
    val delegated: Parsing = self
    val BDDFactory: BDDFactory with Z3BoolExprToBDDTranslator = self.BDDFactory
  }

  private lazy val cutPhiIterative = new DelegatedParsing
    with SMTTranslation
    with IterativeCutEnumerator{
    val delegated: Parsing = self
  }

  private lazy val cutPhiPB = new DelegatedParsing
    with SMTTranslation
    with CutEnumeratorPB{
    val delegated: Parsing = self
  }

  private lazy val cutPhiXFTA =new DelegatedParsing
    with SMTTranslation
    with OpenPSAExporter
    with XFTACutEnumerator{
    val delegated: Parsing = self
  }

  private lazy val cutPhiXSAP = new DelegatedParsing
    with SMTTranslation
    with SMVExporter
    with XSAPCutEnumerator{
    val delegated: Parsing = self
  }

  private lazy val cutPhiHipHOPS = new DelegatedParsing
    with SMTTranslation
    with HipHOPSExporter
    with HipHOPSCutEnumerator{
    val delegated: Parsing = self
  }

  /**
    * Current computation mode
    */
  var defaultComputationMethod: CutComputationMode = IterativeSATCutComputation

  /**
    * Compute cutsets with a given method
    *
    * @param max    bound on cutset cardinality
    * @param method method used
    * @return computed cutsets
    */
  def enumerateCutWith(max: Option[Int], method: CutComputationMode): Iterable[Set[InstBoolIdent]] = {
    defaultComputationMethod = method
    enumerateCutUntil(max)
  }

  /**
    * Compute cutsets with a default method
    *
    * @param max bound on cutset cardinality
    * @return computed cutsets
    */
  def enumerateCutUntil(max: Option[Int]): Iterable[Set[InstBoolIdent]] = {
    val computer = defaultComputationMethod match {
      case XFTACutComputation => cutPhiXFTA
      case XSAPCutComputation => cutPhiXSAP
      case HipHOPSCutComputation => cutPhiHipHOPS
      case IterativeSATCutComputation => cutPhiIterative
      case IterativeSMTCutComputation => cutPhiSMT
      case BDDCutComputation => cutPhiBDD
      case PBCutComputation => cutPhiPB
    }
    computer.enumerateCutUntil(max)
  }

  /**
    * Profiler for timing the cutsets computation according to selected method
    */
  val profilerCutEnumeration = new Profiler[Option[Int], Iterable[Set[InstBoolIdent]]] {
    def runWithProfile(i: Option[Int])(iterations: Int = 1): (Iterable[Set[InstBoolIdent]], ExecutionTimes) = {
      val computer = defaultComputationMethod match {
        case XFTACutComputation => cutPhiXFTA
        case XSAPCutComputation => cutPhiXSAP
        case HipHOPSCutComputation => cutPhiHipHOPS
        case IterativeSATCutComputation => cutPhiIterative
        case IterativeSMTCutComputation => cutPhiSMT
        case BDDCutComputation => cutPhiBDD
        case PBCutComputation => cutPhiPB
      }
      computer.profilerCutEnumeration.runWithProfile(i)(iterations)
    }
  }
}