package analyzers

import java.io.File

import ast.dsl.{CompAlternative, CompInstanceIdent}
import translators.DSLToHipHOPSTranslator
import util.analyzers.{ExecutionTimes, HipHOPSInterface, Profiler, WineInterface}
import util.analyzers.analyzerTypes.Real
import util.io.{FileManager, SystemInfo}
import scala.language.postfixOps
import scala.sys.process.Process
import scala.xml.XML

/**
  * Trait for DSE solving using HIPHOPS
  */
trait DSEProblemSolverHipHOPS extends ConfigurationAnalyzer {
  self: Parsing =>

  type Solution = Map[CompInstanceIdent, CompAlternative]

  /**
    * Solve DSE problem using HIPHOPS
    *
    * @param reliabilityReq reliability requirement if defined, otherwise optimised
    * @return Solution if at least one found by HIPHOPS
    */
  def solveWithHipHOP(reliabilityReq: Option[String] = None): Option[Solution] = {
    val modelFile = createHipHOPSFiles(reliabilityReq)
    val resFile = runHipHOPS(modelFile)
    for (r <- resFile;
         s <- backtranslateFile(r)) yield {
      s
    }
  }

  /**
    * Generate model file
    *
    * @param req the reliability requirement
    * @return the model file
    */
  private def createHipHOPSFiles(req: Option[String] = None): File = {
    val tmpHip = FileManager.addFileToTemp("tempHip.xml")
    DSLToHipHOPSTranslator.translateConfiguration(conf, tmpHip, req)
    tmpHip
  }

  /**
    * Run HIPHOPS on model file
    *
    * @param hiphopFile    HIPHOPS model file
    * @param nbGenerations maximum number of genetations (by default 10)
    * @return result file
    */
  private def runHipHOPS(hiphopFile: File, nbGenerations: Int = 10): Option[File] = {
    val outRepertory = FileManager.addFileToTemp("tempHipOutExp")
    val tooPath = HipHOPSInterface.toolPath.toString()
    SystemInfo.userOs match {
      case SystemInfo.Windows =>
        HipHOPSInterface.startProcess(Process(Seq(
          tooPath,
          hiphopFile.getAbsolutePath,
          "idevns=false",
          s"outputdir=${outRepertory.getAbsolutePath}",
          "optimise=true",
          s"maxGenerations=$nbGenerations"))).exitValue()
      case _ =>
        HipHOPSInterface.startProcess(Process(Seq(WineInterface.toolPath.toString(),
          tooPath,
          hiphopFile.getAbsolutePath,
          "idevns=false",
          s"outputdir=${outRepertory.getAbsolutePath}",
          "optimise=true",
          s"maxGenerations=$nbGenerations"))).exitValue()
    }

    HipHOPSInterface.killInternetExplorer()
    val filter = (f: File) => {
      f.getName.matches("GA.*\\.xml")
    }
    FileManager.locateFirstFile(outRepertory, filter)
  }

  /**
    * Run HIPHOPS on model file and extract resolution time from result
    *
    * @param hiphopFile    HIPHOPS model file
    * @param nbGenerations maximum number of genetations (by default 10)
    * @return result file and execution time in nanoseconds
    */
  private def runHipHOPSWithTime(hiphopFile: File, nbGenerations: Int = 10): (File, Long) = {
    val outRepertory = FileManager.addFileToTemp("tempHipOutExp")
    val tooPath = HipHOPSInterface.toolPath.toString()
    SystemInfo.userOs match {
      case SystemInfo.Windows =>
        HipHOPSInterface.startProcess(Process(Seq(
          tooPath,
          hiphopFile.getAbsolutePath,
          "idevns=false",
          s"outputdir=${outRepertory.getAbsolutePath}",
          "optimise=true",
          s"maxGenerations=$nbGenerations"))).exitValue()
      case _ =>
        HipHOPSInterface.startProcess(Process(Seq(WineInterface.toolPath.toString(),
          tooPath,
          hiphopFile.getAbsolutePath,
          "idevns=false",
          s"outputdir=${outRepertory.getAbsolutePath}",
          "optimise=true",
          s"maxGenerations=$nbGenerations"))).exitValue()
    }
    HipHOPSInterface.killInternetExplorer()
    val filter = (f: File) => {
      f.getName.matches("GA.*\\.xml")
    }
    val result = FileManager.locateFirstFile(outRepertory, filter).get
    val tXml = XML.loadFile(result.getAbsolutePath) \\ "CoitGA" \ "TimeElapsed"
    val t = (tXml.head.text.toDouble * 1e9).toLong
    (FileManager.locateFirstFile(outRepertory, filter).get, t)
  }

  import scala.xml._

  /**
    * Find an alternative of a component instance from its name in HIPHOPS model file
    *
    * @param name HIPHOPS name of the alternative
    * @param inst HIPHOPS name of the instance identifier
    * @return the component instance identifier and alternative of the design space
    */
  private def locateAlternativeFromString(name: String, inst: String): (CompInstanceIdent, CompAlternative) = {
    DSLToHipHOPSTranslator.altFromString(inst, name, conf).get
  }

  /**
    * extract solution from result file if any
    *
    * @param resFile result file
    * @return optionnal solution
    */
  private def backtranslateFile(resFile: File): Option[Solution] = {
    val xml = XML.loadFile(resFile)

    //extract individual of the result file
    val individuals = xml \\ "Population" \ "Individual"

    //extract the selected substitutions and unreliability
    val solutionsStrings = individuals.map(n => {
      val comps = n \\ "Component" \ "Name" map (c => c.text)
      val subs = n \\ "Component" \ "Implementation" \ "Name" map (c => c.text)
      val unr = n \\ "Evaluation" \ "Value" text

      (comps.zip(subs), unr)
    })

    //filter all individuals fulfilling the reliability requirement if any
    val slts = conf.reliability match {
      case Some(req) =>
        solutionsStrings.collect({
          case (slt, unr) if 1 - Real(unr) >= Real(req.repr) => slt.map(p => locateAlternativeFromString(p._2, p._1)).toMap
        })
      case None =>
        solutionsStrings.map(slt => slt._1.map(p => locateAlternativeFromString(p._2, p._1)).toMap)
    }
    slts.headOption
  }

  val designSpaceExplorationHipHopProfiler = new Profiler[Unit, Option[Solution]] {
    override def runWithProfile(i: Unit)(iterations: Int = 1): (Option[Solution], ExecutionTimes) = {
      val (r, preTime) = time({
        createHipHOPSFiles()
      }, iterations)
      val (r2, pTime) = runHipHOPSWithTime(r)
      val (r3, postTime) = time({
        backtranslateFile(r2)
      }, iterations)
      (r3, ExecutionTimes(preTime, pTime, postTime))
    }
  }
}
