package analyzers

import com.microsoft.z3.BoolExpr
import com.microsoft.z3.Status.{SATISFIABLE, UNKNOWN, UNSATISFIABLE}
import util.analyzers.{ExecutionTimes, Profiler, Z3UnknownException}
import util.analyzers.analyzerTypes.Real

import scala.collection.mutable.ListBuffer

/**
  * Trait for DSE solving of KCR modelled problems
  */
trait DSEProblemSolverKCR extends DSEProblemSolver {
  self: Parsing with SMTTranslation with STMDDBuilder with SafetyTheory =>

  /**
    * solution of the DSE problem
    */
  lazy val solution: Option[Solution] = solveConf()

  /**
    * Timed DSE resolution
    */
  val designSpaceExplorationProfilerKCR = new Profiler[Unit, Option[Solution]] {
    def runWithProfile(i: Unit)(iterations: Int = 1): (Option[Solution], ExecutionTimes) = {
      val (_, preTime) = time({
        STMDDD
      }, iterations)
      val (result, pTime) = time({
        solveConf()
      }, iterations)
      (result, ExecutionTimes(preTime, pTime, 0))
    }
  }

  /**
    * Solve the DSE problem of the actual configuration [[conf]]
    *
    * @param withXplain use quickXlain conflict clause minimisation algorithm
    * @return solution if it exists
    */
  def solveConf(withXplain: Boolean = true): Option[Solution] = {
    Solver.reset()
    for (r <- conf.reliability) yield {
      Solver.add(SafeR(STMDDD, Real(r.repr)))
    }
    for (o <- conf.minCardValue) yield {
      Solver.add(SafeOrder(STMDDD, o.repr.toInt))
    }
    Solver.checkAndGetModel(withXplain)
  }

  /**
    * Solve the DSE problem using custom reliability and order requirements
    *
    * @param RObj       optional reliability requirement
    * @param cardObj    optional order requirement
    * @param withXplain use quickXlain conflict clause minimisation algorithm
    * @return solution if it exists
    */
  def solveConfWithReq(RObj: Option[Real],
                       cardObj: Option[Int],
                       withXplain: Boolean = true): Option[Solution] = {
    Solver.reset()
    for (r <- RObj) yield {
      Solver.add(SafeR(STMDDD, r))
    }
    for (o <- cardObj) yield {
      Solver.add(SafeOrder(STMDDD, o))
    }
    Solver.checkAndGetModel(withXplain)
  }

  /**
    * Solver object handling classic Z3 Boolean assertions and custom Safety assertions
    */
  object Solver {

    //list of safety assertions
    val safetyPredicates: ListBuffer[SafetyReqPredicate] = ListBuffer.empty

    //list of Z3 assertions
    val z3Predicates: ListBuffer[BoolExpr] = ListBuffer.empty

    //clear assertion lists
    def reset(): Unit = {
      safetyPredicates.clear()
      z3Predicates.clear()
    }

    /**
      * add a safety assertion to the problem
      *
      * @param p safety requirement
      */
    def add(p: SafetyReqPredicate): Unit = {
      safetyPredicates += p
    }

    /**
      * add classic assertion to the problem
      *
      * @param p Z3 boolean expression
      */
    def add(p: BoolExpr): Unit = {
      z3Predicates += p
    }

    /**
      * Extract the substitution choice from a model
      *
      * @param model model of the substitution variables
      * @return the chosen alternatives on system's components
      */
    def buildReplaceMap(model: Model): Solution = {
      model.map(p => {
        val (myVar, value) = p
        val sub = conf.designSpace.find(compSub => compSub.instance == myVar.inst).get
        val allAlts = sub.init +: sub.alternatives
        myVar.inst -> allAlts(value)
      }).toMap
    }

    /**
      * Lazy integration of the Safety theory to solve DSE problem
      *
      * @param withXplain use clause minimisation for problem solving
      * @return the solution if exists
      */
    def checkAndGetModel(withXplain: Boolean = true): Option[Solution] = {

      //generate substitution variables
      val subVars = generateDecisionVars()

      //create a Z3 solver
      val s = _SMTContext.mkSolver()

      //status of the overall problem resolution
      var status = UNKNOWN

      //result of the problem
      var result: Option[Solution] = None

      //add Z3 assertion to Z3 solver
      s.add(z3Predicates: _*)

      //while problem not solved
      while (status == UNKNOWN) {

        //check SAT instance of the problem
        s.check() match {

          //if Satisfiable check according to Safety theory
          case SATISFIABLE =>

            //extract model from SAT instance as set of Choose predicates
            val model = subVars.map(e => Choose(e, _SMTContext.getValuePos(e, s.getModel.eval(e.z3Var, true)).get))

            //check if correct according to Safety
            checkSat(safetyPredicates ++ model, withXplain) match {

              //satisfiable modulo Safety
              case Left(partialModel) =>

                //problem solved
                status = SATISFIABLE

                //build solution from model
                result = Some(buildReplaceMap(partialModel))

              //if conflict
              case Right(conflict) =>

                //compute conflict and add it to Z3 solver
                s.add(_SMTContext.mkOr(conflict.map(c => c.toZ3): _*))
            }

          //if SAT instance is UNSATISFIABLE
          case UNSATISFIABLE =>

            //the overall problem is UNSAT
            status = UNSATISFIABLE

          //if Z3 cannot solve SAT instance
          case UNKNOWN =>

            //impossible since SAT is decidable
            throw Z3UnknownException(self)
        }
      }

      //free Z3 solver
      s.dispose()

      result
    }
  }

}
