package analyzers

import java.io.File

import ast.dsl.{BoolIdent, CompInstanceIdent, InstBoolIdent, StdDSLSymbol}
import translators.DSLToHipHOPSTranslator
import util.analyzers.{ExecutionTimes, HipHOPSInterface, Profiler, WineInterface}
import util.io.{FileManager, SystemInfo}
import scala.language.postfixOps
import scala.io.Source
import scala.sys.process.Process
import scala.xml.XML

/**
  * Trait for HIPHOPS based computation of cutsets
  */
trait HipHOPSCutEnumerator extends CutEnumerator {
  self: Parsing with SMTTranslation with HipHOPSExporter =>

  /**
    * Generate the HIPHOPS translation of the system
    *
    * @return generated file
    */
  private def createHipHOPSFiles(): File = {
    val tmpHip = FileManager.addFileToTemp("tempHip.xml")
    DSLToHipHOPSTranslator.translateConfiguration(conf, tmpHip)
    tmpHip
  }

  /**
    * Run HIPHOPS generation on the system file
    *
    * @param hiphopFile HIPHOPS system file
    * @param max        bound on cutsets cardinality
    * @return the result file
    */
  private def computeWithHipHOPS(hiphopFile: File, max: Option[Int]): File = {
    val outRepertory = FileManager.addFileToTemp("tempHipOut")
    val toolPath = HipHOPSInterface.toolPath.toString
    val command = SystemInfo.userOs match {
      case SystemInfo.Windows =>
        Seq(toolPath,
          hiphopFile.getAbsolutePath,
          "outputtype=XML",
          s"outputdir=${outRepertory.getAbsolutePath}",
          "-openiex",
          (for (m <- max) yield s"maxcutsets=$m") getOrElse "")
      case _ =>
        Seq(WineInterface.toolPath.toString(),
          toolPath,
          hiphopFile.getAbsolutePath,
          "outputtype=XML",
          s"outputdir=${outRepertory.getAbsolutePath}",
          "-openiex",
          (for (m <- max) yield s"maxcutsets=$m") getOrElse "")
    }
    HipHOPSInterface.startProcess(command).exitValue()
    val filter = (f: File) => {
      f.getName.matches("CutSets.*")
    }
    FileManager.recursiveLocateFirstFile(outRepertory, filter).get
  }

  /**
    * Run HIPHOPS generation on the system file
    *
    * @param hiphopFile HIPHOPS system file
    * @param max        bound on cutsets cardinality
    * @return the result file and the time given by HIPHOPS in nanoseconds
    */
  private def computeWithHipHOPSWithTime(hiphopFile: File, max: Option[Int]): (File, Long) = {
    val outRepertory = FileManager.addFileToTemp("tempHipOut")
    val logFile = FileManager.addFileToTemp("log.txt")
    val toolPath = HipHOPSInterface.toolPath.toString
    val command = SystemInfo.userOs match {
      case SystemInfo.Windows =>
        Seq(toolPath,
          hiphopFile.getAbsolutePath,
          "outputtype=XML",
          s"outputdir=${outRepertory.getAbsolutePath}",
          "-openiex",
          (for (m <- max) yield s"maxcutsets=$m") getOrElse "")
      case _ =>
        Seq(WineInterface.toolPath.toString(),
          toolPath,
          hiphopFile.getAbsolutePath,
          "outputtype=XML",
          s"outputdir=${outRepertory.getAbsolutePath}",
          "-openiex",
          (for (m <- max) yield s"maxcutsets=$m") getOrElse "")
    }
    HipHOPSInterface.startProcess(Process(command) #> logFile).exitValue()
    val tLine = Source.fromFile(logFile).getLines().find(s => s.matches("Time elapsed.*")).get
    //from h:m:s.ms to nanoseconds
    val sRegexp = ".* ([0-9]+\\.[0-9]+)s".r
    val mRegexp = ".* ([0-9]+)m .*".r
    val hRegexp = ".* ([0-9]+)h .*".r
    val h = tLine match {
      case hRegexp(x) => Some((x.toInt * 3600).toDouble)
      case _ => None
    }
    val m = tLine match {
      case mRegexp(x) => Some((x.toInt * 60).toDouble)
      case _ => None
    }
    val s = tLine match {
      case sRegexp(x) => Some(x.toDouble)
      case _ => None
    }
    val t: Double = List(h, m, s).flatten.sum
    val filter = (f: File) => {
      f.getName.matches("CutSets.*")
    }
    (FileManager.recursiveLocateFirstFile(outRepertory, filter).get, (t * 1e9).toLong)
  }

  /**
    * extract cutsets from HIPHOPS result file
    *
    * @param result HIPHOPS result file
    * @return cutsets
    */
  private def backtranslateHipHOPS(result: File): Iterable[Set[InstBoolIdent]] = {
    val xml = XML.loadFile(result)
    val ids = xml \\ "FMEA" \ "Component" \ "Events" \ "BasicEvent" map (n => n \ "@ID" text) filter (s => s.toInt >= 0)
    val names = xml \\ "FMEA" \ "Component" \ "Events" \ "BasicEvent" \ "Name" collect {
      case n if !n.text.contains(172.toChar) =>
        val evtId = n.text.split("\\.")
        InstBoolIdent(BoolIdent(StdDSLSymbol(Symbol(evtId.last))), CompInstanceIdent(evtId.init.map(x => StdDSLSymbol(Symbol(x))).toList))
    }
    val evtMap = ids.zip(names).toMap
    val cutsets = xml \\ "AllCutSets" \ "CutSets" \ "CutSet"
    cutsets.map(c => c \\ "Events" \ "Event" collect {
      case e if (e \ "@ID" text).toInt >= 0 =>
        evtMap(e \ "@ID" text)
    } toSet)
  }

  /**
    * Compute cutsets with HIPHOPS
    *
    * @param max maximal cutset size if defined, no bound otherwise
    * @return the set of minimal cutsets
    */
  def enumerateCutUntil(max: Option[Int]): Iterable[Set[InstBoolIdent]] = {
    val model = createHipHOPSFiles()
    val resFile = computeWithHipHOPS(model, max)
    val slt = backtranslateHipHOPS(resFile)
    slt
  }

  val profilerCutEnumeration = new Profiler[Option[Int], Iterable[Set[InstBoolIdent]]] {
    def runWithProfile(i: Option[Int])(iterations: Int = 1): (Iterable[Set[InstBoolIdent]], ExecutionTimes) = {
      val (r1, preTime) = time({
        createHipHOPSFiles()
      }, iterations)
      val (r2, pTime) = computeWithHipHOPSWithTime(r1, i)
      val (r3, posTime) = time({
        backtranslateHipHOPS(r2)
      }, iterations)
      (r3, ExecutionTimes(preTime, pTime, posTime))
    }
  }

}
