package analyzers

import ast.dsl.InstBoolIdent
import com.microsoft.z3.{BoolExpr => Z3BoolExpr, Expr => Z3Expr}
import com.microsoft.z3.Status._
import util.analyzers.MonotonyCounterExample



trait MonotonyAnalyser extends ConfigurationAnalyzer{

  self: Parsing with SMTTranslation =>

  lazy val counterExample: Option[MonotonyCounterExample] = _isCoherent

  def isCoherent():Boolean ={
    counterExample.isEmpty
  }

  /**
    * Analyser checking if phi is coherent (monotonic) by solving SMT problem
    *
    * @return if phi is coherent
    */
  private def _isCoherent: Option[MonotonyCounterExample] = {

    //generate a Z3 solver
    val s = _SMTContext.mkSolver()

    val eventsList: List[(InstBoolIdent, Z3BoolExpr)] = phi.events.toList

    //generate two vectors of Boolean interpreted variables representing system events,
    val consts: List[(Z3BoolExpr, Z3BoolExpr, Z3BoolExpr)] = eventsList.map(e =>
      (e._2, _SMTContext.mkConst(e._1.toString(), e._2.getSort).asInstanceOf[Z3BoolExpr]
        , _SMTContext.mkConst(e._1.toString() + "2", e._2.getSort).asInstanceOf[Z3BoolExpr]))

    //ensure that any event of the first vector implies the corresponding one in the second vector
    val impliesEvt: Z3BoolExpr = _SMTContext.mkAnd(consts.map(t => _SMTContext.mkImplies(t._2, t._3)): _*)

    val (iniVars, newVars1, newVars2) = consts.unzip3

    //substitute varaible in phi expression by the constants of the first vector
    val phiVar1 = phi.expr.substitute(iniVars.toArray[Z3Expr], newVars1.toArray[Z3Expr]).asInstanceOf[Z3BoolExpr]

    //substitute varaible in phi expression by the constants of the second vector
    val phiVar2 = phi.expr.substitute(iniVars.toArray[Z3Expr], newVars2.toArray[Z3Expr]).asInstanceOf[Z3BoolExpr]

    //assert vector1 => vector2
    s.add(impliesEvt)

    //assert phi(vector1) is true and phi(vector2) is false
    s.add(phiVar1)
    s.add(_SMTContext.mkNot(phiVar2))

    //check if (vector1 => vector2) => (phi(vector1)=>phi(vector2))
    if(s.check() == UNSATISFIABLE){
      None
    }else {
      Some(MonotonyCounterExample(consts.indices.map(i => eventsList(i)._1 -> (s.getModel.eval(consts(i)._2,true), s.getModel.eval(consts(i)._3,true))).toMap))
    }

  }
}
