package analyzers

import ast.dsl.{BoolIdent, CompInstanceIdent, ComponentDeclaration, InstBoolIdent, InstFlowIdent}
import com.microsoft.z3.{BitVecExpr, BoolExpr => Z3BoolExpr, Sort => Z3Sort, Status => Z3Status, Symbol => Z3Symbol}
import translators.DSLSMTTranslator
import util.analyzers._
import util.translators.CacheImplicits._
import util.translators.{EventKey, FlowKey}

/**
  * Created by kdelmas on 27/09/17.
  */

/**
  * Trait for KCR translation to SMT
  */
trait SMTTranslation extends ConfigurationAnalyzer with Z3ExprTranslators {
  self: Parsing =>

  lazy val translator = new DSLSMTTranslator()

  lazy val (_DSLContext, _SMTContext) = translator.trConfig(conf)

  lazy val phi: Phi = getPhi

  /**
    * Get the SMT translation of a KCR component
    *
    * @param comp AST node of the component declaration
    * @return component case class containing mapping from KCR flows ident to SMT translation
    */
  def getCompTranslation(comp: ComponentDeclaration): Component = {

    val inputs: Map[InstFlowIdent, BitVecExpr] = _DSLContext.filter((pair: (FlowKey, BitVecExpr)) =>
      comp.ident.baseId == pair._1.path.head && comp.inputs.map(e => e.ident.baseId).contains(pair._1.path.last)).
      map(kv => InstFlowIdent(kv._1.path.last, CompInstanceIdent(kv._1.path.init)) -> kv._2).toMap

    val outputs: Map[InstFlowIdent, BitVecExpr] = _DSLContext.filter((pair: (FlowKey, BitVecExpr)) =>
      comp.ident.baseId == pair._1.path.head && comp.outputs.map(e => e.ident.baseId).contains(pair._1.path.last)).
      map(kv => InstFlowIdent(kv._1.path.last, CompInstanceIdent(kv._1.path.init)) -> kv._2).toMap

    val events: Map[InstBoolIdent, Z3BoolExpr] = _DSLContext.filter((pair: (EventKey, Z3BoolExpr)) =>
      comp.ident.baseId == pair._1.path.head && comp.body.events.map(e => e.baseId).contains(pair._1.path.last)).
      map(kv => InstBoolIdent(BoolIdent(kv._1.path.last), CompInstanceIdent(kv._1.path.init)) -> kv._2).toMap

    new Component(inputs, outputs, events, _SMTContext)
  }

  /**
    * Compute the structure function of the system by solving an SMT problem
    *
    * @return
    */
  private def getPhi = {

    //select the translation of the events of the system and of the failure condition
    val getPhiInfos: PartialFunction[(EventKey, Z3BoolExpr), (InstBoolIdent, Z3BoolExpr)] = {
      case (key@EventKey(a :: _), c) if a == conf.rootCompId.baseId || a == conf.ident.baseId =>
        val id = key.path.tail
        (InstBoolIdent(BoolIdent(id.last), CompInstanceIdent(id.init)), c)
    }

    val info: Map[InstBoolIdent, Z3BoolExpr] = _DSLContext.collect(getPhiInfos).toMap

    //find  the expression of phi
    val expr = info.filter(pair => pair._1.instId.path == Nil).head._2

    //find the system events
    val events = info.filterNot(pair => pair._1.instId.path == Nil)

    //instantiate a solver
    val s = _SMTContext.mkSolver()

    //sorts events by bounded index
    val eventsList = events.toList.sortBy(_._2.getIndex)

    //create Boolean constant for each system event
    val consts = eventsList.map(e => (_SMTContext.mkSymbol(e._1.toString()), _SMTContext.mkBoolSort())).unzip

    //add a uninterpreted function declaration for phi
    val funcDecl = _SMTContext.mkFuncDecl("phiBool", (1 to eventsList.size).map(e => _SMTContext.mkBoolSort()).toArray[Z3Sort], _SMTContext.mkBoolSort())

    //create the application of the function to the events
    val funcApply = funcDecl.apply(eventsList.map(e => e._2): _*)

    // phi(e) = fc(root(e))
    val equal = _SMTContext.mkEq(funcApply, expr)

    //assert forall e phi(e) = fc(root(e))
    s.add(_SMTContext.mkForall(consts._2.toArray[Z3Sort], consts._1.reverse.toArray[Z3Symbol], equal, 0, null, null, null, null))
    s.check() match {
      case Z3Status.SATISFIABLE =>

        //find definition and store it
        val definition = s.getModel.getFuncInterp(funcDecl).getElse.asInstanceOf[Z3BoolExpr]
        s.dispose()
        Phi(toNNF(definition, events.values.toList, _SMTContext), events, _SMTContext)
      case Z3Status.UNSATISFIABLE => throw Z3UnsatException(this)
      case Z3Status.UNKNOWN => throw Z3UnknownException(this)
    }
  }

  val translationSMTProfiler = new Profiler[Unit, Unit] {
    override def runWithProfile(i: Unit)(iterations: Int = 1): (Unit, ExecutionTimes) = {
      val (result, preTime) = time({
        new DSLSMTTranslator()
      }, iterations)
      val (_, pTime) = time({
        result.trConfig(conf)
        getPhi
      }, iterations)
      (Unit, ExecutionTimes(preTime, pTime, 0))
    }
  }
}
