package analyzers

import ast.dsl.{BasicFlowIdent, CompCallFlowDefinition, CompInstanceIdent, DSLSymbol, FlowConst, InstBoolIdent, InstFlowIdent, TypeDeclaration, TypeIdent, TypeValueFlowEmpty, TypeValueFlowIdent}
import util.analyzers.InfNumericTypes.InfInt
import util.analyzers.InfNumericImplicits._
import util.analyzers.{CharacteristicFun, MDDFactory, MDDPrinter, MissingBaseAlternative}
import util.analyzers.analyzerTypes.{Real, Z3TraceType}
import util.translators.FlowKey
import com.microsoft.z3.{Sort,
  Symbol => Z3Symbol, FuncDecl => Z3FuncDecl,
  Solver => Z3Solver, BoolExpr => Z3BoolExpr,
  Expr => Z3Expr,
  BitVecExpr => Z3BitVecExpr,
  BitVecNum}
import com.microsoft.z3.Status._
import scala.reflect.ClassTag
import scala.util.{Failure, Success}

/**
  * Trait for computing the STMDD with SMT problem solving
  */
trait STMDDBuilderSMT extends STMDDBuilder {
  self: Parsing with SMTTranslation with SubstitutionAnalyzerSMT =>

  lazy val STMDDD: mddFactory.MDD = buildTrTree()

  lazy val (possibleSysTrExpr, dangSysTrExpr, sysTrVars) = buildTrSys()

  lazy val mddFactory = new MDDFactory[OrderedCompInstIdent, Set, STMDDEdgeLabel, OrderedSymbol, Set[STMDDEdgeLabel]](extractor = (a) => a)
    with MDDPrinter[OrderedCompInstIdent, Set, STMDDEdgeLabel, OrderedSymbol, Set[STMDDEdgeLabel]]

  /**
    * Check if it exists at least one trace fulfilling the prefix
    *
    * @param prefix partial evaluation of the system flow varaibles
    * @return it exists at least one trace fulfilling the prefix
    */
  private def atLeastOneOkTr(prefix: List[(InstFlowIdent, Z3Expr)]) = {
    val cst = sysTrVars.map(e => (e._1, _SMTContext.mkConst(e._1.toString, e._2.getSort)))
    val cstPrefix = _SMTContext.mkAnd(prefix.map(e => _SMTContext.mkEq(cst(e._1), e._2)): _*)
    val s = _SMTContext.mkSolver()
    val cstList = cst.toList.sortBy(p => sysTrVars(p._1).getIndex)
    val newVars = cstList.map(p => p._2).toArray
    s.add(cstPrefix)
    val badTrSys = dangSysTrExpr.substituteVars(newVars) match {
      case b: Z3BoolExpr => b
    }
    val possTrSys = possibleSysTrExpr.substituteVars(newVars) match {
      case b: Z3BoolExpr => b
    }
    val cstOkTr = _SMTContext.mkAnd(possTrSys, _SMTContext.mkNot(badTrSys))
    s.add(cstOkTr)
    s.add(cstPrefix)
    val res = s.check() == SATISFIABLE
    s.dispose()
    res
  }

  /**
    * Compute the set of dangerous component traces for a given component instance, such that the
    * prefix is fulfilled
    *
    * @param ofInst the instance component
    * @param prefix the partial valuation of system flows
    * @return the possible traces
    */
  private def getDangTr(ofInst: CompCallFlowDefinition, prefix: List[(InstFlowIdent, BitVecNum)]) = {
    var result = List[List[(InstFlowIdent, BitVecNum)]]()
    val cst = sysTrVars.map(e => (e._1, _SMTContext.mkConst(e._1.toString, e._2.getSort)))
    val cstPrefix = _SMTContext.mkAnd(prefix.map(e => _SMTContext.mkEq(cst(e._1), e._2)): _*)
    val s = _SMTContext.mkSolver()
    val cstList = cst.toList.sortBy(p => sysTrVars(p._1).getIndex)
    val newVars = cstList.map(p => p._2).toArray
    s.add(cstPrefix)
    s.add(dangSysTrExpr.substituteVars(newVars).asInstanceOf[Z3BoolExpr])
    while (s.check() != UNSATISFIABLE) {
      val trace = cst.filter(p => p._1.instId.path.contains(ofInst.rhs.instId)).map(e => {
        (e._1, s.getModel.eval(cst(e._1), true).asInstanceOf[BitVecNum])
      }).toList
      result = result :+ trace
      s.add(_SMTContext.mkNot(_SMTContext.mkAnd(trace.map(p => _SMTContext.mkEq(cst(p._1), p._2)): _*)))
    }
    s.dispose()
    result
  }

  /**
    * Find the order and probability arrays of a component instance
    *
    * @param comInst the instance
    * @return the order and probability arrays
    */
  private def getF(comInst: CompCallFlowDefinition) = {
    substitutions.find(sub => {
      val currentInst = sub.sub.instance.path.mkString(".")
      val searchedInst = comInst.rhs.instId.name
      currentInst == searchedInst
    }) match {
      case Some(sub) =>
        (sub.characAlternatives.map(alt => alt.comp.fCardField), sub.characAlternatives.map(alt => alt.fUnr)) //renvoi les charac
      case None =>
        if (comInst.namespace.locate(comInst.rhs.compId).body.events.isEmpty)
          (List(CharacteristicFun[InfInt](Map.empty, 0)), List(new CharacteristicFun[Real](Map.empty, 1)))
        else
          throw MissingBaseAlternative(comInst.rhs.instId)
    }
  }

  /**
    * Find the type declaration of a flow in a given component instance
    *
    * @param inst the component instance containing the flow
    * @param f the basic flox ident
    * @return the type declaration of the flow
    */
  def typeOf(inst: CompInstanceIdent, f: BasicFlowIdent): TypeDeclaration = {
    locateCompDef(inst).namespace.locate(f) match {
      case Left(tyype) => tyype
      case Right(fdecl) => conf.namespace.locate(fdecl.typeId)
    }
  }

  /**
    * Transform a valuation of flows to a component trace for th given instance
    *
    * @param inst      the instance
    * @param valuation the valuation of the flows
    * @return the corresponding trace
    */
  def toOrderedCompTrace(inst: CompInstanceIdent, valuation: Map[BasicFlowIdent, BitVecNum]): OrderedCompTrace = {
    val newTr = valuation.map(kv => {
      val tyype = typeOf(inst, kv._1)
      val indices = kv._2.getInt.toBinaryString.reverse.zipWithIndex.collect({ case (char, index) if char == '1' => index })
      kv._1 -> indices.map(i => TypeValueFlowIdent(tyype.ident, tyype.values(i)).asInstanceOf[FlowConst]).toSet
    })
    OrderedCompTrace(newTr)
  }

  /**
    * Transform a set of characteristic functions to a mapping of traces to indicators arrays
    *
    * @param inst the instance
    * @param l    the list of characteristic functions
    * @param tag  class tag distinguishing order and reliability characteristic functions
    * @tparam T the type of the characteristic function
    * @return mapping of traces to indicators arrays
    */
  private def toLabel[T](inst: CompInstanceIdent, l: List[CharacteristicFun[T]])(implicit tag: ClassTag[T]): Map[OrderedCompTrace, Array[T]] = {
    l match {
      case Nil => Map[OrderedCompTrace, Array[T]]()
      case h :: Nil => h.func.map(kv => toOrderedCompTrace(inst, kv._1) -> Array(kv._2))
      case h :: t => h.func.map(kv => toOrderedCompTrace(inst, kv._1) -> (kv._2 +: t.map(f => f.getFirst(kv._1)).toArray))
    }
  }

  /**
    * Build the STMDD from SMT problem solving
    *
    * @return the STMDD
    */
  protected def buildTrTree(): mddFactory.MDD = {
    val comps = rootCompInstsInOrder.collect({ case inst if rootDef.namespace.locateDef(inst).nonEmpty =>
      (inst, rootDef.namespace.locateDef(inst).head)
    })
    buildSons(Nil, comps)
  }

  /**
    * Iterative computation of STMDD
    *
    * @param prefix the current partial valuation of the system flows
    * @param comps  the component instances of the system
    * @return a STMDD node
    */
  def buildSons(prefix: List[Z3TraceType], comps: List[(CompInstanceIdent, CompCallFlowDefinition)]): mddFactory.MDD = {

    comps match {
      case Nil =>
        mddFactory.getTerminal('one)

      case (inst, fdef) :: t =>

        //asks node for current component, if not known add it to the variables
        mddFactory.getVarMDD(OrderedCompInstIdent(inst), (l: STMDDEdgeLabel) => Symbol(l.trace.valuationInOrder)) match {
          case Success(x) =>
            x
          case Failure(_) =>
            val characs = getF(fdef)
            val cards = toLabel(inst, characs._1)
            val unrs = toLabel(inst, characs._2)
            val labels = cards.map(kv => STMDDEdgeLabel(kv._1, unrs(kv._1), kv._2)).toList
            mddFactory.addVar(OrderedCompInstIdent(inst), labels.toSet)
        }

        //build the component traces involved in a dangerous traces for the current component
        val dangTr = getDangTr(comps.head._2, prefix.flatten)

        //get all possible traces of the current component
        val labels = mddFactory.getDomain(OrderedCompInstIdent(inst))

        //build the sons
        val sons: Map[STMDDEdgeLabel, mddFactory.MDD] = dangTr.map(tr => {

          //find the label for the curren trace
          val l = labels.find(l => l.trace == toOrderedCompTrace(inst, tr.map(kv => BasicFlowIdent(kv._1.baseId) -> kv._2).toMap)).get

          //if all possible traces are dangerous then terminal 1 otherwise build sons
          if (!atLeastOneOkTr((prefix :+ tr).flatten)) {
            l -> mddFactory.getTerminal('one)
          } else {
            l -> buildSons(prefix :+ tr, t)
          }
        }).toMap

        mddFactory.buildNodeDefaultTrash(OrderedCompInstIdent(inst), sons)
    }
  }

  /**
    * Build two predicates
    * possibleSysTr: true if the valuation of system flows is a possible system trace
    * dangerousSysTr: true if the valuation of system flows is a dangerous system trace
    *
    * @return
    */
  private def buildTrSys() = {

    //system events
    val events: Map[InstBoolIdent, Z3BoolExpr] = phi.events

    //maps system flows to component instance flows
    val systFlow: List[(InstFlowIdent, InstFlowIdent)] = rootCompInstsInOrder.collect({
      case inst if rootDef.namespace.locateDef(inst).nonEmpty =>
        val compCall = rootDef.namespace.locateDef(inst).head
        val compDecl = rootDef.namespace.locate(compCall.rhs.compId)
        (compDecl.inputs ++ compDecl.outputs).map(p => {
          val keyVar = InstFlowIdent(p.ident.baseId, CompInstanceIdent(List(rootDef.ident.baseId, compCall.rhs.instId)))
          val keyIniFlow = InstFlowIdent(p.ident.baseId, CompInstanceIdent(List(compCall.rhs.compId.baseId)))
          (keyVar, keyIniFlow)
        })
    }).flatten

    //generate bounded varaible for each system flow
    val trSysVars: Map[InstFlowIdent, Z3Expr] = systFlow.zipWithIndex.map(p => {
      (p._1._1, _SMTContext.mkBound(p._2 + events.size, _DSLContext(FlowKey(p._1._2.instId.path :+ p._1._2.baseId)).getSort))
    }).toMap

    //bounded var for the system events
    val existsVars = events.toList.sortBy(p => p._2.getIndex).map(p => (_SMTContext.mkSymbol(p._1.toString()), _SMTContext.mkBoolSort()))

    //equalities between bounded flow varaible and system flow definitions
    //  ie for all system flow f:= Expr add SMT equality (= f Expr)
    val equalities = rootCompInstsInOrder.collect({
      case inst if rootDef.namespace.locateDef(inst).nonEmpty =>
        val compCall = rootDef.namespace.locateDef(inst).head
        val compDecl = rootDef.namespace.locate(compCall.rhs.compId)
        val outEq: List[Z3BoolExpr] = compCall.lhs.zipWithIndex.map(p => {
          val keyIniOut = List(rootDef.ident.baseId, compCall.rhs.instId, compDecl.outputs(p._2).ident.baseId)
          val keySystemOut = List(rootDef.ident.baseId, p._1.baseId)
          _SMTContext.mkEq(_DSLContext(FlowKey(keySystemOut)), trSysVars(InstFlowIdent(keyIniOut.last, CompInstanceIdent(keyIniOut.init))))
        })
        val inEq: List[Z3BoolExpr] = compCall.rhs.args.zipWithIndex.map(p => {
          val keyIniIn = InstFlowIdent(compDecl.inputs(p._2).ident.baseId, CompInstanceIdent(List(rootDef.ident.baseId, compCall.rhs.instId)))
          p._1 match {
            case TypeValueFlowIdent(typeId: TypeIdent, valueId: DSLSymbol) =>
              val tyype = rootDef.namespace.locate(typeId)
              _SMTContext.mkEq(_SMTContext.mkBV(Math.pow(2.0, tyype.values.indexOf(valueId).toDouble).toInt, tyype.values.size), trSysVars(keyIniIn))
            case TypeValueFlowEmpty(typeId: TypeIdent) =>
              val tyype = rootDef.namespace.locate(typeId)
              _SMTContext.mkEq(_SMTContext.mkBV(0, tyype.values.size), trSysVars(keyIniIn))
            case id: InstFlowIdent =>
              rootDef.body.namespace.locate(id) match {
                case Left(valueType) =>
                  _SMTContext.mkEq(_SMTContext.mkBV(Math.pow(2.0, valueType.values.indexOf(id.baseId).toDouble).toInt, valueType.values.size), trSysVars(keyIniIn))
                case Right(_) =>
                  val keySystemIn = List(rootDef.ident.baseId, id.baseId)
                  _SMTContext.mkEq(_DSLContext(FlowKey(keySystemIn)), trSysVars(keyIniIn))
              }
            case id: BasicFlowIdent =>
              rootDef.body.namespace.locate(id) match {
                case Left(valueType) =>
                  _SMTContext.mkEq(_SMTContext.mkBV(Math.pow(2.0, valueType.values.indexOf(id.baseId).toDouble).toInt, valueType.values.size), trSysVars(keyIniIn))
                case Right(_) =>
                  val keySystemIn = List(rootDef.ident.baseId, id.baseId)
                  _SMTContext.mkEq(_DSLContext(FlowKey(keySystemIn)), trSysVars(keyIniIn))
              }
            case _ => ???
          }
        })
        inEq ++ outEq
    }).flatten

    val possibleSysTr = _SMTContext.mkExists(existsVars.map(_._2).toArray[Sort],
      existsVars.map(_._1).toArray[Z3Symbol].reverse,
      _SMTContext.mkAnd(equalities: _*),
      0,
      null,
      null,
      null,
      null
    )

    val dangerousSysTr = _SMTContext.mkExists(existsVars.map(_._2).toArray[Sort],
      existsVars.map(_._1).toArray[Z3Symbol].reverse,
      _SMTContext.mkAnd(equalities :+ phi.expr: _*),
      0,
      null,
      null,
      null,
      null
    )
    (possibleSysTr, dangerousSysTr, trSysVars)
  }
}

