package analyzers

import ast.dsl.CompInstanceIdent
import interpreter.TraceInterpreter
import interpreter.TraceInterpreter.{ComponentInterpretation, ComponentTraces}
import net.sf.javabdd.BDD
import util.analyzers._

import scala.util.control.TailCalls.{TailRec, done, tailcall}

/**
  * Trace tree computation based on interpretation
  */
trait STMDDBuilderSim extends STMDDBuilder {
  self: Parsing with SMTTranslation with SubstitutionAnalyzerSim with TraceInterpreter =>

  /**
    * BDD factory for the interpreter and BDD build of phi
    */
  val BDDFactory: BDDFactory with Z3BoolExprToBDDTranslator

  /**
    * the variables of the STMDD (OrderedCompInstIdent)  and their domain (OrderedCompTrace)
    */
  lazy val stmddVars: Map[OrderedCompInstIdent, Map[OrderedCompTrace, STMDDEdgeLabel]] = substitutions.mapValues(v =>
    v.map(kv => kv._1 -> STMDDEdgeLabel(kv._1, kv._2._2, kv._2._1)))

  /**
    * the BDD of the structure function
    */
  lazy val phiBDD : Phi with BDDBuild = new Phi(phi) with BDDBuild {
    val BDDFactory: BDDFactory with Z3BoolExprToBDDTranslator = self.BDDFactory
  }

  lazy val mddFactory = new MDDFactory[
    OrderedCompInstIdent,   //variable type of STMDD is component instance identifiers
    Set,                    //collection for values is Set
    STMDDEdgeLabel,         //the values are edge label (ie trace and indicators arrays)
    OrderedSymbol,          //the terminal values are symbols
    Set[STMDDEdgeLabel]](   //the domain of a variable is a set of edge value
    stmddVars.mapValues(_.values.toSet), //the domains of the variable of the STMDD
    (a) => a)               //an extractor of the collection of values from a domain (here identity)
    with MDDPrinter[OrderedCompInstIdent, Set, STMDDEdgeLabel, OrderedSymbol, Set[STMDDEdgeLabel]]


  lazy val STMDDD: mddFactory.MDD = buildSTMDD()


  /**
    * compute the SMTDD
    *
    * @return
    */
  def buildSTMDD(): mddFactory.MDD = {
    //the interpreter compute the trace knowing that phi is true so compute only the dangerous traces
    systemTraceMDD(evaluateKnowing(phiBDD.phiBDD, phiBDD.varMap))
  }

  /**
    * Build STMDD from the interpretations of components
    *
    * @param systemInter a mapping from component instances to their interpretation
    * @return the STMDD
    */
  private def systemTraceMDD(systemInter: Map[CompInstanceIdent, ComponentInterpretation]): mddFactory.MDD = {
    import InterpreterImplicits._
    val comps = systemInter.collect({
      case (id, tr) if id.path.nonEmpty =>
        (OrderedCompInstIdent(id), toComponentTraces(tr))
    }).toList.sortBy(_._1)
    buildMDD(InstBoolBDDFactory.one(), comps).result
  }

  /**
    * Build a terminal symbol from a edge label
    */
  private val genTerminalSymbol: (STMDDEdgeLabel) => OrderedSymbol = a => Symbol(a.trace.valuationInOrder)

  /**
    * Recursive construction of the STMDD
    *
    * @param preCondition the system event valuations imposed by the parent node
    * @param comps        the list of interpretations
    * @return the STMDD
    */
  private def buildMDD(preCondition: BDD, comps: List[(OrderedCompInstIdent, ComponentTraces)]): TailRec[mddFactory.MDD] = {
    import InterpreterImplicits._

    comps match {
      //since only dangerous traces the terminal is always one or bottom (which is implicitly represented)
      case Nil =>
        done(mddFactory.getTerminal('one))

      case (inst, traces) :: tail =>

        //check if the variable exists
        if(mddFactory.varIsDefined(inst)){

          //build only the sons for the component traces that can occurs
          //(ie such that tr.valid \wedge precondition \neq false)
          val sons: Map[STMDDEdgeLabel, mddFactory.MDD] = traces.collect({
            case tr if !tr.valid.and(preCondition).isZero =>
              stmddVars(inst)(toOrderedCompTrace(tr)) -> tailcall(buildMDD(tr.valid.and(preCondition), tail)).result
          }).toMap

          //check if all possible component trace are considered in the interpretation
          //and target the same son, if yes then return this son, otherwise return the built node
          done(rmIfAllPossible(inst, sons).getOrElse(mddFactory.buildNodeDefaultTrash(inst, sons)))
        }else{
          throw UnknownVariable(inst)
        }
    }
  }

  /**
    * Simplification rule, check if all the possible component traces are considered in the interpretation
    * and that all of then target the same son
    *
    * @param comp the instance
    * @param sons the MDD sons
    * @return the new node if simplification possible None otherwise
    */
  private def rmIfAllPossible(comp: OrderedCompInstIdent, sons: Map[STMDDEdgeLabel, mddFactory.MDD]): Option[mddFactory.MDD] = {

    //inputs of the components
    val inputs = locateCompDef(comp.value).inputs

    //get the possible component traces
    val domain = mddFactory.getDomain(comp)

    //select in a trace the valuation of the inputs of the component
    val buildKey = (label: STMDDEdgeLabel) => {
      label.trace.value.filter(flow => inputs.map(in => in.ident).contains(flow._1))
    }

    //all sons are the same
    val allTheSame = sons.values.toSet.size == 1

    //group the traces by their valuations of the component inputs
    val groupedByInputs = sons.keys.groupBy(buildKey)

    //for a given value of the component inputs, it exists a component trace which is not belonging
    //to the traces labelled on the edge of the node (so not all possible traces for the component are considered)
    val aPossibleTraceNotReachable = groupedByInputs.exists(group => domain.filter(domTr => buildKey(domTr) == group._1).exists(domTr => !group._2.exists(grTrace => grTrace == domTr)))

    //if a possible trace is not considered or the targeted son is not always the same then no simplification
    //otherwise send the son
    if (aPossibleTraceNotReachable || !allTheSame) {
      None
    } else {
      Some(sons.head._2)
    }
  }
}

