package analyzers

import util.analyzers.analyzerTypes._
import com.microsoft.z3.Status._
import com.microsoft.z3.{Status, BoolExpr => Z3BoolExpr}
import translators.translatorsTypes._
import util.analyzers.MeasureTypes._
import util.analyzers.{ProbaMeasure, _}


import scala.collection.mutable
import scalaz.Memo

/**
  * Created by kdelmas on 25/09/17.
  */
trait SafetyTheory {
  self: Parsing with SMTTranslation with STMDDBuilder =>

  /**
    * type representing a clause as a set of predicate
    */
  type Clause = Seq[ModelPredicate]

  /**
    * type representing a model as a mapping of substitution variable to an integer
    * encoding the selected substitution
    */
  type Model = Seq[(SubstitutionVar, Int)]

  /**
    * type representing a mapping of component instance identifier to optional integer
    * encoding the selected substitution
    */
  type SelectionSeq = Seq[(PathIdentType, Option[Int])]

  /**
    * Memoized computation of the unreliability of an STMDD according to a selection
    */
  val computeUnr: ((mddFactory.MDD, SelectionSeq)) => ProbMeasure = Memo.mutableHashMapMemo {
    (pair: (mddFactory.MDD, SelectionSeq)) => {

      val (node, choice) = pair

      node match {

        case n: mddFactory.MDDNode =>

          //choice for the node
          val nodeChoiceIndex = getChoiceIndex(choice, n)

          //choice for sons
          val choiceTail = choice.drop(nodeChoiceIndex + 1)

          choice(nodeChoiceIndex)._2 match {

            case Some(bit) =>

              //if probability is zero then return zero
              if (n.forall(pair => pair._1.unr(bit) == 0)) {
                ProbaMeasure.zero
              } else {
                //compute node probability
                n.foldLeft(ProbaMeasure.zero)((acc, pair) => {
                  acc + computeUnr((pair._2, choiceTail)) * pair._1.unr(bit)
                })
              }

            case None =>

              //if probability is zero then return zero
              if (n.forall(pair => pair._1.maxUnr == 0)) {
                ProbaMeasure.zero
              } else {
                n.foldLeft(ProbaMeasure.zero)((acc, pair) => {
                  acc + computeUnr((pair._2, choiceTail)) * ProbaMeasure(pair._1.maxUnr, pair._1.minUnr, None)
                })
              }
          }

        //in this STMDD implementation only terminal one is kept
        case _:mddFactory.Terminal => ProbaMeasure.one
      }
    }
  }
  /**
    * Recursively check if the cardinality budget can be dispatched
    * for a given mdd node and substitution choice
    *
    * @return the status of the check, the effective cardinality computed and the substitutions informations used to
    *         decide if the cardinality budget is fulfilled
    */
  private val isSafeRec: ((mddFactory.MDD, SelectionSeq, CardMeasure)) => Status = Memo.mutableHashMapMemo {
    import util.analyzers.InfNumericImplicits._
    (t: (mddFactory.MDD, SelectionSeq, CardMeasure)) => {
      val (node, choice, remain) = t
      node match {
        case n: mddFactory.MDDNode =>

          //choice for the node
          val nodeChoiceIndex = getChoiceIndex(choice, n)

          //choice for sons
          val choiceTail = choice.drop(nodeChoiceIndex + 1)

          //try to fulfil objective  on sons
          n.foldLeft(SATISFIABLE)((acc, pair) => {

            //this node order is
            val card = choice(nodeChoiceIndex)._2 match {
              case Some(bit) =>

                //either the trace order for the selected substitution
                CardMeasure(pair._1.card(bit))

              case None =>

                //or the approximation when no substitution selection
                CardMeasure(pair._1.maxCard, pair._1.minCard, None)
            }

            acc match {

              //all previous sons met the objective
              case SATISFIABLE =>

                //the new objective
                val sonRemain = remain - card

                //if the objective is fulfilled on current son send SAT
                //otherwise perform analysis on son
                if (sonRemain <= 0)
                  SATISFIABLE
                else
                  isSafeRec((n.kids(pair._1), choiceTail, sonRemain))

              //at least on son did not meet the objective so the candidate is not acceptable
              case UNSATISFIABLE => UNSATISFIABLE

              //at least one son is not sure to meet the objective
              case UNKNOWN =>

                //the new objective
                val sonRemain = remain - card

                //if the objective is fulfilled on current son send UNKNOWN (since one other son is unknown)
                //otherwise perform analysis on son
                if (sonRemain <= 0)
                  UNKNOWN
                else if (isSafeRec((n.kids(pair._1), choiceTail, sonRemain)) == UNSATISFIABLE)
                  UNSATISFIABLE
                else
                  UNKNOWN
            }
          })

        //in this STMDD implementation only terminal one is kept
        case _: mddFactory.Terminal =>
          if (remain <= 0) {

            //if the objective is met then the path is acceptable
            SATISFIABLE
          } else if (remain > 0) {

            //if the objective is not met for a terminal then the failure condition can be triggered without
            //fulfilling the objective, thus is not acceptable
            UNSATISFIABLE
          } else {

            //otherwise we don't know
            Status.UNKNOWN
          }
      }
    }
  }

  /**
    * Decision procedure of the Safety theory
    *
    * @param cube       the conjonction of Safety predicate on which
    *                   the satisfiability modulo Safety must be decided
    * @param withXplain use quickXplain conflict clause minimisation
    * @return a model if SAT or a conflict clause if UNSAT
    */
  def checkSat(cube: Seq[SafetyPredicate], withXplain: Boolean = true): Either[Model, Clause] = {

    //extract substitution model from the choose predicate in the cube
    val model: Model = cube.collect { case Choose(sub, value) => (sub, value) }

    //check the safety predicate in the cube
    val result = cube.foldLeft((SATISFIABLE, Seq[ModelPredicate]()))((acc, predicate) => {

      acc match {

        //if at least one predicate in UNSAT then the conjunction is UNSAT and c is the conflict clause
        case (UNSATISFIABLE, c) => (UNSATISFIABLE, c)

        //if the previous predicate are not UNSAT
        case (_, c) =>

          predicate match {

            //if the predicate is a SafeOrder predicate
            case r@SafeOrder(stmdd, req) =>

              //check the requirement with current model
              checkOrder(stmdd, req, model) match {

                //if UNSAT then generate a conflict clause
                case UNSATISFIABLE =>
                  if (withXplain)
                    (UNSATISFIABLE, conflict(r, model))
                  else
                    (UNSATISFIABLE, cube.collect { case c: Choose => Neg(c) })

                //if unknown then at least one predicate is unknown
                case UNKNOWN => (UNKNOWN, c)

                //if the predicate is satifiable then transmit the current state
                case SATISFIABLE => acc
              }

            //if the predicate is a SafeUnr predicate same process as above
            case r@SafeR(stmdd, req) =>
              checkR(stmdd, req, model) match {
                case UNSATISFIABLE =>
                  if (withXplain)
                    (UNSATISFIABLE, conflict(r, model))
                  else
                    (UNSATISFIABLE, cube.collect { case c: Choose => Neg(c) })
                case UNKNOWN => (UNKNOWN, c)
                case SATISFIABLE => acc
              }
            case _ => acc
          }
      }
    })

    //if the conjunction is UNSAT then send the conflict clause
    //otherwise if the conjunction is SAT then send the model
    //since the theory is used in lazy mode the conjunction cannot be
    //UNKNOWN (safety theory is decidable)
    if (result._1 == UNSATISFIABLE)
      Right(result._2)
    else if (result._1 == SATISFIABLE)
      Left(model)
    else
      throw new Exception("unknown impossible in lazy use")
  }

  /**
    * Conflit clause generation methode
    *
    * @param r     the Safety predicate to analyse (which must be unsat)
    * @param model the current model
    * @return a conflict clause
    */
  private def conflict(r: SafetyReqPredicate, model: Model): Clause = {

    //initialise the result to the full model
    val result = mutable.ListBuffer(model: _*)
    for (kv <- result) {

      //remove recursively the elements of the model if not necessary to prove that predicate is UNSAT
      r match {
        case SafeOrder(stmdd: mddFactory.MDD, req: Int) =>
          if (checkOrder(stmdd, req, result - kv) == UNSATISFIABLE)
            result -= kv

        case SafeR(stmdd: mddFactory.MDD, req: Real) =>
          if (checkR(stmdd, req, result - kv) == UNSATISFIABLE)
            result -= kv
      }
    }

    //transform the model to a clause (disjunction of choose predicates)
    result.map(p => Neg(Choose(p._1, p._2)))
  }

  /**
    * Check if the reliability requirement is met by a STMDD for a given model
    *
    * @param stmdd the STMDD on which the reliability is computed
    * @param req   the reliability requirement
    * @param model the model of the substitution variables
    * @return SAT if the STMDD satisfies requirement, UNSAT if not and UNKNOWN if the
    *         approximation of the reliability is not accurate enough to decide
    */
  private def checkR(stmdd: mddFactory.MDD, req: Real, model: Model): Status = {
    val result = computeUnr((stmdd, modelToSelectionSeq(stmdd, model)))
    if (result > 1 - req)
      UNSATISFIABLE
    else if (result <= 1 - req)
      SATISFIABLE
    else
      UNKNOWN
  }


  /**
    * Check if the order requirement is met by a STMDD for a given model
    *
    * @param stmdd the STMDD on which the reliability is computed
    * @param req   the order requirement
    * @param model the model of the substitution variables
    * @return SAT if the STMDD satisfies requirement, UNSAT if not and UNKNOWN if the
    *         approximation of the reliability is not accurate enough to decide
    */
  private def checkOrder(stmdd: mddFactory.MDD, req: Int, model: Model): Status = {
    import util.analyzers.InfNumericImplicits._
    isSafeRec((stmdd, modelToSelectionSeq(stmdd, model), CardMeasure(req)))
  }

  /**
    * Transform a model of the substitution variables to a list of selection index in the
    * local indicators table of the STMDD edges
    *
    * @param stmdd the STMDD concerned by the model
    * @param m     the model of the substitution variables
    * @return the selection sequence
    */
  def modelToSelectionSeq(stmdd: mddFactory.MDD, m: Model): SelectionSeq = {
    val modelMap = m.map(p => p._1.inst -> p._2).toMap
    stmdd.variables.toSeq.map(v => v.value.path -> modelMap.get(v.value))
  }

  def getChoiceIndex(choice: SelectionSeq, of: mddFactory.MDDNode): Int = {
    choice.indexWhere(p => p._1 == of.param.value.path)
  }

  /**
    * Base trait for safety theory predicates
    */
  sealed trait SafetyPredicate

  /**
    * Base trait for requirement predicates
    */
  sealed trait SafetyReqPredicate extends SafetyPredicate

  /**
    * Base trait for model predicates
    */
  sealed trait ModelPredicate extends SafetyPredicate {
    def toZ3: Z3BoolExpr
  }

  /**
    * Case class encoding the reliability requirement predicate
    *
    * @param stmdd the STMDD representing the system
    * @param req   the reliability requirement
    */
  case class SafeR(stmdd: mddFactory.MDD, req: Real) extends SafetyReqPredicate

  /**
    * Case class encoding the order requirement predicate
    *
    * @param stmdd the STMDD representing the system
    * @param req   the order requirement
    */
  case class SafeOrder(stmdd: mddFactory.MDD, req: Int) extends SafetyReqPredicate

  /**
    * Case class encoding the choice over a substitution varaible
    *
    * @param subVar the substitution variable
    * @param value  the value of the variable
    */
  case class Choose(subVar: SubstitutionVar, value: Int) extends ModelPredicate {
    def toZ3: Z3BoolExpr = {
      _SMTContext.mkEq(subVar.z3Var, _SMTContext.getValueAtPos(subVar, value).get)
    }
  }

  /**
    * Util case class for negation of Safety predicates
    *
    * @param p the safety predicate to negate
    */
  case class Neg(p: Choose) extends ModelPredicate {
    def toZ3: Z3BoolExpr = _SMTContext.mkNot(p.toZ3)
  }

}




