package analyzers

import ast.dsl.{CompAlternative, CompInstanceIdent, CompSubstitution, ComponentDeclaration, InstBoolIdent}
import interpreter.TraceInterpreter
import net.sf.javabdd.BDD
import util.analyzers.InfNumericTypes.InfInt
import util.analyzers._
import util.analyzers.analyzerTypes.{ProbabilityMapType, Real}

/**
  * Analysis of the substitutions given in the design space of the DSE problem by using the interpreter
  */
trait SubstitutionAnalyzerSim extends SubstitutionAnalyzer with BDDAnalyzer {
  self: Parsing with SMTTranslation with TraceInterpreter =>


  /**
    * Associate to each component instance of the system and each of its trace,
    * the local safety indicator arrays
    */
  lazy val substitutions: Map[OrderedCompInstIdent, Map[OrderedCompTrace, (Array[InfInt], Array[Real])]] = {

    //check that every primary component has at least an initial alternative
    require({
      rootCompInstsInOrder.init.forall(inst => conf.designSpace.exists(sub => sub.instance == inst))
    }, {
      val missing = rootCompInstsInOrder.init.collect(
        { case inst if !conf.designSpace.exists(sub => sub.instance == inst) => inst })
      s"missing substitution for primary components ${missing.mkString(",")}"
    })

    //analyse the alternatives with the interpreter
    conf.designSpace.map(sub => OrderedCompInstIdent(sub.instance) -> analyseSub(sub)).toMap
  }

  /**
    * Build a mapping from alternative failure event  BDD to their probability
    *
    * @param alt the alternative to analyse
    * @return the mapping
    */
  private def eventsToBDDMap(alt: CompAlternative): Map[Int, Real] = {

    //exposition time
    val time: Real = conf.duration.map(time => Real(time.repr)).getOrElse(1)

    //exponential distribution
    val formula = (lambda: Real, time: Real) => Real(1 - math.exp(-(lambda * time).toDouble))

    alt.lambdaValues.map(kv => InstBoolIdent(kv._1, CompInstanceIdent(alt.compId.baseId :: Nil)) -> formula(Real(kv._2.repr), time)).
      map(kv => InstBoolBDDFactory.produceVar(kv._1).`var`() -> kv._2).toMap
  }

  /**
    * Compute the order and probability arrays for a given substitution
    *
    * @param s an AST node ofr component substitution
    * @return the order and probability arrays
    */
  def analyseSub(s: CompSubstitution): Map[OrderedCompTrace, (Array[InfInt], Array[Real])] = {

    //locate the component instance declaration
    val compDef: ComponentDeclaration = locateCompDef(s.init.compId)

    //compute the possible initial component traces with the interpreter
    val iniTraces = makePossibleTraces(compDef)

    //compute the initial component traces order and probability arrays
    val iniEvts = eventsToBDDMap(s.init)
    val iniCharacTr = iniTraces.mapValues(bdd => (computeTraceCard(bdd), computeTraceUnr((bdd, iniEvts))))

    //for all other alternative ensures that traces are included in initial one and
    //compute arrays
    val newTraces = s.alternatives.map(alt => {
      val altTraces = makePossibleTraces(locateCompDef(alt.compId))
      val altEvt = eventsToBDDMap(alt)
      assert(altTraces.keySet.subsetOf(iniTraces.keySet),
        throw IllegalSubstitution(s.instance, alt.compId, altTraces.keySet.diff(iniTraces.keySet).mkString(",")))
      altTraces.mapValues(bdd => (computeTraceCard(bdd), computeTraceUnr((bdd, altEvt))))
    })

    iniTraces.keys.map(tr => tr -> {
      val characterisation = (iniCharacTr +: newTraces).map(alt => {
        for (v <- alt.get(tr)) yield v
      }.getOrElse((InfNumeric[Int](Right(POSITIVE_INFINITY)), Real(0)))).unzip
      (characterisation._1.toArray, characterisation._2.toArray)
    }).toMap
  }

  /**
    * Memoized computation of the order of a BDD
    */
  val computeTraceCard: (BDD) => InfInt = MyMemo.mutableOpenHashMapMemo {
    (bdd: BDD) => computeCardBDD(bdd)
  }

  /**
    * Memoized computation of the probability of a BDD
    */
  val computeTraceUnr: ((BDD, ProbabilityMapType)) => Real = MyMemo.mutableOpenHashMapMemo {
    (p: (BDD, ProbabilityMapType)) => computeProbaBDD(p._1, p._2)
  }
}

