package analyzers

import java.io.{FileOutputStream, PrintStream}

import ast.dsl.{CompInstanceIdent, InstBoolIdent}
import util.analyzers.BDDAnalyzer
import util.analyzers.analyzerTypes.Real

/**
  * Trait for failure rate exportation
  */
trait SymbolicFailureRateComputer extends SafetyIndicatorComputer with AcceptableReliabilityAnalyzer with BDDAnalyzer {
  self: Parsing with SMTTranslation =>

  /**
    * Compute and print symbolic reliability and failure rate formula in a Python script
    *
    * @param fileName output python file
    */
  def printSymbIniRBDD(fileName: String): Unit = {
    val fos = new FileOutputStream(fileName)
    val ps = new PrintStream(fos)
    val time: Real = conf.duration.map(time => Real(time.repr)).getOrElse(1)

    val formula = (lambda: String) => s"(1-exp(-(lambda_$lambda * t)))"
    val iniEvtPr: Map[InstBoolIdent, String] = conf.designSpace.flatMap(sub => sub.init.lambdaValues.map(e => (InstBoolIdent(e._1, CompInstanceIdent(sub.instance.path)), e._2.repr))).toMap
    val variables: Map[String, String] = iniEvtPr.map(p => s"lambda_${p._1.toString().replace(".", "_").replace("|", "")}" -> p._2) + ("t" -> time.toString)

    ps.println("import sympy")
    ps.println("from sympy import exp")
    ps.println("from sympy import Symbol")
    ps.println("from sympy.plotting import plot")
    ps.println("def main():")
    val r = "r"
    val r_t = "r_t"
    val l_t = "lambda_t"
    //model values as comments
    ps.println("\t#In the model the values were:")
    variables.foreach(p => ps.println(s"\t#${p._1} = ${p._2}"))
    //declares symbols
    variables.foreach(p => ps.println(s"\t${p._1} = Symbol('${p._1}')"))
    //map from BDD var identifier to actual names
    val unrMap = iniEvtPr.map(e => (phiCarac.varMap(e._1).`var`(), formula(e._1.toString().replace(".", "_").replace("|", ""))))
    //print reliability formula
    val unr = printSymbProbaBDD(phiCarac.phiBDD, unrMap, ps)
    ps.println(s"\t$r= (1-$unr)")
    //print reliability only depending on t
    ps.println(s"\t$r_t=r.subs(${variables.filter(p => p._1 != "t").map(p => s"(${p._1}, ${p._2})").mkString("[", ",", "]")})")
    //print failure intensity
    ps.println(s"\t$l_t=(-r_t.diff('t')/r_t)")
    //figure of r_t and lambda_t
    ps.println(s"#\tplot($r_t,(t,0,$time),xlabel='time(s)',ylabel='reliability', title='system reliability')")
    ps.println(s"#\tplot($l_t,(t,0,$time),xlabel='time(s)',ylabel='failure intensity', title='system failure intensity')")
    //run main
    ps.println()
    ps.println("if __name__ == \"__main__\":")
    ps.println("\tmain()")
    ps.close()
    fos.close()
  }
}

