package analyzers

import java.io.File

import ast.dsl.{BoolIdent, CompInstanceIdent, InstBoolIdent, StdDSLSymbol}
import util.analyzers.{ExecutionTimes, Profiler, XFTAInterface}
import util.io.FileManager

import scala.io.Source
import scala.sys.process.Process
import scala.xml.dtd.DocType

/**
  * Cutsets computation with XFTA method
  */
trait XFTACutEnumerator extends CutEnumerator {
  self: Parsing with SMTTranslation with OpenPSAExporter =>

  /**
    * Generate the fault tree encoding of the structure function in OpenPSA language and the computation script
    *
    * @param max        bound on cutset cardinality
    * @param ftFile     file receiving fault tree
    * @param scriptFile file receiving computation script
    * @param resultFile file receiving results
    */
  private def createXFTAFiles(max: Option[Int], ftFile: File, scriptFile: File, resultFile: File): Unit = {
    val script =
      <xfta>
        <load>
          <model input={ftFile.getAbsolutePath}/>
        </load>
        <build>
          {if (max.nonEmpty)
            <minimal-cutsets top-event="top" handle="MCS" maximum-order={max.get.toString}/>
        else
            <minimal-cutsets top-event="top" handle="MCS"/>}
        </build>
        <set>
          <option name="print-minimal-cutset-rank" value="off"/>
          <option name="print-minimal-cutset-order" value="off"/>
          <option name="print-minimal-cutset-probability" value="off"/>
          <option name="print-minimal-cutset-contribution" value="off"/>
        </set>
        <print>
          <minimal-cutsets top-event="top" handle="MCS" output={resultFile.getAbsolutePath}/>
        </print>
      </xfta>
    xml.XML.save(scriptFile.getAbsolutePath, script, xmlDecl = true, doctype = DocType("xfta"))
    exportToOpenPSA(ftFile)
  }

  /**
    * Compute cutsets with XFTA by executing script file
    *
    * @param scriptFile script file
    */
  private def computeWithXFTA(scriptFile: File): Unit = {
    val toolPath = XFTAInterface.toolPath
    XFTAInterface.startProcess(Process(s"$toolPath ${scriptFile.getAbsolutePath}")).exitValue()
  }

  /**
    * Extrat cutsets from result file
    *
    * @param resultFile result file
    * @return cutsets
    */
  private def backTranslateXFTA(resultFile: File): Iterable[Set[InstBoolIdent]] = {
    val lines = Source.fromFile(resultFile).getLines().toList
    lines.map(s => {
      s.split("\t").map(e => {
        val splitted = e.split('.')
        val compInstanceIdent = CompInstanceIdent(splitted.init.map(x => StdDSLSymbol(Symbol(x))).toList)
        val boolIdent = BoolIdent(StdDSLSymbol(Symbol(splitted.last)))
        InstBoolIdent(boolIdent, compInstanceIdent)
      }).toSet
    })
  }

  val profilerCutEnumeration = new Profiler[Option[Int], Iterable[Set[InstBoolIdent]]] {
    def runWithProfile(i: Option[Int])(iterations: Int = 1): (Iterable[Set[InstBoolIdent]], ExecutionTimes) = {
      val ftFile = FileManager.addFileToTemp("tmp.xml")
      val scriptFile = FileManager.addFileToTemp("tmpScript.xml")
      val resultFile = FileManager.addFileToTemp("result.txt")
      val (_, preTime) = time({
        createXFTAFiles(i, ftFile, scriptFile, resultFile)
      }, iterations)
      val (_, pTime) = time({
        computeWithXFTA(scriptFile)
      }, iterations)
      val (result, postTime) = time({
        backTranslateXFTA(resultFile)
      }, iterations)
      (result, ExecutionTimes(preTime, pTime, postTime))
    }
  }

  /**
    * Compute cutsets with XFTA
    *
    * @param max maximal cutset size if defined, no bound otherwise
    * @return the set of minimal cutsets
    */
  def enumerateCutUntil(max: Option[Int]): Iterable[Set[InstBoolIdent]] = {
    val ftFile = FileManager.addFileToTemp("tmp.xml")
    val scriptFile = FileManager.addFileToTemp("tmpScript.xml")
    val resultFile = FileManager.addFileToTemp("result.txt")
    createXFTAFiles(max, ftFile, scriptFile, resultFile)
    computeWithXFTA(scriptFile)
    val res = backTranslateXFTA(resultFile)
    res
  }
}
