package analyzers

import java.io.File

import ast.dsl.{BoolIdent, CompInstanceIdent, InstBoolIdent, StdDSLSymbol}
import util.analyzers.{ExecutionTimes, Profiler, PythonInterface, XSAPInterface}
import util.io.FileManager

import scala.io.Source
import scala.sys.process.Process

/**
  * Cutset computation with XSAP tool
  */
trait XSAPCutEnumerator extends CutEnumerator {
  self: Parsing with SMTTranslation with SMVExporter =>

  /**
    * Generate SMV model file and fault injection file
    *
    * @param model SMV model of the system
    * @param fei   fault injection model of the system
    * @return the directory containing the files
    */
  private def createSMVFiles(model: File, fei: File): File = {
    val dir = FileManager.addFileToTemp("out")
    exportToSMV(model)
    val scriptPath = XSAPInterface.toolPath + "/extend_model.py"
    val command = Seq(
      PythonInterface.toolPath.toString(),
      scriptPath.toString,
      model.getAbsolutePath,
      fei.getAbsolutePath,
      s"-d=${dir.getAbsolutePath}"
    )
    XSAPInterface.startProcess(Process(command)).exitValue()
    dir
  }

  /**
    * Compute cutsets with XSAP from model files in outDir
    *
    * @param max    bound on cutset cardinality
    * @param outDir directory containing model files
    * @return the result file
    */
  private def computeWithXSAP(max: Option[Int], outDir: File): File = {
    (for (model <- FileManager.recursiveLocateFirstFile(outDir, (f: File) => {
      f.getName.matches(".*\\.smv")
    })
    ) yield {
      val scriptPath = XSAPInterface.toolPath + "/compute_ft.py"
      val commandPr = Seq(
        PythonInterface.toolPath.toString(),
        scriptPath.toString,
        s"--smv-file=${model.getAbsolutePath}",
        "-p=FC",
        s"-d=${outDir.getAbsolutePath}"
      )
      val command = (for (m <- max) yield commandPr :+ s"-N $m") getOrElse commandPr
      XSAPInterface.startProcess(Process(command)).exitValue()
      new File(outDir, "xsap_compute_ft.out")
    }) getOrElse (throw new Exception(s"no smv files in ${outDir.getAbsolutePath}"))
  }

  /**
    * Extract cutsets from XSAP result file
    *
    * @param result result file
    * @return cutsets
    */
  private def backTranslateXSAP(result: File): Iterable[Set[InstBoolIdent]] = {
    val validNames = phi.events.keySet.flatMap(id => id.baseId.baseId.name +: id.instId.path.flatMap(s => {
      val p = s.name.replace("|", "")
      p.split("[:\\.]")
    }))
    val zero = List[Set[InstBoolIdent]]()
    var inMinialSec = false
    Source.fromFile(result).getLines().foldLeft(zero)((acc, line) => {
      if (!inMinialSec) {
        inMinialSec = line.contains("MINIMAL CUT SET")
        acc

      } else if (line.contains("STATISTICS")) {
        inMinialSec = false
        acc
      } else if (line.matches("[0-9]+\\).*")) {
        acc :+ Set()
      } else if (line.contains("_hist")) {
        val l = line.split("_").filter(s => validNames.contains(s))
        val id = InstBoolIdent(BoolIdent(StdDSLSymbol(Symbol(l.last))), CompInstanceIdent(l.init.map(s => StdDSLSymbol(Symbol(s))).toList))
        acc match {
          case Nil => Set(id) :: Nil
          case h :: Nil => (h + id) :: Nil
          case _ => acc.init :+ (acc.last + id)
        }
      } else {
        acc
      }
    })
  }

  /**
    * Compute minimal cutsets with XSAP tool
    *
    * @param max maximal cutset size if defined, no bound otherwise
    * @return the set of minimal cutsets
    */
  def enumerateCutUntil(max: Option[Int]): Iterable[Set[InstBoolIdent]] = {
    val modelFile = FileManager.addFileToTemp("tmpModel.smv")
    val fei = FileManager.addFileToTemp("tmpModel.fei")
    val outDir = createSMVFiles(modelFile, fei)
    val res = computeWithXSAP(max, outDir)
    modelFile.delete()
    fei.delete()
    backTranslateXSAP(res)
  }

  val profilerCutEnumeration = new Profiler[Option[Int], Iterable[Set[InstBoolIdent]]] {
    def runWithProfile(i: Option[Int])(iterations: Int = 1): (Iterable[Set[InstBoolIdent]], ExecutionTimes) = {
      val modelFile = FileManager.addFileToTemp("tmpModel.smv")
      val fei = FileManager.addFileToTemp("tmpModel.fei")
      val (r1, preTime) = time({
        createSMVFiles(modelFile, fei)
      }, iterations)
      val (r2, pTime) = time({
        computeWithXSAP(i, r1)
      }, iterations)
      val (r3, posTime) = time({
        backTranslateXSAP(r2)
      }, iterations)
      (r3, ExecutionTimes(preTime, pTime, posTime))
    }
  }
}
