/** $Id: DSLAST.scala 1433 2017-10-19 14:06:15Z kdelmas $
  *
  * @author $Author: kdelmas $
  * */

package ast.dsl


/* TODO move typing rules to their own package. */

import namespace.dsl._

/** Locates an ASTNode in a source file. */
sealed trait SourceLocation

/** Location of an ASTNode in some source file
  *
  * @param filename the source filename 
  * @param line     line number in the file
  * @param column   column number in the line
  * */
case class InFile(filename: String, line: Int, column: Int, posInFile: Int, size: Int) extends SourceLocation {
  override def toString: String = s"$filename:$line($column)"
}

/** Unknown source location. */
case object UnknownLocation extends SourceLocation {
  override def toString: String = "location unknown"
}

/** Adds source location information to an ASTNode. */
trait SourceMapped {
  this: ASTNode =>

  /** The source location of the source-mapped entity. */
  private[this] var _location: SourceLocation = UnknownLocation

  private[this] var _otherOpLocation: List[InFile] = Nil

  /** Sets the location. */
  def location_(loc: SourceLocation): Unit = {
    _location = loc
  }

  def addAltLoc(loc: InFile): Unit = {
    _otherOpLocation ::= loc
  }

  def otherOpLocation: List[InFile] = _otherOpLocation

  /** Gets the location. */
  def location: SourceLocation = _location
}

/** Abstract type for all AST nodes. */
sealed trait ASTNode extends SourceMapped

/** Abstract type for identifiers-like nodes. */
sealed trait IdentLike extends ASTNode


/** Abstract trait for symbols handled in the DSL. */
sealed trait DSLSymbol extends ASTNode {
  def symbol: Symbol

  def isPiped: Boolean

  def name: String = symbol.name

  override def toString: String  = name

  def unPiped: String

  def ++(that: DSLSymbol): DSLSymbol = {
    val newSym = Symbol(this.unPiped ++ that.unPiped)
    if (this.isPiped || that.isPiped) {
      PipedDSLSymbol(newSym)
    }
    else {
      StdDSLSymbol(newSym)
    }
  }
}


/** Standard symbols mathching the IDENT rule of the parser. */
case class StdDSLSymbol(symbol: Symbol) extends DSLSymbol {
  def isPiped: Boolean = false

  def unPiped: String = symbol.name
}

/** Piped symbols matching the PIPED_IDENT rule of the parser. */
case class PipedDSLSymbol(symbol: Symbol) extends DSLSymbol {
  def isPiped: Boolean = true

  def unPiped: String = symbol.name.slice(1, symbol.name.length - 1)
}

/** Piped symbols matching the POINTED_IDENT rule of the parser. */
case class PointedDSLSymbol(symbol: Symbol) extends DSLSymbol {
  def isPiped: Boolean = true

  def unPiped: String = symbol.name

  def toStringList: List[String] = symbol.name.split('.').toList
}


/* ============= PACKAGE Identifiers =================== */
case class PackageIdent(path: List[DSLSymbol]) extends ASTNode with IdentLike {
  override def toString: String  = path.view.map(_.name).mkString("::")

  def addPrefix(s: String): String = path.size match {
    case 0 => s
    case _ => this.toString + "::" + s
  }
}

object NilPackageIdent {
  def apply() = PackageIdent(Nil)
}

/* ============= TYPE EXPRESSIONS ================= */

/** Type expressions. */
sealed trait TypeExpr extends ASTNode

sealed trait ScalarTypeExpr extends TypeExpr

sealed trait CompositeTypeExpr extends TypeExpr

/** Type identifiers. 
  * P1 :: ... :: Pn ::  T
  * |  pkgId         | baseId
  */
case class TypeIdent(baseId: DSLSymbol, pkgId: PackageIdent = NilPackageIdent()) extends ScalarTypeExpr with IdentLike {
  override def toString: String = pkgId.addPrefix(baseId.name)
}

/** The type of boolean expressions. */
case object BoolType extends ScalarTypeExpr with IdentLike {
  override def toString: String = "bool"
}

case class TupleType(kids: List[ScalarTypeExpr]) extends CompositeTypeExpr {
  override def toString: String = kids.map(_.toString).mkString("(", ", ", ")")
}

/* ============ VALUE-CARRYING EXPRESSIONS ========== */

/** Type for value-carrying expressions. */
sealed trait Expr extends ASTNode {
  def getType(ns: Namespace): ScalarTypeExpr
}

/* ============= BOOLEAN EXPRESSIONS ================ */

/** Type for Boolean-valued expressions. */
sealed trait BoolExpr extends Expr

/** Basic boolean identifier (with or withtout special characters). */
case class BoolIdent(baseId: DSLSymbol) extends BoolExpr with IdentLike {
  override def toString: String = baseId.name

  def getType(ns: Namespace): ScalarTypeExpr = BoolType
}

/** Component instance prefixed boolean identifier */
case class InstBoolIdent(baseId: BoolIdent, instId: CompInstanceIdent) {
  override def toString: String = (instId.path.map(_.name).mkString(".") + "." + baseId.baseId.name).replace("|", "")
}

/** Typing rule for BoolExpr with K kids which must have the same type. */
trait BoolExprKidsSameTypeRule[K <: Expr] {
  this: BoolExpr =>
  val kid1: K
  val kid2: K

  override def getType(ns: Namespace): ScalarTypeExpr = {
    val t1 = kid1.getType(ns)
    val t2 = kid2.getType(ns)
    if (t1 == t2) {
      BoolType
    } else {
      throw WrongTypeExpr(expr = kid2, found = t2, expected = t1)
    }
  }
}

/** Flow equality. */
case class FlowEq(kid1: FlowExpr, kid2: FlowExpr) extends BoolExpr with BoolExprKidsSameTypeRule[FlowExpr]

/** Flow inequality. */
case class FlowNeq(kid1: FlowExpr, kid2: FlowExpr) extends BoolExpr with BoolExprKidsSameTypeRule[FlowExpr]

/** Flow containment test. */
case class FlowContains(kid1: FlowExpr, kid2: FlowExpr) extends BoolExpr with BoolExprKidsSameTypeRule[FlowExpr]

/** Logical equivalence. */
case class Iff(kid1: BoolExpr, kid2: BoolExpr) extends BoolExpr with BoolExprKidsSameTypeRule[BoolExpr]

/** Logical implication. */
case class Implies(kid1: BoolExpr, kid2: BoolExpr) extends BoolExpr with BoolExprKidsSameTypeRule[BoolExpr]

/** Logical negation. */
case class Not(kid1: BoolExpr) extends BoolExpr {
  override def getType(ns: Namespace): ScalarTypeExpr = kid1.getType(ns)
}

/** Logical conjunction. */
case class And(kid1: BoolExpr, kid2: BoolExpr) extends BoolExpr with BoolExprKidsSameTypeRule[BoolExpr]

/** Logical disjuction. */
case class Or(kid1: BoolExpr, kid2: BoolExpr) extends BoolExpr with BoolExprKidsSameTypeRule[BoolExpr]

/* ============= CARDINALITY EXPRESSIONS ================= */

/** Type for relational operators used in population counts. */
sealed trait RelOp

/** Equal operator. */
case object Eq extends RelOp

/** Greater-than-or-equal operator. */
case object Ge extends RelOp

/** Greater-than operator. */
case object Gt extends RelOp

/** Less-than-or-equal operator. */
case object Le extends RelOp

/** Less-than operator. */
case object Lt extends RelOp

/** Population count: predicates equvalent to """ \Sigma_{kid \in kids }{kid} \langle relop \rangle cst""". */
case class PopCount(kids: List[BoolExpr], relop: RelOp, cst: DecIntLit) extends BoolExpr {
  def getType(ns: Namespace): BoolType.type = {
    kids.foreach(_.getType(ns))
    BoolType
  }
}

/*============= Integer-valued expressions ========== */

/** Base type for integer-valued expressions. */
sealed trait IntExpr extends ASTNode

/** Integer literal in decimal notation. */
case class DecIntLit(repr: String) extends IntExpr {
  override def toString: String = repr
}

/*============= Real-valued expressions ========== */

/** Base type for integer-valued expressions. */
sealed trait RealExpr extends ASTNode

/** Integer literal in decimal notation. */
case class DecRealLit(repr: String) extends RealExpr {
  override def toString: String = repr
}

/* ============= FLOW EXPRESSIONS ================= */

/** Flow-valued expressions. */
sealed trait FlowExpr extends Expr

trait FlowIdentTypeRule {
  this: FlowIdent =>
  def getType(ns: Namespace): ScalarTypeExpr = {
    ns.locate(this) match {
      case Left(typeDecl) => typeDecl.absoluteIdent
      case Right(flowDecl) => ns.locate(flowDecl.typeId).absoluteIdent
    }
  }
}

/** Type for identifier-like flow-valued expressions. */
sealed trait FlowIdent extends FlowExpr with IdentLike with FlowIdentTypeRule


/** Basic flow identifier (without special characters). */
case class BasicFlowIdent(baseId: DSLSymbol) extends FlowIdent {
  override def toString: String = baseId.name
}

/** Flow identifier wrapped in a instance prefix. */
case class InstFlowIdent(baseId: DSLSymbol, instId: CompInstanceIdent) extends FlowIdent {
  override def toString: String = (instId.path.map(_.name).mkString(".") + "." + baseId.name).replace("|", "")
}

sealed trait FlowConst extends FlowIdent

/** The empty set litteral of a given type.
  * P1 :: ... :: Pn :: T   !    empty
  * |       typeId         |
  */
case class TypeValueFlowEmpty(typeId: TypeIdent) extends FlowConst {
  override def toString: String = s"$typeId!empty"
}

/** Flow identifier corresponding to a type constructor
  * P1 :: ... :: Pn :: T   !     v
  * |       typeId         |   | valueId |
  *
  */
case class TypeValueFlowIdent(typeId: TypeIdent, valueId: DSLSymbol) extends FlowConst {
  override def toString: String = s"$typeId!${valueId.name}"
}

trait FlowExprKidsSameTypeRule {
  this: FlowExpr =>
  val kid1: FlowExpr
  val kid2: FlowExpr

  def getType(ns: Namespace): ScalarTypeExpr = {
    val t1 = kid1.getType(ns)
    val t2 = kid2.getType(ns)
    if (t1 == t2) {
      t1
    } else {
      throw WrongTypeExpr(expr = kid2, expected = t1, found = t2)
    }
  }
}

/** Flow union operator. */
case class FlowUnion(kid1: FlowExpr, kid2: FlowExpr) extends FlowExpr with FlowExprKidsSameTypeRule

/** Flow intersection operator. */
case class FlowInter(kid1: FlowExpr, kid2: FlowExpr) extends FlowExpr with FlowExprKidsSameTypeRule

/** Flow intersection operator. */
case class FlowIte(kid1: BoolExpr, kid2: FlowExpr, kid3: FlowExpr) extends FlowExpr {
  def getType(ns: Namespace): ScalarTypeExpr = {
    val t2 = kid2.getType(ns)
    val t3 = kid3.getType(ns)
    if (t2 == t3) {
      t2
    } else {
      throw WrongTypeExpr(expr = kid3, expected = t2, found = t3)
    }
  }
}

/* ============= COMPONENT-RELATED EXPRESSIONS ================= */

/** Base type for component-based expressions. */
sealed trait CompExpr extends ASTNode

/** Component identifier, located in a package. */
case class CompIdent(baseId: DSLSymbol, pkgId: PackageIdent = NilPackageIdent()) extends IdentLike {
  override def toString: String = pkgId.addPrefix(baseId.name)
}

/** Component Instanciation. */
case class CompCall(compId: CompIdent, instId: DSLSymbol, args: List[FlowExpr]) extends CompExpr {
  override def toString: String = s"$compId @ ${instId.name} ( ... )"
}

/** Component instance identifier, representing a path in the call tree relative to some root instance. */
case class CompInstanceIdent(path: List[DSLSymbol]) extends IdentLike {
  override def toString: String = path.view.map(_.name).mkString(".")

  def addSuffix(s: String): String = path.size match {
    case 0 => s
    case _ => this.toString + "." + s
  }
}

/** Represents the information that component instance can be replaced by alternative components. */
case class CompSubstitution(instance: CompInstanceIdent, alternatives: List[CompAlternative], init: CompAlternative) extends ASTNode {
  override def toString: String = s"$instance -> ${alternatives.view.map(_.toString).mkString(", ")}"
}

case class CompAlternative(compId: CompIdent, lambdaValues: List[(BoolIdent, DecRealLit)]) extends ASTNode {
  override def toString: String = {
    val xs = lambdaValues.view.map(x => s"${x._1} = ${x._2}").mkString(", ")
    s"$compId( $xs )"
  }
}

/* ============= DECLARATIONS ================= */
sealed trait VisibilityStatus

case object PublicStatus extends VisibilityStatus

case object PrivateStatus extends VisibilityStatus

/** Type for all declarations allowed in a DSL model. */
sealed trait Declaration extends ASTNode

/** Type for all program-level declarations. */
sealed trait ProgramDeclaration extends Declaration with ParentField[PackageDeclaration]

/** Includes a file. */
case class IncludeDeclaration(filePath: String) extends ProgramDeclaration {
  override def toString: String = s"#include $filePath ;"
}

/** Opens a package. */
case class OpenDeclaration(pkgId: PackageIdent) extends ProgramDeclaration with TargetDeclField[PackageDeclaration] {
  override def toString: String = s"open $pkgId ;"
}

/** Creates the root package from a list of declarations. */
object RootPackage {
  def apply(declarations: List[ProgramDeclaration]): PackageDeclaration = PackageDeclaration(PackageIdent(StdDSLSymbol('root) :: Nil), declarations)
}

/** Declares a package. */
case class PackageDeclaration(
                               ident: PackageIdent,
                               declarations: List[ProgramDeclaration],
                               visibility: VisibilityStatus = PublicStatus
                             ) extends ProgramDeclaration with NamespaceField[PackageNamespace] {
  assert(ident.path.size == 1)

  override def toString: String = s"package $ident ... end;"

  /** Returns the inverse absolute path of the package from itself to the root package as a list of symbols, empty if the package is the root package.
    * */
  def absolutePathInv: List[DSLSymbol] = parentOpt match {
    case Some(p) => ident.path.head :: p.absolutePathInv
    case None => Nil
  }

  /** Returns the absolute path of the package from the root package as a list of symbols, empty if the package is the root package.
    * */
  def absolutePath: List[DSLSymbol] = absolutePathInv.reverse

  // add parent links
  declarations foreach { d => d.parent_(this) }

}

/** Declares a type. */
case class TypeDeclaration(ident: TypeIdent, values: List[DSLSymbol]) extends ProgramDeclaration {

  override def toString: String = s"$ident := { ${values.view.map(_.name).mkString(", ")} } ;"

  def namespace: Namespace = parent.namespace

  def absolutePath: List[DSLSymbol] = parent.absolutePath

  def absolutePathInv: List[DSLSymbol] = parent.absolutePathInv

  def absoluteIdent: TypeIdent = parent.absolutePath match {
    case Nil => ident
    case path =>
      val res = TypeIdent(ident.baseId, PackageIdent(path))
      res.location_(this.ident.location)
      res
  }

}

/** Component body. */
case class ComponentBody(events: List[BoolIdent], locals: List[FlowDeclaration], definitions: List[FlowDefinition]) extends ASTNode with
  ParentField[ComponentDeclaration] with NamespaceField[ComponentBodyNamespace] {
  override def toString: String = "... body ..."

  // sets backlinks to parent
  locals foreach {
    _.parent_(this)
  }
  definitions foreach {
    _.parent_(this)
  }

}

/** Declares a component. */
case class ComponentDeclaration(
                                 ident: CompIdent,
                                 inputs: List[FlowDeclaration],
                                 outputs: List[FlowDeclaration],
                                 body: ComponentBody,
                                 isPrimary: Boolean = false
                               ) extends ProgramDeclaration with NamespaceField[ComponentNamespace] {

  override def toString: String = {
    val isPrimS = if (isPrimary) "primary " else ""
    val inS = inputs.view.map(_.toString).mkString(", ")
    val outS = outputs.view.map(_.toString).mkString(", ")
    s"${isPrimS}component $ident ( $inS ) returns ( $outS) {\n $body`\n };"
  }

  def absolutePath: List[DSLSymbol] = parent.absolutePath

  def absolutePathInv: List[DSLSymbol] = parent.absolutePathInv

  def absoluteIdent: CompIdent = parent.absolutePath match {
    case Nil => ident
    case path =>
          val res = CompIdent(ident.baseId, PackageIdent(path))
          res.location_(this.ident.location)
          res
  }

  // sets backlinks to parent
  inputs foreach {
    _.parent_(this)
  }
  outputs foreach {
    _.parent_(this)
  }
  body.parent_(this)

}


/** Component identifier, located in a package. */
case class ConfigIdent(baseId: DSLSymbol, pkgId: PackageIdent = NilPackageIdent()) extends IdentLike {
  override def toString: String = pkgId.addPrefix(baseId.name)
}

/** Declaration of a configuration for an analysis. */
case class ConfigurationDeclaration(
                                     ident: ConfigIdent,
                                     rootCompId: CompIdent,
                                     minCardValue: Option[DecIntLit],
                                     failureCondition: BoolExpr,
                                     duration: Option[DecRealLit],
                                     reliability: Option[DecRealLit],
                                     designSpace: List[CompSubstitution]
                                   ) extends ProgramDeclaration with NamespaceField[ConfigurationNamespace] {

  override def toString: String = s"configuration ${ident.toString} { ... };"

  def absolutePath: List[DSLSymbol] = parent.absolutePath

  def absolutePathInv: List[DSLSymbol] = parent.absolutePathInv

  def absoluteIdent: ConfigIdent = parent.absolutePath match {
    case Nil => ident
    case path =>
      val res = ConfigIdent(ident.baseId, PackageIdent(path))
      res.location_(this.ident.location)
      res
  }
}


/** Declares a new data-flow. */
case class FlowDeclaration(
                            ident: BasicFlowIdent,
                            typeId: TypeIdent
                          ) extends Declaration with ParentEitherField[ComponentDeclaration, ComponentBody] {

  override def toString: String = s"$ident: $typeId"

  def namespace: Namespace = parent.fold(x => x.namespace, x => x.namespace)

}

/** Type for all data-flow definitions. */
sealed trait FlowDefinition extends ASTNode with ParentField[ComponentBody] {
  def namespace: ComponentBodyNamespace = parent.namespace

  def getType: TupleType
}

/** Definition of a list of flows by component instanciation. */
case class CompCallFlowDefinition(lhs: List[BasicFlowIdent], rhs: CompCall) extends FlowDefinition {

  override def toString: String = s"${lhs.view.map(_.toString).mkString(", ")} := $rhs ;"

  def getType: TupleType = {
    val lhsT = TupleType(lhs.map(_.getType(this.namespace)))
    val argsT = TupleType(rhs.args.map(_.getType(this.namespace)))
    val compDecl = this.namespace.locate(rhs.compId)
    val inT = TupleType(compDecl.inputs.map(_.ident.getType(compDecl.namespace)))
    val outT = TupleType(compDecl.outputs.map(_.ident.getType(compDecl.namespace)))

    if (argsT != inT) {
      throw WrongTypeCompCallArgs(call = rhs, argsType = argsT, inputsType = inT)
    }

    if (lhsT == outT) {
      lhsT
    } else {
      throw WrongTypeFlowDef(d = this, lhsType = lhsT, rhsType = outT)
    }
  }
}

/** Definition of a single flow definition by a flow expression. */
case class SingleFlowDefinition(lhs: BasicFlowIdent, rhs: FlowExpr) extends FlowDefinition {
  override def toString: String = s"${lhs.toString} := ... ;"

  def getType: TupleType = {
    val lhsT = TupleType(lhs.getType(this.namespace) :: Nil)
    val rhsT = TupleType(rhs.getType(this.namespace) :: Nil)
    if (lhsT == rhsT) {
      lhsT
    } else {
      throw WrongTypeFlowDef(d = this, lhsType = lhsT, rhsType = rhsT)
    }
  }

}
