/** $Id: Inliner.scala 1433 2017-10-19 14:06:15Z kdelmas $
  *
  * @author $Author: kdelmas $
  * @date $Date: 2017-10-19 16:06:15 +0200 (jeu. 19 oct. 2017) $ 
  * */

package inlining.dsl

import ast.dsl._
import util.io.{FileManager, LoadDSLFiles}

import scala.collection.mutable

/** Inlining cache that stores new AST elements generated by the inlining process. */
class Inliner(val configuration: ConfigurationDeclaration) {

  val rootComp: ComponentDeclaration = configuration.namespace.locate(configuration.rootCompId)

  // rootComponent must have no inputs closed
  assert(this.rootComp.inputs.isEmpty)

  val cache = new InliningCache()

  /** Inlines a component call. */
  def inline(compCall: CompCallFlowDefinition)(implicit rwrtr: Rewriter): Unit = {

    /** Generates the name of the called components interface variables by prefixing with instId. */
    def mkInterfaceFlow(decl: FlowDeclaration): BasicFlowIdent = {
      BasicFlowIdent(Utils.mkPipedSym(decl.ident.baseId :: compCall.rhs.instId :: rwrtr.instancePrefix, "."))
    }

    /** Creates and caches a single flow definition given a pair (BasicFlowIdent, FlowExpr). */
    def mkSingleFlowDef(x: (BasicFlowIdent, FlowExpr)): Unit = {
      val (lhs, rhs) = x
      val d = SingleFlowDefinition(lhs, rhs)
      this.cache.definitions.add(d)
      this.cache.singleFlowDefinitions.update(lhs, d)
    }

    /** Generates rewritten versions of the lhs variables of a call definition. */
    def newLhsDecls(compCall: CompCallFlowDefinition): List[FlowDeclaration] =
      compCall.lhs.map(rwrtr.component.body.namespace.locate(_)).collect({ case Right(d) => d }).map(rwrtr.rewrite)

    // find called component declaration
    val calledCompDecl = compCall.namespace.locate(compCall.rhs.compId)

    // bind arguments and recurse on sub component if it is not primary
    if (calledCompDecl.isPrimary) {

      // rewrite call expression but do not inline the component

      // extend instance ident
      val newInstId = Utils.mkPipedSym(compCall.rhs.instId :: rwrtr.instancePrefix, ".")

      // create empty rewriter
      val rw = Rewriter(calledCompDecl)

      // rewrite this component definition
      val newCalledComp = rw.rewritePrimary(calledCompDecl)

      // move type declarations from rw cache to our cache
      this.cache.typeDecls ++= rw.cache.typeDecls
      this.cache.primaryCompDecls ++= rw.cache.primaryCompDecls

      // rewrite argument expressions
      val newArgs = compCall.rhs.args.map(rwrtr.rewrite)

      // new rhs expression
      val newRhs = CompCall(newCalledComp.ident, newInstId, newArgs)

      // new lhs variables
      val newLhsFlows = newLhsDecls(compCall).map(_.ident)

      // new definition
      val newDef = CompCallFlowDefinition(newLhsFlows, newRhs)

      // add to cache
      this.cache.definitions.add(newDef)
      this.cache.callDefinitions.update(CompInstanceIdent(newInstId :: Nil), newDef)

    } else {
      // component is not primary, do inlining

      //  bind inputs using SingleFlowDefinitions
      val callInputs = calledCompDecl.inputs.map(mkInterfaceFlow)

      // rewrite argument expressions
      val newArgs = compCall.rhs.args.map(rwrtr.rewrite)

      // bind inputs
      (callInputs zip newArgs) foreach {
        mkSingleFlowDef
      }

      // initialize new context map
      val zero: mutable.Map[BasicFlowIdent, FlowExpr] = mutable.HashMap.empty
      val newArgsMap: mutable.Map[BasicFlowIdent, FlowExpr] = (calledCompDecl.inputs.map(_.ident) zip newArgs).foldLeft(zero) { (acc, current) => acc + current }

      // update context rewriter
      val newRewriter = Rewriter(calledCompDecl, rwrtr.cache, compCall.rhs.instId :: rwrtr.instancePrefix, newArgsMap)

      // rewrite sub component definitions in topological order
      calledCompDecl.namespace.flowDepGraph.compDefsInOrder.foreach {
        this.inline(_)(newRewriter)
      }

      // new lhs variables
      val newLhsFlows = newLhsDecls(compCall).map(_.ident)

      // bind outputs
      val callOutputs = calledCompDecl.outputs.map(mkInterfaceFlow)
      (newLhsFlows zip callOutputs) foreach {
        mkSingleFlowDef(_)
      }

    }
  }

  /** Recursively inlines a flow definition. */
  def inline(definition: FlowDefinition)(implicit rwrtr: Rewriter): Unit = definition match {
    case d: SingleFlowDefinition => inline(d)
    case d: CompCallFlowDefinition => inline(d)
  }

  /** Inlines a single flow definition. */
  def inline(flowDef: SingleFlowDefinition)(implicit rwrtr: Rewriter): Unit = {

    // rewrite lhs
    val newLhsDecl = rwrtr.rewrite(rwrtr.component.body.namespace.locate(flowDef.lhs).right.get)

    // rewrite rhs
    val newRhs = rwrtr.rewrite(flowDef.rhs)

    // create new def 
    val newDef = SingleFlowDefinition(newLhsDecl.ident, newRhs)

    // store results
    this.cache.flowDecls.update(newLhsDecl.ident, newLhsDecl)
    this.cache.definitions.add(newDef)
    this.cache.singleFlowDefinitions.update(newLhsDecl.ident, newDef)
  }


  def getAllUsedCompDef(c: ComponentDeclaration): List[CompIdent] = {
    c.namespace.compInstanceDepGraph.inOrder.toList
      .flatMap(n => rootComp.namespace.locateDef(n.data).map(fdef => fdef.rhs.compId))
  }


  def inlineNonPrimaryAlt(c: ComponentDeclaration)(rwrtr: Rewriter): ComponentDeclaration = {
    val calledByRoot = getAllUsedCompDef(rootComp)
    if (calledByRoot.contains(c.ident))
      throw new Exception(s"cannot use $c as alternative since called by $rootComp")
    else {
      c.namespace.flowDepGraph.compDefsInOrder foreach (fdef => {
        this.inlineForAlt(fdef)(rwrtr)
      })
      // rewrite all output declarations
      c.outputs.foreach {
        decl => {
          val newDecl = rwrtr.rewrite(decl)
          rwrtr.cache.outputDecls.update(newDecl.ident, newDecl)
        }
      }

      c.inputs.foreach {
        decl => {
          val newDecl = rwrtr.rewrite(decl)
          rwrtr.cache.inputDecls.update(newDecl.ident, newDecl)
        }
      }
      // create new root component from cache contents
      val newCompBody = ComponentBody(
        c.body.events.map(rwrtr.rewrite) ++ rwrtr.cache.evtDecls.toList, // no events
        rwrtr.cache.flowDecls.collect({ case (key, value) if !rwrtr.cache.outputDecls.contains(key) && !rwrtr.cache.inputDecls.contains(key) => value }).toList, // collect only decls that are not outputs
        rwrtr.cache.definitions.iterator.toList // collect all definitions
      )

      ComponentDeclaration(
        rwrtr.rewrite(c.ident),
        rwrtr.cache.inputDecls.values.toList,
        rwrtr.cache.outputDecls.values.toList,
        newCompBody,
        isPrimary = true
      )
    }
  }

  /** Recursively inlines a flow definition. */
  def inlineForAlt(definition: FlowDefinition)(implicit rwrtr: Rewriter): Unit = definition match {
    case d: SingleFlowDefinition => inlineAlt(d)
    case d: CompCallFlowDefinition => inlineAlt(d)
  }


  /** Performs the recursive inlining of the root component. */
  def inline(): Option[PackageDeclaration] = {

    // empty initial context
    val rootRwrtr = Rewriter(component = rootComp, cache = this.cache)

    // rewrite all definitions of the root component in topological order
    this.rootComp.namespace.flowDepGraph.compDefsInOrder foreach {
      this.inline(_)(rootRwrtr)
    }

    // rewrite all output declarations of the root component
    this.rootComp.outputs.foreach {
      decl => {
        val newDecl = rootRwrtr.rewrite(decl)
        this.cache.outputDecls.update(newDecl.ident, newDecl)
      }
    }

    // create new root component from cache contents
    val newRootCompBody = ComponentBody(
      Nil, // no events
      this.cache.flowDecls.collect({ case (key, value) if !this.cache.outputDecls.contains(key) => value }).toList, // collect only decls that are not outputs
      this.cache.definitions.iterator.toList // collect all definitions
    )

    val newRootComp = ComponentDeclaration(
      rootRwrtr.rewrite(this.rootComp.ident),
      Nil,
      this.cache.outputDecls.values.toList,
      newRootCompBody
    )

    // rewrite failure condition
    val newFailureCondition = rootRwrtr.rewrite(configuration.failureCondition)

    // pull and rewrite all components needed for the design-space
    val newDesignSpace = configuration.designSpace.foldLeft[List[CompSubstitution]](Nil) {
      (acc, subst) => {
        val newInstance = rootRwrtr.rewrite(subst.instance)

        val newAlts = (subst.init +: subst.alternatives).foldLeft[List[CompAlternative]](Nil) {
          (acc, alt) => {
            val compDecl = configuration.namespace.locate(alt.compId)
            val absId = compDecl.absoluteIdent
            val rw = Rewriter(compDecl)
            val newCompId = this.cache.primaryCompDecls.getOrElseUpdate(absId, {
              if (compDecl.isPrimary) {
                val newComp = rw.rewritePrimary(compDecl)
                this.cache.typeDecls ++= rw.cache.typeDecls
                newComp
              } else {
                inlineNonPrimaryAlt(compDecl)(rw)
              }
            }
            ).ident
            val newLambdaValues = alt.lambdaValues.map({ (x: (BoolIdent, DecRealLit)) => (rw.rewrite(x._1), x._2) })
            val newAlt = CompAlternative(newCompId, newLambdaValues)
            newAlt :: acc
          }
        }.reverse
        val newSubst = CompSubstitution(newInstance, newAlts.tail, newAlts.head)
        newSubst :: acc
      }
    }.reverse


    // create new configuration
    val newConfiguration = configuration.copy(rootCompId = newRootComp.ident, failureCondition = newFailureCondition, designSpace = newDesignSpace)

    // create new RootPackage containing the inlined system
    val newRootPkg = RootPackage((this.cache.typeDecls.values.toList ::: (newRootComp :: this.cache.primaryCompDecls.values.toList)) :+ newConfiguration)

    // dump to file a reload to validate
    var validatedNewRootPkg: Option[PackageDeclaration] = None

    {
      import java.io.{BufferedWriter, File, FileWriter, IOException}

      try {
        val file: File = FileManager.addFileToTemp(s"${configuration.ident}_inlined.kcr")

        if (!file.exists()) {
          file.createNewFile()
        }

        val fw: FileWriter = new FileWriter(file.getAbsoluteFile)
        val bw: BufferedWriter = new BufferedWriter(fw)
        val printer = new printing.dsl.Printer(bw)
        printer.asRoot(newRootPkg)
        bw.close()
        println(s"Inlined configuration written to ${file.getAbsolutePath}.")
        // validate inlined program by reloading
        println(s"validating results...")
        validatedNewRootPkg = Some(LoadDSLFiles(file.getAbsolutePath :: Nil))
        println(s"success!")
      } catch {
        case e: IOException => e.printStackTrace()
      }
    }
    validatedNewRootPkg
  }


  /** Inlines a component call. */
  def inlineAlt(compCall: CompCallFlowDefinition)(implicit rwrtr: Rewriter): Unit = {

    /** Generates the name of the called components interface variables by prefixing with instId. */
    def mkInterfaceFlow(decl: FlowDeclaration): BasicFlowIdent = {
      BasicFlowIdent(Utils.mkPipedSym(decl.ident.baseId :: compCall.rhs.instId :: rwrtr.instancePrefix, "."))
    }

    /** Generates the name of the called components interface variables by prefixing with instId. */
    def mkInstBoolIdent(evt: BoolIdent): BoolIdent = {
      BoolIdent(Utils.mkPipedSym(evt.baseId :: compCall.rhs.instId :: rwrtr.instancePrefix, "."))
    }

    /** Creates and caches a single flow definition given a pair (BasicFlowIdent, FlowExpr). */
    def mkSingleFlowDef(x: (BasicFlowIdent, FlowExpr)): Unit = {
      val (lhs, rhs) = x
      val d = SingleFlowDefinition(lhs, rhs)
      rwrtr.cache.definitions.add(d)
      rwrtr.cache.singleFlowDefinitions.update(lhs, d)
    }

    /** Generates rewritten versions of the lhs variables of a call definition. */
    def newLhsDecls(compCall: CompCallFlowDefinition): List[FlowDeclaration] =
      compCall.lhs.map(rwrtr.component.body.namespace.locate(_)).collect({ case Right(d) => d }).map(rwrtr.rewrite)

    // find called component declaration
    val calledCompDecl = compCall.namespace.locate(compCall.rhs.compId)

    // component is not primary, do inlining

    //  bind inputs using SingleFlowDefinitions
    val callInputs = calledCompDecl.inputs.map(mkInterfaceFlow)


    val newEvents = calledCompDecl.body.events.map(mkInstBoolIdent)

    // rewrite argument expressions
    val newArgs = compCall.rhs.args.map(rwrtr.rewrite)

    // bind inputs
    (callInputs zip newArgs) foreach {
      mkSingleFlowDef
    }

    // initialize new context map
    val zero: mutable.Map[BasicFlowIdent, FlowExpr] = mutable.HashMap.empty
    val newArgsMap: mutable.Map[BasicFlowIdent, FlowExpr] = (calledCompDecl.inputs.map(_.ident) zip newArgs).foldLeft(zero) { (acc, current) => acc + current }

    //add event to inliningCache
    rwrtr.cache.evtDecls ++= newEvents

    //new event map
    val newEvtsMap = mutable.HashMap(calledCompDecl.body.events zip newEvents: _*)

    // update context rewriter
    val newRewriter = rwrtr.copy(component = calledCompDecl, instancePrefix = compCall.rhs.instId :: rwrtr.instancePrefix, argsMap = newArgsMap, evtMap = newEvtsMap)

    // rewrite sub component definitions in topological order
    calledCompDecl.namespace.flowDepGraph.compDefsInOrder.foreach {
      this.inlineForAlt(_)(newRewriter)
    }

    // new lhs variables
    val newLhsFlows = newLhsDecls(compCall).map(_.ident)

    // bind outputs
    val callOutputs = calledCompDecl.outputs.map(mkInterfaceFlow)
    (newLhsFlows zip callOutputs) foreach {
      mkSingleFlowDef(_)
    }

  }

  /** Inlines a single flow definition. */
  def inlineAlt(flowDef: SingleFlowDefinition)(implicit rwrtr: Rewriter): Unit = {

    // rewrite lhs
    val newLhsDecl = rwrtr.rewrite(rwrtr.component.body.namespace.locate(flowDef.lhs).right.get)

    // rewrite rhs
    val newRhs = rwrtr.rewrite(flowDef.rhs)

    // create new def
    val newDef = SingleFlowDefinition(newLhsDecl.ident, newRhs)

    // store results
    rwrtr.cache.flowDecls.update(newLhsDecl.ident, newLhsDecl)
    rwrtr.cache.definitions.add(newDef)
    rwrtr.cache.singleFlowDefinitions.update(newLhsDecl.ident, newDef)
  }

}
