/** $Id: Rewriter.scala 1433 2017-10-19 14:06:15Z kdelmas $
  *
  * @author $Author: kdelmas $
  * @date $Date: 2017-10-19 16:06:15 +0200 (jeu. 19 oct. 2017) $ 
  **/

package inlining.dsl

import ast.dsl._

import scala.collection.mutable

case class Rewriter(
                     /** current component context. */
                     component: ComponentDeclaration,

                     /** The inlining cache. */
                     cache: InliningCache = new InliningCache(),

                     /** Inverse path of comp in the call tree. */
                     instancePrefix: List[DSLSymbol] = Nil,

                     /** Map associating arguments expressions to the component inputs. */
                     argsMap: mutable.Map[BasicFlowIdent, FlowExpr] = mutable.Map.empty,

                     /** Map for event **/
                     evtMap: mutable.Map[BoolIdent, BoolIdent] = mutable.Map.empty,

                     /** tells the rewriter to use the argsMap to subsitute expressions for identifiers or not. */
                     useArgsMap: Boolean = false) {

  /** Returns true if the context of the rewriter. */
  def isRootContext: Boolean = {
    instancePrefix.isEmpty &&
      argsMap.isEmpty &&
      !useArgsMap
  }

  /** Rewrites a type declaration to new type a declaration in which the
    * type identifier is the original type's absolute identifier
    * between pipes, safe to be placed in a root package.
    */
  def rewrite(decl: TypeDeclaration): TypeDeclaration = {
    cache.typeDecls.getOrElseUpdate(
      decl.absoluteIdent,
      {
        val newBaseId = Utils.mkPipedSym(decl.ident.baseId :: decl.absolutePathInv, "::")
        val newId = TypeIdent(newBaseId, PackageIdent(Nil))
        val newDecl = decl.copy(ident = newId)
        newDecl
      }
    )
  }

  def rewrite(id: CompIdent): CompIdent = {
    val compDecl = this.component.namespace.locate(id)
    val newBaseId = Utils.mkPipedSym(compDecl.ident.baseId :: compDecl.absolutePathInv, "::")
    val newId = CompIdent(newBaseId, PackageIdent(Nil))
    newId
  }

  def rewrite(id: CompInstanceIdent): CompInstanceIdent = {
    // extend instance ident
    val newInstId = CompInstanceIdent(Utils.mkPipedSym(id.path.reverse ::: this.instancePrefix, ".") :: Nil)
    newInstId
  }

  /** Rewrites a primary component declaration to new a component
    * declaration in which the component identifier is the original
    * component's absolute identifier between pipes, and all type
    * identifiers are rewritten to their absolute version, that can be
    * placed in a root package.
    */
  def rewritePrimary(decl: ComponentDeclaration): ComponentDeclaration = {

    // we rewrite only  primary components in a root context
    assert(decl.isPrimary && this.isRootContext)

    cache.primaryCompDecls.getOrElseUpdate(
      decl.absoluteIdent,
      {
        // create absolute identifier
        val newBaseId = Utils.mkPipedSym(decl.ident.baseId :: decl.absolutePathInv, "::")
        val newId = CompIdent(newBaseId, PackageIdent(Nil))

        // we consider we are already in an empty context rewriter

        val newInputs = decl.inputs map {
          this.rewrite
        }
        val newOutputs = decl.outputs map {
          this.rewrite
        }

        // rewrite body
        val newLocals = decl.body.locals map {
          this.rewrite
        }
        val newDefs = decl.body.definitions.collect({ case d: SingleFlowDefinition => d }).map(this.rewrite)
        val newBody = ComponentBody(decl.body.events, newLocals, newDefs)
        val newDecl = ComponentDeclaration(newId, newInputs, newOutputs, newBody, isPrimary = true)
        newDecl
      }
    )
  }

  /** Rewrites a single flow definition. */
  def rewrite(flowDef: SingleFlowDefinition): SingleFlowDefinition = {

    // rewrite lhs
    val newLhsDecl = this.rewrite(this.component.body.namespace.locate(flowDef.lhs).right.get)

    // rewrite rhs
    val newRhs = this.rewrite(flowDef.rhs)

    // create new def
    val newDef = SingleFlowDefinition(newLhsDecl.ident, newRhs)

    // store results
    this.cache.flowDecls.update(newLhsDecl.ident, newLhsDecl)
    this.cache.singleFlowDefinitions.update(newLhsDecl.ident, newDef)
    newDef
  }

  /** Rewrites a BasicFlowIdent that was resolved to a type value to an
    * explicit TypeValueFlowIdent.
    */
  def explicit(id: BasicFlowIdent, absTypeDecl: TypeDeclaration): TypeValueFlowIdent = TypeValueFlowIdent(absTypeDecl.ident, id.baseId)

  /** Rewrites a flow declaration in the current context (converts the
    * type to an absolute type and renames the flow by prefixing the
    * flow with instance path and encapsulating it in pipes.
    */
  def rewrite(flowDecl: FlowDeclaration): FlowDeclaration = {
    // get new absolute type
    val newTypeDecl = this.rewrite(component.body.namespace.locate(flowDecl.typeId))
    val newId = BasicFlowIdent(Utils.mkPipedSym(flowDecl.ident.baseId :: instancePrefix, "."))
    // update cache
    cache.flowDecls.getOrElseUpdate(newId, FlowDeclaration(newId, newTypeDecl.ident))
  }

  /** Rewrites a flow expression in the current context. */
  def rewrite(expr: FlowExpr): FlowExpr = expr match {

    case id: FlowIdent => id match {

      case id:BasicFlowIdent => this.component.body.namespace.locate(id) match {

        // dealing with a type value identifier
        case Left(typeDecl) =>
          val newTypeDecl = this.rewrite(typeDecl)
          explicit(id, newTypeDecl) // pipes are not added to type values

        // Dealing with a local flow
        case Right(flowDecl) =>
          val newFlowDecl = this.rewrite(flowDecl)
          newFlowDecl.ident
      }

      case InstFlowIdent(baseId, instId) =>

        // this flow belongs to a subcomponent so its declaration will be produced later down the line
        // just prefix it with instance path
        val newId =
        BasicFlowIdent(Utils.mkPipedSym((baseId :: instId.path.reverse) ::: instancePrefix, "."))

        // return result
        newId

      case TypeValueFlowEmpty(typeId) =>
        val newTypeDecl = this.rewrite(this.component.body.namespace.locate(typeId))
        val newId = TypeValueFlowEmpty(newTypeDecl.ident)
        newId

      case TypeValueFlowIdent(typeId, valueId) =>
        val newTypeDecl = this.rewrite(this.component.body.namespace.locate(typeId))
        val newId = TypeValueFlowIdent(newTypeDecl.ident, valueId)
        newId
    }
    case FlowInter(kid1, kid2) => FlowInter(rewrite(kid1), rewrite(kid2))
    case FlowUnion(kid1, kid2) => FlowUnion(rewrite(kid1), rewrite(kid2))
    case FlowIte(kid1, kid2, kid3) => FlowIte(rewrite(kid1), rewrite(kid2), rewrite(kid3))
  }

  def rewrite(id: BoolIdent): BoolIdent = evtMap.getOrElse(id, {
    id match {
      case BoolIdent(s: PointedDSLSymbol) => BoolIdent(PipedDSLSymbol(Symbol(s"|${s.symbol.name}|")))
      case _ => id
    }
  })

  /** Rewrites a Boolean expression. */
  def rewrite(expr: BoolExpr): BoolExpr = expr match {
    case id: BoolIdent => this.rewrite(id)
    case FlowEq(kid1, kid2) => FlowEq(rewrite(kid1), rewrite(kid2))
    case FlowNeq(kid1, kid2) => FlowNeq(rewrite(kid1), rewrite(kid2))
    case FlowContains(kid1, kid2) => FlowContains(rewrite(kid1), rewrite(kid2))
    case PopCount(kids, relOp, cst) => PopCount(kids.map(this.rewrite), relOp, cst)
    case Not(kid1) => Not(rewrite(kid1))
    case And(kid1, kid2) => And(rewrite(kid1), rewrite(kid2))
    case Or(kid1, kid2) => Or(rewrite(kid1), rewrite(kid2))
    case Iff(kid1, kid2) => Iff(rewrite(kid1), rewrite(kid2))
    case Implies(kid1, kid2) => Implies(rewrite(kid1), rewrite(kid2))
  }
}
