package interpreter

import util.analyzers.analyzerTypes.FlowValue
import analyzers.{ ConfigurationAnalyzer, Parsing}
import ast.dsl._
import namespace.dsl.FlowDepGraph
import net.sf.javabdd.BDD
import util.analyzers._

import scala.collection.generic.CanBuildFrom
import scala.util.control.TailCalls._


/**
  * Created by kdelmas on 23/12/15.
  */
trait DSLInterpreter extends ConfigurationAnalyzer {
  self: Parsing =>
}

object TraceInterpreter {

  /**
    * type representing a valuation of a constant
    */
  type FlowValueInterpretation = Valuation[FlowValue]

  /**
    * type representing an interpretation of a Boolean
    */
  type BoolInterpretation = Set[Valuation[Boolean]]

  /**
    * type representing a component trace as a valuation of its flows
    */
  type ComponentTrace = Valuation[Map[BasicFlowIdent, FlowValue]]

  /**
    * type representing an interpretation of a flow
    */
  type FlowInterpretation = Set[FlowValueInterpretation]

  /**
    * type representing the component traces a a set of component trace
    */
  type ComponentTraces = Set[ComponentTrace]

  /**
    * type representing a component interpreation as a mapping
    * from its flow to their interpretation
    */
  type ComponentInterpretation = Map[BasicFlowIdent, FlowInterpretation]

  /**
    * type representing a mapping from system flows to their interpretation
    */
  type Context = Map[InstFlowIdent, FlowInterpretation]

  /**
    * A valuation for the interpreter is a couple of a concrete value and a valid BDD
    * encoding the system event valuations for which the concrete value is observed
    *
    * @param value the concrete value
    * @param valid the system event BDD
    * @tparam A the type of the concrete value
    */
  case class Valuation[A](value: A, valid: BDD)

}

/**
  * Interpreter of the KCR language
  */
trait TraceInterpreter extends DSLInterpreter {
  self: Parsing =>

  import InterpreterImplicits._
  import TraceInterpreter._

  val InstBoolBDDFactory: InstBoolBDDFactory

  private val evalCache = collection.mutable.HashMap[CompKey, ComponentInterpretation]()

  /**
    * Interpretation cache key
    *
    * @param inst the component instance ident
    * @param bdd  the BDD encoding the system events occurences for which it has been interpreted
    */
  case class CompKey(inst: CompInstanceIdent, bdd: BDD)

  /**
    * Util implicit definition for interpreter
    */
  object InterpreterImplicits {

    /**
      * conversion from component trace to ordered one
      *
      * @param compTrace interpreter component trace
      * @return ordered component trace
      */
    implicit def toOrderedCompTrace(compTrace: ComponentTrace): OrderedCompTrace = {
      OrderedCompTrace(compTrace.value)
    }

    /**
      * Class defining the notion of interpretation product and sum
      *
      * @param value An interpretation (ie a set of valuation)
      * @tparam A the concrete value type
      * @tparam T the collection choosen for gathering valuations
      */
    implicit class RichInterpretation[A, T[C] <: Set[C]](value: T[Valuation[A]]) {

      /**
        * Compute an interpretation obtained by applying a function f taking valuation of this interpretation
        * and another interpretation
        *
        * @param that another interpretation
        * @param f    the function generating the new concrete values
        * @param bf   can build a collection T of D valuations from a collection Set of A valuations
        * @tparam A2 the type of the other concrete values
        * @tparam D  the type of the concrete values generated by f
        * @return the resulting interpretation
        */
      def customProd[A2, D](that: T[Valuation[A2]], f: (A, A2) => D)
                           (implicit bf: CanBuildFrom[Set[Valuation[A]], Valuation[D], T[Valuation[D]]]): T[Valuation[D]] = {

        //foreach valuation in this interpretation
        value.flatMap(l =>

          //collect the new value that can be obtained for a system event valuation
          //(ie such that r.valid \wedge l.valid \neq false)
          that collect {
            case r: Valuation[_] if !r.valid.and(l.valid).isZero =>
              Valuation(f(l.value, r.value), r.valid.and(l.valid))
          }
        )
      }

      /**
        * Compute the union of the valuation of this interpretation and the one of another interpretation
        *
        * @param that another interpration
        * @param bf   can build a collection T of A valuations from a collection Set of A valuations
        * @return the resulting interpretation
        */
      def customSum(that: T[Valuation[A]])
                   (implicit bf: CanBuildFrom[Set[Valuation[A]], Valuation[A], T[Valuation[A]]]): T[Valuation[A]] = {

        //groupe the valuation of the two interpration by their concrete value
        val groupedInter = (value ++ that).groupBy(_.value)

        //for all valuation giving the same concrete value, merge the combinaisons of system events
        //for which it is observed (ie disjunction of their "valid" BDD)
        val result = groupedInter.mapValues(_.foldLeft(InstBoolBDDFactory.zero())((acc, inter) => inter.valid.or(acc)))
          .map(kv => Valuation(kv._1, kv._2)).toSet

        //build the output collection of valuations
        val builder = bf.apply(result)
        builder ++= result
        builder.result()
      }
    }

    /**
      * Extract the basic flow identifier from instance flow identifier
      *
      * @param inst instance flow identifier
      * @return basic flow identifier
      */
    implicit def toBasicFlowIdent(inst: InstFlowIdent): BasicFlowIdent = BasicFlowIdent(inst.baseId)

    /**
      * Extract the basic Boolean identifier from instance Boolean identifier
      *
      * @param inst instance Boolean identifier
      * @return basic Boolean identifier
      */
    implicit def toBoolIdent(inst: InstBoolIdent): BoolIdent = inst.baseId

    private val union = (a: Map[BasicFlowIdent, FlowValue], b: Map[BasicFlowIdent, FlowValue]) => a ++ b

    /**
      * Transform a component interpretation of its input and output flows to a set of component traces
      *
      * @param interpretation the interpretation of the component flows
      * @return the set of component traces
      */
    implicit def toComponentTraces(interpretation: ComponentInterpretation): ComponentTraces = {

      interpretation.foldLeft(Set[ComponentTrace]())((acc, flow) => {

        //for the current flow create a mapping from it to each possible concrete values
        //stored in the interpretation
        val split = flow._2.map(flowVal => Valuation(Map(flow._1 -> flowVal.value), flowVal.valid))

        //then add these mapping to the previous mapping of the component flows
        if (acc.isEmpty) {
          split
        } else {
          acc.customProd(split, union)
        }
      })
    }
  }

  /**
    * Convert a conjunction of system events to a BDD
    *
    * @param pos positive litterals
    * @param neg negative litterals
    * @return the resulting BDD
    */
  def cubeToBDD(pos: Set[InstBoolIdent], neg: Set[InstBoolIdent]): BDD = {
    val simuBDDPos = pos.foldLeft(InstBoolBDDFactory.one())((acc, lit) => {
      acc.and(InstBoolBDDFactory.produceVar(lit))
    })
    neg.foldLeft(simuBDDPos)((acc, lit) => acc.and(InstBoolBDDFactory.produceVar(lit).not()))
  }


  /**
    * Compute the possible interpretations of the component instances
    *
    * @return the interpretation of system components
    */
  def evaluate(): Map[CompInstanceIdent, ComponentInterpretation] = {
    evalCache.clear()
    evalRootComp(InstBoolBDDFactory.one())
    evalCache.map(kv => kv._1.inst -> kv._2).toMap
  }


  /**
    * Compute the possible interpretations of the component instances
    * and focus on the interpretation that are produced for the combinations
    * of system event encoded in the BDD
    *
    * @param piBDD  the BDD of system event
    * @param bddVar mapping from system event to BDD nodes
    * @return the interpretation of system components
    */
  def evaluateKnowing(piBDD: BDD, bddVar: Map[InstBoolIdent, BDD]): Map[CompInstanceIdent, ComponentInterpretation] = {
    evalCache.clear()
    val newPiBDD = InstBoolBDDFactory.importBDD(piBDD, bddVar)
    evalRootComp(newPiBDD)
    evalCache.map(kv => kv._1.inst -> kv._2).toMap
  }

  /**
    * Evaluate a component
    *
    * @param comp            the AST component declaration
    * @param currentInstance the instance analyser
    * @param context         the valuation of system flows
    * @param piBDD           targeted system events valuations
    * @return
    */
  private def evalComp(comp: ComponentDeclaration, currentInstance: CompInstanceIdent, context: Context, piBDD: BDD = InstBoolBDDFactory.one()): TailRec[ComponentInterpretation] = {

    done(evalCache.getOrElseUpdate(CompKey(currentInstance, piBDD), {

      //for each flow definition
      val flowGraph: FlowDepGraph = comp.namespace.flowDepGraph
      val newContext: Context = flowGraph.compDefsInOrder.toList.distinct.foldLeft(context)((localContext: Context, fdef) => {
        fdef match {
          case CompCallFlowDefinition(lhs: List[BasicFlowIdent], rhs: CompCall) =>

            //if defined by instantiation create a new context binding the inputs of the component
            //to their interpretation
            val calledInstance = CompInstanceIdent(currentInstance.path :+ rhs.instId)
            val calledInputs = comp.namespace.locate(rhs.compId).inputs.map(i => InstFlowIdent(i.ident.baseId, calledInstance))
            val calledContext = calledInputs.zip(rhs.args.map(arg => evalFlowExpr(arg, currentInstance, localContext, piBDD).result)).toMap

            //evaluate the component
            val compInter = tailcall(evalComp(comp.namespace.locate(rhs.compId), calledInstance, calledContext, piBDD)).result

            //get the output interpretations and add then to the context
            val calledOutput = comp.namespace.locate(rhs.compId).outputs.map(o => compInter(o.ident))
            localContext ++ lhs.map(f => InstFlowIdent(f.baseId, currentInstance)).zip(calledOutput).toMap

          case SingleFlowDefinition(lhs: BasicFlowIdent, rhs: FlowExpr) =>

            //if defined by expression, evaluate it and store the result in the context
            localContext + (InstFlowIdent(lhs.baseId, currentInstance) -> evalFlowExpr(rhs, currentInstance, localContext, piBDD).result)
        }
      })

      //find the interpretations of the inputs and the ouputs and return then
      (comp.outputs ++ comp.inputs).map(id =>
        id.ident -> newContext.getOrElse(InstFlowIdent(id.ident.baseId, currentInstance),
          throw new Exception(s"unknown in/out $id"))).toMap
    }))
  }

  /**
    * Evaluate the root component
    *
    * @param piBDD the targeted system event valuations
    */
  private def evalRootComp(piBDD: BDD = InstBoolBDDFactory.one()): Unit = {

    val rootInstIdent = CompInstanceIdent(Nil)
    val defTopo = rootDef.namespace.flowDepGraph


    defTopo.compDefsInOrder.toList.distinct match {
      case Nil =>
      case head :: tail =>
        evalCache.getOrElseUpdate(CompKey(rootInstIdent, piBDD), {

          //for each flow definition
          val newContext: Context = (head :: tail).foldLeft(Map[InstFlowIdent, FlowInterpretation]())((localContext: Context, fdef) => {
            fdef match {
              case CompCallFlowDefinition(lhs: List[BasicFlowIdent], rhs: CompCall) =>

                //if defined by instantiation create a new context binding the inputs of the component
                //to their interpretation
                val calledInstance = CompInstanceIdent(rhs.instId :: Nil)
                val calledInputs = rootDef.namespace.locate(rhs.compId).inputs.map(i => InstFlowIdent(i.ident.baseId, calledInstance))
                val calledContext = calledInputs.zip(rhs.args.map(arg => evalFlowExpr(arg, rootInstIdent, localContext, piBDD).result)).toMap

                //evaluate the component
                val compInter = evalPrimaryComp(rootDef.namespace.locate(rhs.compId), calledInstance, calledContext, piBDD)

                //get the output interpretations and add then to the context
                val calledOutput = rootDef.namespace.locate(rhs.compId).outputs.map(o => compInter(o.ident))
                localContext ++ lhs.map(f => InstFlowIdent(f.baseId, rootInstIdent)).zip(calledOutput).toMap

              case SingleFlowDefinition(lhs: BasicFlowIdent, rhs: FlowExpr) =>

                //if defined by expression, evaluate it and store the result in the context
                localContext + (InstFlowIdent(lhs.baseId, rootInstIdent) -> evalFlowExpr(rhs, rootInstIdent, localContext, piBDD).result)
            }
          })

          //find the interpretations of the inputs and the ouputs and return then
          (rootDef.outputs ++ rootDef.inputs).map(id =>
            id.ident -> newContext.getOrElse(InstFlowIdent(id.ident.baseId, CompInstanceIdent(Nil)),
              throw new Exception(s"unknown in/out $id"))).toMap
        })
    }

  }

  private def evalPrimaryComp(comp: ComponentDeclaration, currentInstance: CompInstanceIdent, context: Context, piBDD: BDD = InstBoolBDDFactory.one()): ComponentInterpretation = {

    require(comp.isPrimary)

    evalCache.getOrElseUpdate(CompKey(currentInstance, piBDD), {

      val flowGraph: FlowDepGraph = comp.namespace.flowDepGraph

      //for each flow definition
      val newContext: Context = flowGraph.compDefsInOrder.toList.distinct.foldLeft(context)((localContext: Context, fdef) => {
        fdef match {

          case SingleFlowDefinition(lhs: BasicFlowIdent, rhs: FlowExpr) =>

            //if defined by expression, evaluate it and store the result in the context
            localContext + (InstFlowIdent(lhs.baseId, currentInstance) -> evalFlowExpr(rhs, currentInstance, localContext, piBDD).result)

          case _ =>

            //primary component contains only simple flow definitions
            throw new Exception(s"${comp.ident} is not primary")
        }
      })

      //find the interpretations of the inputs and the ouputs and return then
      (comp.outputs ++ comp.inputs).map(id =>
        id.ident -> newContext.getOrElse(InstFlowIdent(id.ident.baseId, currentInstance),
          throw new Exception(s"unknown in/out $id"))).toMap
    })
  }

  /**
    * Evaluate a KCR boolean expression
    *
    * @param b               the boolean expression
    * @param currentInstance the instance containing this expression
    * @param context         the evaluation context
    * @param piBDD           the targetted system event valuations
    * @return the interpretation of b
    */
  private def evalBoolExpr(b: BoolExpr, currentInstance: CompInstanceIdent, context: Context, piBDD: BDD): TailRec[BoolInterpretation] = {
    import InstBoolBDDFactory.FactoryImplicits._

    b match {
      case b:BoolIdent =>
        val e = InstBoolIdent(b, currentInstance)
        val valuations =
          if (piBDD.and(e.not()).isZero) {
            Set(Valuation(true, piBDD))
          } else if (piBDD.and(e).isZero) {
            Set(Valuation(false, piBDD))
          } else {
            Set(Valuation(true, piBDD.and(InstBoolBDDFactory.produceVar(e))), Valuation(false, piBDD.and(InstBoolBDDFactory.produceVar(e).not())))
          }
        done(valuations)

      case FlowEq(kid1: FlowExpr, kid2: FlowExpr) =>
        val f1 = tailcall(evalFlowExpr(kid1, currentInstance, context, piBDD)).result
        val f2 = tailcall(evalFlowExpr(kid2, currentInstance, context, piBDD)).result
        val eq = (f1: FlowValue, f2: FlowValue) => f1 == f2
        done(f1.customProd(f2, eq))

      case FlowNeq(kid1: FlowExpr, kid2: FlowExpr) =>
        val f1 = tailcall(evalFlowExpr(kid1, currentInstance, context, piBDD)).result
        val f2 = tailcall(evalFlowExpr(kid2, currentInstance, context, piBDD)).result
        val neq = (f1: FlowValue, f2: FlowValue) => f1 != f2
        done(f1.customProd(f2, neq))

      case FlowContains(kid1: FlowExpr, kid2: FlowExpr) =>
        val f1 = tailcall(evalFlowExpr(kid1, currentInstance, context, piBDD)).result
        val f2 = tailcall(evalFlowExpr(kid2, currentInstance, context, piBDD)).result
        val contain = (f1: FlowValue, f2: FlowValue) => f1.subsetOf(f2)
        done(f1.customProd(f2, contain))

      case Iff(kid1: BoolExpr, kid2: BoolExpr) =>
        val b1 = tailcall(evalBoolExpr(kid1, currentInstance, context, piBDD)).result
        val b2 = tailcall(evalBoolExpr(kid2, currentInstance, context, piBDD)).result
        val iff = (b1: Boolean, b2: Boolean) => b1 == b2
        done(b1.customProd(b2, iff))

      case Implies(kid1: BoolExpr, kid2: BoolExpr) =>
        val b1 = tailcall(evalBoolExpr(kid1, currentInstance, context, piBDD)).result
        val b2 = tailcall(evalBoolExpr(kid2, currentInstance, context, piBDD)).result
        val imp = (b1: Boolean, b2: Boolean) => !b1 || b2
        done(b1.customProd(b2, imp))

      case Not(kid1: BoolExpr) =>
        val f = tailcall(evalBoolExpr(kid1, currentInstance, context, piBDD)).result
        done(f.map(inter => inter.copy(value = !inter.value)))

      case And(kid1: BoolExpr, kid2: BoolExpr) =>
        val b1 = tailcall(evalBoolExpr(kid1, currentInstance, context, piBDD)).result
        val b2 = tailcall(evalBoolExpr(kid2, currentInstance, context, piBDD)).result
        val and = (b1: Boolean, b2: Boolean) => b1 && b2
        done(b1.customProd(b2, and))

      case Or(kid1: BoolExpr, kid2: BoolExpr) =>
        val b1 = tailcall(evalBoolExpr(kid1, currentInstance, context, piBDD)).result
        val b2 = tailcall(evalBoolExpr(kid2, currentInstance, context, piBDD)).result
        val or = (b1: Boolean, b2: Boolean) => b1 || b2
        done(b1.customProd(b2, or))

      case count@PopCount(kids: List[BoolExpr], relop: RelOp, cst: DecIntLit) =>
        relop match {
          case Ge =>
            val or = (b1: Boolean, b2: Boolean) => b1 || b2
            val and = (b1: Boolean, b2: Boolean) => b1 && b2
            val bools = kids.map(kid => tailcall(evalBoolExpr(kid, currentInstance, context, piBDD))).combinations(cst.repr.toInt)
            if (bools.isEmpty) {
              done(Set(Valuation(false, InstBoolBDDFactory.one())))
            } else {
              val ands = bools.map(s =>
                if (s.size == 1) {
                  s.head.result
                } else {
                  s.tail.foldLeft(s.head.result)((acc, other) => acc.customProd(other.result, and))
                }).toList
              if (ands.size == 1) {
                done(ands.head)
              } else {
                done(ands.tail.foldLeft(ands.head)((acc, other) => acc.customProd(other, or)))
              }
            }
          case Gt =>
            tailcall(evalBoolExpr(count.copy(relop = Ge, cst = DecIntLit(s"${cst.repr.toInt + 1}")), currentInstance, context, piBDD))
          case Le =>
            tailcall(evalBoolExpr(Not(count.copy(relop = Ge, cst = DecIntLit(s"${cst.repr.toInt + 1}"))), currentInstance, context, piBDD))
          case Lt =>
            tailcall(evalBoolExpr(Not(count.copy(relop = Ge)), currentInstance, context, piBDD))
          case Eq =>
            tailcall(evalBoolExpr(And(count.copy(relop = Ge), Not(count.copy(relop = Ge, cst = DecIntLit(s"${cst.repr.toInt + 1}")))), currentInstance, context, piBDD))
        }
    }
  }

  /**
    * Evaluate a KCR flow expression
    *
    * @param f               the flow expression
    * @param currentInstance the instance containing this expression
    * @param context         the evaluation context
    * @param piBDD           the targetted system event valuations
    * @return the interpretation of b
    */
  private def evalFlowExpr(f: FlowExpr, currentInstance: CompInstanceIdent, context: Context, piBDD: BDD): TailRec[FlowInterpretation] = {
    f match {
      case id:BasicFlowIdent =>
        done(context.getOrElse(InstFlowIdent(id.baseId, currentInstance), throw new Exception(s" unknown flow id $id")))

      case id:InstFlowIdent =>
        done(context.getOrElse(id, throw new Exception(s" unknown flow id $id")))

      case _:TypeValueFlowEmpty =>
        done(Set(Valuation(Set(), piBDD)))

      case c:TypeValueFlowIdent =>
        done(Set(Valuation(Set(c), piBDD)))

      case FlowUnion(kid1: FlowExpr, kid2: FlowExpr) =>
        val f1 = tailcall(evalFlowExpr(kid1, currentInstance, context, piBDD)).result
        val f2 = tailcall(evalFlowExpr(kid2, currentInstance, context, piBDD)).result
        val union = (f1: FlowValue, f2: FlowValue) => f1 ++ f2
        done(f1.customProd(f2, union))

      case FlowInter(kid1: FlowExpr, kid2: FlowExpr) =>
        val f1 = tailcall(evalFlowExpr(kid1, currentInstance, context, piBDD)).result
        val f2 = tailcall(evalFlowExpr(kid2, currentInstance, context, piBDD)).result
        val inter = (f1: FlowValue, f2: FlowValue) => f1.intersect(f2)
        done(f1.customProd(f2, inter))

      case FlowIte(kid1: BoolExpr, kid2: FlowExpr, kid3: FlowExpr) =>
        val b = tailcall(evalBoolExpr(kid1, currentInstance, context, piBDD)).result
        val f1 = tailcall(evalFlowExpr(kid2, currentInstance, context, piBDD)).result
        val f2 = tailcall(evalFlowExpr(kid3, currentInstance, context, piBDD)).result
        val transmit = (_: Boolean, f: FlowValue) => f
        done(b.filter(_.value).customProd(f1, transmit).customSum(b.filter(!_.value).customProd(f2, transmit)))
    }
  }

  //create fictive event e_{flow=value} encoding that the flow flow is equal to value
  private val defineBoolKey = (flow: InstFlowIdent, value: FlowValue) => InstBoolIdent(BoolIdent(StdDSLSymbol(Symbol(s"${flow.baseId.name}=$value"))), CompInstanceIdent(Nil))


  /**
    * Generate the possible traces for a component
    */
  val makePossibleTraces: (ComponentDeclaration) => Map[OrderedCompTrace, BDD] = MyMemo.mutableOpenHashMapMemo {
    (comp: ComponentDeclaration) => {


      //create the all the possible values for the component inputs since it is analysed in isolation
      val inputsPowerSets: Map[InstFlowIdent, Set[FlowValue]] = comp.inputs.map(f => {
        InstFlowIdent(f.ident.baseId, CompInstanceIdent(List(comp.ident.baseId))) -> powerSet(comp.namespace.locate(f.typeId))
      }).toMap

      //associate to each input f possible valuation v the BDD e_{f=v} \wedge_{v'\neg v} e_{f=v'}
      val context: Context = inputsPowerSets.map(flow => {
        flow._1 -> {
          flow._2.map(set => {
            Valuation(
              set,
              cubeToBDD(Set(defineBoolKey(flow._1, set)), (flow._2 - set).map(set2 => defineBoolKey(flow._1, set2))))
          })
        }
      })

      //evaluate the component
      val sim: ComponentTraces = evalComp(comp, CompInstanceIdent(List(comp.ident.baseId)), context).result

      //remove the fictive event from the interpretations
      val purifiedTraces = sim.map(trace => {
        val inputs: Map[InstFlowIdent, (FlowValue, Set[FlowValue])] = trace.value.map(kv => InstFlowIdent(kv._1.baseId, CompInstanceIdent(List(comp.ident.baseId))) -> kv._2).
          filter(kv => inputsPowerSets.contains(kv._1)).map(kv => (kv._1, (kv._2, inputsPowerSets(kv._1))))
        val pairing: BDD = inputs.foldLeft(InstBoolBDDFactory.one())((acc, kv) => acc.and(kv._2._2.foldLeft(InstBoolBDDFactory.one())((acc, value) => acc.and({
          if (value == kv._2._1) {
            InstBoolBDDFactory.produceVar(defineBoolKey(kv._1, value))
          } else {
            InstBoolBDDFactory.produceVar(defineBoolKey(kv._1, value)).not()
          }
        }))))
        val bdd = trace.valid.restrict(pairing)
        trace.copy(valid = bdd)
      })
      purifiedTraces.map(trace => OrderedCompTrace(trace.value) -> trace.valid).toMap
    }
  }

  /**
    * Generate the powerset of the set of value given in a type declaration
    */
  val powerSet: (TypeDeclaration) => Set[FlowValue] = MyMemo.mutableOpenHashMapMemo {
    (tyype: TypeDeclaration) => {
      val values: FlowValue = tyype.values.map(v => {
        if (v.symbol == 'empty) {
          TypeValueFlowEmpty(tyype.ident)
        } else {
          TypeValueFlowIdent(tyype.ident, v)
        }
      }).toSet
      values.subsets().toSet
    }
  }
}


