/** $Id: BuildNamespace.scala 1433 2017-10-19 14:06:15Z kdelmas $
  *
  * @author $Author: kdelmas $
  **/

package namespace.dsl

import ast.dsl._

import scala.annotation.tailrec
import scala.collection.immutable.{HashMap, HashSet}
import scala.util.{Failure, Success, Try}
import scalaz.\/


/** Trampoline that builds and checks the namespace of a program. 
  *   - traverse package hierarchy and resolve package open statements 
  *   - recursively builds the namespace hierarchy 
  *   - checks identifier uses everywhere
  *   - checks for cyclic dependencies at flow levels inside components
  * */
object BuildNamespace extends (PackageDeclaration => \/[NamespaceExceptions, List[NamespaceWarning]]) {
  def apply(rootPackage: PackageDeclaration): \/[NamespaceExceptions, List[NamespaceWarning]] = {
    for {
      openRes <- RunAction(ResolveOpenDeclarations(rootPackage))
      buildRes <- RunAction(BuildPackageNamespace(None, rootPackage))
      checkRes <- RunAction(CheckIdentsPackage(rootPackage))
    } yield {
      openRes ::: buildRes ::: checkRes
    }
  }
}


/** Resolves open statements of a program once namespace maps have been built without error. */
case class ResolveOpenDeclarations(pkg: PackageDeclaration) extends Action {

  override def toString = s"ResolveOpenDeclarations( ${pkg.ident} )"

  def apply(): (List[Action], List[NamespaceException]) = {

    var _exceptions: List[OpenPackageNotFoundException] = Nil

    val opens: HashSet[OpenDeclaration] = HashSet.empty ++ pkg.declarations collect { case o: OpenDeclaration => o }

    opens foreach { o =>
      Try {
        this.locate(o.pkgId.path, pkg)
      } match {
        case Success(p) => o.target_(p)
        case Failure(ParentNotSet(_)) =>
          // root package reached without finding the package
          _exceptions ::= OpenPackageNotFoundException(o)

        case Failure(e) =>
          // unexpected error
          throw e

      }
    }
    val _actions = pkg.declarations collect { case p: PackageDeclaration => ResolveOpenDeclarations(p) }
    (_actions, _exceptions.reverse)
  }

  private[this] def locate(path: List[DSLSymbol], context: PackageDeclaration): PackageDeclaration = {
    locateDown(path, context) getOrElse locate(path, context.parent)
  }

  private[this] def locateDown(path: List[DSLSymbol], context: PackageDeclaration): Option[PackageDeclaration] = path match {

    case head :: tail =>

      /* Use first matching package. There maybe more than one sub-package
       * with the same name at this point, which will trigger an
       * exception later-on. 
       * */
      context.declarations collectFirst { case p: PackageDeclaration if p.ident == head => p } match {
        case Some(p) => locateDown(tail, p)
        case None => None
      }


    case Nil => Some(context)
  }
}

/** Recursively builds the namespace of a package declaration. */
case class BuildPackageNamespace(parent: Option[PackageNamespace], pkg: PackageDeclaration) extends Action {

  override def toString = s"BuildPackageNamespace( ${pkg.ident} )"

  private var _exceptions: List[NamespaceException] = Nil
  private var _packages: PackageMapType = HashMap.empty
  private var _includes: IncludeSetType = HashSet.empty
  private var _opens: OpenSetType = HashSet.empty
  private var _types: TypeMapType = HashMap.empty
  private var _components: ComponentMapType = HashMap.empty
  private var _configurations: ConfigurationMapType = HashMap.empty
  private var _deferredActions: List[Action] = Nil

  @tailrec
  final private def sortDeclarations(decls: List[ProgramDeclaration]): Unit = decls match {

    case decl :: tail =>
      // sort head
      decl match {

        case d: IncludeDeclaration =>
          _includes += d


        case d: OpenDeclaration =>
          _opens += d


        case d: PackageDeclaration =>
          _packages.get(d.ident.path.head) match {
            case None => _packages += (d.ident.path.head -> d)
            case Some(previous) => _exceptions ::= MultipleDeclarationException(d, previous)
          }


        case d: TypeDeclaration =>
          _types.get(d.ident.baseId) match {
            case None => _types += (d.ident.baseId -> d)
            case Some(previous) => _exceptions ::= MultipleDeclarationException(d, previous)
          }


        case d: ComponentDeclaration =>
          _components.get(d.ident.baseId) match {
            case None => _components += (d.ident.baseId -> d)
            case Some(previous) => _exceptions ::= MultipleDeclarationException(d, previous)
          }


        case d: ConfigurationDeclaration =>
          _configurations.get(d.ident.baseId) match {
            case None => _configurations += (d.ident.baseId -> d)
            case Some(previous) => _exceptions ::= MultipleDeclarationException(d, previous)
          }

      }

      // continue sorting tail
      sortDeclarations(tail)

    case Nil =>
      /* we're done */

  }

  def apply(): (List[Action], List[NamespaceException]) = {

    // sort declarations
    sortDeclarations(pkg.declarations)

    // check for identifier shadowing wrt parent namespace 


    // for types
    for (current <- _types.values) {
      Try {
        pkg.parent.namespace.locate(current.ident)
      } match {
        case Success(previous) =>
          if (previous.location != current.location) {
            _exceptions ::= ShadowingDeclarationException(current, previous)
          } else {
            // it is a case of self-shadowing through import of the current
            // package in one of its parent packages
          }

        case Failure(e) => e match {
          case _: ParentNotSet => // root package reached without finding a declaration
          case _: NamespaceNotSet => // root package reached without finding a declaration
          case _: IdentNotFoundException[_] => // root package reached without finding a declaration
          case _ =>
            // unexpected error
            throw e

        }
      }
    }

    // for components
    for (current <- _components.values) {
      Try {
        pkg.parent.namespace.locate(current.ident)
      } match {
        case Success(previous) =>
          if (previous.location != current.location) {
            _exceptions ::= ShadowingDeclarationException(current, previous)
          } else {
            // it is a case of self-shadowing through import of the current
            // package in one of its parent packages
          }

        case Failure(e) => e match {
          case _: ParentNotSet => // root package reached without finding a declaration
          case _: NamespaceNotSet => // root package reached without finding a declaration
          case _: IdentNotFoundException[_] => // root package reached without finding a declaration
          case _ =>
            // unexpected error
            throw e

        }
      }
    }

    // import opened declarations
    _opens foreach { open => importDeclarations(open) }

    // build and set namespace 
    val ns = new PackageNamespace(parent, pkg, _packages, _types, _components, _configurations)

    pkg.namespace_(ns)

    // build next actions
    val pkgs = for (p <- ns.packages.values) yield BuildPackageNamespace(Some(ns), p)

    // build next actions
    val comps = for (p <- ns.components.values) yield BuildComponentNamespaceDown(Some(ns), p)

    // build next actions
    val confs = for (p <- ns.configurations.values) yield BuildConfigurationNamespace(Some(ns), p)

    _deferredActions = comps.toList ::: pkgs.toList ::: confs.toList

    // return next actions and exceptions
    (_deferredActions, _exceptions.reverse)

  }

  /** Imports declarations of an opened package in the local namespace. */
  private[this] def importDeclarations(open: OpenDeclaration): Unit = {
    sortDeclarations(
      open.target.declarations.filter(
        decl => decl match {
          case _: TypeDeclaration => true
          case _: ComponentDeclaration => true
          case _: ConfigurationDeclaration => true
          case _ => false // opens are not transitive on sub-packages
        }
      )
    )
  }
}


/** Builds the namespace of a component declaration. */
case class BuildComponentNamespaceDown(parent: Option[PackageNamespace], component: ComponentDeclaration) extends Action {

  override def toString = s"BuildComponentNamespaceDown( ${component.ident} )"

  def apply(): (List[Action], List[NamespaceException]) = {

    var _exceptions: List[NamespaceException] = Nil
    var _inputs: FlowDeclMapType = HashMap.empty
    var _outputs: FlowDeclMapType = HashMap.empty
    var _declarations: FlowDeclMapType = HashMap.empty
    var _deferredActions: List[Action] = Nil

    component.inputs foreach {
      input =>
        _declarations.get(input.ident.baseId) match {
          case None =>
            _inputs += (input.ident.baseId -> input)
            _declarations += (input.ident.baseId -> input)

          case Some(previous) => _exceptions ::= MultipleDeclarationException(input, previous)
        }
    }

    component.outputs foreach {
      output =>
        _declarations.get(output.ident.baseId) match {
          case None =>
            _outputs += (output.ident.baseId -> output)
            _declarations += (output.ident.baseId -> output)

          case Some(previous) => _exceptions ::= MultipleDeclarationException(output, previous)
        }
    }

    // test for identifier shadowing
    for (current <- _declarations.values) {
      Try {
        component.parent.namespace.locate(current.ident)
      } match {
        case Success(previous) => previous match {
          case Left(typeDecl) => _exceptions ::= ShadowingDeclarationException(current, typeDecl)
          case Right(fdecl) => _exceptions ::= ShadowingDeclarationException(current, fdecl)
        }
        case Failure(e) => e match {
          case _: ParentNotSet => // root package reached without finding a declaration
          case _: NamespaceNotSet => // root package reached without finding a declaration
          case _: IdentNotFoundException[_] => // root package reached without finding a declaration
          case _ =>
            // unexpected error
            throw e

        }
      }
    }

    // build and set namespace
    val ns = new ComponentNamespace(parent, component, _inputs, _outputs, _declarations)

    component.namespace_(ns)

    _deferredActions ::= BuildComponentNamespaceUp(component)
    _deferredActions ::= BuildBodyNamespace(Some(ns), component.body)

    // return next actions and exceptions
    (_deferredActions, _exceptions.reverse)
  }

}


case class BuildComponentNamespaceUp(component: ComponentDeclaration) extends Action {

  override def toString = s"BuildComponentNamespaceUp( ${component.ident} )"

  def apply(): (List[Action], List[NamespaceException]) = {

    var _exceptions: List[NamespaceException] = Nil

    val body = component.body

    component.inputs foreach {
      input =>
        body.namespace.definitions.get(input.ident.baseId) match {
          case None => /* intentionally empty */

          case Some(previous) => _exceptions ::= InputDefinitionException(input, component, previous)
        }
    }

    component.outputs foreach {
      output =>
        body.namespace.definitions.get(output.ident.baseId) match {
          case Some(_) => /* intentionally empty */

          case None => _exceptions ::= MissingDefinitionException(output, component)
        }
    }

    (Nil, _exceptions.reverse)
  }
}

/** Builds the namespace of a component body. */
case class BuildBodyNamespace(parent: Option[ComponentNamespace], body: ComponentBody) extends Action {

  override def toString = s"BuildBodyNamespace( ${parent.get.source.ident} )"

  def apply(): (List[Action], List[NamespaceException]) = {

    var _exceptions: List[NamespaceException] = Nil
    var _events: EventSetType = HashSet.empty
    var _locals: FlowDeclMapType = HashMap.empty
    var _definitions: FlowDefMapType = HashMap.empty
    var _callDefinitions: CallDefMapType = HashMap.empty

    if (body.events.nonEmpty && !body.parent.isPrimary) {
      _exceptions ::= EventsInNonPrimaryComponentException(body.parent.ident)
    } else if (body.parent.isPrimary && body.definitions.collect({ case fdef: CompCallFlowDefinition => fdef }).nonEmpty) {
      _exceptions ::= InstDefInPrimaryComponentException(body.parent.ident)
    }
    else {
      // add event symbols
      _events ++= body.events.map {
        _.baseId
      }
    }

    // add local flow symbols
    body.locals foreach {
      decl => {
        body.parent.namespace.declarations.get(decl.ident.baseId).orElse(
          _locals.get(decl.ident.baseId)) match {
          case None => _locals += (decl.ident.baseId -> decl)
          case Some(previous) => _exceptions ::= MultipleDeclarationException(decl, previous)
        }
      }
    }

    // checks for multiple flow definitions
    def doCheckFlow(d: FlowDefinition, ident: BasicFlowIdent): Unit = _definitions.get(ident.baseId) match {
      case None =>
        _definitions += (ident.baseId -> d)

      case Some(previous) => _exceptions ::= MultipleDefinitionException(ident, previous)
    }

    // checks for multiple component instance definitions
    def doCheckCall(d: CompCallFlowDefinition): Unit = _callDefinitions.get(d.rhs.instId) match {
      case None =>
        _callDefinitions += (d.rhs.instId -> d)

      case Some(_) => _exceptions ::= MultipleInstIdDeclarationException(d.rhs.instId, body.parent)
    }


    body.definitions foreach {

      case d@SingleFlowDefinition(ident, _) =>
        doCheckFlow(d, ident)

      case d@CompCallFlowDefinition(idents, _) =>
        idents.foreach(doCheckFlow(d, _))
        doCheckCall(d)

    }

    body.locals foreach {
      decl =>
        _definitions.get(decl.ident.baseId) match {
          case Some(_) =>
            /* intentionally empty */

          case None => _exceptions ::= MissingDefinitionException(decl, body.parent)
        }
    }

    // check for shadowing
    for (current <- _locals.values) {
      Try {
        body.parent.namespace.locate(current.ident)
      } match {
        case Success(previous) => previous match {
          case Left(typeDecl) => _exceptions ::= ShadowingDeclarationException(current, typeDecl)
          case Right(fdecl) => _exceptions ::= ShadowingDeclarationException(current, fdecl)
        }
        case Failure(e) => e match {
          case _: ParentNotSet => // root package reached without finding a declaration
          case _: NamespaceNotSet => // root package reached without finding a declaration
          case _: IdentNotFoundException[_] => // root package reached without finding a declaration
          case _ =>
            println("unexpected error: " + e)
            sys.exit(1)

        }
      }
    }

    // build and set namespace
    body.namespace_(new ComponentBodyNamespace(parent, body, _events, _locals, _definitions, _callDefinitions))

    (Nil, _exceptions.reverse)
  }

}

case class BuildConfigurationNamespace(
                                        parent: Option[PackageNamespace],
                                        source: ConfigurationDeclaration
                                      ) extends Action {

  override def toString = s"BuildConfigurationNamespace( ${parent.get.source.ident} )"

  def apply(): (List[Action], List[NamespaceException]) = {

    var _exceptions: List[NamespaceException] = Nil
    var _designSpace: DesignSpaceMapType = HashMap.empty

    // check for doublons
    source.designSpace.foreach {
      subst =>
        _designSpace.get(subst.instance) match {
          case None =>
            _designSpace += (subst.instance -> subst)

          case Some(previous) =>
            _exceptions ::= MultipleDefinitionException(subst, previous)
        }
    }

    // build and set namespace
    source.namespace_(new ConfigurationNamespace(parent, source, _designSpace))

    (Nil, _exceptions.reverse)

  }
}
