/** $Id: CheckIdents.scala 1433 2017-10-19 14:06:15Z kdelmas $
  *
  * @author $Author: kdelmas $
  **/

package namespace.dsl

import ast.dsl._

import scala.collection.immutable.{HashSet, Set}
import scala.util.{Failure, Success, Try}

/** Provides utility methods allowing to check if identifiers contained in expressions are declared in a given namespace. */
trait CheckIdents extends Action {

  /** Checks all symbols of some collection, in the context of
    * some program declaration, are all unique. 
    */
  def allUnique[D <: Declaration, E <: NamespaceException](
                                                            symbols: Iterable[DSLSymbol],
                                                            decl: D,
                                                            mkException: (DSLSymbol, D) => E): List[E] = {

    type Result = (Set[DSLSymbol], List[E])
    val zero: Result = (HashSet.empty, Nil)

    def doCheck(acc: Result, id: DSLSymbol): Result = {
      val (ids, exceptions) = acc
      if (ids.contains(id)) {
        (ids, mkException(id, decl) :: exceptions)
      } else {
        (ids + id, exceptions)
      }
    }

    symbols.foldLeft(zero)(doCheck)._2
  }

  /** Checks identifiers found in a TypeDeclaration for doublons. */
  def check(decl: TypeDeclaration): List[NamespaceException] = {
    allUnique(
      decl.values,
      decl,
      (s: DSLSymbol, decl: TypeDeclaration) => MultipleTypeValueDeclarationException(s, decl)
    )
  }

  /** Check type the identifier found in a FlowDeclartion is declared in the namespace. */
  def check(decl: FlowDeclaration): List[NamespaceException] = {
    check(decl.typeId)(decl.namespace)
  }


  /** Check if identifiers found in a FlowDefinition lhs and rhs are all declared. */
  def check(defi: FlowDefinition): List[NamespaceException] = defi match {
    case SingleFlowDefinition(lhs, rhs) =>
      check(lhs)(defi.namespace) :::
        check(rhs)(defi.namespace)


    case CompCallFlowDefinition(lhs, rhs) =>
      lhs.flatMap(id => check(id)(defi.namespace)) :::
        check(rhs)(defi.namespace)


  }

  /** Checks if the identifiers used in the expression are declared in the namespace. */
  def check(expr: TypeExpr)(implicit namespace: Namespace): List[NamespaceException] = checkRec(expr)

  /** Checks if the identifiers used in the expression are declared in the namespace. */
  def check(expr: FlowExpr)(implicit namespace: Namespace): List[NamespaceException] = checkRec(expr)

  /** Checks if the identifiers used in the expression are declared in the namespace. */
  def check(expr: BoolExpr)(implicit namespace: Namespace): List[NamespaceException] = checkRec(expr)

  /** Checks if the identifiers used in the expression are declared in the namespace. */
  def check(expr: CompExpr)(implicit namespace: Namespace): List[NamespaceException] = checkRec(expr)


  /** Recursively traverses the expression and calls namespace.locate on all identifiers. */
  private[this] def checkRec(expr: TypeExpr)(implicit namespace: Namespace): List[NamespaceException] =
    expr match {
      case BoolType => Nil
      case TupleType(kids) => kids.foldLeft(Nil: List[NamespaceException]) { (acc, current) => checkRec(current) ::: acc }.reverse
      case ident: TypeIdent =>
        Try {
          namespace.locate(ident)
        } match {
          case Success(_) => Nil

          case Failure(e@IdentNotFoundException(_, _)) => e :: Nil
          case Failure(e) => throw e
        }

    }


  /** Recursively traverses the expression and calls namespace.locate on all identifiers. */
  private[this] def checkRec(expr: FlowExpr)(implicit namespace: Namespace): List[NamespaceException] = expr match {

    case ident: FlowIdent =>
      Try {
        namespace.locate(ident)
      } match {
        case Success(_) => Nil
        case Failure(e@IdentNotFoundException(_, _)) => e :: Nil
        case Failure(e) => throw e // unexpected exception
      }


    case FlowUnion(kid1, kid2) =>
      checkRec(kid1) ::: checkRec(kid2)


    case FlowInter(kid1, kid2) =>
      checkRec(kid1) ::: checkRec(kid2)


    case FlowIte(kid1, kid2, kid3) =>
      checkRec(kid1) ::: checkRec(kid2) ::: checkRec(kid3)

  }

  /** Recursively traverses the expression and calls namespace.locate on all identifiers. */
  private[this] def checkRec(expr: BoolExpr)(implicit namespace: Namespace): List[NamespaceException] = expr match {

    case ident: BoolIdent =>
      Try {
        namespace.locate(ident)
      } match {
        case Success(_) => Nil
        case Failure(e@IdentNotFoundException(_, _)) => e :: Nil
        case Failure(e) => throw e
      }


    case Not(kid1) =>
      checkRec(kid1)

    case And(kid1, kid2) =>
      checkRec(kid1) ::: checkRec(kid2)

    case Or(kid1, kid2) =>
      checkRec(kid1) ::: checkRec(kid2)

    case Iff(kid1, kid2) =>
      checkRec(kid1) ::: checkRec(kid2)

    case Implies(kid1, kid2) =>
      checkRec(kid1) ::: checkRec(kid2)

    case FlowContains(kid1, kid2) =>
      checkRec(kid1) ::: checkRec(kid2)

    case FlowEq(kid1, kid2) =>
      checkRec(kid1) ::: checkRec(kid2)

    case FlowNeq(kid1, kid2) =>
      checkRec(kid1) ::: checkRec(kid2)

    case PopCount(kids, _, _) =>
      kids.flatMap(kid => checkRec(kid))
  }


  /** Recursively traverses the expression and calls namespace.locate on all identifiers. */
  private[this] def checkRec(expr: CompExpr)(implicit namespace: Namespace): List[NamespaceException] = expr match {

    case CompCall(compId, _, args) =>
      val compIdCheck = Try {
        namespace.locate(compId)
      } match {
        case Success(_) => Nil
        case Failure(e@IdentNotFoundException(_, _)) => e :: Nil
        case Failure(e) => throw e
      }
      // REMARK instId unicity is checked in the enclosing namespace, a component namespace
      compIdCheck ::: args.flatMap(arg => checkRec(arg))
  }
}


/** Checks identifiers used in all ProgramDeclarations of the given package. */
case class CheckIdentsPackage(pkg: PackageDeclaration) extends CheckIdents {

  def apply(): (List[Action], List[NamespaceException]) = {

    implicit val namespace: PackageNamespace = pkg.namespace

    // check type declarations (produce no subsequent actions, just exceptions )
    val typeExceptions = namespace.types.values.flatMap {check}.toList

    // check components and subpackages
    val actions =
      namespace.components.values.map {
        CheckIdentsComponent
      }.toList :::
        namespace.packages.values.map {
          CheckIdentsPackage
        }.toList :::
        namespace.configurations.values.map {
          CheckIdentsConfiguration
        }.toList
    (actions, typeExceptions)

  }

}


/** Checks identifiers used in a ComponentDeclaration. */
case class CheckIdentsComponent(comp: ComponentDeclaration) extends CheckIdents {

  def apply(): (List[Action], List[NamespaceException]) = {

    val exceptions =
      comp.namespace.inputs.values.flatMap {check}.toList :::
        comp.namespace.outputs.values.flatMap {check}.toList

    val actions = CheckIdentsBody(comp.body) :: Nil

    (actions, exceptions)
  }
}


/** Checks identifiers used in a ComponentBody. */
case class CheckIdentsBody(body: ComponentBody) extends CheckIdents {

  def apply(): (List[Action], List[NamespaceException]) = {
    implicit val namespace: ComponentBodyNamespace = body.namespace
    val locals = namespace.locals.values.flatMap {check}.toList
    val defs = namespace.definitions.values.flatMap {check}.toList

    (CheckCyclesComponent(body.parent) :: Nil, locals ::: defs)

  }
}


/** Checks identifiers used in a ComponentBody. */
case class CheckIdentsConfiguration(config: ConfigurationDeclaration) extends CheckIdents {

  def apply(): (List[Action], List[NamespaceException]) = {

    // try to locate the root component with respect to configuration namespace
    val (rootCompOpt, rootExceptions) = Try {
      config.namespace.locate(config.rootCompId)
    } match {
      case Success(d) => (Some(d), Nil)
      case Failure(e@IdentNotFoundException(_, _)) => (None, e :: Nil)
      case Failure(e) => throw e // unexpected exception
    }


    rootCompOpt match {

      // abort if root comp not found
      case None =>
        // no new actions
        (Nil, rootExceptions)

      // if root comp was found, check all remaining names relative to its body namespace
      case Some(rootComp) =>

        // check that the root component has no inputs
        val rootCompNotClosedExceptions = rootComp.inputs.length match {
          case 0 => Nil
          case _ => RootComponentNotClosed(config) :: Nil
        }

        // all subsequent checks are relative to the root component's namespace
        val rootNS = rootComp.body.namespace


        // check failure condition expression
        val fcExceptions = check(config.failureCondition)(rootNS)

        //  check the design space declaration
        val designSpaceExceptions = config.namespace.designSpace.values.map {

          substitution => {

            // try to locate the rhs of the substitution the rootComponent
            val instCheck = Try {
              rootNS.locateDef(substitution.instance)
            } match {

              // ill defined case (should use special scalaz NEList in compinstance idents.
              case Success(Nil) =>
                println("should never happen since instance idents are non empty lists")
                Nil



              // check it is a primary component
              case Success(topCall :: _) =>
                val topComp = topCall.namespace.locate(topCall.rhs.compId)
                if (!topComp.isPrimary) {
                  NonPrimarySubstitutionException(substitution.instance) :: Nil
                } else {
                  Nil
                }


              case Failure(e@IdentNotFoundException(_, _)) =>
                e :: Nil


              case Failure(e) =>
                throw e // unexpected exception

            }

            // try to locate each alternative relative to config namespace
            val altsCheck = substitution.alternatives.flatMap {
              alt => {
                // try to locate the component specified as an alternative
                Try {
                  config.namespace.locate(alt.compId)
                } match {
                  case Success(d) =>
                    // check that lambda parameter values are associated with valid event names
                    val exceptions1 = alt.lambdaValues.flatMap(x => {
                      x._1 match {
                        case id@BoolIdent(pointed: PointedDSLSymbol) =>
                          val openId = pointed.toStringList.map(s => StdDSLSymbol(Symbol(s)))
                          if (!checkPointedEventIdent(openId, d.body.namespace))
                            IdentNotFoundException(id, d.body.namespace) :: Nil
                          else
                            Nil
                        case id@BoolIdent(_) =>
                          Try {
                            d.body.namespace.locate(id)
                          } match {
                            case Success(_) => Nil
                            case Failure(e@IdentNotFoundException(_, _)) => e :: Nil
                            case Failure(e) => throw e // unexpected exception
                          }
                      }

                    })

                    // check that each event has been given a lambda value
                    val idSet = Set.empty ++ alt.lambdaValues.map(x => x._1)

                    var exceptions2: List[NamespaceException] = Nil
                    if (d.isPrimary) {
                      d.body.events.foreach(evtId => {
                        if (!idSet.contains(evtId)) exceptions2 ::= MissingLambdaDefinitionException(evtId, alt)
                      })
                    } else {
                      collectAllSubEvents(d, CompInstanceIdent(Nil)).foreach(evtId =>
                        if (!idSet.contains(evtId))
                          exceptions2 ::= MissingLambdaDefinitionException(evtId, alt))
                    }

                    exceptions1 ::: exceptions2
                  case Failure(e@IdentNotFoundException(_, _)) => e :: Nil
                  case Failure(e) => throw e // unexpected exception
                }
              }
            }

            instCheck ::: altsCheck
          }

        }.toList.flatten

        // no new actions
        (Nil, rootCompNotClosedExceptions ::: fcExceptions ::: designSpaceExceptions)
    }

  }

  def checkPointedEventIdent(id: List[DSLSymbol], ns: ComponentBodyNamespace): Boolean = {
    id match {
      case Nil => false
      case h :: Nil =>
        Try {
          ns.locate(BoolIdent(h))
        } match {
          case Success(_) => true
          case Failure(_: IdentNotFoundException[_]) => false
          case Failure(e) => throw e
        }
      case h :: t =>
        Try {
          ns.locateDef(CompInstanceIdent(List(h)))
        } match {
          case Success(l) =>
            val nextComp = ns.locate(l.head.rhs.compId)
            checkPointedEventIdent(t, nextComp.body.namespace)
          case Failure(_: IdentNotFoundException[_]) => false
          case Failure(e) => throw e
        }
    }
  }

  def collectAllSubEvents(c: ComponentDeclaration, inst: CompInstanceIdent): List[BoolIdent] = {
    if(inst.path.nonEmpty){
      val thisEvts = c.body.events.map(e => BoolIdent(PointedDSLSymbol(Symbol(inst.path.mkString("", ".", ".") + e.baseId.name))))
      thisEvts ++ c.body.definitions.collect({
        case CompCallFlowDefinition(_, rhs) =>
          collectAllSubEvents(c.namespace.locate(rhs.compId), CompInstanceIdent(inst.path :+ rhs.instId))
      }).flatten
    }else{
      val thisEvts = c.body.events
      thisEvts ++ c.body.definitions.collect({
        case CompCallFlowDefinition(_, rhs) =>
          collectAllSubEvents(c.namespace.locate(rhs.compId), CompInstanceIdent(rhs.instId :: Nil))
      }).flatten
    }
  }
}

/** Checks cyclic dependencies in identifiers used in a ComponentDeclaration. */
case class CheckCyclesComponent(comp: ComponentDeclaration) extends Action {

  def apply(): (List[Action], List[NamespaceException]) = {
    val depGraph = new FlowDepGraph(comp)
    val exceptions = depGraph.check
    comp.namespace.flowDepGraph_(depGraph)
    val actions = Nil
    (actions, exceptions)
  }
}


