/** $Id$
  *
  * @author $Author$
  **/

package namespace.dsl

import ast.dsl._

import scala.util.{Failure, Success, Try}
import scalaz.\/

/** Trampoline that builds and checks the namespace of a program. 
  *   - checks for cyclic component dependencies
  * */
object CheckTypes extends (PackageDeclaration => \/[NamespaceExceptions, List[NamespaceWarning]]) {

  def apply(pkg: PackageDeclaration): \/[NamespaceExceptions, List[NamespaceWarning]] = {
    for {
      exceptions <- RunAction(CheckTypesPackage(pkg))
    } yield {
      exceptions
    }
  }

}

/** Check cyclic component definitions from the root package. */
case class CheckTypesPackage(pkg: PackageDeclaration) extends Action {

  def apply(): (List[Action], List[NamespaceException]) = {
    val actions =
      pkg.namespace.packages.values.map {
        CheckTypesPackage
      }.toList :::
        pkg.namespace.components.values.map {
          CheckTypesComponent
        }.toList :::
        pkg.namespace.configurations.values.map {
          CheckTypesConfiguration
        }.toList
    (actions, Nil)
  }

}

/** Check cyclic component definitions from the root package. */
case class CheckTypesComponent(compDecl: ComponentDeclaration) extends Action {

  private[this] def check(flowDef: FlowDefinition): List[NamespaceException] = {
    Try {
      flowDef.getType
    } match {
      case Success(_) => Nil
      case Failure(e) => e match {
        case e: TypingException => e :: Nil
        case _ => throw e // unexpected failure
      }
    }
  }

  def apply(): (List[Action], List[NamespaceException]) = {
    val exceptions = compDecl.body.definitions.foldLeft(Nil: List[NamespaceException]) {
      (acc, current) => {
        check(current) ::: acc
      }
    }.reverse
    (Nil, exceptions)
  }
}


/** Check cyclic component definitions from the root package. */
case class CheckTypesConfiguration(configDecl: ConfigurationDeclaration) extends Action {


  def apply(): (List[Action], List[NamespaceException]) = {

    val rootNamespace = configDecl.namespace.locate(configDecl.rootCompId).namespace

    // check the failure condition
    val fcExceptions = Try {
      configDecl.failureCondition.getType(rootNamespace)
    } match {
      case Success(_) => Nil
      case Failure(e) => e match {
        case e: TypingException => e :: Nil
        case _ => throw e // unexpected failure
      }
    }

    // check that each substitution are correctly typed (number of arguments are the same and their types are matching
    val designSpaceExceptions = configDecl.designSpace.foldLeft(Nil: List[NamespaceException]) {
      (acc, subst) => {
        val compDef = rootNamespace.locateDef(subst.instance).head
        val compDecl = compDef.namespace.locate(compDef.rhs.compId)
        val inT = TupleType(compDecl.inputs.map(_.ident.getType(compDecl.namespace)))
        val outT = TupleType(compDecl.outputs.map(_.ident.getType(compDecl.namespace)))

        val iniDecl = configDecl.namespace.locate(subst.init.compId)
        val initException = {
          if (compDecl != iniDecl)
            WrongDefaultAlternative(subst.instance, subst.init, compDecl.ident) :: Nil
          else
            Nil
        }

        val altExceptions = subst.alternatives.foldLeft(Nil: List[NamespaceException]) {
          (acc, alt) => {
            val altDecl = configDecl.namespace.locate(alt.compId)
            val altInT = TupleType(altDecl.inputs.map(_.ident.getType(altDecl.namespace)))
            val altOutT = TupleType(altDecl.outputs.map(_.ident.getType(altDecl.namespace)))
            if (inT != altInT || outT != altOutT) {
              WrongTypeAlternative(compDecl, inT, outT, altDecl, altInT, altOutT) :: acc
            } else {
              acc
            }
          }
        }.reverse
        altExceptions ::: acc ::: initException
      }
    }.reverse

    (Nil, fcExceptions ::: designSpaceExceptions)
  }
}
