/** $Id: CompInstanceDepGraph.scala 1433 2017-10-19 14:06:15Z kdelmas $
  *
  * @author $Author: kdelmas $
  **/

package namespace.dsl

import ast.dsl._
import util.graph.DepGraphE

import scala.collection.immutable.{HashMap, Map}

/** Class used to build a topological ordering of component of all instance under in a root component . */
class CompInstanceDepGraph(rootComp: ComponentDeclaration) extends DepGraphE {

  /** Node data type. */
  type A = CompInstanceIdent

  /** Node cache */
  private[this] var _dataMap: Map[A, Node] = HashMap.empty

  /** Gets a node from the cache or creates it if not found using a side effect. */
  private[this] def getNode(a: A): Node = _dataMap.get(a) match {
    case Some(v) =>
      v
    case None =>
      val n = Node(a)
      _dataMap += (a -> n)
      n
  }

  /** This method detects cyclic dependencies and returns exceptions, if no exceptions are returned, the graph is topologically sorted. */
  def check: List[NamespaceException] = {

    // traverse the node instance dependency graph by traversing the call tree from the root component
    this.buildGraph(rootComp)

    // check for cycles
    val exceptions = this.sccs.flatMap {
      comp =>
        comp.size match {
          case 0 | 1 => Nil
          case _ => CyclicDependencyException(comp map {
            _.data
          }) :: Nil
        }
    }

    // do topological sort if graph is acyclic
    if (exceptions.isEmpty)
      this.topsort()

    exceptions
  }


  /** Traverses the call tree of the root component using a depth first search to build the graph. */
  private[this] def buildGraph(rootComp: ComponentDeclaration): Unit = {
    buildGraphRec(rootComp, Nil)
  }

  private[this] def buildGraphRec(comp: ComponentDeclaration, instancePath: List[DSLSymbol]): Unit = {
    val currentIdent = CompInstanceIdent(instancePath.reverse)
    val currentNode = getNode(currentIdent)
    comp.namespace.flowDepGraph.compCallsInOrder.toList.reverse.foldLeft[Option[Node]](None) {
      (pred, call) => {
        val newInstancePath = call.rhs.instId :: instancePath
        val calledComp = comp.namespace.locate(call.rhs.compId)
        val calledInstId = CompInstanceIdent(newInstancePath.reverse)
        val calledInstNode = getNode(calledInstId)

        // add sequential call dependency between consecutive calls in a given component
        if (pred.isDefined)
          currentNode.addPred(pred.get)
        currentNode.addPred(calledInstNode)
        buildGraphRec(calledComp, newInstancePath)
        Some(currentNode)
      }
    }
  }
}

