/** $Id: ComponentDepGraph.scala 1433 2017-10-19 14:06:15Z kdelmas $
  *
  * @author $Author: kdelmas $
  **/

package namespace.dsl

import ast.dsl._
import util.graph.DepGraphE

import scala.collection.immutable.{HashMap, Map}

/** Class used to build a topological ordering of component of all instance under in a root component . */
class ComponentDepGraph(rootPkg: PackageDeclaration) extends DepGraphE {

  /** Node data type. */
  type A = CompIdent

  /** Node cache */
  private[this] var _dataMap: Map[A, Node] = HashMap.empty

  /** Gets a node from the cache or creates it if not found using a side effect. */
  private[this] def getNode(a: A): Node = _dataMap.get(a) match {
    case Some(v) => v


    case None =>
      val n = Node(a)
      _dataMap += (a -> n)
      n
  }

  /** This method detects cyclic dependencies and returns exceptions, if no exceptions are returned, the graph is topologically sorted. */
  def check: List[NamespaceException] = {

    // recursively traverse the rootPkg to build the graph
    this.buildGraph(rootPkg)

    // check for cycles
    val exceptions = this.sccs.flatMap {
      comp =>
        comp.size match {
          case 0 | 1 => Nil
          case _ => CyclicDependencyException(comp map {
            _.data
          }) :: Nil
        }
    }

    // do topological sort if graph is acyclic
    if (exceptions.isEmpty)
      this.topsort()

    exceptions
  }


  /** Traverses the call tree of the root component using a depth first search to build the graph. Assumes FlowDepGraph is already built for each component. */
  private[this] def buildGraph(pkg: PackageDeclaration): Unit = {
    // process components in the current pkg declaration namespace
    pkg.namespace.components.values foreach { comp => this.buildGraph(comp) }

    // visit subpackages
    pkg.namespace.packages.values foreach { pkg => this.buildGraph(pkg) }
  }

  private[this] def buildGraph(comp: ComponentDeclaration): Unit = {
    val compId = comp.absoluteIdent
    val currentNode = this.getNode(compId)
    comp.namespace.flowDepGraph.compCallsInOrder foreach {
      call => {
        val calledCompId = call.namespace.locate(call.rhs.compId).absoluteIdent
        val depNode = this.getNode(calledCompId)
        currentNode.addPred(depNode)
      }
    }
  }

}

