/** $Id: FlowDepGraph.scala 1433 2017-10-19 14:06:15Z kdelmas $
  *
  * @author $Author: kdelmas $
  * */

package namespace.dsl

import ast.dsl._
import util.graph.DepGraphE

import scala.collection.immutable.{HashMap, HashSet, Map, Set}
import scala.util.{Either, Left, Right}

class FlowDepGraph(comp: ComponentDeclaration) extends DepGraphE {

  /** Node data type. */
  type A = Either[BoolIdent, FlowIdent]

  /** Node cache */
  private[this] var _dataMap: Map[A, Node] = HashMap.empty

  /** Gets a node from the cache or creates it if not found using a side effect. */
  private[this] def getNode(a: A): Node = _dataMap.get(a) match {
    case Some(v) => v

    case None =>
      val n = Node(a)
      _dataMap += (a -> n)
      n

  }


  /** This method detects cyclic dependencies and returns exceptions, if no exceptions are triggered, the graph is topologically sorted. */
  def check: List[NamespaceException] = {

    // build node predecessor links
    comp.body.definitions foreach {
      buildGraph
    }

    // check for cycles
    val exceptions = this.sccs.flatMap {

      comp =>
        comp.size match {

          // single node in component means no cycle
          case 0 | 1 =>
            Nil

          // filter out boolean idents since they cannot be given definitions
          case _ =>
            CyclicDependencyException(comp.collect({ case Node(Right(id)) => id })) :: Nil

        }
    }

    // do topological sort if graph is acyclic
    if (exceptions.isEmpty) {
      this.topsort()
    }

    // return exceptions
    exceptions
  }

  /** Iterates on all identifiers used in the component in topological order, be they events, type values or local flows. */
  def allIdentsInOrder: Iterator[Either[BoolIdent, FlowIdent]] = inOrder map {
    _.data
  }

  /** Iterates on all flow identifiers used in the component in order, be type values or local flows. */
  def allFlowsInOrder: Iterator[FlowIdent] = inOrder collect { case Node(Right(f)) => f }

  /** Iterates on all basic flow identifiers defined in the component (excludes type values). */
  def compFlowsInOrder: Iterator[BasicFlowIdent] = inOrder collect {
    case Node(Right(f@BasicFlowIdent(_))) if comp.body.namespace.locate(f).isRight => f
  }

  /** Iterates on all flow definitions found in the component. */
  def compDefsInOrder: Iterator[FlowDefinition] = inOrder collect {
    case Node(Right(id@BasicFlowIdent(sym))) if comp.body.namespace.definitions.contains(sym) => comp.body.namespace.locateDef(id).right.get
  }


  /** Iterates in the calls to sub components in topological order. */
  def compCallsInOrder: Iterator[CompCallFlowDefinition] = {

    type AccT = (Set[DSLSymbol], List[CompCallFlowDefinition])

    val zero: AccT = (HashSet.empty, Nil)

    def loopF(acc: AccT, current: FlowDefinition): AccT = {
      val (seen, res) = acc
      current match {
        case defi@CompCallFlowDefinition(_, CompCall(_, instId, _)) if !(seen contains instId) =>
          (seen + instId, defi :: res)

        case _ =>
          acc
      }
    }

    val (_, res) = compDefsInOrder.foldLeft(zero)(loopF)

    res.reverseIterator
  }

  private[this] def buildGraph(d: FlowDefinition): Unit = d match {

    case SingleFlowDefinition(id, lhs) =>
      buildGraphRec(lhs)(getNode(Right(id)) :: Nil)

    case CompCallFlowDefinition(ids, lhs) =>
      buildGraphRec(lhs)(ids.map(id => getNode(Right(id))))

  }

  final private[this] def buildGraphRec(expr: CompCall)(implicit nodes: List[Node]): Unit = expr.args foreach {
    buildGraphRec
  }


  final private[this] def buildGraphRec(expr: FlowExpr)(implicit nodes: List[Node]): Unit = expr match {

    case ident: FlowIdent =>
      val n = getNode(Right(ident))
      nodes foreach {
        _.addPred(n)
      }

    case FlowUnion(kid1, kid2) =>
      buildGraphRec(kid1)
      buildGraphRec(kid2)

    case FlowInter(kid1, kid2) =>
      buildGraphRec(kid1)
      buildGraphRec(kid2)

    case FlowIte(kid1, kid2, kid3) =>
      buildGraphRec(kid1)
      buildGraphRec(kid2)
      buildGraphRec(kid3)
  }

  final private[this] def buildGraphRec(expr: BoolExpr)(implicit nodes: List[Node]): Unit = expr match {

    case ident: BoolIdent =>
      val n = getNode(Left(ident))
      nodes foreach {
        _.addPred(n)
      }

    case And(kid1, kid2) =>
      buildGraphRec(kid1)
      buildGraphRec(kid2)

    case FlowContains(kid1, kid2) =>
      buildGraphRec(kid1)
      buildGraphRec(kid2)

    case FlowEq(kid1, kid2) =>
      buildGraphRec(kid1)
      buildGraphRec(kid2)

    case FlowNeq(kid1, kid2) =>
      buildGraphRec(kid1)
      buildGraphRec(kid2)

    case Iff(kid1, kid2) =>
      buildGraphRec(kid1)
      buildGraphRec(kid2)

    case Implies(kid1, kid2) =>
      buildGraphRec(kid1)
      buildGraphRec(kid2)

    case Not(kid1) =>
      buildGraphRec(kid1)

    case Or(kid1, kid2) =>
      buildGraphRec(kid1)
      buildGraphRec(kid2)

    case PopCount(kids, _, _) =>
      kids foreach {
        buildGraphRec
      }
  }

}
