/** $Id: Namespace.scala 1433 2017-10-19 14:06:15Z kdelmas $
  *
  * @author $Author: kdelmas $
  **/

package namespace.dsl

import ast.dsl._

import scala.annotation.tailrec
import scala.util.{Either, Left, Right}


/** Base type for all DSL namespace management classes. 
  * *
  * All Identifiers are resolved relatively to the lexical scope associated with the namespace object on which methods are called.
  *
  */
sealed trait Namespace {

  /** Type of the ASTNode represented by this namespace. */
  type S <: ASTNode

  /** The ast node described by this namespace. */
  val source: S

  /** The parent namespace (Optional). */
  val parentOpt: Option[Namespace]

  /** Gets the root namespace from any namespace. */
  def root: Namespace = parentOpt match {
    case Some(p) => p.root
    case None => this
  }

  /** True iff the namespace is the root namespace, ie has no parent. */
  def isRoot: Boolean = parentOpt.isDefined

  /** Gets a type declaration from its identifier (triggers exception if not found). */
  def locate(ident: TypeIdent): TypeDeclaration

  /** Gets a package declaration from its identifier (triggers exception if not found). */
  def locate(ident: PackageIdent): PackageDeclaration

  /** Gets a component declaration from its identifier (triggers exception if not found). */
  def locate(ident: CompIdent): ComponentDeclaration

  /** Gets a boolean identifier declaration (triggers exception if not found). */
  def locate(ident: BoolIdent): BoolIdent

  /** Gets a data flow decalaration from its identifier (triggers exception if not found). */
  def locate(ident: FlowIdent): Either[TypeDeclaration, FlowDeclaration]

  /** Gets the definition of a flow identifier, either a flow definition
    * or a type declaration which contains the identifier (triggers exception if not found). */
  def locateDef(ident: FlowIdent): Either[TypeDeclaration, FlowDefinition]

  /** Gets the declaration of a type identifier (triggers exception if not found). */
  def locate(ident: ConfigIdent): ConfigurationDeclaration

  /** Gets the call stack defining a component instance (relative to current namespace) from its identifier (triggers exception if not found). */
  def locateDef(ident: CompInstanceIdent): List[CompCallFlowDefinition]

}

/** Namespace management in a package scope.
  * */
class PackageNamespace(
                        val parentOpt: Option[PackageNamespace],
                        val source: PackageDeclaration,
                        val packages: PackageMapType,
                        val types: TypeMapType,
                        val components: ComponentMapType,
                        val configurations: ConfigurationMapType
                      ) extends Namespace with ComponentDepGraphField {

  type S = PackageDeclaration

  /** Locates a package declaration. */
  def locate(ident: PackageIdent): PackageDeclaration = {
    val path: List[DSLSymbol] = ident.path
    locatePackageDown(path, this).getOrElse(parentOpt.map(_.locate(ident)).getOrElse(throw IdentNotFoundException(ident, this)))
  }

  /** Recursively locates a subpackage in the given context. */
  @tailrec
  final private[this] def locatePackageDown(path: List[DSLSymbol], context: PackageNamespace): Option[PackageDeclaration] = path match {
    case Nil => Some(context.source)
    case head :: tail => context.packages.get(head) match {
      case Some(pkg) => locatePackageDown(tail, pkg.namespace)
      case None => None
    }
  }

  /** Locates a type declaration. */
  def locate(ident: TypeIdent): TypeDeclaration =
    locateDown(ident, this).getOrElse(parentOpt.map(_.locate(ident)).getOrElse(throw IdentNotFoundException(ident, this)))


  /** Recursively locates a type declaration in the given context. */
  private[this] def locateDown(ident: TypeIdent, context: PackageNamespace): Option[TypeDeclaration] = {
    locate(ident.pkgId).namespace.types.get(ident.baseId)
  }

  /** Locates an operator declaration. */
  def locate(ident: CompIdent): ComponentDeclaration = {
    locateDown(ident, this).getOrElse(parentOpt.map(_.locate(ident)).getOrElse(throw IdentNotFoundException(ident, this)))
  }

  /** Recursively locates an operator declaration in the given context. */
  private[this] def locateDown(ident: CompIdent, context: PackageNamespace): Option[ComponentDeclaration] = {
    locate(ident.pkgId).namespace.components.get(ident.baseId)
  }

  /** Locate a boolean ident declaration (must fail because they are local to ComponentBody. */
  def locate(ident: BoolIdent): BoolIdent = throw MeaninglessLocate(ident,source)


  /** Locates a flow declaration (which can only be a type value constructor since we are in a package namespace). */
  def locate(ident: FlowIdent): Either[TypeDeclaration, FlowDeclaration] = {
    locateDown(ident, this).map(Left(_)).getOrElse(parentOpt.map(_.locate(ident)).getOrElse(throw IdentNotFoundException(ident, this)))
  }

  /** Recursively locates a stream declaration in the given context. */
  private[this] def locateDown(ident: FlowIdent, context: PackageNamespace): Option[TypeDeclaration] = ident match {

    case BasicFlowIdent(baseId) =>
      // search for all types that define this symbol in the current context
      val matchingTypes = context.types.values.filter(d => d.values.contains(baseId)).toList
      // return only if single match, otherwise there is ambiguity and we cannot know what to return
      if (matchingTypes.size == 1) Some(matchingTypes.head) else None

    case TypeValueFlowEmpty(typeId) =>
      Some(this.locate(typeId))

    case TypeValueFlowIdent(typeId, valueId) =>
      val typeDef = this.locate(typeId)
      if (typeDef.values.contains(valueId)) Some(typeDef) else None

    case InstFlowIdent(_, _) =>
      throw  MeaninglessLocate(ident,source)

  }

  /** Locates a type value or flow definition (which can only be a type value constructor since we are in a package namespace). */
  def locateDef(ident: FlowIdent): Either[TypeDeclaration, FlowDefinition] =
    locateDown(ident, this).map(Left(_)).getOrElse(parentOpt.map(_.locateDef(ident)).getOrElse(throw IdentNotFoundException(ident, this)))


  /** Locates a configuration declaration. */
  def locate(ident: ConfigIdent): ConfigurationDeclaration =
    locateDown(ident, this).getOrElse(parentOpt.map(_.locate(ident)).getOrElse(throw IdentNotFoundException(ident, this)))

  /** Recursively locates an operator declaration in the given context. */
  private[this] def locateDown(ident: ConfigIdent, context: PackageNamespace): Option[ConfigurationDeclaration] =
    locate(ident.pkgId).namespace.configurations.get(ident.baseId)

  /** Gets a component instance definition from its identifier (triggers exception if not found). */
  def locateDef(id: CompInstanceIdent): List[CompCallFlowDefinition] = {
    throw  MeaninglessLocate(id,source)
  }

}

/** Namespace management in a component interface.
  * */
class ComponentNamespace(
                          val parentOpt: Option[PackageNamespace],
                          val source: ComponentDeclaration,
                          val inputs: FlowDeclMapType,
                          val outputs: FlowDeclMapType,
                          val declarations: FlowDeclMapType
                        ) extends Namespace with FlowDepGraphField with CompInstanceDepGraphField with UsedTypesSetField {

  type S = ComponentDeclaration

  /** Locates a package declaration. */
  def locate(ident: PackageIdent): PackageDeclaration = source.parent.namespace.locate(ident)

  /** Locates a type declaration. */
  def locate(ident: TypeIdent): TypeDeclaration = source.parent.namespace.locate(ident)

  /** Locates a type declaration. */
  def locate(ident: CompIdent): ComponentDeclaration = source.parent.namespace.locate(ident)

  /** Locate a boolean ident declaration (must fail because they are local to ComponentBody. */
  def locate(ident: BoolIdent): BoolIdent = throw  MeaninglessLocate(ident,source)

  /** Locates a constant or stream declaration. */
  def locate(ident: FlowIdent): Either[TypeDeclaration, FlowDeclaration] = locateDown(ident, this) match {
    case Some(decl) => decl
    case None => source.parent.namespace.locate(ident) // search for a type constructor above
  }


  /** Recursively locates a stream declaration in the given context. */
  private[this] def locateDown(ident: FlowIdent, context: ComponentNamespace): Option[Either[TypeDeclaration, FlowDeclaration]] = ident match {
    // look for an component interface flow
    case BasicFlowIdent(baseId) => context.declarations.get(baseId).map(Right(_))

    // look for a type constructor value in the surrounding package
    case ident@TypeValueFlowEmpty(_) => Some(source.parent.namespace.locate(ident))

    // look for a type constructor value in the surrounding package
    case ident@TypeValueFlowIdent(_, _) => Some(source.parent.namespace.locate(ident))

    // instance relative flows are only meaningfull in component bodies, fail.
    case ident@InstFlowIdent(baseId, instId) => instId.path.size match {
      case 0 => context.declarations.get(baseId).map(Right(_))
      case _ => throw  MeaninglessLocate(ident,source)
    }
  }

  /** Locates a flow definition (can only be in a type declaration since we are at the component interface level which does not have definitions) . */
  def locateDef(ident: FlowIdent): Either[TypeDeclaration, FlowDefinition] = source.parent.namespace.locateDef(ident)

  /** Locates a configuration declaration. */
  def locate(ident: ConfigIdent): ConfigurationDeclaration = source.parent.namespace.locate(ident)

  /** Gets a component instance definition from its identifier (triggers exception if not found). */
  def locateDef(ident: CompInstanceIdent): List[CompCallFlowDefinition] = source.body.namespace.locateDef(ident)

}

/** Namespace management in a component body.
  * */
class ComponentBodyNamespace(
                              val parentOpt: Option[ComponentNamespace],
                              val source: ComponentBody,
                              val events: EventSetType,
                              val locals: FlowDeclMapType,
                              val definitions: FlowDefMapType,
                              val callDefinitions: CallDefMapType
                            ) extends Namespace with UsedTypesSetField {

  type S = ComponentBody

  /** Locates a package declaration. */
  def locate(ident: PackageIdent): PackageDeclaration = source.parent.namespace.locate(ident)

  /** Locates a type declaration. */
  def locate(ident: TypeIdent): TypeDeclaration = source.parent.namespace.locate(ident)

  /** Locates an operator declaration. */
  def locate(ident: CompIdent): ComponentDeclaration = source.parent.namespace.locate(ident)

  /** Locate a boolean ident declaration. */
  def locate(ident: BoolIdent): BoolIdent = if (events.contains(ident.baseId)) {
    ident
  } else {
    throw IdentNotFoundException(ident, this)
  }

  /** Locates a constant or stream declaration. */
  def locate(ident: FlowIdent): Either[TypeDeclaration, FlowDeclaration] = locateDown(ident, this) match {
    case Some(decl) => decl
    case None => source.parent.namespace.locate(ident)
  }


  /** Recursively locates a stream declaration in the given context. */
  private[this] def locateDown(ident: FlowIdent, context: ComponentBodyNamespace): Option[Either[TypeDeclaration, FlowDeclaration]] = ident match {

    case BasicFlowIdent(baseId) => context.locals.get(baseId).map(Right(_))

    // return none to look for a type constructor value in the surrounding package
    case TypeValueFlowEmpty(_) => None

    // return none to look for a type constructor value in the surrounding package
    case TypeValueFlowIdent(_, _) => None

    case ident@InstFlowIdent(baseId, instId) => locateDef(instId) match {
      case topCall :: _ =>
        val topComp = topCall.namespace.locate(topCall.rhs.compId)
        val fakeId = BasicFlowIdent(baseId)
        fakeId.location_(ident.location)
        Some(topComp.body.namespace.locate(fakeId))

      case Nil =>
        context.locals.get(baseId).map(Right(_))

    }
  }

  /** Locates a constant or stream definition. */
  def locateDef(ident: FlowIdent): Either[TypeDeclaration, FlowDefinition] = locateDefDown(ident, this) match {
    case Some(decl) => decl
    case None => source.parent.namespace.locateDef(ident) // search for a type constructor in the surrounding package
  }

  /** Recursively locates a stream declaration in the given context. */
  private[this] def locateDefDown(ident: FlowIdent, context: ComponentBodyNamespace): Option[Either[TypeDeclaration, FlowDefinition]] = ident match {
    // look for an component interface flow 
    case BasicFlowIdent(baseId) => context.definitions.get(baseId).map(Right(_))

    // return none to look for a type constructor value in the surrounding package
    case TypeValueFlowEmpty(_) => None

    // return non to look for a type constructor value in the surrounding package
    case TypeValueFlowIdent(_, _) => None

    case ident@InstFlowIdent(baseId, instId) => locateDef(instId) match {
      case topCall :: _ =>
        val topComp = topCall.namespace.locate(topCall.rhs.compId)
        val fakeId = BasicFlowIdent(baseId)
        fakeId.location_(ident.location)
        Some(topComp.body.namespace.locateDef(fakeId))

      case Nil => context.definitions.get(baseId).map(Right(_))
    }

  }

  /** Locates a configuration declaration. */
  def locate(ident: ConfigIdent): ConfigurationDeclaration = source.parent.namespace.locate(ident)

  /** Gets a component instance definition from its identifier (triggers exception if not found). */
  def locateDef(ident: CompInstanceIdent): List[CompCallFlowDefinition] = locateInstDefDown(ident.path, this, Nil) getOrElse {
    throw IdentNotFoundException(ident, this)
  }

  /** Recursively locates a stream declaration in the given context. */
  @tailrec
  private[this] def locateInstDefDown(ident: List[DSLSymbol], context: ComponentBodyNamespace, callStack: List[CompCallFlowDefinition]):
  Option[List[CompCallFlowDefinition]] = ident match {
    case Nil => Some(callStack)
    case head :: tail => context.callDefinitions.get(head) match {
      case None => None
      case Some(call) => locateInstDefDown(tail, locate(call.rhs.compId).body.namespace, call :: callStack)
    }
  }
}


/** Namespace management in a component body.
  * */
class ConfigurationNamespace(
                              val parentOpt: Option[PackageNamespace],
                              val source: ConfigurationDeclaration,
                              val designSpace: DesignSpaceMapType
                            ) extends Namespace {

  type S = ConfigurationDeclaration

  // in the failure condition expression all type identifiers, flow identifiers, etc, are considered relative to the root component's body namespace
  // in the design space alternatives:
  // - component instances are relative to the root component namespace
  // - alternatives components are also relative to the root component namespace
  // in the 
  //
  // fill that name space with maps relating symbols to their definitions
  def rootCompNamespace: ComponentBodyNamespace = parentOpt.get.locate(source.rootCompId).body.namespace

  /** Gets a type declaration from its identifier (triggers exception if not found). */
  def locate(ident: TypeIdent): TypeDeclaration = source.parent.namespace.locate(ident)

  /** Gets a package declaration from its identifier (triggers exception if not found). */
  def locate(ident: PackageIdent): PackageDeclaration = source.parent.namespace.locate(ident)

  /** Gets a component declaration from its identifier (triggers exception if not found). */
  def locate(ident: CompIdent): ComponentDeclaration = source.parent.namespace.locate(ident)

  /** Gets a boolean identifier declaration (triggers exception if not found). */
  def locate(ident: BoolIdent): BoolIdent = source.parent.namespace.locate(ident)

  /** Gets a data flow decalaration from its identifier (triggers exception if not found). */
  def locate(ident: FlowIdent): Either[TypeDeclaration, FlowDeclaration] = source.parent.namespace.locate(ident)

  /** Gets the definition of a flow identifier, either a flow definition
    * or a type declaration which contains the identifier (triggers exception if not found). */
  def locateDef(ident: FlowIdent): Either[TypeDeclaration, FlowDefinition] = source.parent.namespace.locateDef(ident)

  /** Gets the declaration of a type identifier (triggers exception if not found). */
  def locate(ident: ConfigIdent): ConfigurationDeclaration = source.parent.namespace.locate(ident)

  /** Gets a component instance definition from its identifier (triggers exception if not found). */
  def locateDef(ident: CompInstanceIdent): List[CompCallFlowDefinition] = source.parent.namespace.locateDef(ident)

}


