/** $Id: NamespaceException.scala 1433 2017-10-19 14:06:15Z kdelmas $
  *
  * @author $Author: kdelmas $
  **/

package namespace.dsl

import ast.dsl._

/** Base type for all namespace exceptions. */
sealed trait NamespaceException extends DSLException {
  def isFatal: Boolean
}

/** Non-fatal namespace exceptions from which recovery is possible. */
sealed trait NamespaceWarning extends NamespaceException {
  def isFatal = false
}

/** Fatal namespace exceptions from which recovery is not possible. */
sealed trait NamespaceFatal extends NamespaceException {
  def isFatal = true
}


/** Encapsulates a list of namespace exceptions. */
case class NamespaceExceptions(exceptions: List[NamespaceException]) extends Exception {
  override def toString: String = exceptions.map(_.toString).mkString("\n")
}

/** Thrown when the declaration of an identifier could not be found. */
case class UndeclaredIdentifierException[A <: IdentLike](id: A) extends NamespaceFatal {
  override def toString = s"${id.location.toString}:\n\tidentifier '${id.toString}' is not declared."
}

/** Thrown when the definition of an identifier could not be found. */
case class UndefinedIdentifierException[A <: IdentLike](id: A) extends NamespaceFatal {
  override def toString = s"${id.location.toString}:\n\tidentifier '${id.toString}' is not defined."
}

/** Thrown when an identifier is declared multiple times. */
case class MultipleDeclarationException[A <: ASTNode](current: A, previous: A) extends NamespaceFatal {
  override def toString = s"${current.location.toString}:\n\tidentifier '${current.toString}' already declared as '${previous.toString}' in ${previous.location.toString}."

}

/** Thrown when a value identifer is multiply declared in some TypeDeclaration. */
case class MultipleTypeValueDeclarationException(symbol: DSLSymbol, decl: TypeDeclaration) extends NamespaceFatal {
  override def toString = s"${decl.location.toString}:\n\ttype value '${symbol.name}' is multiply declared within type declaration '${decl.ident}'"
}

/** Thrown when a value identifer is multiply declared in some TypeDeclaration. */
case class MultipleInstIdDeclarationException(instId: DSLSymbol, decl: ComponentDeclaration) extends NamespaceFatal {
  override def toString = s"${decl.location.toString}:\n\tinstance name '${instId.name}' is multiply declared within component declaration '${decl.ident}'"
}

/** Thrown when an identifier is defined multiple times. */
case class MultipleDefinitionException[A <: ASTNode](current: A, previous: A) extends NamespaceFatal {
  override def toString = s"${current.location.toString}:\n\tidentifier '${current.toString}' already defined as '${previous.toString}' in ${previous.location.toString}."
}

/** Thrown when a flow of some component is given a definition (for input flows only). */
case class InputDefinitionException(flow: FlowDeclaration, component: ComponentDeclaration, definition: FlowDefinition) extends NamespaceFatal {
  override def toString = s"\n${flow.location.toString}:\n\tinput flow '${flow.ident.toString}'is given a definition at\n\t'${definition.location.toString}'\n\tin component '${component.ident}' in ${component.location.toString}\n"
}


/** Thrown when a flow of some component is not given a definition (for local and output flows only). */
case class MissingDefinitionException(flow: FlowDeclaration, component: ComponentDeclaration) extends NamespaceFatal {
  override def toString = s"\n${flow.location.toString}:\n\tdefinition for flow '${flow.ident.toString}' is missing\n\tin component '${component.ident.toString}' in ${component.location.toString}\n"
}

/** Thrown when a flow of some component is not given a definition (for local and output flows only). */
case class MissingLambdaDefinitionException(event: BoolIdent, alternative: CompAlternative) extends NamespaceFatal {
  override def toString = s"\n${alternative.location.toString}:\n\tlambda value for event '${event.toString}' is missing\n\tin alternative '${alternative.toString}'\n"
}

/** Thrown when the package given in an open declaration could not be located in the context namespacex. */
case class OpenPackageNotFoundException(o: OpenDeclaration) extends NamespaceFatal {
  override def toString = s"\n${o.location.toString}:\n\ttarget package in '$o' could not be found\n"
}

/** Thrown when a declaration shadows a previous declaration declared in an enclosing scope. */
case class ShadowingDeclarationException(current: Declaration, previous: Declaration) extends NamespaceWarning {
  override def toString = s"\n${current.location.toString}:\n\t'${current.toString}' shadows previous declaration \n\t'${previous.toString}' in ${previous.location.toString}\n"
}

/** Thrown when the declaration of an identifier (package, type, constant, stream) cannot be found in the namespace */
case class IdentNotFoundException[I <: IdentLike](ident: I, context: Namespace) extends NamespaceFatal {
  override def toString = s"\n${ident.location.toString}:\n\t'${ident.toString}' declaration or definition could not be found.\n"
}

/** Thrown when the declaration of an identifier (package, type, constant, stream) cannot be found in the namespace */
case class EventsInNonPrimaryComponentException(ident: CompIdent) extends NamespaceFatal {
  override def toString = s"\n${ident.location.toString}:\n\t'${ident.toString}' component is not declared as primary but contains events.\n"
}

/** Thrown when the declaration of an identifier (package, type, constant, stream) cannot be found in the namespace */
case class InstDefInPrimaryComponentException(ident: CompIdent) extends NamespaceFatal {
  override def toString = s"\n${ident.location.toString}:\n\t'${ident.toString}' component is declared as primary but contains instanciations.\n"
}

/** Thrown when a non-primary component is used as the left hand side of an alternative in a design space description. */
case class NonPrimaryAlternativeException(alternative: CompAlternative) extends NamespaceFatal {
  override def toString = s"\n${alternative.location.toString}:\n\t'$alternative' component is not primary but is used as alternative.\n"
}


/** Thrown when a non-primary component is used in the right hand side of an alternative in a design space description. */
case class NonPrimarySubstitutionException(ident: CompInstanceIdent) extends NamespaceFatal {
  override def toString = s"\n${ident.location.toString}:\n\t'${ident.toString}' component instance is not primary but is used as substitution target.\n"
}

/** Thrown when searching an identifier in declaration that cannot contain it */
case class MeaninglessLocate(ident: IdentLike, in:Declaration) extends NamespaceFatal {
  override def toString = s"Searching for identifier '${ident.toString}'  in declaration $in is meaningless.\n"
}

/** Thrown when trying to access a dependency graph that does not exist within some namespace. */
case class DepGraphNotSet[N <: Namespace](namespace: N, graphKind: String) extends NamespaceFatal {
  override def toString = s"${namespace.source.location.toString}:\n\t $graphKind dependency graph not yet set for ${namespace.source.toString}"
}

/** Exception representing a circular dependency error between ASTNodes. */
case class CyclicDependencyException[A <: SourceMapped](cycle: List[A]) extends NamespaceFatal {
  override def toString: String = {
    val head = cycle.head
    val ss = (head :: cycle.reverse).reverse.map(d => s"${d.location.toString}:\n\t${d.toString} depends on").mkString("\n")
    "Cyclic dependency:\n" + ss.mkString
  }
}

/** Thrown when the root component of a configuration is not closed, ie. still contains free inputs. */
case class RootComponentNotClosed(config: ConfigurationDeclaration) extends NamespaceFatal {
  override def toString: String = {
    s"${config.location.toString}:\n\tThe root component ${config.rootCompId.toString} of configuration ${config.toString} is not closed (ie. has free inputs)"
  }
}

sealed trait TypingException extends NamespaceFatal

case class WrongTypeExpr(expr: Expr, found: ScalarTypeExpr, expected: ScalarTypeExpr) extends TypingException {
  override def toString = s"${expr.location.toString}:\n\tWrong type for expression '${expr.toString}', expected '${expected.toString}', found '${found.toString}'."
}

case class WrongTypeCompCallArgs(call: CompCall, argsType: TupleType, inputsType: TupleType) extends TypingException {
  override def toString = s"${call.location.toString}:\n\tWrong type for arguments of call '${call.toString}'\n\texpected '${inputsType.toString}', found '${argsType.toString}'."
}

case class WrongTypeFlowDef(d: FlowDefinition, lhsType: TupleType, rhsType: TupleType) extends TypingException {
  override def toString = s"${d.location.toString}:\n\tType mismatch for definition '${d.toString}'\n\tlhs: '${lhsType.toString}', rhs: '${rhsType.toString}'."
}


case class WrongTypeAlternative(
                                 initial: ComponentDeclaration,
                                 inType: TupleType,
                                 outType: TupleType,
                                 alternative: ComponentDeclaration,
                                 altInType: TupleType,
                                 altOutType: TupleType) extends TypingException {
  override def toString = s"${alternative.location.toString}:\n\tType mismatch for component alternative '${initial.ident.toString}' -> '${alternative.ident.toString}'\n\tinputs types: '${inType.toString}' vs. '$altInType'\n\toutput types: '${outType.toString}' vs. '${altOutType.toString}' ."
}

case class NoAlternative(initial: CompInstanceIdent, conf: ConfigurationDeclaration) extends NamespaceFatal {
  override def toString = s"${conf.location.toString}:\n\tNo alternative for instance ${initial.toString} in configuration ${conf.ident.toString}"
}

case class WrongDefaultAlternative(inst: CompInstanceIdent, alt: CompAlternative, initial: CompIdent) extends TypingException {
  override def toString = s"${alt.location}:\n\tWrong type for ${inst.toString}->${alt.compId.toString}\n\tthe initial alternative component must be ${initial.toString}."
}
