/** $Id: CLIParser.scala 1433 2017-10-19 14:06:15Z kdelmas $
  *
  * @author $Author: kdelmas $
  **/
package parsers.dsl

import ast.dsl._
import util.parsing._

import scala.language.implicitConversions
//import scala.util.{ Try, Success, Failure }

/** A parser for parsing identifiers given as commande line parameters to the tool. */

object CLIParser extends RegexPackratParser {

  /** Set of language keywords (lexicographically sorted). */
  private val keywordStrings: List[String] = List("comp", "condition", "configuration",
    "count", "defs", "design", "else", "empty", "end", "evts", "extern",
    "failure", "if", "imported", "inter", "leaf", "locs", "mincard",
    "open", "package", "private", "public", "returns", "root", // TODO ??? "#include",
    "space", "then", "type", "union").sortWith(_ < _)

  /** Double colon token. */
  private val colcol = "::"

  /** Regular expression parser for keywords, used to separate keywords from identifiers. */
  private lazy val keyword = keywords( """[^_a-zA-Z0-9]""".r, keywordStrings)

  /** Regular expression for identifiers. */
  private lazy val identRegex =
    """([a-zA-Z]|[_][a-zA-Z])[_a-zA-Z0-9]*""".r

  /** Regular expression for uppercase identifiers. */
  private lazy val uidentRegex =
    """([A-Z]|[_][A-Z])[_a-zA-Z0-9]*""".r

  /** Regular expression for lowercase identifiers. */
  private lazy val lidentRegex =
    """([a-z]|[_][a-z])[_a-zA-Z0-9]*""".r

  /** Parser which succeeds when an identifier is not a keyword. */
  private lazy val ident: PackratParser[String] = not(keyword) ~> identRegex | failure("identifier expected")

  /** String literals. */
  private lazy val stringLiteral: PackratParser[String] = regex("(\"[^\\\\\"]*(\\\\\"[^\\\\\"]*)*\")".r)

  /** Regular expression for zero. */
  private lazy val zeroRegex =
    """0""".r

  /** Regular expression for decimal integers. */
  private lazy val decIntRegex =
    """[1-9][0-9]*""".r

  /** Regular expression for octal integers. */
  private lazy val octIntRegex =
    """O[0-7]+""".r

  /** Regular expression for hexadecimal integers. */
  private lazy val hexIntRegex =
    """0[xX][0-9a-fA-F]+""".r

  private lazy val int =
    """[1-9][0-9]*([eE][\+\-]?[0-9]+)?""".r

  /** Regular expression for floating point numbers. */
  private lazy val floatingPointNumber =
    """-?(\d+(\.\d*)?|\d*\.\d+)([eE][+-]?\d+)?[fFdD]?""".r

  private lazy val pathIdent: PackratParser[List[String]] = rep1sep(ident, "::")

  private lazy val instIdent: PackratParser[List[String]] = rep1sep(ident, ".")


  /** Parser for a configuration identifier. */
  lazy val configIdent: PackratParser[ConfigIdent] = {
    pathIdent ^^ {
      case head :: Nil => ConfigIdent(StdDSLSymbol(Symbol(head)))
      case idents =>
        val revIdents = idents.reverse
        val ident = StdDSLSymbol(Symbol(revIdents.head))
        val pkgPath = revIdents.tail.reverse map { (s: String) => StdDSLSymbol(Symbol(s)) }
        ConfigIdent(ident, PackageIdent(pkgPath))
    }
  }

  /** Parses an operator identifier from a string. */
  def parseConfigIdent(s: String): scala.util.Try[ConfigIdent] = {
    parse(phrase(configIdent), s) match {
      case Success(parseTree, _) =>
        scala.util.Success(parseTree)

      case _ =>
        scala.util.Failure(new Exception("CLIParser: incorrect configuration identifier string"))

    }
  }


  /** Parser for a component identifier. */
  lazy val compIdent: PackratParser[CompIdent] = {
    pathIdent ^^ {
        case head :: Nil => CompIdent(StdDSLSymbol(Symbol(head)))
        case idents =>
          val revIdents = idents.reverse
          val ident = StdDSLSymbol(Symbol(revIdents.head))
          val pkgPath = revIdents.tail.reverse map { (s: String) => StdDSLSymbol(Symbol(s)) }
          CompIdent(ident, PackageIdent(pkgPath))
    }
  }

  /** Parses a component identifier from a string. */
  def parseCompIdent(s: String): scala.util.Try[CompIdent] = {
    parse(phrase(compIdent), s) match {
      case Success(parseTree, _) =>
        scala.util.Success(parseTree)

      case _ =>
        scala.util.Failure(new Exception("CLIParser: incorrect component identifier string"))

    }
  }
}
