/** $Id: ParseFiles.scala 1430 2017-10-17 09:16:44Z kdelmas $
  *
  * @author $Author: kdelmas $
  * @date $Date: 2017-10-17 11:16:44 +0200 (mar. 17 oct. 2017) $ 
  **/

package parsers.dsl

import java.io.{File, FileInputStream}

import ast.dsl.{IncludeDeclaration, ProgramDeclaration}
import org.antlr.v4.runtime.{ANTLRInputStream, CommonTokenStream}

import scala.annotation.tailrec
import scala.collection.immutable.HashMap
import scala.util.{Failure, Success, Try}
import scalaz.\/
import scalaz.syntax.either._


object ParseFiles extends (List[String] => List[ParsingException] \/ Map[String, List[ProgramDeclaration]]) {

  /** Parses a list of scade files, extending the given existing set of declarations (or creating a new one if none given).
    *
    * @param files    List of files to parse.
    * @param existing A map storing the result of a previous parse.
    * @return A map containing the parsed declarations keyed under the file name.
    **/
  def apply(files: List[String]): List[ParsingException] \/ Map[String, List[ProgramDeclaration]] = {
    parseRec(files, Nil, HashMap.empty)
  }

  /** Resolves a relative include path wrt the path of the file that contains it.
    *
    * @param file The file that contains the include statement.
    * @param path The included path to be resolved.
    * @returns A canonical path.
    */
  private[this] def resolveInclude(file: File, path: String): String = {
    val trimmed = path.stripPrefix("\"").stripSuffix("\"").trim
    val resolved = new File(file.getParent, trimmed).getCanonicalPath
    println(s"#include '${path}' -> '${resolved}'")
    resolved
  }


  /** Iterates on the given list of files, pulls any includes automatically. */
  @tailrec
  private[this] final def parseRec(toParse: List[String], errors: List[ParsingException], result: Map[String, List[ProgramDeclaration]]): List[ParsingException] \/ Map[String, List[ProgramDeclaration]] = toParse match {

    // done parsing all files
    case Nil => if (errors.isEmpty) result.right else errors.left

    // still some files to parse
    case path :: remainingPaths => {

      val file = new File(path)

      file.exists match {

        case false => {
          parseRec(remainingPaths, FileNotFound(path) :: errors, result)
        }

        case true => {

          val cannPath = file.getCanonicalPath

          result.contains(cannPath) match {

            // file already loaded
            case true => {
              println(s"file '${file.getCanonicalPath}' already loaded, skipped.")
              parseRec(remainingPaths, errors, result)
            }

            // load a new file
            case false => {
              println(s"loading '${cannPath}'")
              Try {
                val input = new ANTLRInputStream(new FileInputStream(file))
                val lexer = new DSLLexer(input)
                val tokens = new CommonTokenStream(lexer)
                val parser = new DSLParser(tokens)
                val listener = new ScalaDSLListener
                listener.filename = path
                parser.addParseListener(listener)
                ThrowingErrorListener.error = None
                parser.addErrorListener(ThrowingErrorListener)
                val prog = parser.program()
                val declarations = listener.declarations
                // add any include statement to the parsing queue
                var includes: List[String] = Nil
                val resultDecls = declarations map {
                  decl =>
                    decl match {
                      case IncludeDeclaration(path) => {
                        val resolved = resolveInclude(file, path)
                        includes ::= resolved
                        IncludeDeclaration(resolved)
                      }
                      case d => d
                    }
                }
                (resultDecls, includes)
              } match {
                case Success((decls, includes)) => parseRec(includes ::: remainingPaths, errors, result + (cannPath -> decls))
                case Failure(e) =>
                  parseRec(remainingPaths, ParserError(ThrowingErrorListener.error.get) :: errors, result)
              }
            }
          }
        }
      }
    }
  }
}
