package perf

import java.io._
import java.util.Locale

import analyzers._
import ast.dsl.ConfigIdent
import util.analyzers._
import util.io.FileManager.GridCut
import util.io.LatexWriter.RowCutPerf
import util.io.{FileManager, LatexWriter, LoadDSLFiles, SystemInfo}

import scala.collection.mutable.ListBuffer
import scala.concurrent.ExecutionContext.Implicits.global
import scala.concurrent.duration._
import scala.language.postfixOps
import scala.util.{Failure, Success, Try}

/**
  * Created by kdelmas on 24/02/17.
  */
object CutEnumerationPerformancePerf {

  //default timeout for cutset analysis set to 100 seconds
  private val defaultTimeOut = 100 seconds

  //cutset size bounds for each benchmark
  private val cutSizeMap = FileManager.cutEnumerationBench.collect {
    case GridCut => GridCut -> List(3, 4, 5, 6)
    case bench => bench -> List(1, 2, 3, 7)
  }.toMap

  //find available computation methods according to available external tools
  private val methods = SystemInfo.findAvailableCutComputationMethods()
    .filterNot(m => m == BDDCutComputation || m == IterativeSMTCutComputation || m == PBCutComputation)

  /**
    * Build the analyser for a given configuration given in a KCR file
    * @param file KCR file
    * @param confName configuration name
    * @return an analyser for cutset computation if parsing succeeded, None otherwise
    */
  private def makeAnalyserFor(file: File, confName: String): Option[Parsing with SMTTranslation with CutEnumeratorAll] = {

   Try{
     new FileParsing
      with SystemConsole
       with SMTTranslation
       with CutEnumeratorAll {
       val fileNames = file.getAbsolutePath :: Nil
       val confId = confName
       val BDDFactory = new BDDFactory with Z3BoolExprToBDDTranslator
     }
   } match {
     case Success(x) => Some(x)
     case Failure(e)=> None
   }
  }


  /**
    * Perform cutset computation with available methods and store execution times in Latex file
    * @param timeout timeout for enumeration
    */
  private def cutGenerationPerf(timeout: FiniteDuration) : Unit = {

    //output file
    val out = FileManager.addFileToTest("cutGenerationPerformance.tex")

    //build an analyser per benchmark
    val analyzers = FileManager.cutEnumerationBench.map(bench =>
      bench -> makeAnalyserFor(new File(bench.path.toAbsolute.toString()), bench.conf)).collect({
      case (p, Some(a)) => p -> a
      case (p, None) => throw new Exception(s"cannot create analyzer for $p")
    })

    //rows of the array
    val rows = ListBuffer[RowCutPerf]()

    //for each benchmark and each cutset size bounds perform analysis with available methods
    for ((bench, a) <- analyzers) {
      for (cutSize <- cutSizeMap(bench)) {

        var row = RowCutPerf(bench.name, cutSize)
        for (m <- methods.filterNot(m => m == HipHOPSCutComputation && bench == GridCut)) {
          val f = TimeoutTask({
            a.defaultComputationMethod = m
            a.profilerCutEnumeration.runWithProfile(Some(cutSize))()._2.processTime
          },
            timeout, {
              println(s"timeout for computing sutsets of size $cutSize with method $m for ${bench.path.name}")
              SystemInfo.stopToolFor(m)
              timeout.toNanos
            })
          val r = f.get()
          val result = if (r == timeout.toNanos)
            s"$$> ${timeout.toSeconds}$$"
          else
            "%3.3f" formatLocal(Locale.UK, r * 1e-9)
          FileManager.cleanTemp()

          m match {
            case HipHOPSCutComputation => row = row.copy(HipHop = result)
            case XSAPCutComputation => row = row.copy(xSAP = result)
            case XFTACutComputation => row = row.copy(XFTA = result)
            case IterativeSATCutComputation => row = row.copy(KCR = result)
            case _ =>
          }
        }
        rows += row
      }
    }

    //write result in output file
    val writer = new FileWriter(out)
    LatexWriter.writePerCut("Execution time for cutsets enumeration", rows.toList, writer)
    writer.close()
    println(s"[INFO] Execution time for cutsets enumeration written in $out")
  }

  /**
    * Run cutset enumeration performance analysis
    * @param userTimeout optinal user-defined timeout for cutset enumeration
    */
  def apply(userTimeout: Option[FiniteDuration] = None) : Unit = {
    val timeout: FiniteDuration = userTimeout.getOrElse(defaultTimeOut)
    cutGenerationPerf(timeout)
  }
}
