package perf

/**
  * Created by kdelmas on 07/07/17.
  */

import java.io._
import java.util.Locale

import analyzers.{Parsing, SMTTranslation, STMDDBuilderSim, SubstitutionAnalyzerSim, _}
import ast.dsl.ConfigIdent
import interpreter.TraceInterpreter
import util.analyzers._
import util.io.FileManager.{FuelExp, HBSExp, QuadExp, RosaceExp}
import util.io.LatexWriter.RowDSEPerf
import util.io.{FileManager, LoadDSLFiles, _}

import scala.collection.mutable.ListBuffer
import scala.concurrent.ExecutionContext.Implicits.global
import scala.concurrent.duration._
import scala.language.postfixOps
import scala.util.{Failure, Success, Try}


object DesignExplorationPerf {

  //default timeout for exploration problem resolution
  private val defaultTimeout = 600 seconds

  //reliability and order requirement for the exploration problems
  private val reqList = List((1, 0.999), (1, 0.99999), (1, 0.9999999), (2, 0.999999999))

  //find available computation methods according to available external tools
  private val methods = SystemInfo.findAvailableExpMethods()

  /**
    * Build the analyser for a given configuration given in a KCR file
    * @param file KCR file
    * @param confName configuration name
    * @return an analyser for exploration problem resolution if parsing succeeded, None otherwise
    */
  private def makeAnalyserFor(file: File, confName: String): Option[Parsing with SMTTranslation
    with SubstitutionAnalyzerSim
    with STMDDBuilderSim
    with TraceInterpreter
    with DSEProblemSolverAll] = {

    Try{
      new FileParsing
        with SystemConsole
        with SMTTranslation
        with SubstitutionAnalyzerSim
        with STMDDBuilderSim
        with TraceInterpreter
        with DSEProblemSolverAll {
        val BDDFactory = new BDDFactory with Z3BoolExprToBDDTranslator
        val InstBoolBDDFactory = new InstBoolBDDFactory
        val fileNames = file.getAbsolutePath :: Nil
        val confId = confName
      }
    } match {
      case Success(x) => Some(x)
      case Failure(_) => None
    }
  }


  /**
    * Run the resolution of benchmark problems without order requirements and
    * store the results in Latex file
    * @param timeout timeout for resolution
    */
  private def designExplorationPerfWithoutCard(timeout: FiniteDuration) : Unit = {
    val out = FileManager.addFileToTest("designExplorationPerformanceWithoutCard.tex")
    var lines = ListBuffer[RowDSEPerf]()
    for (m <- methods) {
      for (req <- reqList) {
        var line = RowDSEPerf(m.toString, 1, req._2)
        for (b <- FileManager.explorationBench) {
          val a = makeAnalyserFor(new File(b.path.toAbsolute.toString()), b.conf).get
          val f = TimeoutTask({
            val start = System.nanoTime()

            //solve the problem without order requirement (None)
            val res = a.solveWith(m, Some(req._2), None)
            (System.nanoTime() - start, if (res.isDefined) "SAT" else "UNSAT")
          }, timeout, {
            SystemInfo.stopToolFor(m)
            (timeout.toNanos, "UNKNOWN")
          })
          val res = f.get()
          val r = if (res._1 == timeout.toNanos)
            s"$$> ${timeout.toSeconds}$$"
          else {
            val status = if (m == SMTExploration) s"(${res._2})" else ""
            ("%3.3f" formatLocal(Locale.UK, res._1 * 1e-9)) + status
          }

          FileManager.cleanTemp()
          b match {
            case RosaceExp => line = line.copy(Rosace = s"$r")
            case FuelExp => line = line.copy(Fuel = s"$r")
            case HBSExp => line = line.copy(HBS = s"$r")
            case QuadExp => line = line.copy(Quad = s"$r")
          }
        }
        lines += line
      }
    }
    val writer = new FileWriter(out)
    LatexWriter.writePerDSE("design space resolution time without order", lines.toList, writer)
    writer.close()
    println(s"[INFO] Design space resolution time without order written in $out")
  }

  /**
    * Run the resolution of benchmark problems with order requirements and
    * store the results in Latex file
    * @param timeout timeout for resolution
    */
  private def designExplorationPerfWithCard(timeout: FiniteDuration) : Unit = {
    val out = FileManager.addFileToTest("designExplorationPerformanceWithCard.tex")
    var lines = ListBuffer[RowDSEPerf]()
    for (req <- reqList) {
      var row = RowDSEPerf(SMTExploration.toString, req._1, req._2)
      for (b <- FileManager.explorationBench) {
        val a = makeAnalyserFor(new File(b.path.toAbsolute.toString()), b.conf).get
        val f = TimeoutTask({
          val start = System.nanoTime()

          //solve the problem with cardinality requirement, so only with SMT-based method
          val res = a.solveConfWithReq(Some(req._2), Some(req._1))
          (System.nanoTime() - start, if (res.isDefined) "SAT" else "UNSAT")
        }, timeout, {
          (timeout.toNanos, "UNKNOWN")
        })
        val res = f.get()
        val r = if (res._1 == timeout.toNanos)
          s"$$> ${timeout.toSeconds}$$"
        else
          ("%3.3f" formatLocal(Locale.UK, res._1 * 1e-9)) + s"(${res._2})"
        FileManager.cleanTemp()
        b match {
          case RosaceExp => row = row.copy(Rosace = s"$r")
          case FuelExp => row = row.copy(Fuel = s"$r")
          case HBSExp => row = row.copy(HBS = s"$r")
          case QuadExp => row = row.copy(Quad = s"$r")
        }
      }
      lines += row
    }
    val writer = new FileWriter(out)
    LatexWriter.writePerDSE("design space resolution time with order", lines.toList, writer)
    writer.close()
    println(s"[INFO] Design space resolution time with order written in $out")
  }

  /**
    * Run the resolution of benchmark problems with order requirements but without clause minimisation and
    * store the results in Latex file
    * @param timeout timeout for resolution
    */
  private def designExplorationPerfWithoutQuickExplain(timeout: FiniteDuration) : Unit = {
    val out = FileManager.addFileToTest("designExplorationPerformanceWithoutQuickXplain.tex")
    var lines = ListBuffer[RowDSEPerf]()
    for (req <- reqList) {
      var row = RowDSEPerf(SMTExploration.toString, req._1, req._2)
      for (b <- FileManager.explorationBench) {
        val a = makeAnalyserFor(new File(b.path.toAbsolute.toString()), b.conf).get
        val f = TimeoutTask({
          val start = System.nanoTime()

          //solve the problem without clause minimisation
          val res = a.solveConfWithReq(Some(req._2), None, withXplain = false)
          (System.nanoTime() - start, if (res.isDefined) "SAT" else "UNSAT")
        }, timeout, {
          (timeout.toNanos, "UNKNOWN")
        })
        val res = f.get()
        val r = if (res._1 == timeout.toNanos)
          s"$$> ${timeout.toSeconds}$$"
        else
          ("%3.3f" formatLocal(Locale.UK, res._1 * 1e-9)) + s"(${res._2})"
        FileManager.cleanTemp()
        b match {
          case RosaceExp => row = row.copy(Rosace = s"$r")
          case FuelExp => row = row.copy(Fuel = s"$r")
          case HBSExp => row = row.copy(HBS = s"$r")
          case QuadExp => row = row.copy(Quad = s"$r")
        }
      }
      lines += row
    }
    val writer = new FileWriter(out)
    LatexWriter.writePerDSE("design space resolution time without quickXplain", lines.toList, writer)
    writer.close()
    println(s"[INFO] Design space resolution time without quickXplain written in $out")
  }

  def apply(userTimeout: Option[FiniteDuration] = None) :  Unit = {
    val timeout=userTimeout.getOrElse(defaultTimeout)
    designExplorationPerfWithoutCard(timeout)
    designExplorationPerfWithCard(timeout)
    designExplorationPerfWithoutQuickExplain(timeout)
  }
}
