package perf

import java.io.{File, FileOutputStream, PrintStream}

import analyzers._
import ast.dsl.{CompAlternative, CompInstanceIdent, ConfigurationDeclaration, Transformer}
import interpreter.TraceInterpreter
import monix.execution.Scheduler.Implicits.global
import old.UnreliabilityTheorySim
import tool.MainCLI
import util.analyzers.{BDDFactory, _}
import util.io.{FileManager, SystemInfo}

import scala.collection.parallel.ForkJoinTaskSupport
import scala.concurrent.duration._
import scala.language.postfixOps
import scala.util.Random

/**
  * Created by kdelmas on 29/08/17.
  */
object ReliabilityComputationBDDvsSTTPerf {

  //default timeout for reliability computation
  private val defaultTimeout = 1000 seconds

  /**
    * Build a candidate by random selection of the substitution performed on the initial system
    * @param conf the problem configuration
    * @param bias only consider the "bias" last alternatives for each component (since in benchmark
    *             models alternatives are sorted by their "complexity", force to consider more complex
    *             system)
    * @return the substitution map describing the candidate
    */
  private def genRandomCandidate(conf: ConfigurationDeclaration, bias: Int): Map[CompInstanceIdent, CompAlternative] = {
    conf.designSpace.map(sub => sub.instance -> {
      if (bias < sub.alternatives.size)
        sub.alternatives(sub.alternatives.size - bias + Random.nextInt(bias))
      else
        sub.alternatives(Random.nextInt(sub.alternatives.size))
    }).toMap
  }


  /**
    * Generate the reliability of a given a system by BDD method
    *
    * @param replaceMap the substitution chosen for components
    * @param conf       the problem configuration
    * @return the mean time of reliability computation
    */
  private def generateAnalyserRBDDOf(conf: ConfigurationDeclaration, replaceMap: Map[CompInstanceIdent, CompAlternative]): Parsing
    with SMTTranslation
    with ReliabilityComputer =
    synchronized {
      val tempFile = FileManager.addFileToTemp(s"tempBDD.kcr")
      Transformer.writeSolutionFile(tempFile, conf.parent, conf, replaceMap)
     new FileParsing
        with SystemConsole
        with SMTTranslation
        with ReliabilityComputer {
        val BDDFactory = new BDDFactory with Z3BoolExprToBDDTranslator
        val fileNames = tempFile.getAbsolutePath :: Nil
        val confId = conf.ident.baseId.symbol.name
      }
    }


  /**
    * Generate the reliability of a given a system by BDD method
    *
    * @param analyser the problem analyser
    * @return the mean time of reliability computation
    */
  private def computeRBDDOf(
                             analyser: Parsing
                               with SMTTranslation
                               with ReliabilityComputer): Long = {
    val (_, timeBDD) = analyser.profilerReliabilityComputation.runWithProfile()()
    timeBDD.processTime
  }

  /**
    * Generate the reliability of a given a system by BDD method
    *
    * @param replaceMap the substitution chosen for components
    * @param conf       the problem configuration
    * @return the mean time of reliability computation
    */
  private def computeRSTTOf(
                             conf: Parsing with SMTTranslation
                               with TraceInterpreter
                               with SubstitutionAnalyzerSim
                               with STMDDBuilderSim
                               with UnreliabilityTheorySim,
                             replaceMap: Map[CompInstanceIdent, CompAlternative]): Long = {
    val start = System.nanoTime()
    conf.computeUnrWithMemo(replaceMap)
    val timeSTT = System.nanoTime() - start
    timeSTT
  }

  /**
    * Perform the reliability computation on the benchmarks with STMDD and BDD
    * @param bench the benchmark
    * @param nbCandidate the number of candidate derived from the initial system
    * @param bias the bias to select only the "bias" last alternatives for each component
    * @param timeout the timeout for reliability computation
    * @param nbCore the number of physical cores used to perform the analyses
    */
  private def reliabilityPerf(bench: FileManager.ExplorationBench, nbCandidate: Int, bias: Int, timeout: FiniteDuration, nbCore: Int): Unit = {


    val destinationFile = FileManager.addFileToTest(s"${bench.name}_reliability_computation.dat")

    //build the reliability analyser
    val STTAnalyser = new FileParsing
      with SystemConsole
      with SMTTranslation
      with TraceInterpreter
      with SubstitutionAnalyzerSim
      with STMDDBuilderSim
      with UnreliabilityTheorySim {
      val BDDFactory = new BDDFactory with Z3BoolExprToBDDTranslator
      val InstBoolBDDFactory = new InstBoolBDDFactory
      val fileNames = bench.path.toString() :: Nil
      val confId = bench.conf
    }

    //force STMDD computation
    STTAnalyser.STMDDD

    //generate the candidate to analyse
    val candidates = (1 to nbCandidate).map(i => (i, genRandomCandidate(STTAnalyser.conf, bias)))

    //build a parallel collection of the candidate and limit the number of concurrent thread
    //to the numbre of available cores
    val parCandidates = candidates.par
    val e = new scala.concurrent.forkjoin.ForkJoinPool(nbCore)
    parCandidates.tasksupport = new ForkJoinTaskSupport(e)

    //result storage
    val data = scala.collection.concurrent.TrieMap[Int, (Double, Double)]()

    //compute sequentially the reliability computation time
    candidates.foreach(p => {
      data(p._1) = (0.0, computeRSTTOf(STTAnalyser, p._2) * 1e-9)
    })


    val fos = new FileOutputStream(destinationFile)
    val writer = new PrintStream(fos)

    //run the reliability computation with BDD method and write result to file
    ("BDD" :: "STMDD" :: Nil).zipWithIndex.foreach(p => writer.println(s"#column${p._2} ${p._1}"))
    parCandidates.foreach(p => {
      val (i, candidate) = p
      val a = generateAnalyserRBDDOf(STTAnalyser.conf, candidate)
      val t = TimeoutTask(computeRBDDOf(a), timeout, {
        timeout.toNanos
      })
      data(i) = data(i).copy(_1 = t.get() * 1e-9)
      synchronized {writer.println(s"${data(i)._1} ${data(i)._2}")}
    })
    writer.println(s"#speedup:${data.map(p => p._2._1 / p._2._2).sum / data.size}")
    writer.close()
    fos.close()
    println(s"[INFO] STMDD vs BDD reliability computation written in ${destinationFile.getPath}")
  }

  def apply(userTimeout: Option[FiniteDuration] = None) : Unit = {
    val timeout: FiniteDuration = userTimeout.getOrElse(defaultTimeout)
    FileManager.explorationBench.foreach(b => {
      reliabilityPerf(b, 5000, 4, timeout, SystemInfo.usableCores)
    })
  }
}
