/** $Id: Printer.scala 1433 2017-10-19 14:06:15Z kdelmas $
  *
  * @author $Author: kdelmas $
  **/

package printing.dsl

import java.io.Writer

import ast.dsl._

/** Printer for ASTNodes */
class Printer(out: Writer) {

  /** Writes the kid expression to out, adding parenthesis if needed depending on parent-kid precedence order. */
  def applyWithParen(parent: FlowExpr, kid: FlowExpr): Unit = {
    if (Precedence.mustParen(parent, kid)) {
      out.write("( ")
      apply(kid)
      out.write(" )")
    } else {
      apply(kid)
    }
  }

  /** Writes the kid expression to out, adding parenthesis if needed depending on parent-kid precedence order. */
  def applyWithParen(parent: BoolExpr, kid: BoolExpr): Unit = {
    if (Precedence.mustParen(parent, kid)) {
      out.write("( ")
      apply(kid)
      out.write(" )")
    } else {
      apply(kid)
    }
  }

  /** Writes the kid expression to out, adding parenthesis if needed depending on parent-kid precedence order. */
  def applyWithParen(parent: FlowExpr, kid: BoolExpr): Unit = {
    if (Precedence.mustParen(parent, kid)) {
      out.write("( ")
      apply(kid)
      out.write(" )")
    } else {
      apply(kid)
    }
  }

  /** Writes the kid expression to out, adding parenthesis if needed depending on parent-kid precedence order. */
  def applyWithParen(parent: BoolExpr, kid: FlowExpr): Unit = {
    if (Precedence.mustParen(parent, kid)) {
      out.write("( ")
      apply(kid)
      out.write(" )")
    } else {
      apply(kid)
    }
  }

  /** Writes a package path given as a list */
  def printPathPrefix(path: List[DSLSymbol]): Unit = printSep(path, { s: DSLSymbol => apply(s) }, sep = "::")

  /** Writes a list of A:s to out, using the printA function, separated by a sep. */
  def printSep[A](l: List[A], printA: A => Unit, sep: String): Unit = l match {
    case Nil =>
    case head :: tail =>
      printA(head)
      tail.foreach {
        a => {
          out.write(sep)
          printA(a)
        }
      }
  }

  def apply(node: DSLSymbol): Unit = out.write(node.name)

  def apply(node: PackageIdent): Unit = out.write(node.toString)

  def apply(node: TypeIdent): Unit = out.write(node.toString)

  def apply(node: CompIdent): Unit = out.write(node.toString)

  def apply(node: CompInstanceIdent): Unit = out.write(node.toString)

  def apply(node: FlowIdent): Unit = out.write(node.toString)

  def apply(node: BoolIdent): Unit = out.write(node.toString)

  def apply(node: ConfigIdent): Unit = out.write(node.toString)


  /** Writes a program declaration to out. */
  def apply(node: ProgramDeclaration): Unit = node match {
    case n: IncludeDeclaration => apply(n)
    case n: OpenDeclaration => apply(n)
    case n: PackageDeclaration => apply(n)
    case n: TypeDeclaration => apply(n)
    case n: ComponentDeclaration => apply(n)
    case n: ConfigurationDeclaration => apply(n)
  }


  def apply(node: IncludeDeclaration): Unit = {
    out.write("#include ")
    out.write(node.filePath)
    out.write(" ;\n")
  }

  def apply(node: OpenDeclaration): Unit = {
    out.write("open ")
    apply(node.pkgId)
    out.write(" ;\n")
  }

  /** Writes a package declaration as the root package (omits the package name). */
  def asRoot(node: PackageDeclaration): Unit = {
    node.declarations foreach {
      decl => {
        out.write("\n")
        apply(decl)
        out.write("\n")
      }
    }
  }

  def apply(node: PackageDeclaration): Unit = {
    out.write("package ")
    apply(node.ident)
    out.write("\n")
    node.declarations foreach {
      decl => {
        out.write("\n")
        apply(decl)
        out.write("\n")
      }
    }
    out.write("end;\n")
  }

  def apply(node: TypeDeclaration): Unit = {
    out.write("type ")
    apply(node.ident)
    out.write(" := { ")
    printSep(node.values, { (s: DSLSymbol) => apply(s) }, ", ")
    out.write(" };")
  }

  def apply(node: ComponentDeclaration): Unit = {
    if (node.isPrimary)
      out.write("primary comp ")
    else
      out.write("comp ")
    apply(node.ident)
    out.write(" (")
    printDeclarationList(l = node.inputs, sep = ";")
    out.write(") returns (")
    printDeclarationList(l = node.outputs, sep = ";")
    out.write(") ")
    apply(node.body)
  }


  def printDeclarationList(l: List[FlowDeclaration], start: String = "", sep: String, end: String = ""): Unit = {
    if(l.nonEmpty){
        val init = List[(TypeIdent, List[FlowDeclaration])]()
        val grouped = l.foldLeft(init)((acc, fdecl) => {
          if(acc.isEmpty){
            (fdecl.typeId, fdecl :: Nil) :: Nil
          }else{
            if (acc.last._1 == fdecl.typeId)
              acc.init :+ (fdecl.typeId, acc.last._2 :+ fdecl)
            else
              acc :+ (fdecl.typeId, fdecl :: Nil)
          }
        })
        out.write(start)
        grouped.foreach(lt => {
          printSep(lt._2, { (d: FlowDeclaration) => apply(d.ident) }, ", ")
          out.write(" : ")
          apply(lt._1)
          if (lt != grouped.last)
            out.write(sep)
        })
        out.write(end)
    }
  }


  def apply(node: ComponentBody): Unit = {

    out.write("{")

    node.events match {
      case Nil => /* intentionally empty */

      case l =>
        out.write("\n\nevents:\n\t ")
        printSep(l, { (id: BoolIdent) => apply(id) }, ", ")
        out.write(" ;\n")
    }

    node.locals match {
      case Nil => /* intentionally empty */

      case l =>
        out.write("\nlocs:\n ")
        printDeclarationList(l, "\t", " ;\n\t", " ;\n")
    }

    node.definitions match {
      case Nil =>
        /* intentionally empty */

      case l =>
        out.write("\ndefs:\n")
        printSep(l, { (d: FlowDefinition) => {
          out.write("\t"); apply(d)
        }
        }, "\n")
    }

    out.write("\n\n};\n")
  }


  def apply(node: FlowDeclaration): Unit = {
    apply(node.ident)
    out.write(": ")
    apply(node.typeId)
  }

  def apply(node: FlowDefinition): Unit = node match {
    case SingleFlowDefinition(lhs, rhs) =>
      apply(lhs)
      out.write(" := ")
      apply(rhs)
      out.write(" ;")

    case CompCallFlowDefinition(lhs, rhs) =>
      printSep(lhs, { (s: BasicFlowIdent) => apply(s) }, sep = ", ")
      out.write(" := ")
      apply(rhs)
      out.write(" ;")

  }

  def apply(node: CompCall): Unit = {
    apply(node.compId)
    out.write("@")
    apply(node.instId)
    out.write("(")
    printSep(node.args, { (e: FlowExpr) => apply(e) }, ", ")
    out.write(")")
  }

  def apply(node: FlowExpr): Unit = node match {
    case n: FlowIdent =>
      apply(n)

    case FlowUnion(kid1, kid2) =>
      applyWithParen(node, kid1)
      out.write(" union ")
      applyWithParen(node, kid2)

    case FlowInter(kid1, kid2) =>
      applyWithParen(node, kid1)
      out.write(" inter ")
      applyWithParen(node, kid2)

    case FlowIte(kid1, kid2, kid3) =>
      out.write("if ")
      applyWithParen(node, kid1)
      out.write(" then ")
      applyWithParen(node, kid2)
      out.write("\n\t\t else ")
      applyWithParen(node, kid3)
  }


  def apply(node: BoolExpr): Unit = node match {
    case BoolIdent(symbol) =>
      apply(symbol)

    case FlowEq(kid1, kid2) =>
      applyWithParen(node, kid1)
      out.write(" = ")
      applyWithParen(node, kid2)

    case FlowNeq(kid1, kid2) =>
      applyWithParen(node, kid1)
      out.write(" != ")
      applyWithParen(node, kid2)

    case FlowContains(kid1, kid2) =>
      applyWithParen(node, kid1)
      out.write(" ? ")
      applyWithParen(node, kid2)

    case Iff(kid1, kid2) =>
      applyWithParen(node, kid1)
      out.write(" <=> ")
      applyWithParen(node, kid2)

    case Implies(kid1, kid2) =>
      applyWithParen(node, kid1)
      out.write(" => ")
      applyWithParen(node, kid2)

    case Or(kid1, kid2) =>
      applyWithParen(node, kid1)
      out.write(" # ")
      applyWithParen(node, kid2)

    case And(kid1, kid2) =>
      applyWithParen(node, kid1)
      out.write(" & ")
      applyWithParen(node, kid2)

    case Not(kid1) =>
      out.write("~")
      applyWithParen(node, kid1)

    case PopCount(kids, relOp, cst) =>
      out.write("count( ")
      printSep(kids, { e: BoolExpr => apply(e) }, ", ")
      out.write(" ) ")
      apply(relOp)
      out.write(" ")
      apply(cst)
  }

  def apply(node: RelOp): Unit = node match {
    case Eq => out.write("=")
    case Ge => out.write(">=")
    case Gt => out.write(">")
    case Le => out.write("<=")
    case Lt => out.write("<")
  }

  def apply(node: IntExpr): Unit = node match {
    case DecIntLit(repr) => out.write(repr)
  }

  def apply(node: RealExpr): Unit = node match {
    case DecRealLit(repr) => out.write(repr)
  }


  def apply(node: ConfigurationDeclaration): Unit = {
    out.write("configuration ")
    apply(node.ident)
    out.write(" {\n")

    out.write("\n\troot := ")
    apply(node.rootCompId)
    out.write(";")

    if (node.minCardValue.isDefined) {
      out.write("\n\tminorder := ")
      apply(node.minCardValue.get)
      out.write(";")
    }


    out.write("\n\tfailure condition := ")
    apply(node.failureCondition)
    out.write(";")

    if (node.duration.isDefined) {
      out.write("\n\tduration := ")
      apply(node.duration.get)
      out.write(";")
    }

    if (node.reliability.isDefined) {
      out.write("\n\tminreliability := ")
      apply(node.reliability.get)
      out.write(";")
    }

    node.designSpace match {
      case Nil => /* intentionally empty */

      case l =>
        out.write("\n\tdesign space {\n")
        printSep(l, { (s: CompSubstitution) => apply(s) }, "\n\n")
        out.write("\n\t};")
    }

    out.write("\n};")
  }

  def apply(node: CompSubstitution): Unit = {
    out.write("\t\t")
    apply(node.instance)
    out.write(" -> { ")
    out.write("\n\t\t\tinit ")
    printSep(node.init +: node.alternatives, { (alt: CompAlternative) => {
      apply(alt)
    }
    }, ",\n\t\t\t")
    out.write(" };")
  }

  def apply(node: CompAlternative): Unit = {
    apply(node.compId)
    out.write("( ")
    printSep(
      node.lambdaValues,
      (x: (BoolIdent, DecRealLit)) => {
        out.write(" ")
        apply(x._1)
        out.write(" = ")
        apply(x._2)
      },
      ", "
    )
    out.write(" )")
  }
}

