package translators

import java.io.FileWriter

import com.microsoft.z3.enumerations.Z3_ast_kind._
import com.microsoft.z3.enumerations.Z3_decl_kind._
import com.microsoft.z3.{BoolExpr => Z3BoolExpr}
import util.analyzers.{UnknownZ3AST, WrongBoolExpr}

/**
  * Created by kdelmas on 30/06/17.
  */
object DSLCircuitGraphTranslator {

  val secInputs = "Inputs:"
  val secLatches = "Latches:"
  val secGates = "Gates:"
  val secOutputs = "Outputs:"


  def boolExprToCG(expr: Z3BoolExpr, varMap: Map[Z3BoolExpr, String], writer: FileWriter) {
    val hasBeenDefined = collection.mutable.Map[Z3BoolExpr, String]()
    var counter = 0

    writer.write(secInputs + "\n")
    varMap.foreach(kv => {
      writer.write(kv._2 + ";\n")
      hasBeenDefined(kv._1) = kv._2
    })
    writer.write(secGates + "\n")

    def generateFreshName(): String = {
      val r = s"gate$counter"
      counter += 1
      r
    }

    def boolExprToCGReq(expr: Z3BoolExpr): String = expr.getASTKind match{
        case Z3_VAR_AST =>
          hasBeenDefined(expr)

        case Z3_APP_AST =>
          expr.getFuncDecl.getDeclKind match {
            case Z3_OP_AND =>
              hasBeenDefined.getOrElseUpdate(expr, {
                val args = expr.getArgs.collect({
                  case b: Z3BoolExpr =>
                    boolExprToCGReq(b)
                })
                val name = generateFreshName()
                writer.write(s"$name := ${args.head} & ${args.last};\n")
                name
              })

            case Z3_OP_OR =>
              hasBeenDefined.getOrElseUpdate(expr, {
                val args = expr.getArgs.collect({
                  case b: Z3BoolExpr =>
                    boolExprToCGReq(b)
                })
                val name = generateFreshName()
                writer.write(s"$name := ${args.head} # ${args.last};\n")
                name
              })

            case Z3_OP_NOT =>
              val arg = expr.getArgs.collect({
                case b: Z3BoolExpr => boolExprToCGReq(b)
              })
              s"~${arg.head}"

            case Z3_OP_IMPLIES =>
              hasBeenDefined.getOrElseUpdate(expr, {
                val args = expr.getArgs.collect({
                  case b: Z3BoolExpr =>
                    boolExprToCGReq(b)
                })
                val name = generateFreshName()
                writer.write(s"$name := ${args.head} -> ${args.last};\n")
                name
              })

            case Z3_OP_IFF =>
              hasBeenDefined.getOrElseUpdate(expr, {
                val args = expr.getArgs.collect({
                  case b: Z3BoolExpr =>
                    boolExprToCGReq(b)
                })
                val name = generateFreshName()
                writer.write(s"$name := ${args.head} <-> ${args.last};\n")
                name
              })

            case Z3_OP_ITE =>
              hasBeenDefined.getOrElseUpdate(expr, {
                val args = expr.getArgs.collect({
                  case b: Z3BoolExpr =>
                    boolExprToCGReq(b)
                })
                val name = generateFreshName()
                writer.write(s"$name :=  if ${args.head} then  ${args(1)} else ${args.last};\n")
                name
              })

            case Z3_OP_TRUE =>
              "true"

            case Z3_OP_FALSE =>
              "false"

            case _ => throw WrongBoolExpr(expr.toString)
          }

        case _ => throw UnknownZ3AST(expr.toString)

    }

    val output = boolExprToCGReq(expr)
    writer.write(s"$secOutputs \n $output;")
  }
}
