package translators


import util.analyzers.analyzerTypes._
import com.microsoft.z3.enumerations.Z3_ast_kind._
import com.microsoft.z3.enumerations.Z3_decl_kind._
import com.microsoft.z3.{BoolExpr => Z3BoolExpr}
import util.analyzers.{UnknownEvent, UnknownZ3AST, WrongBoolExpr}

import scala.xml._

/**
  * Created by kdelmas on 06/02/17.
  */
object DSLOpenPSATranslator {
  val treeDef = "define-fault-tree"
  val gateDef = "define-gate"
  val gateCall = "gate"
  val basicEventCall = "basic-event"
  val houseEventCall = "house-event"
  val basicEventDef = "define-basic-event"
  val houseEventDef = "define-house-event"
  val parameterCall = "parameter"
  val parameterDef = "define-parameter"
  val idKey = "name"


  sealed trait Law

  case class Exponential(lambda: Real, time: Option[Double]) extends Law

  private var gateNum = 0

  def generateFreshGateName: String = {
    gateNum += 1
    s"gate$gateNum"
  }

  def createBasicEventDef(name: String, law: Elem): Elem = {
    <define-basic-event name={name}>
      {law}
    </define-basic-event>
  }

  def createBasicEventCall(name: String): Elem = {
      <basic-event name={name}/>
  }

  def createHouseEventDef(name: String, value: Elem): Elem = {
    <define-house-event name={name}>
      {value}
    </define-house-event>
  }

  def createHouseEventCall(name: String): Elem = {
      <house-event name={name}/>
  }

  def createConstantValue[T](value: T): Elem = {
    value match {
      case v: Int => <int value={v.toString}/>

      case v: Boolean => <constant value={v.toString}/>

      case v: Double => <float value={v.toString}/>

      case v: Real =>
        val n = implicitly[Numeric[Real]]
          <float value={n.toDouble(v).toString}/>
    }
  }

  def createMissionTime(value: Option[Double]): Elem = {
      <mission-time value={(for (v <- value) yield v.toString) getOrElse "1.0"}/>
  }

  def createLaw(law: Law): Elem = {
    law match {
      case Exponential(l, t) =>
        <exponential>
          {createConstantValue(l)}{createMissionTime(t)}
        </exponential>
    }
  }

  def createAnd(sons: List[Elem]): Elem = {
    sons match {
      case Nil => createConstantValue(false)
      case h :: Nil => h
      case _ =>
        <define-gate name={generateFreshGateName}>
          <and>
            {sons}
          </and>
        </define-gate>
    }
  }

  def createITE(i: Elem, t: Elem, e: Elem): List[Elem] = {
    val tTr = createAnd(i :: t :: Nil)
    val eTr = createAnd(createNot(i) :: e :: Nil)
    val iteTr = createOr(tTr :: eTr :: Nil map createCall)
    iteTr :: tTr :: eTr :: Nil
  }

  def createIFF(l: Elem, r: Elem): List[Elem] = {
    val pTr = createAnd(l :: r :: Nil)
    val nTr = createAnd(createNot(l) :: createNot(r) :: Nil)
    val iffTr = createOr(pTr :: nTr :: Nil map createCall)
    iffTr :: pTr :: nTr :: Nil
  }

  def createImplies(l: Elem, r: Elem): Elem = {
    createOr(createNot(l) :: r :: Nil)
  }

  def createOr(sons: List[Elem]): Elem = {
    sons match {
      case Nil => createConstantValue(true)
      case h :: Nil => h
      case _ =>
        <define-gate name={generateFreshGateName}>
          <or>
            {sons}
          </or>
        </define-gate>
    }
  }

  def createNot(son: Elem): Elem = {
    <not>
      {son}
    </not>
  }

  def createAtLeast(sons: List[Elem], min: Int): Elem = {
    sons match {
      case Nil => createConstantValue(false)
      case h :: Nil => h
      case _ =>
        <define-gate name={generateFreshGateName}>
          <atleast min={min.toString}>
            {sons}
          </atleast>
        </define-gate>
    }
  }

  def createGateCall(name: String): Elem = {
      <gate name={name}/>
  }

  def createParamDef(name: String, value: Elem): Elem = {
    <define-parameter name={name}>
      {value}
    </define-parameter>
  }

  def createParamCall(name: String): Elem = {
      <parameter name={name}/>
  }

  def createCall(e: Elem): Elem = {
    e.label match {
      case s if s == gateDef =>
        createGateCall(e.attribute(idKey).get.head.text)
      case s if s == parameterDef =>
        createParamCall(e.attribute(idKey).get.head.text)
      case s if s == basicEventDef =>
        createBasicEventCall(e.attribute(idKey).get.head.text)
      case s if s == houseEventDef =>
        createHouseEventCall(e.attribute(idKey).get.head.text)
      case s if s == gateCall | s == parameterCall | s == basicEventCall | s == houseEventCall => e
    }
  }


  def boolExprToPSA(expr: Z3BoolExpr, varMap: Map[Z3BoolExpr, (String, Real, Option[Double])]): Elem = {
    val hasBeenDefined = collection.mutable.Map[Z3BoolExpr, Elem]()
    gateNum = 0
    val sons = scala.collection.mutable.ListBuffer[Elem]()

    def boolExprToPSARec(expr: Z3BoolExpr): Elem = {
      expr.getASTKind match {
        case Z3_VAR_AST =>
          val eDef = hasBeenDefined.getOrElseUpdate(expr, {
            val params = varMap.getOrElse(expr, throw UnknownEvent(expr.getFuncDecl.getName.toString))
            val r = createBasicEventDef(params._1, createLaw(Exponential(params._2, params._3)))
            sons += r
            r
          })
          createCall(eDef)
        case Z3_APP_AST =>
          expr.getFuncDecl.getDeclKind match {
            case Z3_OP_AND =>
              val gDef = hasBeenDefined.getOrElseUpdate(expr, {
                val args = expr.getArgs.collect({
                  case b: Z3BoolExpr => boolExprToPSARec(b)
                })
                val r = createAnd(args.toList)
                sons += r
                r
              })
              createCall(gDef)
            case Z3_OP_OR =>
              val gDef = hasBeenDefined.getOrElseUpdate(expr, {
                val args = expr.getArgs.collect({
                  case b: Z3BoolExpr => boolExprToPSARec(b)
                })
                val r = createOr(args.toList)
                sons += r
                r
              })
              createCall(gDef)
            case Z3_OP_NOT =>
              val args = expr.getArgs.collect({
                case b: Z3BoolExpr => boolExprToPSARec(b)
              })
              createNot(args.head)
            case Z3_OP_IMPLIES =>
              val gDef = hasBeenDefined.getOrElseUpdate(expr, {
                val args = expr.getArgs.collect({
                  case b: Z3BoolExpr => boolExprToPSARec(b)
                })
                val r = createImplies(args.head, args.last)
                sons += r
                r
              })
              createCall(gDef)
            case Z3_OP_IFF =>
              val gDef = hasBeenDefined.getOrElseUpdate(expr, {
                val args = expr.getArgs.collect({
                  case b: Z3BoolExpr => boolExprToPSARec(b)
                })
                val gDef = createIFF(args.head, args.last)
                sons ++= gDef
                gDef.head
              })
              createCall(gDef)
            case Z3_OP_ITE =>
              val gDef = hasBeenDefined.getOrElseUpdate(expr, {
                val args = expr.getArgs.collect({
                  case b: Z3BoolExpr => boolExprToPSARec(b)
                })
                val gDef = createITE(args.head, args(1), args(2))
                sons ++= gDef
                gDef.head
              })
              createCall(gDef)
            case Z3_OP_TRUE =>
              val hDef = hasBeenDefined.getOrElseUpdate(expr, {
                val r = createHouseEventDef("hTrue", createConstantValue(true))
                sons += r
                r
              })
              createCall(hDef)

            case Z3_OP_FALSE =>
              val hDef = hasBeenDefined.getOrElseUpdate(expr, {
                val r = createHouseEventDef("hFalse", createConstantValue(false))
                sons += r
                r
              })
              createCall(hDef)
            case _ => throw WrongBoolExpr(expr.toString)
          }
        case _ => throw UnknownZ3AST(expr.toString)
      }
    }

    boolExprToPSARec(expr)
    val root = sons.last
    val gates = sons.init
    val top = {
      <define-gate name="top">
        {root.child}
      </define-gate>
    }

    <open-psa>
      {gates}{top}
    </open-psa>
  }
}
