package translators

import ast.dsl.{BoolExpr, IntExpr, _}
import com.microsoft.z3.{BitVecExpr, AST => Z3AST, BoolExpr => Z3BoolExpr, Context => Z3Context, Expr => Z3Expr, IntExpr => Z3IntExpr, Symbol => Z3Symbol, _}
import namespace.dsl.Namespace
import translators.translatorsTypes._
import util.translators._

/**
  * Created by kdelmas on 23/12/15.
  */

/**
  * Trait translating a KCR program to a SMT problem
  */
class DSLSMTTranslator {

  import util.translators.CacheImplicits._

  private val _DSLContext = new MutableCache

  private val _SMTContext = SMTContext()

  private val _tupleName = "Tuple"

  private val _accessorName = "field"

  private val _phi = StdDSLSymbol('phi)

  private val _empty = StdDSLSymbol('empty)

  /**
    * Translate a custom tuple to a Z3 tuple
    *
    * @param t custom tuple
    * @return Z3 tuple
    */
  private def tupleToZ3Tuple(t: Tuple) = {
    val z3Accessors = t.sorts.indices.map(i => _SMTContext.mkSymbol(_accessorName + i))
    _SMTContext.mkTupleSort(
      _SMTContext.mkSymbol(t.name),
      z3Accessors.toArray,
      t.sorts.toArray[Sort])
  }

  /**
    * Gives the SMT accessors of a given tuple
    *
    * @param t custom tuple
    * @return accessors list sorted by field
    */
  private def tupleAccessors(t: Tuple) = t.sorts.indices.map(i => _SMTContext.mkSymbol(_accessorName + i))

  /**
    * translate a population count
    *
    * @param kids      the Boolean expressions
    * @param relop     the operator
    * @param cst       the integer constant
    * @param namespace the current namespace
    * @return SMT encoding of pop count
    */
  private def _trPopCount(kids: List[BoolExpr],
                          relop: RelOp,
                          cst: Int,
                          namespace: Namespace): Z3BoolExpr = relop match {
    case Ge =>
      val z3Kids = kids.map(e => _trBoolExpr(e, namespace))
      _SMTContext.atLeastNBDD(z3Kids, cst)
    case Gt =>
      _trPopCount(kids, Ge, cst + 1, namespace)
    case Le =>
      _SMTContext.mkNot(_trPopCount(kids, Gt, cst, namespace))
    case Lt =>
      _SMTContext.mkNot(_trPopCount(kids, Ge, cst, namespace))
    case Eq =>
      _SMTContext.mkAnd(_trPopCount(kids, Le, cst, namespace), _trPopCount(kids, Ge, cst, namespace))
  }

  /**
    * Translate an AST integer representation to an integer
    *
    * @param node AST integer representation
    * @return corresponding integer
    */
  private def _trIntExpr(node: IntExpr): Int = node match {
    case DecIntLit(repr: String) => repr.toInt
  }

  /**
    * translate a boolean expression
    *
    * @param node      AST Boolean expression
    * @param namespace the current namespace
    * @return SMT Boolean expression
    */
  private def _trBoolExpr(node: BoolExpr, namespace: Namespace): Z3BoolExpr = node match {

    case And(kid1: BoolExpr, kid2: BoolExpr) =>
      _SMTContext.mkAnd(_trBoolExpr(kid1, namespace), _trBoolExpr(kid2, namespace))

    case id: BoolIdent =>
      _DSLContext(generateUniqueKey(id, namespace))

    case FlowContains(kid1: FlowExpr, kid2: FlowExpr) =>
      _SMTContext.mkEq(_trFlowExpr(kid1, namespace), _SMTContext.mkBVAND(_trFlowExpr(kid1, namespace), _trFlowExpr(kid2, namespace)))

    case FlowEq(kid1: FlowExpr, kid2: FlowExpr) =>
      _SMTContext.mkEq(_trFlowExpr(kid1, namespace), _trFlowExpr(kid2, namespace))

    case FlowNeq(kid1: FlowExpr, kid2: FlowExpr) =>
      _SMTContext.mkNot(_SMTContext.mkEq(_trFlowExpr(kid1, namespace), _trFlowExpr(kid2, namespace)))

    case Iff(kid1: BoolExpr, kid2: BoolExpr) =>
      _SMTContext.mkIff(_trBoolExpr(kid1, namespace), _trBoolExpr(kid2, namespace))

    case Implies(kid1: BoolExpr, kid2: BoolExpr) =>
      _SMTContext.mkImplies(_trBoolExpr(kid1, namespace), _trBoolExpr(kid2, namespace))

    case Not(kid1: BoolExpr) =>
      _SMTContext.mkNot(_trBoolExpr(kid1, namespace))

    case Or(kid1: BoolExpr, kid2: BoolExpr) =>
      _SMTContext.mkOr(_trBoolExpr(kid1, namespace), _trBoolExpr(kid2, namespace))

    case PopCount(kids: List[BoolExpr], relOp: RelOp, cst: DecIntLit) =>
      _trPopCount(kids, relOp, _trIntExpr(cst), namespace)
  }

  /**
    * translate a flow expression
    *
    * @param node      AST flow expression
    * @param namespace the current namespace
    * @return SMT Bitvector expression
    */
  private def _trFlowExpr(node: FlowExpr, namespace: Namespace): BitVecExpr = node match {

    case FlowInter(kid1: FlowExpr, kid2: FlowExpr) =>
      _SMTContext.mkBVAND(_trFlowExpr(kid1, namespace), _trFlowExpr(kid2, namespace))

    case id: FlowIdent => _trFlowIdent(id, namespace)

    case FlowIte(kid1: BoolExpr, kid2: FlowExpr, kid3: FlowExpr) =>
      _SMTContext.mkIte(_trBoolExpr(kid1, namespace), _trFlowExpr(kid2, namespace), _trFlowExpr(kid3, namespace))

    case FlowUnion(kid1: FlowExpr, kid2: FlowExpr) =>
      _SMTContext.mkBVOR(_trFlowExpr(kid1, namespace), _trFlowExpr(kid2, namespace))
  }

  /**
    * translate a flow identifier or constant
    *
    * @param node      AST flow identifier
    * @param namespace the current namespace
    * @return SMT Bitvector expression
    */
  private def _trFlowIdent(node: FlowIdent, namespace: Namespace): BitVecExpr = {
    val key = generateUniqueKey(node, namespace)
    node match {

      case id: BasicFlowIdent =>
        namespace.locate(id) match {
          case Left(valueType) =>

            //constant value then create a bitvector constant
            _DSLContext.getOrElseUpdate(key, generateValue(id.baseId, valueType))

          case Right(_) =>
            //flow identifier then get expression from context
            _DSLContext(key)
        }

      case id: InstFlowIdent =>
        namespace.locate(id) match {
          case Left(valueType) =>

            //constant value then create a bitvector constant
            _DSLContext.getOrElseUpdate(key, generateValue(id.baseId, valueType))

          case Right(_) =>

            //flow identifier then get expression from context
            _DSLContext(key)
        }

      case TypeValueFlowIdent(typeId: TypeIdent, valueId: DSLSymbol) =>

        //constant value then create a bitvector constant
        _DSLContext.getOrElseUpdate(key, generateValue(valueId, namespace.locate(typeId)))

      case TypeValueFlowEmpty(typeId: TypeIdent) =>

        //constant value then create a bitvector constant
        _DSLContext.getOrElseUpdate(key, generateEmptyElement(namespace.locate(typeId)))
    }
  }

  /**
    * Translate of a component
    *
    * @param comp AST component declaration
    */
  private def _trCompDeclaration(comp: ComponentDeclaration): Unit = {

    //translate inputs as bounded bitvectors variables
    comp.inputs.zipWithIndex.foreach(p =>
      _DSLContext.update(generateUniqueKey(p._1.ident, comp.namespace) ->
        _SMTContext.mkBitVecBound(p._2 + comp.body.events.size, _SMTContext.mkBitVecSort(comp.namespace.locate(p._1.typeId).values.size)))
    )

    //translate events as bounded Boolean variables
    comp.body.events.zipWithIndex.foreach(p => _DSLContext.update(generateUniqueKey(p._1, comp.namespace) -> _SMTContext.mkBoolBound(p._2)))


    // inline all the events of the subcomponents
    val inlinedEvents = comp.namespace.flowDepGraph.compCallsInOrder.toList.flatMap(call => {
      val key = List(comp.absoluteIdent.baseId, call.rhs.instId)

      val inline: PartialFunction[(EventKey, Z3BoolExpr), PathIdentType] = {
        case (ekey@EventKey(a :: _), _) if a == call.rhs.compId.baseId =>
          key ++ ekey.path.tail
      }
      _DSLContext.collect(inline).toList
    })

    //build a bounded Boolean variable for each of the inline events
    inlinedEvents.zipWithIndex.foreach(p => _DSLContext.update(EventKey(p._1) -> _SMTContext.mkBoolBound(p._2 + comp.body.events.size + comp.inputs.size)))

    //translate flow definitions
    comp.namespace.flowDepGraph.compDefsInOrder.toList.distinct.foreach(d => _trFlowDefinition(d, comp.body.namespace))

    //define the bitvector sorts of the outputs
    val outputSorts = comp.outputs.map(e => _SMTContext.mkBitVecSort(comp.namespace.locate(e.typeId).values.size))

    //define the tuple sort gathering the outputs of the component
    val tupleSort = Tuple(_tupleName + comp.outputs.size + outputSorts.map(s => s.getSize).mkString("_"), outputSorts)

    //add the translations of the component outputs to the context
    val outDefs = comp.outputs.map(e => _DSLContext(FlowKey(List(comp.absoluteIdent.baseId, e.ident.baseId))))

    //add the output tuple sort to the context
    _DSLContext.update(generateUniqueKey(tupleSort) -> tupleSort)

    //associate to the component to the tuple containing the definition of itd outputs
    _DSLContext.update(generateUniqueKey(comp.ident) -> tupleToZ3Tuple(tupleSort).mkDecl().apply(outDefs: _*).asInstanceOf[DatatypeExpr])
  }

  /**
    * Translate a flow definition
    *
    * @param node      AST node of flow definition
    * @param namespace the current namespace
    */
  private def _trFlowDefinition(node: FlowDefinition, namespace: Namespace): Unit = {

    //get component identifier
    val compPr = namespace.source match {
      case source: ComponentDeclaration => source.absoluteIdent.baseId
      case source: ComponentBody => source.parent.absoluteIdent.baseId
    }

    node match {
      case CompCallFlowDefinition(lhs: List[BasicFlowIdent], rhs: CompCall) =>

        val compExpr = _DSLContext(CompKey(List(rhs.compId.baseId)))

        val getParentEvt: PartialFunction[(EventKey, Z3BoolExpr), (Z3BoolExpr, Z3BoolExpr)] = {
          case (ekey@EventKey(a :: b :: _), expr) if a == compPr && b == rhs.instId =>
            val sonKey = rhs.compId.baseId :: ekey.path.tail.tail
            (_DSLContext(EventKey(sonKey)), expr)
        }

        //find the declaration of the component
        val cDecl = namespace.locate(rhs.compId)

        //find the event of the parent component inlined from this component
        val subEvtMap = _DSLContext.collect(getParentEvt)

        //find the input and associate the translation of the arguments passed in the component instantiation
        val sonInputs = namespace.locate(rhs.compId).inputs.map(e => e.ident.baseId)
        val argsMap = rhs.args.zipWithIndex.map(p => (_DSLContext(FlowKey(List(rhs.compId.baseId, sonInputs(p._2)))), _trFlowExpr(p._1, namespace)))

        //find the expression of the ouput and substitute the varaible by the args and inlined events variables
        val listVars = (subEvtMap ++ argsMap).unzip
        val newCompExpr = compExpr.substitute(listVars._1.toArray, listVars._2.toArray)

        //biuld the tuple of output expressions
        val outSort = _DSLContext(generateUniqueKey(cDecl.outputs.size, cDecl.outputs.map(o => namespace.locate(o.typeId).values.size)))
        val outAccess = tupleToZ3Tuple(outSort).getFieldDecls

        //link it to the left hand flows of the definition
        lhs.zipWithIndex.foreach(p => _DSLContext.update(generateUniqueKey(p._1, namespace) -> outAccess(p._2).apply(newCompExpr).asInstanceOf[BitVecExpr]))

      case SingleFlowDefinition(lhs: BasicFlowIdent, rhs: FlowExpr) =>

        //associate to the flow to the translation of the flow expression
        _DSLContext.update(generateUniqueKey(lhs, namespace) -> _trFlowExpr(rhs, namespace))
    }
  }

  /**
    * Translate a configuration
    *
    * @param conf the AST configuration
    */
  def trConfig(conf: ConfigurationDeclaration): (ImmutableCache, SMTContext) = {

    //locate the root component
    val rootComp = conf.namespace.locate(conf.rootCompId)

    //translate all component instances of the system
    rootComp.namespace.compInstanceDepGraph.inOrder.foreach(comp => {
      if (rootComp.namespace.locateDef(comp.data).nonEmpty) {
        val cDecl = rootComp.namespace.locate(rootComp.namespace.locateDef(comp.data).head.rhs.compId)
        if (_DSLContext.get(CompKey(List(cDecl.ident.baseId))).isEmpty)
          _trCompDeclaration(cDecl)
      } else
        _trCompDeclaration(rootComp)
    })

    //translate the alternative components
    val altComps = conf.designSpace.flatMap(sub => sub.alternatives.map(alt => alt.compId)).distinct
    altComps.foreach(comp => {
      if (_DSLContext.get(CompKey(List(comp.baseId))).isEmpty)
        _trCompDeclaration(rootComp.namespace.locate(comp))
    })

    //translate the failure condition
    _DSLContext.update(EventKey(List(conf.ident.baseId, _phi)) -> _trBoolExpr(conf.failureCondition, rootComp.namespace))

    //return the translation context and the Z3 SMT context
    (_DSLContext.toImmutable, _SMTContext)
  }

  private def generateUniqueKey(comp: CompIdent) = CompKey(List(comp.baseId))

  /**
    * Generate a unique cache key for the flow identifiers
    *
    * @param flow      flow identifier
    * @param namespace evaluation namespace
    * @return the cache key
    */
  def generateUniqueKey(flow: FlowIdent, namespace: Namespace): FlowKey = flow match {
    case id@BasicFlowIdent(baseId: DSLSymbol) =>
      val compPr = namespace.source match {
        case source: ComponentDeclaration => source.absoluteIdent.baseId
        case source: ComponentBody => source.parent.absoluteIdent.baseId
      }
      namespace.locate(id) match {
        case Left(valueType) => FlowKey(List(valueType.ident.baseId, baseId))
        case Right(_) => FlowKey(List(compPr, baseId))
      }
    case id@InstFlowIdent(baseId: DSLSymbol, _: CompInstanceIdent) =>
      val compPr = namespace.source match {
        case source: ComponentDeclaration => source.absoluteIdent.baseId
        case source: ComponentBody => source.parent.absoluteIdent.baseId
      }
      namespace.locate(id) match {
        case Left(valueType) => FlowKey(List(valueType.ident.baseId, baseId))
        case Right(_) => FlowKey(List(compPr, baseId))
      }
    case TypeValueFlowEmpty(typeId: TypeIdent) =>
      FlowKey(List(typeId.baseId, _empty))
    case TypeValueFlowIdent(typeId: TypeIdent, valueId: DSLSymbol) =>
      FlowKey(List(typeId.baseId, valueId))
  }

  private def generateUniqueKey(tuple: Tuple) = {
    TypeKey(List(StdDSLSymbol(Symbol(s"${_tupleName}${tuple.sorts.size}${tuple.sorts.map(s => s"BV${s.getSize}").mkString("_", "_", "")}"))))
  }

  private def generateUniqueKey(size: Int, sortsSize: List[Int]) = {
    TypeKey(List(StdDSLSymbol(Symbol(s"${_tupleName}$size${sortsSize.map(i => s"BV$i").mkString("_", "_", "")}"))))
  }

  private def generateUniqueKey(event: BoolIdent, namespace: Namespace) = namespace.source match {

    case source: ComponentDeclaration =>
      EventKey(List(source.absoluteIdent.baseId, event.baseId))

    case source: ComponentBody =>
      EventKey(List(source.parent.absoluteIdent.baseId, event.baseId))
  }

  private def generateValue(value: DSLSymbol, tyype: TypeDeclaration): BitVecExpr = {
    _SMTContext.mkBV(Math.pow(2.0, tyype.values.indexOf(value)).toInt, tyype.values.size)
  }

  private def generateEmptyElement(tyype: TypeDeclaration): BitVecExpr = {
    _SMTContext.mkBV(0, tyype.values.size)
  }
}
