package translators

import java.io.Writer

import ast.dsl._

/**
  * Created by kdelmas on 10/02/17.
  */
object DSLToSMVTranslator {
  private val module = "MODULE"
  private val varSec = "VAR"
  private val ivarSec = "IVAR"
  private val assignSec = "ASSIGN"
  private val defineSec = "DEFINE"
  private val initOf = (s: String) => s"init($s)"
  private val nextOf = (s: String) => s"next($s)"
  private val True = "TRUE"
  private val False = "FALSE"
  private val newLine = System.lineSeparator()
  private val idOf: String => String = s => {
    val x = s.replace("|", "").replace(".", "_").replace("::", "_")

    if (x == "in")
      "entry"
    else
      x
  }

  private val eventModule = "eventModule"
  private val eventState = "eventState"
  private val eventTrig = "eventTrig"
  private val hasOccurred = "hasOccurred"
  private val evtHasOccured: BoolIdent => String = e => s"$e.$hasOccurred"

  private val faultExtension = "FAULT EXTENSION "
  private val extendModule: String => String = s => s"EXTENSION OF MODULE $s"
  private val slice = (name: String, vars: List[String]) => s"SLICE $name AFFECTS ${vars.mkString(",")} WITH"
  private val mode = (name: String, value: String, affected: String, event: String) =>
    s"""MODE $name : Permanent StuckAtByValue_D(
       |data term << $value,
       |data input << $affected,
       |data varout >> $affected,
       |event failure >> $event);
    """.stripMargin

  def makeFEI(p: ConfigurationDeclaration, to: Writer): Unit = {
    to.write(faultExtension + idOf(p.rootCompId.baseId.name) + newLine)
    to.write(extendModule(eventModule) + newLine)
    to.write(slice(eventModule + "_faults", eventState :: Nil) + newLine)
    to.write(mode(eventModule + "_" + hasOccurred, "TRUE", eventState, eventTrig))
    to.write(newLine)
  }


  def translateConfiguration(p: ConfigurationDeclaration, to: Writer): Unit = {
    to.write(s"$module  $eventModule  $newLine")
    to.write(varSec + newLine)
    to.write(s"$eventState : boolean;" + newLine)
    to.write(ivarSec + newLine)
    to.write(s"$eventTrig : boolean;" + newLine)
    to.write(assignSec + newLine)
    to.write(s"${initOf(eventState)} := FALSE;" + newLine)
    to.write(s"${nextOf(eventState)} := $eventState;" + newLine)
    to.write(defineSec + newLine)
    to.write(s"$hasOccurred := $eventState;" + newLine + newLine)

    val root = p.namespace.locate(p.rootCompId)
    val toTranslate = root.namespace.flowDepGraph.compCallsInOrder.map(call => root.namespace.locate(call.rhs.compId)).toSet
    toTranslate.foreach(c => translateComp(c, to))
    translateComp(root, to, isRoot = true)
    to.write(defineSec + newLine)
    to.write("FC := ")
    to.write(translateBoolExpr(p.failureCondition, root, None))
    to.write(";" + newLine)
  }

  private def translateComp(c: ComponentDeclaration, to: Writer, isRoot: Boolean = false): Unit = {
    if (!isRoot) {
      to.write(s"$module ${idOf(c.ident.baseId.name)}")
      if (c.inputs.nonEmpty) {
        val in = c.inputs.flatMap(d => {
          val tyype = c.namespace.locate(d.typeId)
          val modes = tyype.values.map(v => TypeValueFlowIdent(tyype.ident, v))
          modes.map(v => basicFlowIdentValueId(d.ident, v))
        }).mkString("(", ",", ")")
        to.write(in)
      }
      to.write(newLine)
    } else {
      to.write(s"$module main " + newLine)
    }
    to.write(newLine)

    if (c.body.events.nonEmpty) {
      to.write(varSec + newLine)
      c.body.events.foreach(e => to.write(s"${e.baseId.name} : $eventModule;" + newLine))
    }
    c.namespace.flowDepGraph.compDefsInOrder.toList.distinct.foreach(fdef => {
      translateFlowDefinition(fdef, to)
    })
    to.write(newLine)
  }

  private def translateFlowDefinition(f: FlowDefinition, to: Writer): Unit = {
    f match {
      case SingleFlowDefinition(lhs, rhs) =>
        to.write(defineSec + newLine)
        val trF = translateFlowExpr(rhs, f.parent.parent, Some(lhs))
        trF.map(p => s"${basicFlowIdentValueId(lhs, p._1)} := ${p._2};" + newLine).foreach(to.write)

      case CompCallFlowDefinition(lhs, rhs) =>
        to.write(varSec + newLine)
        to.write(s"${idOf(rhs.instId.name)} : ${idOf(rhs.compId.baseId.name)}${
          val args = rhs.args.collect({
            case id: BasicFlowIdent =>
              val tyype = typeOf(f.parent.parent, id)
              val modes = tyype.values.map(v => TypeValueFlowIdent(tyype.ident, v))
              modes.map(v => basicFlowIdentValueId(id, v))
            case InstFlowIdent(id, _) =>
              val bID = BasicFlowIdent(id)
              val tyype = typeOf(f.parent.parent, bID)
              val modes = tyype.values.map(v => TypeValueFlowIdent(tyype.ident, v))
              modes.map(v => basicFlowIdentValueId(bID, v))
            case e => throw ExprNotHandledByTranslation(e, "SMV")
          }).flatten
          if (args.nonEmpty) args.mkString("(", ",", ")") else ""
        };" + newLine)
        to.write(defineSec + newLine)
        val iniComp = f.parent.parent.parent.namespace.locate(rhs.compId)
        val outAssoc = lhs.zip(iniComp.outputs)
        outAssoc.flatMap(p => {
          val tyype = typeOf(f.parent.parent, p._1)
          val fmodes = tyype.values.map(v => TypeValueFlowIdent(tyype.ident, v))
          fmodes.map(v => s"${basicFlowIdentValueId(p._1, v)} := ${idOf(rhs.instId.name)}.${basicFlowIdentValueId(p._2.ident, v)};" + newLine)
        }).foreach(to.write)
    }
  }

  private def mkAnd(b1: String, b2: String): String = {
    b1 match {
      case True => b2
      case False => False
      case v1 =>
        b2 match {
          case True => v1
          case False => False
          case v2 if v2.compareTo(v1) == 0 => v2
          case v2 => s"($v1 & $v2)"
        }
    }
  }

  private def mkOr(b1: String, b2: String): String = {
    b1 match {
      case True => True
      case False => b2
      case v1 =>
        b2 match {
          case True => True
          case False => v1
          case v2 if v2.compareTo(v1) == 0 => v2
          case v2 => s"($v1 | $v2)"
        }
    }
  }

  private def mkNot(b: String): String = {
    b match {
      case True => False
      case False => True
      case x =>
        val p = "\\(! (.*)\\)".r
        if (x.matches("\\(! .*\\)")) {
          val p(r) = x
          r
        } else {
          s"(! $x)"
        }
    }
  }

  private def mkIte(i: String, t: String, e: String): String = {
    i match {
      case True => t
      case False => e
      case _ =>
        t match {
          case True =>
            e match {
              case True => True
              case False => i
              case _ => s"($i ? $t : $e)"
            }
          case False =>
            e match {
              case True => mkNot(i)
              case False => False
              case _ => s"($i ? $t : $e)"
            }
          case _ => s"($i ? $t : $e)"
        }
    }
  }

  private def mkEq(l: String, r: String): String = {
    l match {
      case True => r
      case False => mkNot(r)
      case x =>
        r match {
          case True => x
          case False => mkNot(x)
          case _ =>
            s"($l = $r)"
        }
    }
  }

  private def typeOf(c: ComponentDeclaration, b: FlowIdent) = {
    c.body.namespace.locate(b) match {
      case Left(tyype) => tyype
      case Right(fdecl) => c.parent.namespace.locate(fdecl.typeId)
    }
  }

  private def basicFlowIdentValueId(f: BasicFlowIdent, v: TypeValueFlowIdent): String = {
    s"${idOf(f.baseId.name)}_is_${idOf(v.valueId.name)}"
  }

  private def translateFlowExpr(f: FlowExpr, c: ComponentDeclaration, isDef: Option[BasicFlowIdent]): Map[TypeValueFlowIdent, String] = {
    f match {
      case FlowIte(i, t, e) =>
        val iS = translateBoolExpr(i, c, isDef)
        val eS = translateFlowExpr(e, c, isDef)
        val tS = translateFlowExpr(t, c, isDef)
        eS.map(p => p._1 -> mkIte(iS, tS(p._1), p._2))

      case FlowInter(f1, f2) =>
        val f1S = translateFlowExpr(f1, c, isDef)
        val f2S = translateFlowExpr(f2, c, isDef)
        f1S.map(p => p._1 -> mkAnd(p._2, f2S(p._1)))

      case FlowUnion(f1, f2) =>
        val f1S = translateFlowExpr(f1, c, isDef)
        val f2S = translateFlowExpr(f2, c, isDef)
        f1S.map(p => p._1 -> mkOr(p._2, f2S(p._1)))

      case id: BasicFlowIdent =>
        val tyype = typeOf(c, id)
        tyype.values.map(s => TypeValueFlowIdent(tyype.ident, s)).map(v => v -> basicFlowIdentValueId(id, v)).toMap

      case id: InstFlowIdent =>
        val bId = BasicFlowIdent(id.baseId)
        val tyype = typeOf(c, id)
        tyype.values.map(s => TypeValueFlowIdent(tyype.ident, s)).map(v => v -> basicFlowIdentValueId(bId, v)).toMap

      case v: TypeValueFlowIdent =>
        val tyype = typeOf(c, v)
        tyype.values.map(s => TypeValueFlowIdent(tyype.ident, s)).map(x => x -> {
          if (v == x) True else False
        }).toMap

      case empty: TypeValueFlowEmpty =>
        val tyype = typeOf(c, empty)
        tyype.values.map(s => TypeValueFlowIdent(tyype.ident, s)).map(v => v -> False).toMap


      case _ => throw ExprNotHandledByTranslation(f, "SMV")
    }
  }

  private def translateBoolExpr(b: BoolExpr, c: ComponentDeclaration, isDef: Option[BasicFlowIdent]): String = {
    b match {
      case And(b1, b2) =>
        val sB1 = translateBoolExpr(b1, c, isDef)
        val sB2 = translateBoolExpr(b2, c, isDef)
        mkAnd(sB1, sB2)

      case Or(b1, b2) =>
        val sB1 = translateBoolExpr(b1, c, isDef)
        val sB2 = translateBoolExpr(b2, c, isDef)
        mkOr(sB1, sB2)

      case Not(b1) =>
        mkNot(translateBoolExpr(b1, c, isDef))

      case PopCount(kids, op, cst) =>
        val sonsTr = kids.map(b => translateBoolExpr(b, c, isDef))
        val opTr = op match {
          case Ge => ">="
          case Gt => ">"
          case Le => "<="
          case Lt => "<"
          case Eq => "="
        }
        val cstTr = cst.repr
        s"count(${sonsTr.mkString(",")}) $opTr $cstTr"

      case FlowEq(f1, f2) =>
        val f1S = translateFlowExpr(f1, c, isDef)
        val f2S = translateFlowExpr(f2, c, isDef)
        val eq = f1S.map(p => p._1 -> mkEq(p._2, f2S(p._1))).foldLeft(True)((acc, v) => mkAnd(acc, v._2))
        eq

      case FlowNeq(f1, f2) => translateBoolExpr(Not(FlowEq(f1, f2)), c, isDef)

      case FlowContains(f1, f2) =>
        val f1S = translateFlowExpr(f1, c, isDef)
        val f2S = translateFlowExpr(f2, c, isDef)
        val imp = f1S.map(p => p._1 -> mkOr(mkNot(p._2), f2S(p._1))).foldLeft(True)((acc, v) => mkAnd(acc, v._2))
        imp

      case e: BoolIdent =>
        if(isDef.isDefined){
          evtHasOccured(e)
        }else{
          throw ForbiddenInFC(e)
        }

      case _ => throw new Exception(s"unhandled expression $b for SMV translation")
    }
  }
}
