package ui

import java.awt.{BorderLayout, LayoutManager}
import java.io.FileWriter
import java.nio.file.Files
import javax.swing.JPanel

import analyzers._
import util.analyzers.analyzerTypes._
import ast.dsl.{ConfigIdent, ConfigurationDeclaration}
import interpreter.TraceInterpreter
import namespace.dsl.NamespaceException
import util.analyzers._
import util.io.{FileManager, LoadDSLFiles}

import scala.collection.mutable
import scala.util.{Failure, Success, Try}

/**
  * Created by kdelmas on 03/02/17.
  */
case class CodeContainer(code: KCRTextPane, config: Config, mlayout: LayoutManager = new BorderLayout(), uiConsole: ConsoleTextPane) extends JPanel(mlayout) {

  val analyzers: mutable.Map[String, FullAnalyzer] = mutable.Map.empty

  def time[A](f: () => A): (A, Double) = {
    val start = System.nanoTime()
    val r = f()
    val t = (System.nanoTime() - start) * 1e-9
    (r, t)
  }

  def updateRequest(console: ConsoleTextPane, conf: ConfigIdent): Unit = {
    if (!analyzers.contains(conf.baseId.symbol.name) || code.modificationPending) {
      for (analyser <- buildAnalyzer(conf.baseId.symbol.name)) analyzers(conf.baseId.symbol.name) = analyser
    } else {
      console.printInfo("already up-to-date")
    }
  }

  def buildAnalyzer(confName: String): Option[FullAnalyzer] = {
    Try{
      code.getAndUpdateAvailableConfig
      code.modificationProcessed()
      new FileParsing
        with DelegatedConsole
        with SMTTranslation
        with AcceptableCardAnalyzer
        with MonotonyAnalyser
        with CutEnumeratorAll
        with ReliabilityComputer
        with AcceptableReliabilityAnalyzer
        with TraceInterpreter
        with SubstitutionAnalyzerSim
        with STMDDBuilderSim
        with DSEProblemSolverAll
        with SolutionAnalyzer
        with CardinalityComputer
        with SymbolicFailureRateComputer
        with MDDDrawer
        with ArchitectureDrawer
        with TraceDrawer
        with OpenPSAExporter
        with SMVExporter
        with HipHOPSExporter
        with CircuitGraphExporter {
        val fileNames: List[String] = code.file.getAbsolutePath :: Nil
        val confId: String = confName
        val BDDFactory = new BDDFactory with Z3BoolExprToBDDTranslator
        val InstBoolBDDFactory = new InstBoolBDDFactory
        val console: Console = uiConsole
      }
    } match {
      case Success(analyzer)=>
        uiConsole.printInfo(s"successful build")
        Some(analyzer)
      case Failure(e) =>
         uiConsole.printError(s"build failed: $e")
        None
    }
  }

  def getAnalyser(conf: ConfigIdent): Option[FullAnalyzer] = {
    analyzers.get(conf.baseId.symbol.name) match {
      case Some(analyser) if !code.modificationPending => Some(analyser)
      case Some(_) if code.modificationPending =>
        val newAnalyzer = buildAnalyzer(conf.baseId.symbol.name)
        for (analyser <- newAnalyzer) analyzers(conf.baseId.symbol.name) = analyser
        newAnalyzer
      case None =>
        val newAnalyzer = buildAnalyzer(conf.baseId.symbol.name)
        for (analyser <- newAnalyzer) analyzers(conf.baseId.symbol.name) = analyser
        newAnalyzer

    }
  }

  def getSystemDrawers(console: ConsoleTextPane, conf: ConfigIdent): Option[ArchitectureDrawer with MDDDrawer] = {
    for (a <- getAnalyser(conf)) yield a
  }

  def getPhiDrawer(console: ConsoleTextPane, conf: ConfigIdent): Option[PhiDrawer] = {
    for (a <- getAnalyser(conf)) yield {
      new Phi(a.phi) with BDDBuild with PhiDrawer {
        val BDDFactory: BDDFactory with Z3BoolExprToBDDTranslator = a.BDDFactory
      }
    }
  }
}
