package ui

/**
  * Created by kdelmas on 05/01/17.
  */

import java.awt.{Color, Font}
import javax.swing.JScrollPane

import analyzers._
import ast.dsl._
import net.sf.javabdd.BDD
import util.analyzers.{BDDBuild, Phi}

import scala.collection.mutable


trait Drawer

trait ArchitectureView extends Drawer {

  protected lazy val diagram: Diagram = GSFactory.createDiagram
  protected lazy val diagramView: DiagramView = GSFactory.createDiagramView

  protected val nodeFont = new Font("Helvetica", Font.CENTER_BASELINE, 3)

  protected val nodeBrush = GradientColor(Color.WHITE, Color.GRAY, 0)
  protected val okNodeBrush = GradientColor(Color.WHITE, Color.GREEN, 0)
  protected val unusedNodeBrush = GradientColor(Color.WHITE, new Color(200, 200, 200), 0)
  protected val failNodeBrush = GradientColor(Color.WHITE, Color.RED, 0)
  protected lazy val layeredLayout: LayerLayout = GSFactory.createLayerLayout()

  diagramView.setDiagram(diagram)

  lazy val architectureView = new JScrollPane(diagramView)
}

trait MDDDrawer extends Drawer {
  self: STMDDBuilder =>

  private val mddDiagram: Diagram = GSFactory.createDiagram
  private val mddDiagramView = GSFactory.createDiagramView
  mddDiagramView.setDiagram(mddDiagram)
  private val mddLayout = GSFactory.createLayerLayout(Vertical)
  private val MDDNodeMap = mutable.Map.empty[mddFactory.MDD, BasicNode]
  private val mddNodeBrush = PlainColor(Color.WHITE)
  private val mddNodeFont = new Font("Arial", Font.CENTER_BASELINE, 3)

  lazy val mddView: JScrollPane = {
    MDDNodeMap.clear()
    createMDDView(STMDDD)
    mddLayout.arrangeDiagram(mddDiagram)
    new JScrollPane(mddDiagramView)
  }

  private def createMDDView(node: mddFactory.MDD): BasicNode = {
    node match {
      case n: mddFactory.MDDNode =>
        val nShapeNode = MDDNodeMap.getOrElseUpdate(n, {
          val newNode = createNode(n)
          mddDiagram.addElement(newNode)
          newNode
        })
        val sonsShapeNodes = n.map(p => (p._1, MDDNodeMap.getOrElse(p._2, {
          createMDDView(p._2)
        })))
        sonsShapeNodes.foreach(p => {
          mddDiagram.addElement(createLink(nShapeNode, p._2, p._1.trace.toString))
        })
        nShapeNode

      case t: mddFactory.Terminal =>
        MDDNodeMap.getOrElseUpdate(t, {
          val newNode = createNode(t)
          mddDiagram.addElement(newNode)
          newNode
        })
    }
  }

  private def createNode(mdd: mddFactory.MDD): BasicNode = {
    val node = mdd match {
      case n: mddFactory.MDDNode => GSFactory.createBasicNode(n.param.value.toString.replace("|", ""))
      case _: mddFactory.Terminal => GSFactory.createBasicNode(" 1 ")
    }
    node.setFont(mddNodeFont)
    node.setBackground(mddNodeBrush)
    node.setShape(Rectangle)
    node.fitToContent()
    node
  }

  private def createLink(source: DiagramNode, dest: DiagramNode, name: String): DiagramLink = {
    val l = GSFactory.createLink(mddDiagram, source, dest)
    l.setLabel(name.replace("|", ""))
    l.setDestinationShape(Triangle)
    l.setDestinationShapeSize(mddNodeFont.getSize / 2)
    l.setDestinationShapeColor(PlainColor(Color.BLACK))
    l
  }
}

trait PhiDrawer extends Drawer {
  self: Phi with BDDBuild =>

  private val phiDrawerNodeMap = mutable.Map.empty[BDD, BasicNode]
  private val layout = GSFactory.createLayerLayout(Vertical)
  private val diagram = GSFactory.createDiagram
  private val diagramView = GSFactory.createDiagramView
  diagramView.setDiagram(diagram)

  private val nodeBrush = PlainColor(Color.WHITE)
  private val nodeFont = new Font("Arial", Font.CENTER_BASELINE, 3)

  lazy val phiView: JScrollPane = {
    createBDDView(phiBDD, varMap.mapValues(_.`var`()).map(_.swap))
    layout.arrangeDiagram(diagram)
    new JScrollPane(diagramView)
  }

  private def createBDDView(bdd: BDD, varMap: Map[Int, InstBoolIdent]): BasicNode = {
    if (bdd.isOne) {
      phiDrawerNodeMap.getOrElseUpdate(bdd, {
        val newNode = createNode(bdd, varMap)
        diagram.addElement(newNode)
        newNode
      })
    } else if (bdd.isZero) {
      phiDrawerNodeMap.getOrElseUpdate(bdd, {
        val newNode = createNode(bdd, varMap)
        diagram.addElement(newNode)
        newNode
      })
    } else {
      val root = phiDrawerNodeMap.getOrElseUpdate(bdd, {
        val newNode = createNode(bdd, varMap)
        diagram.addElement(newNode)
        newNode
      })
      val high = phiDrawerNodeMap.getOrElseUpdate(bdd.high(), {
        val newNode = createBDDView(bdd.high(), varMap)
        diagram.addElement(newNode)
        newNode
      })
      val low = phiDrawerNodeMap.getOrElseUpdate(bdd.low(), {
        val newNode = createBDDView(bdd.low(), varMap)
        diagram.addElement(newNode)
        newNode
      })
      diagram.addElement(createLink(root, high, "1"))
      diagram.addElement(createLink(root, low, "0"))
      root
    }
  }

  private def createNode(bdd: BDD, varMap: Map[Int, InstBoolIdent]): BasicNode = {
    val name = {
      if (bdd.isZero) " 0 "
      else if (bdd.isOne) " 1 "
      else varMap(bdd.`var`()).toString.replace("|", "")
    }
    val node = GSFactory.createBasicNode(name)
    node.setFont(nodeFont)
    node.setBackground(nodeBrush)
    node.fitToContent()
    node
  }

  private def createLink(source: DiagramNode, dest: DiagramNode, name: String): DiagramLink = {
    val l = GSFactory.createLink(diagram, source, dest)
    l.setLabel(name.replace("|", ""))
    l.setDestinationShape(Triangle)
    l.setDestinationShapeSize(nodeFont.getSize / 2)
    l.setDestinationShapeColor(PlainColor(Color.BLACK))
    l
  }
}

trait ArchitectureDrawer extends ArchitectureView {
  self: Parsing =>
  private val failureNode = createShapeNode("Failure")
  protected lazy val (compMap, flowMap) = initTables()

  def draw() : Unit = {
    //trigger archi draw
    compMap
  }

  protected case class InterfaceMapping(in: Map[BasicFlowIdent, (DiagramNode, BasicFlowIdent)], out: Map[BasicFlowIdent, (DiagramNode, BasicFlowIdent)], compIdent: Option[CompInstanceIdent] = None)

  private def getFlowIdentOf(fExpr: ast.dsl.Expr): Set[BasicFlowIdent] = {
    fExpr match {
      case x: BasicFlowIdent => Set(x)
      case x: InstFlowIdent => Set(BasicFlowIdent(x.baseId))
      case _: FlowConst => Set()
      case FlowUnion(kid1, kid2) => getFlowIdentOf(kid1) ++ getFlowIdentOf(kid2)
      case FlowInter(kid1, kid2) => getFlowIdentOf(kid1) ++ getFlowIdentOf(kid2)
      case FlowIte(kid1, kid2, kid3) => getFlowIdentOf(kid1) ++ getFlowIdentOf(kid2) ++ getFlowIdentOf(kid3)
      case _: BoolIdent => Set()
      case _: InstBoolIdent => Set()
      case FlowEq(kid1, kid2) => getFlowIdentOf(kid1) ++ getFlowIdentOf(kid2)
      case FlowNeq(kid1, kid2) => getFlowIdentOf(kid1) ++ getFlowIdentOf(kid2)
      case FlowContains(kid1, kid2) => getFlowIdentOf(kid1) ++ getFlowIdentOf(kid2)
      case Iff(kid1, kid2) => getFlowIdentOf(kid1) ++ getFlowIdentOf(kid2)
      case Implies(kid1, kid2) => getFlowIdentOf(kid1) ++ getFlowIdentOf(kid2)
      case And(kid1, kid2) => getFlowIdentOf(kid1) ++ getFlowIdentOf(kid2)
      case Or(kid1, kid2) => getFlowIdentOf(kid1) ++ getFlowIdentOf(kid2)
      case Not(kid1) => getFlowIdentOf(kid1)
      case PopCount(kids, _, _) => kids.toSet.flatMap(getFlowIdentOf)
    }
  }

  private def createShapeNode(name: String): BasicNode = {
    val node = GSFactory.createBasicNode(name.replace("|", ""))
    node.setShape(Rectangle)
    node.setFont(nodeFont)
    node.setBackground(nodeBrush)
    node.fitToContent()
    node
  }

  private def createLink(source: DiagramNode, dest: DiagramNode, name: String): DiagramLink = {
    val l = GSFactory.createLink(diagram, source, dest)
    l.setLabel(name.replace("|", ""))
    l.setDestinationShape(Triangle)
    l.setDestinationShapeSize(nodeFont.getSize / 2)
    l.setDestinationShapeColor(PlainColor(Color.BLACK))
    l
  }

  private def createContainerNode(name: String): ContainerNode = {
    val node = GSFactory.createContainerNode(name.replace("|", ""))
    node.setFont(nodeFont)
    node.fitToContent()
    node
  }


  protected def initTables(): (Map[CompInstanceIdent, DiagramNode], Map[(InstFlowIdent, InstFlowIdent), DiagramLink]) = {

    val nodeMap = mutable.Map.empty[InterfaceMapping, DiagramNode]
    val compMapBuilder = mutable.Map.empty[CompInstanceIdent, DiagramNode]
    val flowMapBuilder = mutable.Map.empty[(InstFlowIdent, InstFlowIdent), DiagramLink]

    locateCompDef(conf.rootCompId).body.definitions.foreach({
      case CompCallFlowDefinition(lhs, rhs) =>
        val container = createContainerNode(rhs.instId.name.replace("|", ""))
        val cdecl = locateCompDef(rhs.compId)
        val inputs = cdecl.inputs.map(_.ident)
        val outputs = cdecl.outputs.map(_.ident)
        // attention correct seulement si list de flow ident dans inputs
        assert(rhs.args.forall(_.isInstanceOf[FlowIdent]))
        val inputMap = inputs.zip(rhs.args.flatMap(getFlowIdentOf)).toMap.map(kv => kv._1 -> (createShapeNode(kv._1.baseId.name), kv._2))
        inputMap.foreach(p => {
          diagram.addElement(p._2._1)
          container.addChild(p._2._1)
        })
        val outMap = outputs.zip(lhs).toMap.map(kv => kv._1 -> (createShapeNode(kv._1.baseId.name), kv._2))
        outMap.foreach(p => {
          diagram.addElement(p._2._1)
          container.addChild(p._2._1)
        })
        val mapping = InterfaceMapping(inputMap, outMap, Some(CompInstanceIdent(rhs.instId :: Nil)))
        nodeMap(mapping) = container
        compMapBuilder(CompInstanceIdent(rhs.instId :: Nil)) = container
      case expr@SingleFlowDefinition(lhs, rhs) =>
        val logicNode = createShapeNode(expr.toString)
        val mapping = InterfaceMapping(getFlowIdentOf(rhs).map(f => f -> (logicNode, f)).toMap, Map(lhs -> (logicNode, lhs)))
        nodeMap(mapping) = logicNode
    })

    nodeMap(InterfaceMapping(getFlowIdentOf(conf.failureCondition).map(f => f -> (failureNode, f)).toMap, Map())) = failureNode

    val links = nodeMap.flatMap(pair => {
      val mapping = pair._1
      nodeMap.collect({
        case (otherMapping, _) if mapping.out.values.map(x => x._2).toSet.intersect(otherMapping.in.values.map(x => x._2).toSet).nonEmpty =>
          mapping.out.collect({ case (out, f) if otherMapping.in.exists(pair => pair._2._2 == f._2) =>
            val otherIn = otherMapping.in.find(pair => pair._2._2 == f._2).get
            val link = createLink(f._1, otherIn._2._1, f._2.baseId.name)
            val in = otherMapping.in.find(pair => pair._2._2 == f._2).get
            val sourceComp = mapping.compIdent.getOrElse(CompInstanceIdent(Nil))
            val destComp = otherMapping.compIdent.getOrElse(CompInstanceIdent(Nil))
            flowMapBuilder((InstFlowIdent(out.baseId, sourceComp), InstFlowIdent(in._1.baseId, destComp))) = link
            link
          })
      }).flatten
    })
    nodeMap.values.foreach(diagram.addElement)
    links.foreach(diagram.addElement)
    layeredLayout.arrangeDiagram(diagram)
    (compMapBuilder.toMap, flowMapBuilder.toMap)
  }
}

trait TraceDrawer extends ArchitectureView {
  self: Parsing with STMDDBuilder with ArchitectureDrawer =>
  private lazy val initialLinkName = flowMap.values.map(link => link -> link.getLabel).toMap

  def next(): Unit = {
    reinitialize()
    currentPath = next(STMDDD, currentPath)
  }

  def reinitialize(): Unit = {
    initialLinkName.foreach(p => {
      p._1.setLabel(p._2)
      p._1.destination.setBackground(unusedNodeBrush)
      p._1.origin.setBackground(unusedNodeBrush)
    })
  }

  def restart(): Unit = {
    reinitialize()
    currentPath = Nil
  }

  def stop(): Unit = {
    initialLinkName.foreach(p => {
      p._1.setLabel(p._2)
      p._1.destination.setBackground(nodeBrush)
      p._1.origin.setBackground(nodeBrush)
    })
    currentPath = Nil
  }

  private var currentPath: List[(STMDDEdgeLabel, mddFactory.MDD)] = Nil

  private def next(node: mddFactory.MDD, currentPath: List[(STMDDEdgeLabel, mddFactory.MDD)]): List[(STMDDEdgeLabel, mddFactory.MDD)] = {
    node match {
      case n: mddFactory.MDDNode =>
        currentPath match {
          case Nil =>
            val (trace, kid) = n.toList.head
            flowMap.collect({ case ((source, _), link) if source.instId == n.param.value && trace.trace.value.contains(BasicFlowIdent(source.baseId)) =>
              val md = trace.trace.value(BasicFlowIdent(source.baseId))
              if (md.nonEmpty) {
                val fm = trace.trace.value(BasicFlowIdent(source.baseId)).collect({ case TypeValueFlowIdent(_, const) => const })
                link.setLabel(s"${initialLinkName(link)}=${fm.mkString(",")}")
                link.destination.setBackground(failNodeBrush)
                link.origin.setBackground(failNodeBrush)
              } else {
                link.setLabel(s"${initialLinkName(link)}")
                link.destination.setBackground(okNodeBrush)
                link.origin.setBackground(okNodeBrush)
              }
            })
            next(kid, Nil) match {
              case Nil =>
                if (n.tail.nonEmpty) {
                  n.tail.head :: Nil
                } else {
                  Nil
                }
              case son => (trace, kid) :: son
            }
          case h :: t =>
            val (trace, kid) = h
            flowMap.collect({ case ((source, _), link) if source.instId == n.param.value && trace.trace.value.contains(BasicFlowIdent(source.baseId)) =>
              val md = trace.trace.value(BasicFlowIdent(source.baseId))
              if (md.nonEmpty) {
                val fm = trace.trace.value(BasicFlowIdent(source.baseId)).collect({ case TypeValueFlowIdent(_, const) => const })
                link.setLabel(s"${initialLinkName(link)}=${fm.mkString(",")}")
                link.destination.setBackground(failNodeBrush)
                link.origin.setBackground(failNodeBrush)
              } else {
                link.setLabel(s"${initialLinkName(link)}")
                link.destination.setBackground(okNodeBrush)
                link.origin.setBackground(okNodeBrush)
              }
            })
            next(kid, t) match {
              case Nil =>
                val currentIndex = n.toSeq.indexOf(h)
                if (currentIndex != n.size - 1) {
                  n.toSeq(currentIndex + 1) :: Nil
                } else {
                  Nil
                }
              case son => h :: son
            }
        }
      case _: mddFactory.Terminal =>
        Nil
    }
  }
}
