package ui

import ast.dsl.{And, BasicFlowIdent, BoolExpr, BoolIdent, CompCall, CompCallFlowDefinition, CompIdent, CompInstanceIdent, CompSubstitution, ComponentBody, ComponentDeclaration, ConfigurationDeclaration, FlowConst, FlowContains, FlowDeclaration, FlowDefinition, FlowEq, FlowExpr, FlowIdent, FlowInter, FlowIte, FlowNeq, FlowUnion, Iff, Implies, InFile, IncludeDeclaration, IntExpr, Not, OpenDeclaration, Or, PackageDeclaration, PopCount, ProgramDeclaration, RealExpr, SingleFlowDefinition, TypeDeclaration}

/**
  * Created by kdelmas on 29/09/17.
  */
class Highlighter(val kCRTextPane: KCRTextPane) {

  def highlightComment(): Unit = {
    sealed trait State
    case class LineComment(start: Int) extends State
    case class BlockComment(start: Int, last: Char = ' ') extends State
    case class Code(last: Char = ' ') extends State
    val ini: State = Code()
    kCRTextPane.doc.getText(0, kCRTextPane.doc.getLength).zipWithIndex.foldLeft(ini)((state, elem) => {
      val (char, index) = elem
      state match {
        case LineComment(start) if char == '\n' =>
          kCRTextPane.CommentToken(start, index - start).applyStyle()
          Code()
        case l: LineComment => l
        case BlockComment(start, last) if last == '*' && char == '/' =>
          kCRTextPane.CommentToken(start, index - start + 1).applyStyle()
          Code()
        case BlockComment(start, _) => BlockComment(start, char)
        case Code(last) if last == '/' && char == '/' => LineComment(index - 1)
        case Code(last) if last == '/' && char == '*' => BlockComment(index - 1)
        case Code(_) => Code(char)
      }
    })
  }

  def apply(node: CompInstanceIdent)(implicit pkg: PackageDeclaration): Unit = node.location match {
    case InFile(_, _, _, pos, size) => kCRTextPane.CompInstToken(pos, size).applyStyle()
    case _ =>
  }

  def apply(node: CompIdent)(implicit pkg: PackageDeclaration): Unit = node.location match {
    case InFile(_, _, _, pos, size) => kCRTextPane.CompInstToken(pos, size).applyStyle()
    case _ =>
  }


  def apply(node: FlowIdent)(implicit pkg: PackageDeclaration): Unit = {
    node match {
      case cst: FlowConst => cst.location match {
        case InFile(_, _, _, pos, size) => kCRTextPane.ConstantToken(pos, size).applyStyle()
        case _ =>
      }
        cst.otherOpLocation.foreach(loc => kCRTextPane.ConstantToken(loc.posInFile, loc.size).applyStyle())
      case f: BasicFlowIdent =>
        if (pkg.declarations.collect({ case TypeDeclaration(_, v) => v }).flatten.contains(f.baseId))
          f.location match {
            case InFile(_, _, _, pos, size) => kCRTextPane.ConstantToken(pos, size).applyStyle()
            case _ =>
          }
      case _ =>
    }
  }


  /** Writes a program declaration to out. */
  def apply(node: ProgramDeclaration)(implicit pkg: PackageDeclaration): Unit = node match {
    case n: IncludeDeclaration => apply(n)
    case n: OpenDeclaration => apply(n)
    case n: PackageDeclaration => apply(n)
    case n: TypeDeclaration => apply(n)
    case n: ComponentDeclaration => apply(n)
    case n: ConfigurationDeclaration => apply(n)
  }

  def apply(node: OpenDeclaration): Unit = node.location match {
    case InFile(_, _, _, pos, size) => kCRTextPane.KeywordToken(pos, size).applyStyle()
    case _ =>
  }

  def apply(node: IncludeDeclaration): Unit = node.location match {
    case InFile(_, _, _, pos, size) => kCRTextPane.KeywordToken(pos, size).applyStyle()
    case _ =>
  }


  /** Writes a package declaration as the root package (omits the package name). */
  def asRoot(node: PackageDeclaration)(implicit pkg: PackageDeclaration): Unit = {
    node.declarations foreach {
      decl => {
        apply(decl)
      }
    }
  }

  def highlight(node: PackageDeclaration): Unit = {
    kCRTextPane.BasicToken(0, kCRTextPane.doc.getLength).applyStyle()
    node.declarations foreach {
      decl => {
        apply(decl)(node)
      }
    }
    highlightComment()
  }

  def apply(node: TypeDeclaration)(implicit pkg: PackageDeclaration): Unit = {
    node.location match {
      case InFile(_, _, _, pos, size) => kCRTextPane.KeywordToken(pos, size).applyStyle()
      case _ =>
    }
    node.values.foreach(_.location match {
      case InFile(_, _, _, pos, size) => kCRTextPane.ConstantToken(pos, size).applyStyle()
      case _ =>
    })
  }

  def apply(node: ComponentDeclaration)(implicit pkg: PackageDeclaration): Unit = {
    node.location match {
      case InFile(_, _, _, pos, size) => kCRTextPane.KeywordToken(pos, size).applyStyle()
      case _ =>
    }
    node.otherOpLocation.foreach(loc => kCRTextPane.KeywordToken(loc.posInFile, loc.size).applyStyle())
    node.inputs.foreach(apply)
    node.outputs.foreach(apply)
    apply(node.body)
  }


  def apply(node: ComponentBody)(implicit pkg: PackageDeclaration): Unit = {
    node.otherOpLocation.foreach(loc => kCRTextPane.KeywordToken(loc.posInFile, loc.size).applyStyle())
    node.events.foreach(apply)
    node.locals.foreach(apply)
    node.definitions.foreach(apply)

  }


  def apply(node: FlowDeclaration)(implicit pkg: PackageDeclaration): Unit = {
    apply(node.ident)
  }

  def apply(node: FlowDefinition)(implicit pkg: PackageDeclaration): Unit = node match {
    case SingleFlowDefinition(lhs, rhs) =>
      apply(lhs)
      apply(rhs)

    case CompCallFlowDefinition(lhs, rhs) =>
      lhs.foreach(apply)
      apply(rhs)

  }

  def apply(node: CompCall)(implicit pkg: PackageDeclaration): Unit = {
    node.location match {
      case InFile(_, _, _, pos, size) => kCRTextPane.CompInstToken(pos, size).applyStyle()
      case _ =>
    }
    val instId = CompInstanceIdent(node.instId :: Nil)
    instId.location_(node.instId.location)
    apply(instId)
    apply(node.compId)
    node.args.foreach(apply)
  }

  def apply(node: FlowExpr)(implicit pkg: PackageDeclaration): Unit = node match {
    case n: FlowIdent =>
      apply(n)

    case FlowUnion(kid1, kid2) =>
      node.location match {
        case InFile(_, _, _, pos, size) =>
          kCRTextPane.KeywordToken(pos, size).applyStyle()
        case _ =>
      }
      apply(kid1)
      apply(kid2)

    case FlowInter(kid1, kid2) =>
      node.location match {
        case InFile(_, _, _, pos, size) =>
          kCRTextPane.KeywordToken(pos, size).applyStyle()
        case _ =>
      }
      apply(kid1)
      apply(kid2)

    case FlowIte(kid1, kid2, kid3) =>
      node.location match {
        case InFile(_, _, _, pos, size) =>
          kCRTextPane.KeywordToken(pos, size).applyStyle()
        case _ =>
      }
      node.otherOpLocation.foreach({
        case InFile(_, _, _, pos, size) => kCRTextPane.KeywordToken(pos, size).applyStyle()
      })

      apply(kid1)
      apply(kid2)
      apply(kid3)

  }


  def apply(node: BoolExpr)(implicit pkg: PackageDeclaration): Unit = node match {
    case BoolIdent(_) =>
    case FlowEq(kid1, kid2) =>
      node.location match {
        case InFile(_, _, _, pos, size) =>
          kCRTextPane.KeywordToken(pos, size).applyStyle()
        case _ =>
      }
      apply(kid1)
      apply(kid2)

    case FlowNeq(kid1, kid2) =>
      node.location match {
        case InFile(_, _, _, pos, size) =>
          kCRTextPane.KeywordToken(pos, size).applyStyle()
        case _ =>
      }
      apply(kid1)
      apply(kid2)

    case FlowContains(kid1, kid2) =>
      node.location match {
        case InFile(_, _, _, pos, size) =>
          kCRTextPane.KeywordToken(pos, size).applyStyle()
        case _ =>
      }
      apply(kid1)
      apply(kid2)

    case Iff(kid1, kid2) =>
      node.location match {
        case InFile(_, _, _, pos, size) =>
          kCRTextPane.KeywordToken(pos, size).applyStyle()
        case _ =>
      }
      apply(kid1)
      apply(kid2)

    case Implies(kid1, kid2) =>
      node.location match {
        case InFile(_, _, _, pos, size) =>
          kCRTextPane.KeywordToken(pos, size).applyStyle()
        case _ =>
      }
      apply(kid1)
      apply(kid2)

    case Or(kid1, kid2) =>
      node.location match {
        case InFile(_, _, _, pos, size) =>
          kCRTextPane.KeywordToken(pos, size).applyStyle()
        case _ =>
      }
      apply(kid1)
      apply(kid2)

    case And(kid1, kid2) =>
      node.location match {
        case InFile(_, _, _, pos, size) =>
          kCRTextPane.KeywordToken(pos, size).applyStyle()
        case _ =>
      }
      apply(kid1)
      apply(kid2)

    case Not(kid1) => apply(kid1)

    case PopCount(kids, _, cst) =>
      node.location match {
        case InFile(_, _, _, pos, size) =>
          kCRTextPane.KeywordToken(pos, size).applyStyle()
        case _ =>
      }
      kids.foreach(apply)
      apply(cst)
      node.otherOpLocation.foreach({
        case InFile(_, _, _, pos, size) => kCRTextPane.KeywordToken(pos, size).applyStyle()
      })

  }


  def apply(node: IntExpr)(implicit pkg: PackageDeclaration): Unit = node.location match {
    case InFile(_, _, _, pos, size) => kCRTextPane.NumericToken(pos, size).applyStyle()
    case _ =>
  }

  def apply(node: RealExpr)(implicit pkg: PackageDeclaration): Unit = node.location match {
    case InFile(_, _, _, pos, size) => kCRTextPane.NumericToken(pos, size).applyStyle()
    case _ =>
  }


  def apply(node: ConfigurationDeclaration)(implicit pkg: PackageDeclaration): Unit = {

    node.location match {
      case InFile(_, _, _, pos, size) => kCRTextPane.KeywordToken(pos, size).applyStyle()
      case _ =>
    }

    node.otherOpLocation.foreach(loc => kCRTextPane.KeywordToken(loc.posInFile, loc.size).applyStyle())

    if (node.minCardValue.isDefined) {
      apply(node.minCardValue.get)
    }

    apply(node.failureCondition)

    if (node.duration.isDefined) {
      apply(node.duration.get)
    }

    if (node.reliability.isDefined) {
      apply(node.reliability.get)
    }

    node.designSpace match {
      case Nil => /* intentionally empty */

      case l => l.foreach(apply)
    }
  }

  def apply(node: CompSubstitution)(implicit pkg: PackageDeclaration): Unit = {
    node.location match {
      case InFile(_, _, _, pos, size) => kCRTextPane.KeywordToken(pos, size).applyStyle()
      case _ =>
    }
    node.otherOpLocation.foreach(loc => kCRTextPane.KeywordToken(loc.posInFile, loc.size).applyStyle())
  }
}
