package ui

import java.awt.Color
import java.awt.event.{ActionEvent, InputEvent, KeyEvent}
import java.io.{File, PrintWriter}
import javax.swing.event.{UndoableEditEvent, UndoableEditListener}
import javax.swing.text.{DefaultStyledDocument, Style, StyleConstants, StyledDocument}
import javax.swing.undo.{CannotRedoException, CannotUndoException, UndoManager}
import javax.swing.{AbstractAction, JTextPane, KeyStroke}

import ast.dsl.{ConfigIdent, ConfigurationDeclaration, RootPackage}
import parsers.dsl.{ParseFiles, ParserError}

import scala.io.Source
import scalaz.{-\/, \/-}

/**
  * Created by kdelmas on 16/01/17.
  */
class KCRTextPane(var file: File, val tabComponent: ButtonTabComponent, val doc: StyledDocument = new DefaultStyledDocument(), val console: ConsoleTextPane) extends JTextPane(doc) {
  val UNDO_ACTION = "undo"
  val REDO_ACTION = "redo"
  val SAVE_ACTION = "save"
  val undoMgr = new UndoManager()

  private var nonUserUpdate: Boolean = false

  private var needUpdate = false

  private var hasBeenModified = false

  def modificationProcessed(): Unit = {
    hasBeenModified = false
  }


  def modificationPending = hasBeenModified

  private def execUpdate(f: () => Unit): Unit = {
    nonUserUpdate = true
    getDocument.removeUndoableEditListener(undoableEditListener)
    f()
    nonUserUpdate = false
    getDocument.addUndoableEditListener(undoableEditListener)
  }

  def save() = {
    val pw = new PrintWriter(file)
    write(pw)
    tabComponent.setText(file.getName)
  }


  private def findAvailableConfig: List[ConfigIdent] = {
    val res = for {
      declMap <- ParseFiles(file.getAbsolutePath :: Nil)
    } yield {
      RootPackage(declMap.values.toList.flatten)
    }

    res match {
      case \/-(prog) =>
        prog.declarations.collect({ case conf: ConfigurationDeclaration => conf.ident })
      case _ => Nil
    }
  }

  private val undoableEditListener = new UndoableEditListener() {
    def undoableEditHappened(pEvt: UndoableEditEvent) {
      if (!nonUserUpdate) {
        needUpdate = true
        undoMgr.addEdit(pEvt.getEdit)
        tabComponent.setText(file.getName + "*")
      }
      hasBeenModified = true
    }
  }

  // Add undo/redo actions
  getActionMap.put(UNDO_ACTION, new AbstractAction(UNDO_ACTION) {
    def actionPerformed(e: ActionEvent) = {
      try {
        if (undoMgr.canUndo()) {
          undoMgr.undo()
        }
      } catch {
        case e: CannotUndoException => e.printStackTrace();
      }
    }
  })
  getActionMap.put(REDO_ACTION, new AbstractAction(REDO_ACTION) {
    def actionPerformed(pEvt: ActionEvent) = {
      try {
        if (undoMgr.canRedo()) {
          undoMgr.redo()
        }
      } catch {
        case e: CannotRedoException => e.printStackTrace();
      }
    }
  })
  getActionMap.put(SAVE_ACTION, new AbstractAction(SAVE_ACTION) {
    def actionPerformed(e: ActionEvent) = save()
  })

  // Create keyboard accelerators for undo/redo actions (Ctrl+Z/Ctrl+Y)
  getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_Z, InputEvent.CTRL_DOWN_MASK),
    UNDO_ACTION)
  getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_Y, InputEvent.CTRL_DOWN_MASK),
    REDO_ACTION)
  getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_S, InputEvent.CTRL_DOWN_MASK),
    SAVE_ACTION)

  val basicStyle = addStyle("IdentStyle", getLogicalStyle)
  StyleConstants.setForeground(basicStyle, Color.black)

  val opStyle = addStyle("OpStyle", getLogicalStyle)
  StyleConstants.setForeground(opStyle, new Color(0, 0, 153))
  StyleConstants.setBold(opStyle, true)

  val numStyle = addStyle("numStyle", getLogicalStyle)
  StyleConstants.setForeground(numStyle, new Color(0, 0, 200))

  val kwStyle = addStyle("kwStyle", getLogicalStyle)
  StyleConstants.setForeground(kwStyle, new Color(0, 0, 153))
  StyleConstants.setBold(kwStyle, true)

  val cmtStyle = addStyle("cmtStyle", getLogicalStyle)
  StyleConstants.setForeground(cmtStyle, Color.red)
  StyleConstants.setItalic(cmtStyle, true)

  val cstStyle = addStyle("cstStyle", getLogicalStyle)
  StyleConstants.setForeground(cstStyle, new Color(0, 102, 0))
  StyleConstants.setBold(cstStyle, true)

  val cmpStyle = addStyle("cmpStyle", getLogicalStyle)
  StyleConstants.setForeground(cmpStyle, new Color(153, 0, 153))
  StyleConstants.setItalic(cmpStyle, true)

  load()

  var lastConf: List[ConfigIdent] = findAvailableConfig

  def getAndUpdateAvailableConfig: List[ConfigIdent] = {
    if (modificationPending) {
      lastConf = findAvailableConfig
    }
    lastConf
  }


  trait Token {
    val sourcePos: Int
    val length: Int
    val style: Style

    def applyStyle(): Unit = {
      doc.setCharacterAttributes(sourcePos, length, style, true)
    }
  }

  case class BasicToken(sourcePos: Int, length: Int) extends Token {
    val style = basicStyle
  }

  case class NumericToken(sourcePos: Int, length: Int) extends Token {
    val style = numStyle
  }

  case class OperatorToken(sourcePos: Int, length: Int) extends Token {
    val style = opStyle
  }

  case class KeywordToken(sourcePos: Int, length: Int) extends Token {
    val style = kwStyle
  }

  case class CommentToken(sourcePos: Int, length: Int) extends Token {
    val style = cmpStyle
  }

  case class ConstantToken(sourcePos: Int, length: Int) extends Token {
    val style = cstStyle
  }

  case class CompInstToken(sourcePos: Int, length: Int) extends Token {
    val style = cmpStyle
  }

  def ifNecessaryUpdate(): Unit = {
    if (needUpdate) {
      execUpdate(() => {
        highlight()
      })
      needUpdate = false
    }
  }

  def load(): Unit = {
    execUpdate(() => {
      for (line <- Source.fromFile(file).getLines()) {
        doc.insertString(doc.getLength, line + "\n", null)
      }
    })
  }

  def reload(): Unit = {
    execUpdate(() => {
      doc.remove(0, doc.getLength)
      for (line <- Source.fromFile(file).getLines()) {
        doc.insertString(doc.getLength, line + "\n", null)
      }
      highlight()
    })
  }

  highlight()

  private def highlight(): Unit = {
    execUpdate(() => {
      console.clear()
      save()
      val res = for {
        declMap <- ParseFiles(file.getAbsolutePath :: Nil)
      } yield {
        RootPackage(declMap.values.toList.flatten)
      }
      res match {
        case \/-(prog) =>
          val highlighter = new Highlighter(this)
          highlighter.highlight(prog)
        case -\/(errors) =>
          errors.foreach({ case ParserError(e) => console.printError(e.getMessage); case _ => })
      }
    })
  }

  // Add listener for undoable events
  getDocument.addUndoableEditListener(undoableEditListener)
}
