package ui

import java.awt._
import java.awt.event._
import java.io.{File, FileWriter, PrintWriter}
import java.nio.file.Files
import javax.swing._
import javax.swing.event._
import javax.swing.filechooser.FileFilter

import analyzers._
import util.analyzers.analyzerTypes._
import ast.dsl.{ConfigIdent, RootPackage, Transformer}
import parsers.dsl.{ParseFiles, ParserError}
import util.analyzers._
import util.io.{FileManager, SystemInfo}

import scala.concurrent.CancellationException
import scala.concurrent.ExecutionContext.Implicits.global
import scala.reflect.io.Path
import scala.util.{Failure, Success}
import scalaz.{-\/, \/-}

/**
  *
  * Created by kdelmas on 13/01/17.
  */


case class DiagramContainer(drawer: Drawer, mlayout: LayoutManager = new BorderLayout()) extends JPanel(mlayout)


class UI() {
  val jFrame = new JFrame()
  jFrame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE)
  jFrame.addWindowListener(new WindowListener {
    def windowDeiconified(e: WindowEvent): Unit = {}

    def windowClosing(e: WindowEvent): Unit = {}

    def windowClosed(e: WindowEvent): Unit = {
      val current = new File(".")
      current.listFiles().filter(f => f.getName.contains("inlined.kcr")).foreach(_.delete())
      jFrame.dispose()
      System.exit(0)
    }

    def windowActivated(e: WindowEvent): Unit = {}

    def windowOpened(e: WindowEvent): Unit = {}

    def windowDeactivated(e: WindowEvent): Unit = {}

    def windowIconified(e: WindowEvent): Unit = {}
  })

  //cannot do anything if the java binding is not there
  if (getToolPathOf(Z3JavaSO).isEmpty) {
    JOptionPane.showMessageDialog(jFrame, s"cannot find ${Z3JavaSO.exeName} in ${Z3JavaSO.LIBRARY}", "Error", JOptionPane.ERROR_MESSAGE)
    jFrame.dispatchEvent(new WindowEvent(jFrame, WindowEvent.WINDOW_CLOSED))
  }

  val uiDim = getScreenSize

  val menuBar = new JMenuBar()

  val fileMenu = new JMenu("File")
  fileMenu.setVisible(true)
  menuBar.add(fileMenu)

  val fc = new JFileChooser()
  fc.setCurrentDirectory(new File("."))
  val openFile = new JMenuItem("open file")
  val save = new JMenuItem("save file")
  val saveAs = new JMenuItem("save file as")
  val newFile = new JMenuItem("new file")
  val consoleResult = new ConsoleTextPane()
  val resultScroll = new JScrollPane(consoleResult,
    ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,
    ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER
  )
  val consoleScrollPane = new JScrollPane(consoleResult)
  val config = new Config()
  openFile.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      fc.setFileFilter(KCRFileFilter)
      val returnVal = fc.showOpenDialog(jFrame)
      if (returnVal == JFileChooser.APPROVE_OPTION) {
        openTabFromFile(fc.getSelectedFile)
      }
    }
  })
  val analysisMenu = new JMenu("Analysis")
  save.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      for (pane <- getSelectedKCRPane) yield {
        pane.code.save()
      }
    }
  })
  val computeMCS = new JMenuItem("compute cutsets")
  saveAs.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      fc.setFileFilter(KCRFileFilter)
      val returnVal = fc.showOpenDialog(jFrame)
      if (returnVal == JFileChooser.APPROVE_OPTION) {
        val file = fc.getSelectedFile
        for (code <- getSelectedKCRPane) yield {
          val pw = new PrintWriter(file)
          code.code.write(pw)
          code.code.file = file
          code.code.tabComponent.setText(code.code.file.getName)
        }
      }
    }
  })
  val checkMinCard = new JMenuItem("check cutsets cardinality")
  newFile.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      val file = FileManager.addFileToTemp("untitled.kcr")
      Files.deleteIfExists(file.toPath)
      file.createNewFile()
      val tabComponent = new ButtonTabComponent(file.getName, tabbedPane)
      val codeEditor = new KCRTextPane(file, tabComponent, console = consoleResult)
      val lineNumber = new TextLineNumber(codeEditor, 2)
      lineNumber.setPreferredSize(new Dimension(lineNumber.getPreferredSize.width, codeEditor.getPreferredSize.height))
      val panel = new JPanel(new BorderLayout())
      panel.add(codeEditor, BorderLayout.CENTER)
      panel.add(lineNumber, BorderLayout.LINE_START)
      val codeScrollPane = new JScrollPane(panel)
      codeScrollPane.setMinimumSize(new Dimension(uiDim.width, uiDim.height / 2))
      val tabeIndex = tabbedPane.getTabCount
      val topContainer = CodeContainer(codeEditor, config, uiConsole = consoleResult)
      topContainer.add(codeScrollPane)
      tabbedPane.addTab(file.getName, topContainer)
      tabbedPane.setTabComponentAt(tabeIndex, tabComponent)
      tabbedPane.setSelectedIndex(tabeIndex)
    }
  })
  fileMenu.add(openFile)
  fileMenu.add(newFile)
  fileMenu.add(save)
  fileMenu.add(saveAs)

  val editMenu = new JMenu("Edit")
  editMenu.setVisible(true)
  menuBar.add(editMenu)

  val reformat = new JMenuItem("reformat")
  reformat.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      consoleResult.clear()
      for (kcrPane <- getSelectedKCRPane) yield {
        val res = for {
          declMap <- ParseFiles(kcrPane.code.file.getAbsolutePath :: Nil)
        } yield {
          RootPackage(declMap.values.toList.flatten)
        }
        res match {
          case \/-(prog) =>
            val writer = new FileWriter(kcrPane.code.file, false)
            val p = new printing.dsl.Printer(writer)
            p.asRoot(prog)
            writer.flush()
            writer.close()
            val lastPos = kcrPane.code.getCaretPosition
            kcrPane.code.reload()
            if (lastPos < kcrPane.code.doc.getLength)
              kcrPane.code.setCaretPosition(lastPos)
          case -\/(error) =>
            error match {
              case l: Traversable[_] =>
                l.foreach({
                  case e: ParserError =>
                    consoleResult.printError(e.e.getMessage)
                  case e => consoleResult.printError(e.toString)
                })
              case e => consoleResult.printError(e.toString)
            }
        }


      }
    }
  })
  editMenu.add(reformat)


  val computeMinCard = new JMenuItem("compute minimum cardinality")
  consoleResult.setEditable(false)
  consoleResult.setFocusable(false)
  consoleResult.setOpaque(false)
  val computeR = new JMenuItem("compute reliability")

  resultScroll.setPreferredSize(new Dimension(uiDim.width, uiDim.height / 5))
  val checkR = new JMenuItem("check reliability")
  val checkCoherence = new JMenuItem("check coherence")
  val runExploration = new JMenuItem("run exploration")
  val viewMenu = new JMenu("View")
  analysisMenu.setVisible(false)
  fileMenu.getAccessibleContext.setAccessibleDescription(
    "Configure analysis")
  val displayPhiBDD = new JMenuItem("display phi BDD")
  computeMCS.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      for (kcrPane <- getSelectedKCRPane;
           c <- getCurrentConf;
           a <- kcrPane.getAnalyser(c)
      ) yield {
        abortableCode {
          a.defaultComputationMethod = config.cutComputationMode
          val (cuts, t) = a.profilerCutEnumeration.runWithProfile(config.mcsMaxSize)()
          val sizes = cuts.toList.map(s => s.size).groupBy(s => s).mapValues(_.size).toList.sortBy(_._1)
          config.mcsMaxSize match {
            case Some(k) =>
              consoleResult.printResult(s"system cutsets of size lesser or equal than $k ${if (config.verboseMode) s"(computed with ${config.cutComputationMode} method in $t s)" else ""}:")
            case None =>
              consoleResult.printResult(s"system cutsets ${if (config.verboseMode) s"(computed with ${config.cutComputationMode} method in $t )" else ""}:")
          }
          if (cuts.isEmpty) {
            consoleResult.printResult(s"No cutsets")
          } else {
            consoleResult.printResult(s"population by size:")
            sizes.foreach(p => consoleResult.printResult(s"${p._1} -> ${p._2}"))
            consoleResult.printResult(s"total: ${cuts.size}")
            if (config.verboseMode) {
              cuts.toList.sortBy(cut => cut.size).foreach(cut => consoleResult.printResult(cut.map(evt => evt.toString()).mkString(",")))
            }
          }
          FileManager.cleanTemp()
        }
      }
    }
  })
  analysisMenu.add(computeMCS)
  val displaySTT = new JMenuItem("display trace tree")
  checkMinCard.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      for (kcrPane <- getSelectedKCRPane;
           c <- getCurrentConf;
           a <- kcrPane.getAnalyser(c)) yield {
        a.conf.minCardValue match {
          case Some(obj) =>
            abortableCode {
              val (r, t) = time(() => a.cardOK)
              for (c <- r) yield consoleResult.printResult(s"mincard req satisfied ${if (config.verboseMode) s"($t s)" else ""}: $c")
            }
          case None =>
            consoleResult.printError(s"[Error] no minimal cardinality requirement defined in configuration ${configField.getSelectedItem}")
        }
      }
    }
  })
  analysisMenu.add(checkMinCard)
  val displayArchitecture = new JMenuItem("display architecture")
  computeMinCard.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      for (kcrPane <- getSelectedKCRPane;
           c <- getCurrentConf;
           a <- kcrPane.getAnalyser(c)) yield {
        val (r, t) = time(() => a.cardinality)
        r match {
          case InfNumeric(Left(card)) =>
            abortableCode {
              consoleResult.printResult(s"minimal cardinality of cutsets is $card ${if (config.verboseMode) s"($t s)" else ""}")
            }
          case InfNumeric(Right(_)) => consoleResult.printResult(s"the failure condition is unreachable (no cutsets)")
        }
      }
    }
  })
  analysisMenu.add(computeMinCard)
  val configField = new JComboBox[ConfigIdent]()
  computeR.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      for (kcrPane <- getSelectedKCRPane;
           c <- getCurrentConf;
           a <- kcrPane.getAnalyser(c)) yield {
        abortableCode {
          val (r, t) = time(() => a.reliability)
          consoleResult.printResult(s"system reliability is ${if (config.verboseMode) s"($t s)" else ""}: ${r.toDouble}")
        }
      }
    }
  })
  analysisMenu.add(computeR)
  val simulationMenu = new JMenu("Simulation")
  checkR.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      for (kcrPane <- getSelectedKCRPane;
           c <- getCurrentConf;
           a <- kcrPane.getAnalyser(c)) yield {
        a.conf.reliability.map(s => BigDecimal(s.repr)) match {
          case Some(obj: BigDecimal) =>
            abortableCode {
              val (r, t) = time(() => a.reliability)
              consoleResult.printResult(s"reliability req satisfied ${if (config.verboseMode) s"($t s)"}: ${r >= obj}")
            }
          case None =>
            consoleResult.printError(s"[Error] no reliability requirement defined in configuration ${getCurrentConf}")
        }
      }
    }
  })
  analysisMenu.add(checkR)
  val traceTreeWalker = new JMenuItem("walk on trace tree")
  checkCoherence.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      for (kcrPane <- getSelectedKCRPane;
           c <- getCurrentConf;
           a <- kcrPane.getAnalyser(c)) yield {
        abortableCode {
          val (r, t) = time(() => a.isCoherent)
          if(r){
            consoleResult.printResult(s"the system is monotonic  ${if (config.verboseMode) s"($t s)" else ""}")
          }else{
            consoleResult.printResult(s"the system is not monotonic  ${if (config.verboseMode) s"($t s)" else ""}: \n ${a.counterExample.get}")
          }
          consoleResult.printResult(s"the system is ${if (!r) "not " else ""}monotonic ${if (config.verboseMode) s"($t s)" else ""}")
        }
      }
    }
  })
  analysisMenu.add(checkCoherence)
  val optionMenu = new JMenu("Option")
  runExploration.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      for (kcrPane <- getSelectedKCRPane;
           c <- getCurrentConf;
           a <- kcrPane.getAnalyser(c)) yield {
        abortableCode {
          consoleResult.printInfo("starting exploration phase")
          val (solution, t) = a.designSpaceExplorationProfiler.runWithProfile(config.explorationMethod)()
          solution match {
            case Some(slt) =>
              if (slt.isEmpty)
                consoleResult.printResult(s"All substitutions are solutions (trivial problem) ${if (config.verboseMode) s"(computed with ${config.explorationMethod} method in $t s)" else ""}")
              else {
                consoleResult.printResult(s"solution found ${if (config.verboseMode) s"(computed with ${config.explorationMethod} method in $t s)" else ""}:")
                slt.foreach(x => consoleResult.printResult(s"${x._1} -> ${x._2}"))
              }
              val sltR = Measure[Real](1) - a.computeUnrWithMemo(slt)
              val sltCard = a.computeCard(slt)
              val toPrintR = (for (e <- sltR.exact) yield s"solution reliability is ${e.toDouble}") getOrElse {
                s"solution reliability in [${sltR.low.toDouble}, ${sltR.up.toDouble}]"
              }
              val toPrintCard = (for (e <- sltCard.exact) yield s"solution cardinality is $e") getOrElse {
                s"solution reliability in [${sltCard.low}, ${sltCard.up}]"
              }
              consoleResult.printResult(toPrintR)
              consoleResult.printResult(toPrintCard)
              for (sltFile <- config.printSolutionIn) yield {
                Transformer.writeSolutionFile(sltFile, a.conf.parent, a.conf, slt)
                consoleResult.printResult(s"solution system exported in KCR file: ${sltFile.getName}")
              }
            case None =>
              consoleResult.printResult(s"no solution found ${if (config.verboseMode) s"(computed with ${config.explorationMethod} method in $t s)" else ""}")
          }
          consoleResult.printInfo("end of exploration phase")
          FileManager.cleanTemp()
        }
      }
    }
  })
  analysisMenu.add(runExploration)

  menuBar.add(analysisMenu)
  val verbose = new JCheckBoxMenuItem("verbose mode")
  viewMenu.setVisible(false)
  fileMenu.getAccessibleContext.setAccessibleDescription("Configure views")
  val cutSize = new JCheckBoxMenuItem("limit cutsets size")
  displayPhiBDD.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      for (kcrPane <- getSelectedKCRPane) yield {
        for (c <- getCurrentConf;
             drawer <- kcrPane.getPhiDrawer(consoleResult, c)) yield {
          val tabIndex = tabbedPane.getTabCount
          val title = s"view of ${kcrPane.code.file.getName.replace("*", "")} BDD"
          val diagramContainer = DiagramContainer(drawer)
          diagramContainer.add(drawer.phiView)
          tabbedPane.addTab(title, diagramContainer)
          tabbedPane.setTabComponentAt(tabIndex, new ButtonTabComponent(title, tabbedPane))
          tabbedPane.setSelectedIndex(tabIndex)
        }
      }
    }
  })
  viewMenu.add(displayPhiBDD)
  val cutTh = new JTextField()
  displaySTT.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      for (kcrPane <- getSelectedKCRPane) yield {
        for (c <- getCurrentConf;
             drawer <- kcrPane.getSystemDrawers(consoleResult, c)) yield {
          val tabIndex = tabbedPane.getTabCount
          val title = s"view of ${kcrPane.code.file.getName.replace("*", "")} STT"
          val drawerContainer = DiagramContainer(drawer)
          drawerContainer.add(drawer.mddView)
          tabbedPane.addTab(title, drawerContainer)
          tabbedPane.setTabComponentAt(tabIndex, new ButtonTabComponent(title, tabbedPane))
          tabbedPane.setSelectedIndex(tabIndex)
        }
      }
    }
  })
  viewMenu.add(displaySTT)
  val exportSolution = new JCheckBoxMenuItem("export KCR solution")
  displayArchitecture.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      for (kcrPane <- getSelectedKCRPane) yield {
        for (c <- getCurrentConf;
             drawer <- kcrPane.getSystemDrawers(consoleResult, c)) yield {
          val tabIndex = tabbedPane.getTabCount
          val title = s"view of ${kcrPane.code.file.getName.replace("*", "")}"
          val drawerContainer = DiagramContainer(drawer)
          drawer.draw()
          drawerContainer.add(drawer.architectureView)
          tabbedPane.addTab(title, drawerContainer)
          tabbedPane.setTabComponentAt(tabIndex, new ButtonTabComponent(title, tabbedPane))
          tabbedPane.setSelectedIndex(tabIndex)
        }
      }
    }
  })
  viewMenu.add(displayArchitecture)

  menuBar.add(viewMenu)
  val cutGenMethode = new JMenuItem("cut generation method")

  configField.addPopupMenuListener(new PopupMenuListener {

    def popupMenuWillBecomeVisible(e: PopupMenuEvent): Unit = refreshConfigList(configField)

    def popupMenuWillBecomeInvisible(e: PopupMenuEvent): Unit = {}

    def popupMenuCanceled(e: PopupMenuEvent): Unit = {}
  })
  val explorationMethode = new JMenuItem("exploration generation method")
  val exportSage = new JMenuItem("export reliability and failure rate to python")
  simulationMenu.setVisible(false)
  val exportSMV = new JMenuItem("export system to SMV")
  traceTreeWalker.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      simuAdds.setVisible(true)
      for (drawer <- getSelectedDrawer) {
        drawer match {
          case d: TraceDrawer => d.stop()
          case _ =>
        }
      }
    }
  })
  simulationMenu.add(traceTreeWalker)
  menuBar.add(simulationMenu)
  val exportCG = new JMenuItem("export phi to circuit graph")
  optionMenu.setVisible(false)
  val exportPhi = new JMenuItem("export phi BDD")
  verbose.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      config.verboseMode = verbose.getState
    }
  })
  optionMenu.add(verbose)
  val exportPhiPSA = new JMenuItem("export phi as OpenPSA tree")
  cutSize.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      if (cutSize.getState) {
        cutTh.setVisible(true)
      } else {
        cutTh.setVisible(false)
        config.mcsMaxSize = None
      }
    }
  })
  optionMenu.add(cutSize)
  val exportHipHOPS = new JMenuItem("export system as HipHOPS")
  cutTh.setVisible(false)
  cutTh.addFocusListener(new FocusListener {
    def focusGained(e: FocusEvent): Unit = {}

    def focusLost(e: FocusEvent): Unit = {
      try {
        config.mcsMaxSize = Some(cutTh.getText().toInt)
      } catch {
        case nfe: NumberFormatException => cutTh.setText("")
      }
    }
  })
  optionMenu.add(cutTh)
  val exportMDDIn = new JMenuItem("export trace tree")
  exportSolution.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      if (exportSolution.getState) {
        fc.setFileFilter(KCRFileFilter)
        val returnVal = fc.showOpenDialog(jFrame)
        if (returnVal == JFileChooser.APPROVE_OPTION) {
          config.printSolutionIn = Some(fc.getSelectedFile)
        } else {
          exportSolution.setState(false)
        }
      } else {
        config.printSolutionIn = None
      }
    }
  })
  optionMenu.add(exportSolution)
  val toolPathMenu = new JMenu("Path")
  cutGenMethode.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      JOptionPane.showInputDialog(jFrame,
        "choose cutsets generation method",
        "cutsets generation",
        JOptionPane.QUESTION_MESSAGE,
        null,
        SystemInfo.findAvailableCutComputationMethods().sortBy(m => m.toString).toArray,
        config.cutComputationMode) match {
        case a: CutComputationMode => config.cutComputationMode = a
        case _ =>
      }
    }
  })
  optionMenu.add(cutGenMethode)
  val addsPanel = new JPanel(new FlowLayout())
  explorationMethode.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      JOptionPane.showInputDialog(jFrame,
        "choose exploration method",
        "exploration methods",
        JOptionPane.QUESTION_MESSAGE,
        null,
        SystemInfo.findAvailableExpMethods().sortBy(m => m.toString).toArray,
        config.explorationMethod) match {
        case a: DesignSpaceExplorationMethod => config.explorationMethod = a
        case _ =>
      }
    }
  })
  optionMenu.add(explorationMethode)
  val buildAdds = new JPanel(new FlowLayout())
  exportSage.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      fc.resetChoosableFileFilters()
      fc.setFileFilter(PythonFilter)
      val returnVal = fc.showOpenDialog(jFrame)
      if (returnVal == JFileChooser.APPROVE_OPTION) {
        for (kcrPane <- getSelectedKCRPane;
             c <- getCurrentConf;
             a <- kcrPane.getAnalyser(c)) yield {
          val start = System.nanoTime()
          a.printSymbIniRBDD(fc.getSelectedFile.getAbsolutePath)
          val end = System.nanoTime()
          consoleResult.printInfo(s"reliability and failure rate formula exported in python script: ${fc.getSelectedFile.getName} ${if (verbose.getState) s"(in ${(end - start) * 1e-9}s)" else ""}")
        }
      }
    }
  })
  optionMenu.add(exportSage)
  val simuAdds = new JPanel(new FlowLayout())
  exportSMV.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      fc.resetChoosableFileFilters()
      fc.setFileFilter(SMVFileFilter)
      val returnVal = fc.showOpenDialog(jFrame)
      if (returnVal == JFileChooser.APPROVE_OPTION) {
        for (kcrPane <- getSelectedKCRPane;
             c <- getCurrentConf;
             a <- kcrPane.getAnalyser(c)) yield {
          a.exportToSMV(fc.getSelectedFile)
          consoleResult.printInfo(s"system exported to SMV in: ${fc.getSelectedFile.getName}")
        }
      }
    }
  })
  optionMenu.add(exportSMV)
  val configLabel = new JLabel("configuration:")
  exportCG.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      fc.resetChoosableFileFilters()
      fc.setFileFilter(CGFileFilter)
      val returnVal = fc.showOpenDialog(jFrame)
      if (returnVal == JFileChooser.APPROVE_OPTION) {
        for (kcrPane <- getSelectedKCRPane;
             c <- getCurrentConf;
             a <- kcrPane.getAnalyser(c)) yield {
          a.exportToCG(fc.getSelectedFile)
          consoleResult.printInfo(s"system exported to CG in: ${fc.getSelectedFile.getName}")
        }
      }
    }
  })
  optionMenu.add(exportCG)
  val buildButton = new JButton("build")
  exportPhi.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      fc.setFileFilter(DOTFileFilter)
      val returnVal = fc.showOpenDialog(jFrame)
      if (returnVal == JFileChooser.APPROVE_OPTION) {
        for (kcrPane <- getSelectedKCRPane;
             c <- getCurrentConf;
             a <- kcrPane.getAnalyser(c)) yield {
          a.phiCarac.printDot(fc.getSelectedFile.getAbsolutePath)
          consoleResult.printInfo(s"phi BDD exported in DOT file: ${fc.getSelectedFile.getName}")
        }
      }
    }
  })
  optionMenu.add(exportPhi)
  val clear = new JButton("clear")
  exportPhiPSA.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      fc.setFileFilter(XMLFileFilter)
      val returnVal = fc.showOpenDialog(jFrame)
      if (returnVal == JFileChooser.APPROVE_OPTION) {
        for (kcrPane <- getSelectedKCRPane;
             c <- getCurrentConf;
             a <- kcrPane.getAnalyser(c)) yield {
          a.exportToOpenPSA(fc.getSelectedFile)
          consoleResult.printInfo(s"phi Open PSA exported in XML file: ${fc.getSelectedFile.getName}")
        }
      }
    }
  })
  optionMenu.add(exportPhiPSA)
  val highlight = new JButton("highlight")
  exportHipHOPS.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      fc.setFileFilter(XMLFileFilter)
      val returnVal = fc.showOpenDialog(jFrame)
      if (returnVal == JFileChooser.APPROVE_OPTION) {
        for (kcrPane <- getSelectedKCRPane;
             c <- getCurrentConf;
             a <- kcrPane.getAnalyser(c)) yield {
          a.exportToHipHOPS(fc.getSelectedFile)
          consoleResult.printInfo(s"system exported as HipHOPS system in XML file: ${fc.getSelectedFile.getName}")
        }
      }
    }
  })
  optionMenu.add(exportHipHOPS)
  val nextButton = new JButton("next")
  exportMDDIn.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      fc.setFileFilter(DOTFileFilter)
      val returnVal = fc.showOpenDialog(jFrame)
      if (returnVal == JFileChooser.APPROVE_OPTION) {
        for (kcrPane <- getSelectedKCRPane;
             c <- getCurrentConf;
             a <- kcrPane.getAnalyser(c)) yield {
          a.mddFactory.printMDD(a.STMDDD, fc.getSelectedFile.getAbsolutePath)
          consoleResult.printInfo(s"Trace tree exported in DOT file: ${fc.getSelectedFile.getName}")
        }
      }
    }
  })
  optionMenu.add(exportMDDIn)

  menuBar.add(optionMenu)
  val reinit = new JButton("reinit")
  toolPathMenu.setVisible(true)

  addPathMenuItem(PythonInterface, toolPathMenu)
  addPathMenuItem(ClaspInterface, toolPathMenu)
  addPathMenuItem(SharpCDCLInterface, toolPathMenu)
  addPathMenuItem(XFTAInterface, toolPathMenu)
  addPathMenuItem(XSAPInterface, toolPathMenu)
  addPathMenuItem(HipHOPSInterface, toolPathMenu)
  addPathMenuItem(WineInterface, toolPathMenu)
  addPathMenuItem(PBEncoderInterface, toolPathMenu)
  addPathMenuItem(Z3JavaSO, toolPathMenu, true)

  menuBar.add(toolPathMenu)
  val stopSim = new JButton("stop")
  val tabbedPane = new JTabbedPane()
  buildAdds.setVisible(false)
  val container = jFrame.getContentPane
  simuAdds.setVisible(false)
  addsPanel.add(buildAdds)
  addsPanel.add(simuAdds)
  val headerContainer = new JPanel(new BorderLayout())
  buildAdds.add(configLabel)
  buildAdds.add(configField)

  def openTabFromFile(file: File): Unit = {
    val tabComponent = new ButtonTabComponent(file.getName, tabbedPane)
    val codeEditor = new KCRTextPane(file, tabComponent, console = consoleResult)
    val lineNumber = new TextLineNumber(codeEditor, 2)
    lineNumber.setPreferredSize(new Dimension(lineNumber.getPreferredSize.width, codeEditor.getPreferredSize.height))
    val panel = new JPanel(new BorderLayout())
    panel.add(codeEditor, BorderLayout.CENTER)
    panel.add(lineNumber, BorderLayout.LINE_START)
    val codeScrollPane = new JScrollPane(panel)
    codeScrollPane.setMinimumSize(new Dimension(uiDim.width, uiDim.height / 2))
    val tabeIndex = tabbedPane.getTabCount
    val topContainer = CodeContainer(codeEditor, config, uiConsole = consoleResult)
    topContainer.add(codeScrollPane)
    tabbedPane.addTab(file.getName, topContainer)
    tabbedPane.setTabComponentAt(tabeIndex, tabComponent)
    tabbedPane.setSelectedIndex(tabeIndex)
  }

  buildButton.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      for (kcrPane <- getSelectedKCRPane; c <- getCurrentConf) yield {
        consoleResult.clear()
        kcrPane.updateRequest(consoleResult, c)
      }
    }
  })
  buildAdds.add(buildButton)

  //TODO does not works if code run sub Process (still running)
  def abortableCode(toDo: => Unit): Unit = {
    val dialog = new JFrame()
    dialog.getContentPane.setLayout(new BorderLayout())
    val stopButton = new JButton("Stop")
    dialog.getContentPane.add(new JLabel(s"Stop computation"), BorderLayout.PAGE_START)
    dialog.getContentPane.add(stopButton, BorderLayout.CENTER)
    dialog.pack()
    val x = jFrame.getLocation.getX.toInt + jFrame.getSize.width / 2
    val y = jFrame.getLocation.getY.toInt + jFrame.getSize.height / 2
    dialog.setLocation(x, y)
    dialog.setVisible(true)
    Thread.sleep(100) //if let some time to the window to appear if the computation is too fast
    val c = Cancellable(toDo)
    stopButton.addActionListener(new ActionListener {
      def actionPerformed(e: ActionEvent): Unit = {
        c.cancel()
        config.cutComputationMode match {
          case HipHOPSCutComputation =>
            HipHOPSInterface.destroyAll()
          case XSAPCutComputation =>
            XSAPInterface.destroyAll()
            XSAPInterface.killXSAP()
          case XFTACutComputation =>
            XFTAInterface.destroyAll()
          case IterativeSATCutComputation =>
            ClaspInterface.destroyAll()
          case PBCutComputation =>
            PBEncoderInterface.destroyAll()
            SharpCDCLInterface.destroyAll()
          case _ =>
        }
        config.explorationMethod match {
          case GeneticExploration => HipHOPSInterface.destroyAll()
          case _ =>
        }
      }
    })
    val f = c.future onComplete {
      case Success(r) =>
        dialog.setVisible(false)
      case Failure(e: CancellationException) =>
        dialog.setVisible(false)
        consoleResult.printInfo(s"computation aborted")
      case Failure(e) =>
        dialog.setVisible(false)
        consoleResult.printError(s"An exception occured: ${e.getMessage} \n trace:${e.getStackTrace.mkString("\n")}")
    }
  }

  clear.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      consoleResult.clear()
    }
  })
  buildAdds.add(clear)

  def getCurrentConf: Option[ConfigIdent] = {
    if (configField.getSelectedItem != null) {
      Some(configField.getSelectedItem.asInstanceOf[ConfigIdent])
    } else
      None
  }

  highlight.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      for (codePanel <- getSelectedKCRPane) yield {
        codePanel.code.ifNecessaryUpdate()
      }
    }
  })
  buildAdds.add(highlight)

  def refreshConfigList(comboBox: JComboBox[ConfigIdent]): Unit = {
    comboBox.removeAllItems()
    for (kcr <- getSelectedKCRPane; s <- kcr.code.getAndUpdateAvailableConfig) {
      comboBox.addItem(s)
    }
  }

  nextButton.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      for (drawer <- getSelectedDrawer) yield {
        drawer match {
          case x: TraceDrawer =>
            x.next()
          case _ =>
        }
      }
    }
  })
  simuAdds.add(nextButton)

  def getSelectedKCRPane: Option[CodeContainer] = {
    tabbedPane.getSelectedComponent match {
      case x: CodeContainer => Some(x)
      case _ => None
    }
  }

  reinit.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      for (drawer <- getSelectedDrawer) yield {
        drawer match {
          case x: TraceDrawer =>
            x.restart()
          case _ =>
        }
      }
    }
  })
  simuAdds.add(reinit)

  def addPathMenuItem(t: ToolInterface, menu: JMenu, isLabel: Boolean = false): Unit = {
    val pathName = getToolPathOf(t) match {
      case Some(p) =>
        p.toString()
      case None =>
        consoleResult.printWarning(s"no path for tool ${t.exeName}")
        "None"
    }
    if (isLabel) {
      val pathSet = new JLabel(s" ${t.exeName} path: $pathName")
      menu.add(pathSet)
    } else {
      val pathSet = new JMenuItem(s"${t.exeName} path: $pathName")
      pathSet.addActionListener(new ActionListener {
        def actionPerformed(e: ActionEvent): Unit = {
          fc.resetChoosableFileFilters()
          fc.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES)
          val returnVal = fc.showOpenDialog(jFrame)
          if (returnVal == JFileChooser.APPROVE_OPTION) {
            if (!t.setToolPath(fc.getSelectedFile.getAbsolutePath)) {
              JOptionPane.showMessageDialog(jFrame, s"cannot find ${t.mustContain} in ${fc.getSelectedFile.getAbsolutePath}")
            } else {
              //              updateAvailableMethod()
              pathSet.setText(s"${t.exeName} path: ${t.toolPath.toAbsolute}")
              consoleResult.printInfo(s"path for ${t.exeName} successfully added")
            }
          }
        }
      })
      menu.add(pathSet)
    }
  }

  stopSim.addActionListener(new ActionListener {
    def actionPerformed(e: ActionEvent): Unit = {
      simuAdds.setVisible(false)
      for (drawer <- getSelectedDrawer) yield {
        drawer match {
          case x: TraceDrawer =>
            x.stop()
          case _ =>
        }
      }
    }
  })
  simuAdds.add(stopSim)

  def getToolPathOf(t: ToolInterface): Option[Path] = {
    try {
      Some(t.toolPath)
    } catch {
      case ToolNotFound(msg) =>
        None
    }
  }

  def getScreenSize: Dimension = {
    val screen = GraphicsEnvironment.getLocalGraphicsEnvironment.getDefaultScreenDevice.getDisplayMode
    new Dimension(screen.getHeight, screen.getWidth)
  }

  def getSelectedDrawer: Option[Drawer] = {
    tabbedPane.getSelectedComponent match {
      case DiagramContainer(d, _) => Some(d)
      case _ => None
    }
  }

  tabbedPane.addChangeListener(new ChangeListener {
    def stateChanged(e: ChangeEvent): Unit = {
      configField.removeAllItems()
      if (getSelectedKCRPane.nonEmpty) {
        buildAdds.setVisible(true)
        viewMenu.setVisible(true)
        simulationMenu.setVisible(false)
        refreshConfigList(configField)
        optionMenu.setVisible(true)
        simuAdds.setVisible(false)
        analysisMenu.setVisible(true)
      } else if (getSelectedDrawer.nonEmpty) {
        if (getSelectedDrawer.get.isInstanceOf[MDDDrawer]) {
          simulationMenu.setVisible(true)
        }
        viewMenu.setVisible(false)
        analysisMenu.setVisible(false)
        optionMenu.setVisible(false)
        buildAdds.setVisible(false)
      }
      else {
        consoleResult.clear()
        buildAdds.setVisible(false)
        analysisMenu.setVisible(false)
        viewMenu.setVisible(false)
        simulationMenu.setVisible(false)
        optionMenu.setVisible(false)
        simuAdds.setVisible(false)
        //        BDDFactory.reset()
        //        InstBoolBDDFactory.reset()
      }
    }
  })

  def time[A](f: () => A): (A, Double) = {
    val start = System.nanoTime()
    val r = f()
    val t = (System.nanoTime() - start) * 1e-9
    (r, t)
  }

  container.setLayout(new BorderLayout())

  object KCRFileFilter extends FileFilter {

    def getDescription: String = "KCR files"

    def accept(pathname: File): Boolean = pathname.getName.contains(".kcr") || pathname.isDirectory
  }

  headerContainer.add(menuBar, BorderLayout.PAGE_START)
  headerContainer.add(addsPanel, BorderLayout.CENTER)

  resultScroll.getViewport.add(consoleResult)

  container.add(tabbedPane, BorderLayout.CENTER)
  container.add(headerContainer, BorderLayout.PAGE_START)
  container.add(resultScroll, BorderLayout.PAGE_END)
  jFrame.setSize(uiDim)

  jFrame.setVisible(true)

  object XMLFileFilter extends FileFilter {

    def getDescription: String = "xml files"

    def accept(pathname: File): Boolean = pathname.getName.contains(".xml") || pathname.isDirectory
  }

  object DOTFileFilter extends FileFilter {

    def getDescription: String = "dot files"

    def accept(pathname: File): Boolean = pathname.getName.contains(".dot") || pathname.isDirectory
  }

  object SMVFileFilter extends FileFilter {

    def getDescription: String = "smv files"

    def accept(pathname: File): Boolean = pathname.getName.contains(".smv") || pathname.isDirectory
  }

  object CGFileFilter extends FileFilter {

    def getDescription: String = "circuit graph file"

    def accept(pathname: File): Boolean = pathname.getName.contains(".cg") || pathname.isDirectory
  }

  object PythonFilter extends FileFilter {

    def getDescription: String = "python file"

    def accept(pathname: File): Boolean = pathname.getName.contains(".py") || pathname.isDirectory
  }

}







