package util.analyzers

import java.io.PrintStream

import analyzers._
import util.analyzers.analyzerTypes._
import net.sf.javabdd.BDD
import util.analyzers.InfNumericImplicits._
import util.analyzers.InfNumericTypes._

import scala.collection.mutable
import scalaz.Memo._

/**
  * Created by kdelmas on 04/11/16.
  */

trait BDDAnalyzer {


  /**
    * Compute the probability to reach one terminal from root ie
    *
    * @param bdd    BDD
    * @param unrMap probability of each event attached to a BDD node
    * @return Probability to reach one terminal from root
    */
  def computeProbaBDD(bdd: BDD, unrMap: ProbabilityMapType): Real = {
    lazy val memoComputeUnrBDD: ((BDD, ProbabilityMapType)) => Real = mutableHashMapMemo {
      p => {
        val (bdd, unrMap) = p
        if (bdd.isOne) {
          1.0
        } else if (bdd.isZero) {
          0.0
        } else {
          //p(N)= p(e)*p(high)+(1-p(e))*p(low)
          memoComputeUnrBDD((bdd.high(), unrMap)) * unrMap(bdd.`var`()) + memoComputeUnrBDD((bdd.low(), unrMap)) * (1 - unrMap(bdd.`var`()))
        }
      }
    }
    memoComputeUnrBDD((bdd, unrMap))
  }

  /**
    * Compute symbolic probability to  reach one terminal from root
    *
    * @param bdd    BDD
    * @param unrMap probability of each event attached to a BDD node
    * @return symbolic probability to  reach one terminal from root
    */
  def printSymbProbaBDD(bdd: BDD, unrMap: Map[Int, String], ps: PrintStream): String = {
    val printSymbUnrBDDCache = mutable.OpenHashMap[BDD, String]()
    var cpt = 0
    val freshName = () => {
      cpt += 1; s"subformula$cpt"
    }
    lazy val memoPrintSymbUnrBDD: (BDD) => String = mutableMapMemo(printSymbUnrBDDCache) {
      bdd => {
        if (bdd.isOne) {
          "1"
        } else if (bdd.isZero) {
          "0"
        } else {
          if (bdd.high().isOne) {
            if (bdd.low().isZero) {
              unrMap(bdd.`var`())
            } else {
              val low = memoPrintSymbUnrBDD(bdd.low())
              val name = freshName()
              ps.print(s"\t$name=(${unrMap(bdd.`var`())} + (1-${unrMap(bdd.`var`())})*$low)\n")
              name
            }
          } else if (bdd.high().isZero) {
            if (bdd.low().isOne) {
              s"(1-${unrMap(bdd.`var`())})"
            } else {
              val name = freshName()
              val low = memoPrintSymbUnrBDD(bdd.low())
              ps.print(s"\t$name=((1-${unrMap(bdd.`var`())})*$low)\n")
              name
            }
          } else {
            val name = freshName()
            if (bdd.low().isOne) {
              val high = memoPrintSymbUnrBDD(bdd.high())
              ps.print(s"\t$name=($high*${unrMap(bdd.`var`())}+(1-${unrMap(bdd.`var`())}))\n")
              name
            } else if (bdd.low().isZero) {
              val high = memoPrintSymbUnrBDD(bdd.high())
              ps.print(s"\t$name=($high*${unrMap(bdd.`var`())})\n")
              name
            } else {
              val high = memoPrintSymbUnrBDD(bdd.high())
              val low = memoPrintSymbUnrBDD(bdd.low())
              ps.print(s"\t$name=($high*${unrMap(bdd.`var`())}+$low*(1-${unrMap(bdd.`var`())}))\n")
              memoPrintSymbUnrBDD(bdd.high())
              name
            }
          }
        }
      }
    }
    memoPrintSymbUnrBDD(bdd)
  }

  /**
    * Compute the order of a BDD ie the minimal number of event needed to reach one terminal
    *
    * @param bdd the BDD
    * @return the order of the BDD
    */
  def computeCardBDD(bdd: BDD): InfInt = {

    lazy val memoComputeCardBDD: (BDD) => InfInt = mutableHashMapMemo {
      bdd => {
        if (bdd.isOne) {
          //=> <1>, if node is 1 then 0 events needed to reach one
          0
        } else if (bdd.isZero) {
          //=> <0>, if node is 0 then impossible to reach terminal 0
          POSITIVE_INFINITY
        } else {
          //=> <x ? ?>
          if (bdd.high().isOne) {
            //=> <x 1 ?>
            if (bdd.low().isOne) {
              //=> <x 1 1> impossible
              throw WrongBDDSimplification(bdd.toString)
            } else if (bdd.low().isZero) {
              //=> <x 1 0>
              1
            } else {
              //=> <x 1 y>
              val one: InfInt = 1
              List(one, memoComputeCardBDD(bdd.low())).min
            }
          } else if (bdd.high().isZero) {
            //=> <x 0 ?>
            if (bdd.low().isOne) {
              //=> <x 0 1>
              0
            } else if (bdd.low().isZero) {
              //=> <x 0 0>  impossible
              throw WrongBDDSimplification(bdd.toString)
            } else {
              //=> <x 0 z>
              memoComputeCardBDD(bdd.low())
            }
          } else {
            //=> <x y ?>
            if (bdd.low().isOne) {
              //=> <x y 1>
              0
            } else if (bdd.low().isZero) {
              //=> <x y 0>
              1 + memoComputeCardBDD(bdd.high())
            } else {
              //=> <x y z>
              List(1 + memoComputeCardBDD(bdd.high()), memoComputeCardBDD(bdd.low())).min
            }
          }
        }
      }
    }
    memoComputeCardBDD(bdd)
  }

  /**
    * Compute the minimal cutsets in a BDD thanks to Rauzy algorithm
    *
    * @param bdd the BDD
    * @return the cutsets as set of sets of index
    */
  def computeCutsBDD(bdd: BDD): Set[Set[Int]] = {
    lazy val memoComputeCutsBDD: (BDD) => Set[Set[Int]] = mutableHashMapMemo {
      bdd => {
        if (bdd.isOne) {
          //=> <1>, if terminal 1 then the cutsets are {\emptyset\}
          Set(Set.empty)
        } else if (bdd.isZero) {
          //=> <0>, if terminal 0 then cutsets are \emptyset
          Set.empty
        } else {
          //=> <x ? ?>
          val S0 = memoComputeCutsBDD(bdd.low())
          val MCLF1 = memoComputeCutsBDD(bdd.high().orWith(bdd.low()))
          val S1 = {
            if (S0.nonEmpty && MCLF1.nonEmpty && S0.map(_.size).min <= MCLF1.map(_.size).max) {
              // implementation of MCLF1 'divided by' S0
              MCLF1.filterNot(setS1 => S0.exists(setS0 => setS0.subsetOf(setS1)))
            } else {
              MCLF1
            }
          }
          S1.map(set => set + bdd.`var`()) ++ S0
        }
      }
    }
    memoComputeCutsBDD(bdd)
  }

  /**
    * Compute minimal cutsets up to a given cardinality thanks to Rauzy algorithm
    *
    * @param bdd BDD
    * @param max bound of cutsets cardinality
    * @return cutsets
    */
  def computeCutsBDDUntil(bdd: BDD, max: Int): Set[Set[Int]] = {

    lazy val memoComputeCutsBDDUntil: ((BDD, Int)) => Set[Set[Int]] = mutableHashMapMemo {
      p => {
        val (bdd, remain) = p
        if (bdd.isOne) {
          //=> <1>, if terminal 1 then the cutsets are {\emptyset\}
          Set(Set.empty)
        } else if (bdd.isZero) {
          //=> <0>, if terminal 0 then cutsets are \emptyset
          Set.empty
        } else {
          //=> <x ? ?>
          //compute cutsets if bound not reached
          if (remain >= 0) {
            val S0 = memoComputeCutsBDDUntil((bdd.low(), remain))

            //compute high son if remaining cardinality bound is at least one
            if (remain > 0) {
              val MCLF1 = memoComputeCutsBDDUntil((bdd.high().orWith(bdd.low()), remain - 1))
              val S1 = {
                if (S0.nonEmpty && MCLF1.nonEmpty && S0.map(_.size).min <= MCLF1.map(_.size).max) {
                  // implementation of MCLF1 'divided by' S0
                  MCLF1.filterNot(setS1 => S0.exists(setS0 => setS0.subsetOf(setS1)))
                } else {
                  MCLF1
                }
              }
              S1.map(set => set.union(Set(bdd.`var`()))).union(S0)
            }
            else
              S0
          } else {
            Set.empty
          }
        }
      }
    }
    memoComputeCutsBDDUntil((bdd, max))
  }
}
