package util.analyzers

import java.io.{ByteArrayOutputStream, FileOutputStream, PrintStream}

import ast.dsl.InstBoolIdent
import net.sf.javabdd.BDD
import com.microsoft.z3.{Sort => Z3Sort,
  Symbol => Z3Symbol, FuncDecl => Z3FuncDecl,
  Solver => Z3Solver, BoolExpr => Z3BoolExpr,
  Expr => Z3Expr,
  BitVecExpr => Z3BitVecExpr,
  BitVecNum}
/**
  * Trait for encoding structure function as a BDD
  */
trait BDDBuild extends PhiAnalyzer {
  self: Phi =>

  lazy val (phiBDD, varMap) = _computeBDD()

  val BDDFactory: BDDFactory with Z3BoolExprToBDDTranslator

  /**
    * Compute the BDD of the structure function from its Boolean expression
    *
    * @return the BDD and the mapping of system events to BDD nodes
    */
  private def _computeBDD(): (BDD, Map[InstBoolIdent, BDD]) = {
    val eventsList: List[(InstBoolIdent, Z3BoolExpr)] = events.toList.sortBy(p => p._2.getIndex)
    val varMap: List[(InstBoolIdent, Z3BoolExpr, BDD)] = eventsList.map(e => (e._1, e._2, BDDFactory.produceVar()))
    (BDDFactory.boolExprToBdd(expr, varMap.map(t => (t._2, t._3)).toMap, context), varMap.map(t => (t._1, t._3)).toMap)
  }

  /**
    * Export BDD to dot format
    *
    * @param fileName the output file
    */
  def printDot(fileName: String): Unit = {
    val fos = new FileOutputStream(fileName)
    val ps = new PrintStream(fos)
    val baos = new ByteArrayOutputStream()
    val psBaos = new PrintStream(baos)

    val old = System.out
    var dotBDD = ""
    if (phiBDD.isZero)
      dotBDD = "digraph G {\n0 [shape=box, label=\"0\", style=filled, shape=box, height=0.3, width=0.3]\n}\n"
    else if (phiBDD.isOne)
      dotBDD = "digraph G {\n0 [shape=box, label=\"1\", style=filled, shape=box, height=0.3, width=0.3]\n}\n"
    else {
      //since JavaBDD dot export is print in standard out, rediredting it to a file
      System.setOut(psBaos)
      phiBDD.printDot()
      //reinitialise standard output
      System.setOut(old)
      dotBDD = baos.toString()
    }
    for ((k, v) <- varMap) {
      dotBDD = dotBDD.replace("[label=\"" + v.`var`() + "\"]", "[label=\"" + k + "\"]")
    }
    val splittedString = dotBDD.split("\n")
    for (s <- splittedString)
      ps.println(s)
    ps.close()
    psBaos.close()
  }
}
