/**
  * Created by kdelmas on 05/02/16.
  */
package util.analyzers

import ast.dsl.InstBoolIdent
import net.sf.javabdd.{BDDPairing, BDD => JavaBDD, BDDFactory => JavaBDDFactory}

import scala.collection.mutable

/**
  * BDDFactory where variables labelled on BDD node are InstBoolIdent
  */
class InstBoolBDDFactory extends GenBDDFactory[InstBoolIdent] {
  protected val _factory: JavaBDDFactory = initFactory(500, 500)
  protected val varMap: mutable.HashMap[InstBoolIdent, Int] = mutable.HashMap.empty
  protected var _varCount = 0
  protected val nbOfVar: Int = 500
}

/**
  * base trait for BDD factories
  * @tparam Var the type of variable labelled on BDD nodes
  * @tparam MyBDD the BDD representation which must be a subtype of JavaBDD type
  */
trait BaseBDDFactory[Var, MyBDD <: JavaBDD] {

  /**
    * initialisation of the cache and number of node of the factory
    * @param numberOfVar maximum number of variables in BDDs
    * @param cacheSize initial size of the cache table containing BDD nodes
    * @return
    */
  def initFactory(numberOfVar: Int, cacheSize: Int): JavaBDDFactory

  /**
    * Produce a BDD node labelled with a given variable
    * @param variable the variable
    * @return a BDD node
    */
  def produceVar(variable: Var): MyBDD

  /**
    * Return the ith BDD node in the table
    * @param i the index of the BDD node
    * @return BDD node
    */
  def getIthVar(i: Int): MyBDD

  /**
    * Clean the cache and the index table
    */
  def reset()

  /**
    * Return the zero terminal
    * @return zero terminal
    */
  def zero(): MyBDD

  /**
    * Return the one terminal
    * @return one terminal
    */
  def one(): MyBDD

  /**
    * BDD AND
    * @param left BDD
    * @param right BDD
    * @return the resulting BDD
    */
  def andBDD(left: MyBDD, right: MyBDD): MyBDD

  /**
    * BDD OR
    * @param left BDD
    * @param right BDD
    * @return the resulting BDD
    */
  def orBDD(left: MyBDD, right: MyBDD): MyBDD

  /**
    * BDD negation
    * @param arg initial BDD
    * @return negated BDD
    */
  def notBDD(arg: MyBDD): MyBDD

  /**
    * Replace all BDD nodes labelled by a given to variable to another one
    * @param replace the initial variable to replace
    * @param by the new variable
    * @param in the BDD
    * @return the modified BDD
    */
  def replaceVar(replace: Var, by: Var, in: MyBDD): MyBDD

  /**
    * Build a MDDNode labelled by a variable where the high and low sons are given
    * @param variable the variable labelling the BDD
    * @param high the high son
    * @param low the low son
    * @return the resulting BDD
    */
  def mkNode(variable: Var, high: MyBDD, low: MyBDD): MyBDD

  /**
    * Free native data structure if exists
    */
  def dispose()

  /**
    * @return the mapping from BDD to labelled variable
    */
  def getVarMap: Map[MyBDD, Var]

  /**
    * Import a BDD in this factory from one coming from another factory
    * @param bdd the other factory BDD
    * @param bddVar the map from BDD node to variables
    * @tparam OtherBDD the type of the other BDD
    * @return the BDD imported in this factory
    */
  def importBDD[OtherBDD <: JavaBDD](bdd: OtherBDD, bddVar: Map[Var, OtherBDD]): MyBDD = {
    //add fresh vars in this factory
    bddVar.foreach(kv => kv._1 -> produceVar(kv._1).`var`())
    //copy node by node the initial bdd by replacing old variables by freshly created one
    importRec(bdd, bddVar.map(p => p._2.`var`() -> p._1))
  }

  /**
    * Build a BDD in this factory from one coming from another factory
    * @param bdd the other factory BDD
    * @tparam OtherBDD the type of the other BDD
    * @return the BDD imported in this factory
    */
  private def importRec[OtherBDD <: JavaBDD](bdd: OtherBDD, bddMap: Map[Int, Var]): MyBDD = {
    if (bdd.isOne)
      one()
    else if (bdd.isZero)
      zero()
    else
      mkNode(bddMap(bdd.`var`()), importRec(bdd.high(), bddMap), importRec(bdd.low(), bddMap))
  }

}

/**
  * Class for a classic BDD Factory (variable are integer and BDD are JavaBDD)
  */
class BDDFactory extends BaseBDDFactory[Int, JavaBDD] {

  //the number of variable produced by the factory
  private var _varCount = 0

  //initial maximum number of variables
  val nbOfVar: Int = 500

  //the JavaBDD factory used to delegate BDD computation
  protected val _factory: JavaBDDFactory = initFactory(nbOfVar, 1000)

  /**
    * Initialise a JavaBDD factory
    * @param numberOfVar maximum number of variables in BDDs
    * @param cacheSize initial size of the cache table containing BDD nodes
    * @return  JavaBDD factory
    */
  def initFactory(numberOfVar: Int, cacheSize: Int): JavaBDDFactory = {
    var numberOfNodes = numberOfVar * numberOfVar
    numberOfNodes = Math.max(cacheSize, numberOfNodes)
    val fac = JavaBDDFactory.init("java", numberOfNodes, cacheSize)
    if (fac.varNum() < numberOfVar)
      fac.setVarNum(numberOfVar)
    fac.autoReorder(JavaBDDFactory.REORDER_WIN2)
    fac
  }

  /**
    * Build a variable labelled by the current number of variable produced by the factory
    *
    * @return  BDD
    */
  def produceVar(): JavaBDD = {
    _varCount = _varCount + 1
    if ((_varCount % nbOfVar) == 0)
      _factory.setVarNum(_varCount + nbOfVar)

    _factory.ithVar(_varCount)
  }

  /**
    * Increment the number of produced variable and return the corresponding BDD
    * @param variable the variable
    * @return a BDD node
    */
  def produceVar(variable: Int): JavaBDD = {
    if (variable > _varCount)
      _varCount = variable
    _factory.ithVar(variable)
  }

  /**
    * Return the ith var of the JavaBDD factory
    * @param i the index of the BDD node
    * @return BDD node
    */
  def getIthVar(i: Int): JavaBDD = {
    _factory.ithVar(i)
  }

  /**
    * Reinitialise the JavaBDD factory to initial variable number
    */
  def reset(): Unit = {
    _varCount = 0
    _factory.reset()
    _factory.setVarNum(nbOfVar)

  }

  /**
    * @return zero terminal
    */
  def zero(): JavaBDD = {
    _factory.zero()
  }

  /**
    *
    * @return one terminal
    */
  def one(): JavaBDD = {
    _factory.one()
  }

  /**
    * Delegate BDD AND to JavaBDD factory
    * @param left BDD
    * @param right BDD
    * @return the resulting BDD
    */
  def andBDD(left: JavaBDD, right: JavaBDD): JavaBDD = {
    left.and(right)
  }

  /**
    * Delegate BDD OR to JavaBDD factory
    * @param left BDD
    * @param right BDD
    * @return the resulting BDD
    */
  def orBDD(left: JavaBDD, right: JavaBDD): JavaBDD = {
    left.or(right)
  }

  /**
    * Delegate BDD NOT to JavaBDD
    * @param arg initial BDD
    * @return negated BDD
    */
  def notBDD(arg: JavaBDD): JavaBDD = {
    arg.id().not()
  }

  /**
    * Use JavaBDD replace to build new BDD where a given variable is replaced by another one
    * @param replace the initial variable to replace
    * @param by the new variable
    * @param in the BDD
    * @return the modified BDD
    */
  def replaceVar(replace: Int, by: Int, in: JavaBDD): JavaBDD = {
    val pair = _factory.makePair(replace, by)
    in.replace(pair)
  }

  /**
    * Build a BDD bu applying the formula v.high + (neg v).low
    * @param variable the variable labelling the BDD
    * @param high the high son
    * @param low the low son
    * @return the resulting BDD
    */
  def mkNode(variable: Int, high: JavaBDD, low: JavaBDD): JavaBDD = {
    getIthVar(variable).and(high).or(getIthVar(variable).not().and(low))
  }

  /**
    * Send dispose signal to JavaBDD factory
    */
  def dispose(): Unit = {
    _factory.done()
  }

  /**
    * Build the map from BDD identifier to BDD
    * @return the mapping from BDD to labelled variable
    */
  def getVarMap: Map[JavaBDD, Int] = {
    (0 to _varCount).map(i => getIthVar(i) -> i).toMap
  }
}

/**
  * Generic BDD factory over the type of BDD variables
  * @tparam Var the type of variable labelled on BDD nodes
  */
trait GenBDDFactory[Var] extends BaseBDDFactory[Var, JavaBDD] {

  object FactoryImplicits {
    /**
      * Transform variable to their BDD
      * @param variable the variable labelled on the root node
      * @return the BDD "< v, 1, 0>"
      */
    implicit def toJavaBDD(variable: Var): JavaBDD = produceVar(variable)
  }

  protected var _varCount: Int
  protected val _factory: JavaBDDFactory
  protected val nbOfVar: Int
  protected val varMap: mutable.HashMap[Var, Int]


  /**
    *  Initialise a JavaBDD factory
    * @param numberOfVar maximum number of variables in BDDs
    * @param cacheSize initial size of the cache table containing BDD nodes
    * @return
    */
  def initFactory(numberOfVar: Int, cacheSize: Int): JavaBDDFactory = {
    var numberOfNodes = numberOfVar * numberOfVar
    numberOfNodes = Math.max(cacheSize, numberOfNodes)
    val fac = JavaBDDFactory.init("java", numberOfNodes, cacheSize)
    if (fac.varNum() < numberOfVar)
      fac.setVarNum(numberOfVar)
    fac.autoReorder(JavaBDDFactory.REORDER_WIN2)
    fac
  }

  /**
    * Try to find the BDD of the variable in the correspondence table
    * and return it if existing, otherwise generate a new BDD
    * @param variable the variable
    * @return a BDD node
    */
  def produceVar(variable: Var): JavaBDD = {
    varMap.get(variable) match {
      case Some(i) =>
        getIthVar(i)
      case None =>
        _varCount = _varCount + 1

        if ((_varCount % nbOfVar) == 0)
          _factory.setVarNum(_varCount + nbOfVar)
        val bdd = _factory.ithVar(_varCount)
        varMap += (variable -> bdd.`var`() )
        bdd
    }
  }

  /**
    * Reinitialise the JavaBDD factory to initial variable number
    */
  def reset(): Unit = {
    varMap.clear()
    _factory.reset()
    _varCount = 0
    _factory.setVarNum(nbOfVar)
  }

  /**
    *
    * @return zero terminal
    */
  def zero(): JavaBDD = {
    _factory.zero()
  }

  /**
    *
    * @return one terminal
    */
  def one(): JavaBDD = {
    _factory.one()
  }

  /**
    * Return the BDD labelled by the ith variable
    * @param i the index of the BDD node
    * @return BDD node
    */
  def getIthVar(i: Int): JavaBDD = {
    _factory.ithVar(i)
  }

  /**
    * Delegate BDD AND to JavaBDD factory
    * @param left BDD
    * @param right BDD
    * @return the resulting BDD
    */
  def andBDD(left: JavaBDD, right: JavaBDD): JavaBDD = {
    left.and(right)
  }

  /**
    * Delegate BDD OR to JavaBDD factory
    * @param left BDD
    * @param right BDD
    * @return the resulting BDD
    */
  def orBDD(left: JavaBDD, right: JavaBDD): JavaBDD = {
    left.or(right)
  }

  /**
    * Delegate BDD Not to JavaBDD factory
    * @param arg initial BDD
    * @return negated BDD
    */
  def notBDD(arg: JavaBDD): JavaBDD = {
    arg.id().not()
  }

  /**
    *
    * @param replace the initial variable to replace
    * @param by the new variable
    * @param in the BDD
    * @return the modified BDD
    */
  def replaceVar(replace: Var, by: Var, in: JavaBDD): JavaBDD = {
    {
      for (i1 <- varMap.get(replace);
           i2 <- varMap.get(by))
        yield {
          in.replace(_factory.makePair(i1, i2))
        }
    }.getOrElse(throw new Exception(s"unknown variable $replace or $by"))
  }

  /**
    * Build a BDD bu applying the formula v.high + (neg v).low
    * @param variable the variable labelling the BDD
    * @param high the high son
    * @param low the low son
    * @return the resulting BDD
    */
  def mkNode(variable: Var, high: JavaBDD, low: JavaBDD): JavaBDD = {
    varMap.get(variable) match {
      case Some(index) =>
        getIthVar(index).and(high).or(getIthVar(index).not().and(low))
      case None =>
        throw new Exception(s"unknown variable $variable")
    }

  }

  /**
    * Send dispose signal to JavaBDD factory
    */
  def dispose(): Unit = {
    _factory.done()
    varMap.clear()
    _varCount = 0
  }

  /**
    *
    * @return the mapping from BDD to labelled variable
    */
  def getVarMap: Map[JavaBDD, Var] = {
    varMap.map(kv => getIthVar(kv._2) -> kv._1).toMap
  }
}
