package util.analyzers

import ast.dsl.InstBoolIdent
import util.analyzers.InfNumericTypes.InfInt
import com.microsoft.z3.{Sort => Z3Sort,
  Symbol => Z3Symbol, FuncDecl => Z3FuncDecl,
  Solver => Z3Solver, BoolExpr => Z3BoolExpr,
  Expr => Z3Expr,
  BitVecExpr => Z3BitVecExpr,
  BitVecNum}
import com.microsoft.z3.Status._
/**
  * Trait for order computation
  */
trait CardinalityAnalysis extends PhiAnalyzer {
  self: Phi =>

  import InfNumericImplicits._

  lazy val cardinality: InfInt = compute()

  /**
    * Compute the order of a system by solving an SMT problem
    *
    * @return the order of the system
    */
  private def compute(): InfInt = {

    //initial order
    var result = 0

    //generate Z3 solver
    val s = context.mkSolver()

    val eventsList: List[(InstBoolIdent, Z3BoolExpr)] = events.toList

    //generate Boolean interpreted constant for each system event
    val consts: (List[Z3BoolExpr], List[Z3BoolExpr]) = eventsList.map(e => (e._2, context.mkBoolConst(e._1.toString()))).unzip

    //substitute Z3 variables by system event and assert if (ie assert phi(evt))
    s.add(expr.substitute(consts._1.toArray[Z3Expr], consts._2.toArray[Z3Expr]).asInstanceOf[Z3BoolExpr])

    s.check() match {

      //if UNSAT then phi cannot be true
      case UNSATISFIABLE =>

        //free the solver and return that an infinite number of event is needed to trigger phi
        s.dispose()
        POSITIVE_INFINITY

      //if SAT then at least on valuation of the system events can trigger phi
      case SATISFIABLE =>

        //add restore point
        s.push()

        //add cardinality constraint
        s.add(context.atMostN(consts._2, result))

        //while cannot find any cutset of size "result" triggering phi, then increment "result"
        while (s.check() == UNSATISFIABLE) {
          result = result + 1
          s.pop()
          s.push()
          s.add(context.atMostN(consts._2, result))
        }
        if (s.check() == UNKNOWN)
          throw Z3UnknownException(this)
        else {
          s.dispose()
          result
        }
      case UNKNOWN => throw Z3UnknownException(this)
    }
  }
}
