package util.analyzers

import util.analyzers.MDDType._

import scala.collection.mutable.{ArrayBuffer => Bucket, OpenHashMap => CacheMap}

/**
  * Base trait for node factory based on hashConsing
  * @tparam E the type of the nodes
  */
trait Factory[E] {

  // mapping from hashcode to nodes
  final protected val map = new CacheMap[StructuralHashType, Bucket[E]]

  final val cache = this

  def clear(): Unit = map.clear()
}

/**
  * Factory for nodes containing only a parameter
  */
object FactoryP {

  /**
    * Hashcode computation for leaf node
    * @param param the parameter labelled on the node
    * @tparam P the type of the parameter
    * @return hashcode
    */
  def sHashCode[P](param: P): StructuralHashType = 41 * (41 + param.hashCode) + this.hashCode

  /**
    * Structural equality
    * @param n the node
    * @param param the other node parameter
    * @tparam E the type of other node which must be subtype of leaf node
    * @tparam P the parameter type
    * @return structural equality
    */
  def sEquals[E <: NodeP[P], P](n: E, param: P): Boolean = n.param == param
}

/**
  * Factory for nodes containing only a parameter
  * @tparam E the type of the nodes
  * @tparam P the type of parameter
  */
trait FactoryP[E <: NodeP[P], P] extends Factory[E] {

  /**
    * create a node from a parameter
    * @param param parameter
    * @return node containing parameter
    */
  def make(param: P): E

  /**
    * Return node containing the given parameter if already create
    * otherwise make it
    * @param param the parameter
    * @return the node containing the parameter
    */
  def apply(param: P): E = getOrElseUpdate(param, make(param))

  /**
    * Return node containing the given parameter if already create
    * otherwise make it
    * @param param the parameter
    * @param create the building function
    * @return the node containing the parameter
    */
  def getOrElseUpdate(param: P, create: => E): E = {
    val hash = FactoryP.sHashCode(param)
    val res = map.get(hash) match {
      case None =>
        val x = create
        val bucket = new Bucket[E]()
        bucket += x
        map.update(hash, bucket)
        x
      case Some(bucket) => bucket.find(x => FactoryP.sEquals(x, param)) match {
        case Some(result) => result
        case None =>
          val x = create
          bucket += x
          x
      }
    }
    res
  }
}

/**
  * Extractor from parameter node
  * @tparam E the node type
  * @tparam P the parameter type
  */
trait ExtractorP[E <: NodeP[P], P] {
  def unapply(e: E): Option[P] = Some(e.param)
}

//------------------------------------------
/**
  * Factory for node containing parameter and a map of kids
  */
object FactoryNP {

  /**
    * Hashcode computation for leaf node
    * @param kids the kids of the node
    * @param param the parameter of the node
    * @tparam K the type of kids
    * @tparam P the type of node parameter
    * @tparam V the type of labels of the edges
    * @return hashcode
    */
  def sHashCode[K, P, V <: Node](kids: Map[K, V], param: P): StructuralHashType = 41 * (41 * kids.foldLeft(1) { (res, x) => 41 * res + x.hashCode } + param.hashCode) + this.hashCode

  /**
    * Structural equality
    * @param n the node to compare
    * @param kids the kids of the other node
    * @param param the parameter of the other node
    * @tparam E the type of the node
    * @tparam K the type of edge labels
    * @tparam P the type of the parameter
    * @tparam V the type of the kids
    * @return is structurally equals
    */
  def sEquals[E <: NodeNP[K, V, P], K, P, V <: Node](n: E, kids: Map[K, V], param: P): Boolean = n.kids == kids && n.param == param
}

/**
  * Hashcode computation for leaf node
  * @tparam T the type of reduced nodes
  * @tparam E the type of the nodes
  * @tparam K the type of edge labels
  * @tparam P the type of the parameter
  * @tparam V the type of the kids
  * @tparam M the type of the collection gathering kids
  */
trait FactoryNP[T, E <: T with NodeNP[K, V, P], K, P, V <: Node, M[K, V] <: Map[K, V]] extends Factory[E] {

  /**
    * Generate nodes
    * @param kids the kids of the future node
    * @param param the parameter of the future node
    * @return the resulting node
    */
  def make(kids: M[K, V], param: P): E

  /**
    * Apply reduction rule
    * @param kids kids of the node
    * @param param the parameter of the node
    * @return a possible reduced node
    */
  def reduce(kids: M[K, V], param: P): Option[T]

  /**
    * Generate a node
    * @param kids the future node kids
    * @param param the future node parameter
    * @return the node
    */
  def apply(kids: M[K, V], param: P): T = reduce(kids, param) match {
    case Some(x) => x
    case None => getOrElseUpdate(kids, param, make(kids, param))
  }

  /**
    * Return node containing the given parameter if already create
    * otherwise make it
    * @param kids the kids of the node
    * @param param the parameter
    * @param create the building function
    * @return the node containing the parameter
    */
  def getOrElseUpdate(kids: M[K, V], param: P, create: => E): E = {
    val hash = FactoryNP.sHashCode(kids, param)
    val res = map.get(hash) match {
      case None =>
        val x = create
        val bucket = new Bucket[E]()
        bucket += x
        map.update(hash, bucket)
        x
      case Some(bucket) => bucket.find(x => FactoryNP.sEquals(x, kids, param)) match {
        case Some(result) => result
        case None =>
          val x = create
          bucket += x
          x
      }
    }
    res
  }
}

/**
  * Extractor from parameter node
  * @tparam E the type of the node
  * @tparam K the type of edge label
  * @tparam P the type of parameter
  * @tparam V the type of kids
  */
trait ExtractorNP[E <: NodeNP[K, V, P], K, P, V <: Node] {
  def unapply(e: E): Option[(Map[K, V], P)] = Some((e.kids, e.param))
}

//------------------------------------------
/**
  * Factory for node containing list of kids
  */
object FactoryListP {

  /**
    * Structural hashcode
    * @param kids the kids of the node
    * @param param the parameter of the node
    * @tparam P the type of the parameter
    * @tparam K  the type of the kids
    * @return the hashcode
    */
  def sHashCode[P, K <: Node](kids: List[K], param: P): StructuralHashType = 41 * (41 * kids.foldLeft(1) { (res, x) => 41 * res + x.hashCode } + param.hashCode) + this.hashCode

  /**
    * Structural equality
    * @param n the node
    * @param kids the kids of the other node
    * @param param the parameter of the other node
    * @tparam E the type of the node
    * @tparam P the type of the parameter
    * @tparam K the type of the kids
    * @return the structural equality
    */
  def sEquals[E <: NodeListP[K, P], P, K <: Node](n: E, kids: List[K], param: P): Boolean = n.kids == kids && n.param == param
}

/**
  * Factory for node containing list of kids
  * @tparam T the type of simplified node
  * @tparam E the type of the nodes
  * @tparam P the type of the parameter
  * @tparam K the type of kids
  */
trait FactoryListP[T, E <: T with NodeListP[K, P], P, K <: Node] extends Factory[E] {

  /**
    * Generate a node from parameters
    * @param kids the kids of the future node
    * @param param the parameter of the future node
    * @return the node
    */
  def make(kids: List[K], param: P): E

  /**
    * Apply reduction rule
    * @param kids kids of the node
    * @param param the parameter of the node
    * @return a possible reduced node
    */
  def reduce(kids: List[K], param: P): Option[T]

  /**
    * Generate a node
    * @param kids the future node kids
    * @param param the future node parameter
    * @return the node
    */
  def apply(kids: List[K], param: P): T = reduce(kids, param) match {
    case Some(x) => x
    case None => getOrElseUpdate(kids, param, make(kids, param))
  }

  /**
    * Return node containing the given parameter if already create
    * otherwise make it
    * @param kids the kids of the node
    * @param param the parameter
    * @param create the building function
    * @return the node containing the parameter
    */
  def getOrElseUpdate(kids: List[K], param: P, create: => E): E = {
    val hash = FactoryListP.sHashCode(kids, param)
    val res = map.get(hash) match {
      case None =>
        val x = create
        val bucket = new Bucket[E]()
        bucket += x
        map.update(hash, bucket)
        x
      case Some(bucket) => bucket.find(x => FactoryListP.sEquals(x, kids, param)) match {
        case Some(result) => result
        case None =>
          val x = create
          bucket += x
          x
      }
    }
    res
  }
}

/**
  * Extractor from parameter node
  * @tparam E the type of the node
  * @tparam K the type of the kids
  * @tparam P the type of parameter
  */
trait ExtractorListP[E <: NodeListP[K, P], P, K <: Node] {
  def unapply(e: E): Option[(List[K], P)] = Some((e.kids, e.param))
}
