package util.analyzers

import java.io.{FileOutputStream, PrintStream}

import scala.collection.immutable.{SortedMap, SortedSet}
import scala.util.control.TailCalls._
import scala.util.{Failure, Success, Try}
import scalaz._


/**
  * Created by kdelmas on 21/09/16.
  */


/**
  * MDD factory class
  *
  * @param vars the variables of the MDDs of the factory
  * @param extractor    an extractor function from Domain type to a collection of values
  * @tparam Var Variable type
  * @tparam F   collection type for values
  * @tparam Val values type
  * @tparam TL  terminal type
  * @tparam Dom variables' domain type
  */
class MDDFactory[Var <: Ordered[Var],
                F[V] <: Traversable[V],
                Val <: Ordered[Val],
                TL <: Ordered[TL], Dom]
                (vars: Map[Var, Dom] = Map[Var, Dom](),
                 extractor: (Dom) => F[Val]) extends GenMDDFactory[Var, F, Val, TL, Dom] {

  protected var domains: SortedMap[Var, Domain] =
    SortedMap.empty[Var, Domain] ++ vars.map(p => p._1 -> Domain(p._1, extractor(p._2), Some(p._2)))
}

/**
  * Base trait for MDD factory
  *
  * @tparam Var Variable type
  * @tparam F   collection type for values
  * @tparam Val values type
  * @tparam TL  terminal type
  * @tparam Dom variables' domain type
  */
trait GenMDDFactory[Var <: Ordered[Var], F[V] <: Traversable[V], Val <: Ordered[Val], TL <: Ordered[TL], Dom] {

  implicit object DomainOrdering extends scala.math.Ordering[Domain] {
    def compare(x: Domain, y: Domain): Int = x.compare(y)
  }

  implicit object TerminalOrdering extends scala.math.Ordering[ValuedTerminal] {
    def compare(x: ValuedTerminal, y: ValuedTerminal): Int = x.param.compare(y.param)
  }

  //mappin of variable to their domain
  protected var domains: SortedMap[Var, Domain]

  /**
    * Domain class gathering the possible values of a variable, implements the traversable trait
    *
    * @param variable   the variavle
    * @param domain     the collection of possible values
    * @param initialDom the initial domain given by the user
    */
  case class Domain private[GenMDDFactory](variable: Var, domain: F[Val], initialDom: Option[Dom] = None) extends Traversable[Val] with Ordered[Domain] {

    def foreach[U](f: (Val) => U): Unit = toList.foreach(f)

    override def toList: List[Val] = domain.toList.sorted

    def compare(that: Domain): Int = variable.compare(that.variable)

    def isAcceptableValue(value: Val): Boolean = domain.exists(v => v == value)

    def indexOf(value: Val): Option[Int] = {
      if (isAcceptableValue(value)) {
        Some(toList.indexOf(value))
      } else {
        None
      }
    }

    override def toString(): String = s"$variable"
  }

  /**
    * trait for MDD where terminals is the terminals of the MDD,
    */
  sealed trait MDD extends Node {

    //the terminals of this mDD
    val terminals: SortedSet[ValuedTerminal]

    //the variables labelled on the node of this MDD
    val variables: SortedSet[Var]

    /**
      * F.applyCase(t1->G1,..,tp->Gp) apply MDD Gi if F=ti, thus we must have p=|terminals|
      *
      * @param altsMap gives for each terminal ti of F the corresponding alternative
      * @return the resulting MDD after application of alternatives
      */
    def applyCase(altsMap: Map[ValuedTerminal, MDD]): MDD = {
      val missingTerminals = terminals.diff(altsMap.keySet).foldLeft(Map[Terminal, MDD]())((acc, term) => acc + (term -> term))
      applyCaseTailRec(altsMap ++ missingTerminals).result
    }

    /**
      * apply the case operator by giving the alternative in order according to the mdd terminals order
      *
      * @param alt alternative fo terminal i
      * @return the resulting MDD
      */
    def applyCaseInOrder(alt: MDD*): MDD = applyCase(terminals.toList.zip(alt).toMap)

    protected[GenMDDFactory] val applyCaseTailRec: (Map[Terminal, MDD]) => TailRec[MDD]

    /**
      * apply the case operator by setting all terminals alternatives which are not in altsMap to the default value
      *
      * @param altsMap the alternative map
      * @param default the default value
      * @return the resulting MDD
      */
    def applyCaseWithDefault(altsMap: Map[ValuedTerminal, MDD], default: ValuedTerminal): MDD = {
      applyCase(terminals.toList.map(terminal => terminal -> altsMap.getOrElse(terminal, default)).toMap)
    }

    /**
      * apply the case operator by setting all terminals alternatives which are not in altsMap to the trash terminal
      *
      * @param altsMap the alternative map
      * @return the resulting MDD
      */
    def applyCaseWithDefaultTrash(altsMap: Map[ValuedTerminal, MDD]): MDD = {
      val missingTerminals = terminals.diff(altsMap.keySet).foldLeft(Map[Terminal, MDD]())((acc, term) => acc + (term -> TrashTerminal))
      applyCaseTailRec(altsMap ++ missingTerminals).result
    }

    /**
      * adds all the path to the terminal as forbidden values and returns the result
      *
      * @param terminal the terminal to remove
      * @return
      */
    def removeTerminal(terminal: ValuedTerminal): MDD = {
      if (terminals.contains(terminal)) {
        applyCase(Map(terminal -> TrashTerminal))
      } else {
        this
      }
    }

    /**
      * adds all the path to the terminal as forbidden values and returns the result
      *
      * @param toReplace the terminals to remove
      * @return
      */
    def removeTerminals(toReplace: Traversable[ValuedTerminal]): MDD = {
      val replaceMap = toReplace.collect({ case t: ValuedTerminal if terminals.contains(t) => t -> TrashTerminal }).toMap
      if (replaceMap.isEmpty) {
        this
      } else {
        applyCase(replaceMap)
      }
    }

    /**
      * coFactor(v,vi) gives the resulting MDD where v=vi
      *
      * @param label the label to assign
      * @param value the value to assign
      * @return the coFactor of this BDD
      */
    protected[GenMDDFactory] def coFactor(label: Var, value: Val): MDD
  }


  sealed trait Terminal extends MDD

  /**
    * case class representing the Terminals
    *
    * @param param actual value of the terminal
    */
  class ValuedTerminal private[ValuedTerminal](final val param: TL) extends Terminal with NodeP[TL] {

    //terminal in a terminal is himself
    val terminals = SortedSet(this)

    //no variables for a terminal
    val variables: SortedSet[Var] = SortedSet.empty

    /**
      * If the selector F is a terminal T then the alternative is the one corresponding to T
      *
      * @return the resulting MDD after application of alternatives
      */
    protected[GenMDDFactory] val applyCaseTailRec: Map[Terminal, MDD] => TailRec[MDD] = alt => done(alt(this))


    /**
      * The cofactor of a terminal is himself
      *
      * @param label the label to assign
      * @param value the value to assign
      * @return the coFactor of this BDD
      */
    protected[GenMDDFactory] def coFactor(label: Var, value: Val): MDD = this


    override def toString: String = s"Terminal($param) "

  }


  object TrashTerminal extends Terminal {

    //terminal in a terminal is himself
    val terminals: SortedSet[ValuedTerminal] = SortedSet.empty

    //no variables for a terminal
    val variables: SortedSet[Var] = SortedSet.empty

    /**
      * If the selector F is a terminal T then the alternative is the one corresponding to T
      *
      * @return the resulting MDD after application of alternatives
      */
    protected[GenMDDFactory] val applyCaseTailRec: Map[Terminal, MDD] => TailRec[Terminal] = _ => done(this)

    /**
      * The cofactor of a terminal is himself
      *
      * @param label the label to assign
      * @param value the value to assign
      * @return the coFactor of this BDD
      */
    protected[GenMDDFactory] def coFactor(label: Var, value: Val): MDD = this

    override def toString: String = s"TrashTerminal "
  }

  private object ValuedTerminal extends FactoryP[ValuedTerminal, TL] with ExtractorP[ValuedTerminal, TL] {
    def make(param: TL) = new ValuedTerminal(param)
  }

  /**
    * case class for Node of a MDD
    *
    * @param param the decision label of this Node
    * @param kids  the kids for each possible value of topVar (accessible from the main MDDMap)
    */
  class MDDNode private[MDDNode](final val param: Var, final val kids: SortedMap[Val, MDD]) extends MDD with NodeNP[Val, MDD, Var] with Traversable[(Val, MDD)] {
    //the terminals of a node is the one contained in its sons
    val terminals: SortedSet[ValuedTerminal] = kids.values.foldLeft(SortedSet.empty[ValuedTerminal])((acc, p) => acc ++ p.terminals)

    val variables: SortedSet[Var] = SortedSet(param +: kids.flatMap(p => p._2.variables).toSeq: _*)

    private val kidsArray = kids.toArray

    def foreach[U](f: ((Val, MDD)) => U): Unit = {
      kidsArray.foreach(a => f(a))
    }

    /**
      * If the selector is a node then we apply the formula of
      * "Algorithms for Discrete Function Manipulation" (doi: 10.1109/ICCAD.1990.129849)
      * F.applyCase(t1->G1,..,tp->Gp)=
      * (v, v1->F.coFactor(v,v1).applyCase(t1->G1.cofactor(v,v1),...,tp->Gp.cofactor(v,v1)),
      * ...,
      * vn->F.coFactor(v,vn).applyCase(t1->G1.cofactor(v,vn),...,tp->Gp.cofactor(v,vn)))
      * where v is the lowest var in {F,G1,...,Gp} and its domain is {v1..vn} and terminal is a subset of {t1,...,tp}.
      * Indeed for an applyCase, the MDD is build only if for each possible terminal the alternative is specified
      *
      * @return the resulting MDD after application of alternatives
      */
    protected[GenMDDFactory] val applyCaseTailRec: (Map[Terminal, MDD]) => TailRec[MDD] = Memo.weakHashMapMemo {
      altsMap => {
        if (altsMap.forall(p => p._1 == p._2)) {
          done(this)
        } else if (altsMap.values.toSet.size == 1) {
          done(altsMap.head._2)
        } else {
          val minVar = (altsMap.collect({ case (_, mdd: MDDNode) => mdd.param }).toList :+ param).min
          domains.get(minVar) match {
            case Some(dom: Domain) =>
              val newAltList = dom.map(value => value -> tailcall(coFactor(minVar, value).
                applyCaseTailRec(altsMap.map(p => p._1 -> p._2.coFactor(minVar, value))))).toMap
              done(MDDNode(
                newAltList.foldLeft(SortedMap[Val, MDD]())((acc, kv) => {
                  kv._2.result match {
                    case TrashTerminal => acc
                    case t: ValuedTerminal => acc + (kv._1 -> t)
                    case n: MDDNode => acc + (kv._1 -> n)
                  }
                }),
                minVar))
            case None => throw UnknownVariable(minVar)
          }
        }
      }
    }

    /**
      * For a node(w,kid1,...kidp) the coFactor over v=vi could be :
      * -himself if the v < w (since v will not appear in kids)
      * -kidi if v = w (since the decision is done on this label)
      * -(w,kid1.coFactor(v=vi),...,kidp.coFactor(v=vi)) if v > w (since v could appear in kids we must rebuild a node
      * where the decision v=vi has been applied)
      *
      * @param variable the label to assign
      * @param value    the value to assign
      * @return the coFactor of this BDD
      */
    protected[GenMDDFactory] def coFactor(variable: Var, value: Val): MDD = {
      if (variable < param) {
        this
      } else if (variable > param) {
        MDDNode(kids.map(p => p._1 -> p._2.coFactor(variable, value)), param)
      } else {
        kids.getOrElse(value, TrashTerminal)
      }
    }

    override def toString: String = {
      domains.get(param) match {
        case Some(_) =>
          s"Node( $param, kids:${kids.map(p => s"${p._1}->${p._2}").mkString(",")}) "
        case None => throw UnknownVariable(param)
      }
    }
  }

  private object MDDNode extends FactoryNP[MDD, MDDNode, Val, Var, MDD, SortedMap] with ExtractorNP[MDDNode, Val, Var, MDD] {
    def make(kids: SortedMap[Val, MDD], param: Var): MDDNode = new MDDNode(param, kids)


    /**
      * Simplify a MDD, since two isomorphic nodes have the same hashCode by construction,
      * all kids pointing to isomorphic graphs, point on the same entry in the table
      * thus if all the kids point on the same entry the node is not necessary and
      * the unique son is returned
      *
      * @param kids  the kids of a node
      * @param param the variable labelled on the node
      * @return the node replacing the parent node if the simplification can be done
      */
    def reduce(kids: SortedMap[Val, MDD], param: Var): Option[MDD] = {
      if (kids.values.size == domains(param).size) {
        kids.values.find(kid1 => kids.values.forall(kid2 => kid1.eq(kid2)))
      } else if (kids.values.isEmpty) {
        Some(TrashTerminal)
      } else {
        None
      }
    }
  }

  /**
    * returns the MDD of a variable of the MDD factory domain where each edge label is mapped to
    *  terminals according to f function
    *
    * @param variable the variable to represents as a MDD
    * @param f the terminal mapping function
    * @return
    */
  def getVarMDD(variable: Var, f: (Val) => TL): Try[MDD] = {
    domains.get(variable) match {
      case Some(d: Domain) => Success(buildVarMDD(d, f))
      case None => Failure(UnknownVariable(variable))
    }
  }

  /**
    * check if the variable is defined in the MDDFactory
    *
    * @param variable the variable to represents as a MDD
    * @return
    */
  def varIsDefined(variable: Var): Boolean = domains.contains(variable)

  /**
    * returns the MDD of a variable of the MDD factory domain where each edge label is mapped to
    * to terminal given in map function
    *
    * @param variable the variable to represents as a MDD
    * @param terminalMap the terminal mapping to variable values
    * @return
    */
  def getVarMDD(variable: Var, terminalMap: Map[Val, TL]): Try[MDD] = {
    domains.get(variable) match {
      case Some(d: Domain) =>
        require(d.domain.forall(v => terminalMap.contains(v)),
          s"$terminalMap is not a total function over ${d.domain} ")
        Success(buildVarMDD(d, terminalMap))
      case None => Failure(UnknownVariable(variable))
    }
  }

  /**
    * return the MDD of the ith var of the factory
    *
    * @param index index of the variable
    * @return the mdd if the index is inside domaines range
    */
  def getIthVarMDD(index: Int, f: (Val) => TL): Try[MDD] = {
    if (index >= 0 && index <= domains.size) {
      Success(buildVarMDD(domains.values.toList(index), f))
    } else {
      Failure(OutOfVariableRange(domains.size,index))
    }
  }

  private def buildVarMDD(d: Domain, f: (Val) => TL): MDD = {
    MDDNode(d.foldLeft(SortedMap[Val, MDD]())((acc, value) => acc + (value -> ValuedTerminal(f(value)))), d.variable)
  }

  /**
    * return a MDD terminal if existing or build one (only one object per value)
    *
    * @param value the value of the terminal
    * @return
    */
  def getTerminal(value: TL): ValuedTerminal = {
    ValuedTerminal(value)
  }

  def getDomain(variable: Var): F[Val] = {
    val d = domains.getOrElse(variable, throw UnknownVariable(variable))
    d.domain
  }

  def getIniDomain(variable: Var): Option[Dom] = {
    for (d <- domains.get(variable); ini <- d.initialDom) yield ini
  }

  def addVar(variable: Var, domain: F[Val]): Unit = {
    domains = domains + (variable -> Domain(variable, domain))
  }

  def addVars(vars: Map[Var, F[Val]]): Unit = {
    vars.foreach(p => addVar(p._1, p._2))
  }

  def addVar(variable: Var, domain: Dom, extractor: (Dom) => F[Val]): Unit = {
    domains = domains + (variable -> Domain(variable, extractor(domain), Some(domain)))
  }

  def addVars(vars: Map[Var, Dom], extractor: (Dom) => F[Val]): Unit = {
    vars.foreach(p => addVar(p._1, p._2, extractor))
  }

  def reset(): Unit = domains = SortedMap.empty

  def buildNodeDefaultTrash(variable: Var, f: Map[Val, MDD]): MDD = {
    {
      for (d <- domains.get(variable))
        yield {
          f.foreach({
            case (value, _: Terminal) if !d.domain.exists(value == _) =>
              throw UnknownValue(variable,value)
            case (_, MDDNode(_, son)) if son <= variable =>
              throw BadOrdering(variable,son)
            case  (value, _:MDDNode) if !d.domain.exists(value == _) =>
              throw UnknownValue(variable,value)
            case _ =>
          })
          MDDNode(SortedMap(f.toList: _*), variable)
        }
    } getOrElse (throw UnknownVariable(variable))
  }
}

/**
  * Printer trait should be mixed with a factory
  *
  * @tparam Var the type representing node variables
  * @tparam F   the collection types of values associated to a variable
  * @tparam Val the type of variables possible values
  */
trait MDDPrinter[Var <: Ordered[Var], F[V] <: Traversable[V], Val <: Ordered[Val], TL <: Ordered[TL], Dom] {
  self: GenMDDFactory[Var, F, Val, TL, Dom] =>

  /**
    * prints the MDD to dot format
    *
    * @param toPrint  the MDD to print
    * @param fileName the output file name
    */
  def printMDD(toPrint: MDD, fileName: String): Unit = {
    val os = new FileOutputStream(fileName)
    val writer = new PrintStream(os)
    writer.println("digraph my_MDD {")
    printMDDRec((toPrint, writer)).result
    writer.println("}")
  }

  /**
    * private recursive and memoized function which prints the node and terminals iff they have not been printed
    * before for the current Printstream
    */
  private val printMDDRec: ((MDD, PrintStream)) => TailRec[Unit] = Memo.weakHashMapMemo {
    case (t: ValuedTerminal, writer: PrintStream) =>
      writer.println(s""" ${t.hashCode} [shape=box, label=\" ${t.param} \", style=filled, shape=box, height=0.3, width=0.3]""")
      done(())
    case (n: MDDNode, writer: PrintStream) =>
      writer.println(s"""${n.hashCode} [label=\" ${n.param}\"]""")
      n.kids.foreach { p =>
        tailcall(printMDDRec((p._2, writer))).result
        writer.println(s""" ${n.hashCode}->${p._2.hashCode()} [label=\" ${p._1} \"]""")
      }
      done(())
    case (TrashTerminal, _) => done(())
  }

  /**
    * prints the MDD to dot format and discards incoming edge to terminals in terminals set
    *
    * @param toPrint   the MDD to print
    * @param fileName  the output file name
    * @param terminals the terminals to discard
    */
  def printMDDIgnoreTerm(toPrint: MDD, fileName: String, terminals: Set[ValuedTerminal]): Unit = {
    val os = new FileOutputStream(fileName)
    val writer = new PrintStream(os)
    writer.println("digraph my_MDD {")
    printMDDRecIgnoreTerm((toPrint, writer, terminals)).result
    writer.println("}")
  }

  private val printMDDRecIgnoreTerm: ((MDD, PrintStream, Set[ValuedTerminal])) => TailRec[Unit] = Memo.weakHashMapMemo {
    case (t: ValuedTerminal, writer: PrintStream, terminals: Set[ValuedTerminal]) =>
      if (!terminals.contains(t)) {
        writer.println(s""" ${t.hashCode} [shape=box, label=\" ${t.param} \", style=filled, shape=box, height=0.3, width=0.3]""")
      }
      done(())
    case (n: MDDNode, writer: PrintStream, terminals: Set[ValuedTerminal]) =>
      writer.println(s"""${n.hashCode} [label=\" ${n.param}\"]""")
      n.kids.foreach { p =>
        tailcall(printMDDRecIgnoreTerm((p._2, writer, terminals))).result
        p._2 match {
          case kid: ValuedTerminal =>
            if (!terminals.contains(kid)) {
              writer.println(s""" ${n.hashCode}->${kid.hashCode} [label=\" ${p._1} \"]""")
            }
          case kid: MDDNode => writer.println(s""" ${n.hashCode}->${kid.hashCode} [label=\" ${p._1} \"]""")
          case TrashTerminal => done(())
        }
      }
      done(())
    case (TrashTerminal, _, _) => done(())
  }
}

sealed trait MDDException extends Exception

case class UnknownValue[Var, Val](variable:Var, value:Val) extends MDDException {
  override def toString: String = s" unknown value $value for variable $variable"
}

case class UnknownVariable[Var](variable: Var) extends MDDException {
   override def toString: String = s"unknown variable $variable"
}

case class OutOfVariableRange(domainsSize:Int, index:Int) extends MDDException {
  override def toString : String = s"variable range of MDD factory is $domainsSize, given index is $index"
}

case class BadOrdering[Var](father:Var, son:Var) extends MDDException {
  override def toString : String = s"$father >= $son but $son is used in a son of $father "
}



