package util.analyzers

import scala.collection.immutable.Set

/**
  * Case class encoding Set of Sets and ensuring that any element cannot be contained in another one
  *
  * @param holder initial set
  * @tparam T the element type, must be subtype of Set
  * @tparam E the type of Set elements
  */
case class InclusionSet[T <: Set[E], E](holder: Set[T] = Set[T]()) extends Set[T] {

  /**
    * Delegates contain method to holder
    *
    * @param elem element to find
    * @return if the element belong to [[holder]]
    */
  override def contains(elem: T): Boolean = holder.contains(elem)

  /**
    * Add an element iff this one is not a subset of the other one
    *
    * @param elem element to add
    * @return the resulting set
    */
  override def +(elem: T): InclusionSet[T, E] = {
    if (!holder.exists(e => e.subsetOf(elem)))
      InclusionSet(holder.filterNot(e => elem.subsetOf(e)) + elem)
    else
      this
  }

  /**
    * Delegate remove to [[holder]]
    *
    * @param elem element to remove
    * @return resulting set
    */
  override def -(elem: T): InclusionSet[T, E] = {
    InclusionSet(holder - elem)
  }

  /**
    * Delegate iterator to [[holder]]
    *
    * @return
    */
  override def iterator: Iterator[T] = {
    holder.iterator
  }
}
