package util.analyzers

/**
  * Created by kdelmas on 26/02/16.
  */

/**
  * Base trait for representing infinite values
  */
sealed trait Infinity

/**
  * Object representing positive infinity
  */
case object POSITIVE_INFINITY extends Infinity {
  override def toString: String = {
    "+inf"
  }
}

/**
  * Object representing negative infinity
  */
case object NEGATIVE_INFINITY extends Infinity {
  override def toString: String = {
    "-inf"
  }
}

sealed trait Sign

case object POSITIVE extends Sign

case object NEGATIVE extends Sign

case object ZERO extends Sign

/**
  * Useful conversion from infinity to infinite numeric
  */
object InfNumericImplicits {
  implicit def Inf2InfNumeric[T](v: Infinity)(implicit num: Numeric[T]): InfNumeric[T] = InfNumeric(Right(v))(num)

  implicit def Int2InfNumeric[T](v: T)(implicit num: Numeric[T]): InfNumeric[T] = InfNumeric(Left(v))(num)

  implicit object InfIntIsNumeric extends InfNumericIsNumeric[Int] {
    val num: Numeric[Int] = implicitly
  }

}

/**
  * Particular type of infinite numeric
  */
object InfNumericTypes {
  type InfInt = InfNumeric[Int]
  type InfFloat = InfNumeric[Float]
  type InfDouble = InfNumeric[Double]
  type InfLong = InfNumeric[Long]
  type InfReal = InfNumeric[util.analyzers.analyzerTypes.Real]
}

/**
  * Case class representing infinite numeric (classic number plus infinity)
  *
  * @param value either a number or an infinity
  * @param num   an implicit numeric object
  * @tparam T the type of the values
  */
case class InfNumeric[T] private(value: Either[T, Infinity])(implicit num: Numeric[T]) extends Ordered[InfNumeric[T]] {


  /**
    * @return the sign of this
    */
  def getSign: Sign = {
    value match {
      case Right(POSITIVE_INFINITY) => POSITIVE
      case Right(NEGATIVE_INFINITY) => NEGATIVE
      case Left(x) if num.zero == x => ZERO
      case Left(x) if num.compare(x, num.zero) >= 1 => POSITIVE
      case _ => NEGATIVE
    }
  }

  /**
    * Addition with an other number
    *
    * @param that other number
    * @return the addition
    */
  def +(that: InfNumeric[T]): InfNumeric[T] = {
    value match {
      case Right(POSITIVE_INFINITY) => that.value match {
        case Right(POSITIVE_INFINITY) => InfNumeric(Right(POSITIVE_INFINITY))
        case Right(NEGATIVE_INFINITY) => throw UndeterminedForm(s"$POSITIVE_INFINITY + $NEGATIVE_INFINITY")
        case Left(_) => InfNumeric(Right(POSITIVE_INFINITY))
      }
      case Right(NEGATIVE_INFINITY) => that.value match {
        case Right(POSITIVE_INFINITY) => throw UndeterminedForm(s"$POSITIVE_INFINITY + $NEGATIVE_INFINITY")
        case Right(NEGATIVE_INFINITY) => InfNumeric(Right(NEGATIVE_INFINITY))
        case Left(_) => InfNumeric(Right(NEGATIVE_INFINITY))
      }
      case Left(x) => that.value match {
        case Right(y) => InfNumeric(Right(y))
        case Left(y) => InfNumeric(Left(num.plus(x, y)))
      }
    }
  }

  /**
    * Subtraction with an other number
    *
    * @param that other number
    * @return the addition
    */
  def -(that: InfNumeric[T]): InfNumeric[T] = {
    value match {
      case Right(POSITIVE_INFINITY) =>
        that.value match {
          case Right(POSITIVE_INFINITY) => throw UndeterminedForm(s"$POSITIVE_INFINITY - $POSITIVE_INFINITY")
          case Right(NEGATIVE_INFINITY) => InfNumeric(Right(POSITIVE_INFINITY))
          case Left(_) => InfNumeric(Right(POSITIVE_INFINITY))
        }
      case Right(NEGATIVE_INFINITY) =>
        that.value match {
          case Right(POSITIVE_INFINITY) => InfNumeric(Right(NEGATIVE_INFINITY))
          case Right(NEGATIVE_INFINITY) => throw UndeterminedForm(s"$NEGATIVE_INFINITY - $NEGATIVE_INFINITY")
          case Left(_) => InfNumeric(Right(NEGATIVE_INFINITY))
        }
      case Left(x) =>
        that.value match {
          case Right(POSITIVE_INFINITY) => InfNumeric(Right(NEGATIVE_INFINITY))
          case Right(NEGATIVE_INFINITY) => InfNumeric(Right(POSITIVE_INFINITY))
          case Left(y) => InfNumeric(Left(num.minus(x, y)))
        }
    }
  }

  /**
    * Multiplication with an other number
    *
    * @param that other number
    * @return the addition
    */
  def *(that: InfNumeric[T]): InfNumeric[T] = {
    value match {
      case Right(_) if getSign == that.getSign => this
      case Right(_) if that.getSign != ZERO => InfNumeric(Right(NEGATIVE_INFINITY))
      case Right(_) => throw UndeterminedForm(s"$value * ${that.value}")
      case Left(x) => that.value match {
        case Right(_) if getSign == that.getSign => that
        case Right(_) if that.getSign != ZERO => InfNumeric(Right(NEGATIVE_INFINITY))
        case Right(_) => throw UndeterminedForm(s"$value * ${that.value}")
        case Left(y) => InfNumeric(Left(num.times(x, y)))
      }
    }
  }

  /**
    * @return negation of this
    */
  def negate(): InfNumeric[T] = {
    value match {
      case Right(POSITIVE_INFINITY) => InfNumeric(Right(NEGATIVE_INFINITY))
      case Right(NEGATIVE_INFINITY) => InfNumeric(Right(POSITIVE_INFINITY))
      case Left(x) => InfNumeric(Left(num.negate(x)))
    }
  }

  /**
    * Compare two numbers
    * @param that the other number
    * @return 1 if this > that
    *         0 if this = that
    *         -1 if this < that
    */
  def compare(that: InfNumeric[T]): Int = {
    value match {
      case Right(NEGATIVE_INFINITY) => that.value match {
        case Right(NEGATIVE_INFINITY) => 0
        case Right(POSITIVE_INFINITY) => -1
        case Left(_) => -1
      }
      case Right(POSITIVE_INFINITY) => that.value match {
        case Right(NEGATIVE_INFINITY) => 1
        case Right(POSITIVE_INFINITY) => 0
        case Left(_) => 1
      }
      case Left(x) => that.value match {
        case Right(NEGATIVE_INFINITY) => 1
        case Right(POSITIVE_INFINITY) => -1
        case Left(y) => num.compare(x, y)
      }
    }
  }

  override def toString: String = {
    value match {
      case Left(x) => x.toString
      case Right(x) => x.toString
    }
  }
}

/**
  * Base trait used to defined Numeric implicit objects
  * @tparam T the type of the values
  */
trait InfNumericIsNumeric[T] extends Numeric[InfNumeric[T]] {
  val num: Numeric[T]

  def negate(in: InfNumeric[T]): InfNumeric[T] = {
    in.value match {
      case Left(x) => InfNumeric[T](Left(num.negate(x)))(num)
      case Right(NEGATIVE_INFINITY) => InfNumeric[T](Right(POSITIVE_INFINITY))(num)
      case Right(POSITIVE_INFINITY) => InfNumeric[T](Right(NEGATIVE_INFINITY))(num)
    }
  }

  def fromInt(x: Int): InfNumeric[T] =
    InfNumeric[T](Left(num.fromInt(x)))(num)

  def toInt(x: InfNumeric[T]): Int = x.value match {
    case Left(a) => num.toInt(a)
    case Right(POSITIVE_INFINITY) => Int.MaxValue
    case Right(NEGATIVE_INFINITY) => Int.MinValue
  }

  def toLong(x: InfNumeric[T]): Long = x.value match {
    case Left(a) => num.toInt(a)
    case Right(POSITIVE_INFINITY) => Long.MaxValue
    case Right(NEGATIVE_INFINITY) => Long.MinValue
  }

  def toFloat(x: InfNumeric[T]): Float = x.value match {
    case Left(a) => num.toInt(a)
    case Right(POSITIVE_INFINITY) => Float.MaxValue
    case Right(NEGATIVE_INFINITY) => Float.MinValue
  }

  def toDouble(x: InfNumeric[T]): Double = x.value match {
    case Left(a) => num.toInt(a)
    case Right(POSITIVE_INFINITY) => Double.MaxValue
    case Right(NEGATIVE_INFINITY) => Double.MinValue
  }

  def plus(x: InfNumeric[T], y: InfNumeric[T]): InfNumeric[T] =
    x + y

  def minus(x: InfNumeric[T], y: InfNumeric[T]): InfNumeric[T] =
    x - y

  def times(x: InfNumeric[T], y: InfNumeric[T]): InfNumeric[T] =
    x * y

  def compare(x: InfNumeric[T], y: InfNumeric[T]): Int =
    x.compare(y)
}


case class UndeterminedForm(computation: String) extends Exception {
  override def toString: String = s"$computation is an undetermined form"
}



