/** $Id$
  *
  * @author rdelmas
  *         last edited by $Author$
  **/

package util.analyzers

import java.io._

import ast.dsl.InstBoolIdent
import com.microsoft.z3.enumerations.Z3_ast_kind._
import com.microsoft.z3.enumerations.Z3_decl_kind._
import com.microsoft.z3.{AST => Z3AST, BitVecNum => Z3BitVecNum, BoolExpr => Z3BoolExpr, DatatypeSort => Z3DataTypeSort, Expr => Z3Expr}
import util.io.{FileManager, SystemInfo}
import util.sat.{FormulaFileWriter, FormulaLike}

import scala.collection.{immutable, mutable}
import scala.io.Source

/**
  * Translates the structure function of a system represented by a z3 expression to a SAT-solver friendly formula in DIMACS or OPB format for prime implicant enumeration using external solvers.
  **/
class IterativeMCSEnumerator(phi: Phi) {

  private val z3ToDslCache: immutable.Map[Z3BoolExpr, InstBoolIdent] = phi.events.map(_.swap)
  private val dslToDimacsCache: mutable.Map[InstBoolIdent, Int] = new mutable.OpenHashMap[InstBoolIdent, Int]()
  private val z3ToDimacsCache: mutable.Map[Z3BoolExpr, Int] = new mutable.OpenHashMap[Z3BoolExpr, Int]()
  private val cnfFormula: FormulaLike = new FormulaLike {}

  def generateEnumerateScript():File={
    val result=FileManager.addFileToTemp("enumerateUntil.sh")
    val writer= new FileWriter(result)

    writer.write(
      s"""#!/bin/bash
         |
         |num=$$1
         |modelFile=$$2
         |myOutput='temp.opb'
         |realVars="$$(sed '/^[^*]/d; s/* //; /^[^0-9]/d; s/ -> .*$$//' $$modelFile)"
         |nbRealVars=$$(echo "$$realVars" | wc -l)
         |mySub=$$( sed 's#\\(.*\\)#s/x\\1 /a\\1 /g#' <<< "$$realVars" |tr '\\n' ';')
         |myConfig='--enum-mode=auto --models 0 --quiet=0 -t 4,split'
         |for ((i=1; i<=num;i++))
         |do
         |	cat $$modelFile "$$(sed 's/\\.opb//g' <<< "$$modelFile")_activate_card_$$i.opb"> $$myOutput
         |	echo "$$blockingClauses" >> $$myOutput
         |	blockingClauses=$$(printf '%s\\n%s\\n' "$$blockingClauses" "$$(${ClaspInterface.toolPath} $$myConfig $$myOutput | awk ' /^v / {inside=1;} /^c Answer: [0-9]*$$/   {inside=0; print "\\n"; next} inside {printf sep""$$0; sep=" "; next} {printf ""}'| sed '/^$$/d' |sed 's/ s .*// ' |sed 's/v//g; s/ x/ +1 ~x/g; s/ -x[0-9]*//g; s/$$/ >=1;/;s/^ *//g')")
         |	blockingClauses="$$(sed "$$mySub" <<< "$$blockingClauses")"
         |	blockingClauses="$$(sed 's/+1 x[0-9]* //g' <<< "$$blockingClauses" |sed 's/+1 ~x[0-9]* //g' | sed 's/a/x/g;s/^ *//g;s/  */ /g'  )"
         |
         |done
         |rm -f $$myOutput
         |sed 's/ x/ -/g;s/~x//g;s/+1 //g;s/>=1;//g; s/^ *//g;/^\\s*$$/d;s/  */ /g' <<< "$$blockingClauses"
       """.stripMargin)
    result.setExecutable(true)
    writer.close()
    result
  }

  /** Generates DIMACS and OPB files encoding the prime implicant enumeration problem to solve them using external solvers. */
  def generateFiles(maxOrder: Int, file: File): Unit = {

    this.toCnf(phi)


    // generate model for one-shot enumeration with renaming approach form the JELIA paper.
    val targetFormula: FormulaLike = new FormulaLike {}

    // dump rewritten formula to an OPB file
    FormulaFileWriter.Opb(targetFormula, file.getAbsolutePath + "_translation")

    // add backtranslation map as header to the formula
    val freeLits = dslToDimacsCache.values.toList
    cnfFormula.addHeaderComment("translation map :")
    for ((k, v) <- dslToDimacsCache) {
      cnfFormula.addHeaderComment(s"${v.toString} -> ${k.toString()}")
    }

    // dump original formula without cardinality constraints to a cnf file
    FormulaFileWriter.Cnf(cnfFormula, file.getAbsolutePath + "_original")


    // add switchable cardinality constraints if a maxOrder was specified
    var _switches: List[Int] = Nil
    for (i <- 1 to maxOrder) {
      val switch = cnfFormula.addAtMostK(freeLits, i)(switchable = true).get
      cnfFormula.addHeaderComment(s"switch for card $i : $switch")
      _switches ::= switch
    }

    // generate files containing unit clauses switching cardinality constraints on for each cut order from 1 to maxOrd.
    val switchList = _switches.reverse
    for (i <- 1 to maxOrder) {
      util.io.WriteToFile(s"${file.getAbsolutePath}_iterative_activate_card_$i.opb") {
        out => {
          switchList.zipWithIndex.foreach {
            pair => {
              val (switch, j) = pair
              out.write(s"+1 ${if (i != (j + 1)) "~" else ""}x$switch >= 1;\n")
            }
          }
        }
      }
    }
    // dump formula to an OPB file
    FormulaFileWriter.Opb(cnfFormula, file.getAbsolutePath + "_iterative")
  }

  def apply(maxOrder: Int, file: File): Iterable[Set[InstBoolIdent]] = {
    generateFiles(maxOrder, file)
    val resFile = FileManager.addFileToTemp("res.txt")
    val scriptFile=generateEnumerateScript()
    val cmd = Seq(scriptFile.getAbsolutePath, maxOrder.toString, s"${file.getAbsolutePath}_iterative.opb")
    ClaspInterface.startProcess(SystemInfo.runUnixCmd(cmd.mkString(" ")) #> resFile).exitValue()
    parse(resFile)
  }

  def run(maxOrder: Int, file: File): File = {
    val resFile = FileManager.addFileToTemp("res.txt")
    val scriptFile=generateEnumerateScript()
    val cmd = Seq(scriptFile.getAbsolutePath, maxOrder.toString, s"${file.getAbsolutePath}_iterative.opb")
    ClaspInterface.startProcess(SystemInfo.runUnixCmd(cmd.mkString(" ")) #> resFile).exitValue()
    resFile
  }

  def parse(resFile: File): Iterable[Set[InstBoolIdent]] = {
    val dimacsToDsl = dslToDimacsCache.map(_.swap)
    Source.fromFile(resFile).getLines().map(s => s.split(" ").collect({ case s: String if s.length > 0 => dimacsToDsl(s.toInt) }).toSet).toList
  }

  /** Translates the structure function to CNF. */
  private def toCnf(phi: Phi): Unit = {

    // translate free variables first to map them to low dimacs indices
    this.z3ToDslCache.keys foreach { v => this.toCnf(v) }

    // translate the expression of Phi to CNF and add resulting literal as as unit clause
    this.cnfFormula.addClause(this.toCnf(phi.expr))()

  }

  /** Recursively translates a Z3Expr to a CNF formula (memoization of [[toCnfRec]]). */
  private def toCnf(node: Z3Expr): Int = node match {
    case node: Z3BoolExpr =>
      z3ToDimacsCache.getOrElseUpdate(node, toCnfRec(node))
    case _ =>
      throw new Exception(s"formula ${node.toString} is not purely boolean")
  }

  /** Recursively translates the structure function to a CNF formula (does memoization of [[toCnfRec]]). */
  private
  def toCnfRec(node: Z3BoolExpr): Int = node.getASTKind match {

    // translate as free variable, update translation cache
    case Z3_VAR_AST =>
      val fresh = this.cnfFormula.freshVar
      val dsl = this.z3ToDslCache(node)
      this.dslToDimacsCache.update(dsl, fresh)
      fresh

    // functional definition translate using gate variables
    case Z3_APP_AST =>
      val args = node.getArgs

      node.getFuncDecl.getDeclKind match {
        case Z3_OP_NOT =>
          assert(args.length == 1)
          -toCnf(args(0))

        case Z3_OP_AND =>
          val newArgs = args.toList.map(toCnf)
          val fresh = this.cnfFormula.freshVar
          this.cnfFormula.defineAnd(fresh, newArgs)()
          fresh

        case Z3_OP_OR =>
          val newArgs = args.toList.map(toCnf)
          val fresh = this.cnfFormula.freshVar
          this.cnfFormula.defineOr(fresh, newArgs)()
          fresh

        case Z3_OP_IMPLIES =>
          assert(args.length == 2)
          val newArg0 = toCnf(args(0))
          val newArg1 = toCnf(args(1))
          val fresh = this.cnfFormula.freshVar
          this.cnfFormula.defineImplies(fresh, newArg0, newArg1)()
          fresh

        case Z3_OP_IFF =>
          assert(args.length == 2)
          val newArg0 = toCnf(args(0))
          val newArg1 = toCnf(args(1))
          val fresh = this.cnfFormula.freshVar
          this.cnfFormula.defineIff(fresh, newArg0, newArg1)()
          fresh

        case Z3_OP_XOR =>
          assert(args.length == 2)
          val newArg0 = toCnf(args(0))
          val newArg1 = toCnf(args(1))
          val fresh = this.cnfFormula.freshVar
          this.cnfFormula.defineIff(fresh, newArg0, -newArg1)()
          fresh

        case Z3_OP_ITE =>
          assert(args.length == 3)
          val newArg0 = toCnf(args(0))
          val newArg1 = toCnf(args(1))
          val newArg2 = toCnf(args(2))
          val fresh = this.cnfFormula.freshVar
          this.cnfFormula.defineIte(fresh, newArg0, newArg1, newArg2)()
          fresh

        case otherKind => throw new Exception(s"in ${node.toString}: unexpected operator Z3_decl_kind = ${otherKind.toString}")
      }
    case otherKind => throw new Exception(s"in ${node.toString}: unexpected AST kind Z3_ast_kind = ${otherKind.toString}")
  }
}
