package util.analyzers

import util.analyzers.analyzerTypes._
import util.analyzers.InfNumericTypes._

/**
  * Created by kdelmas on 04/11/16.
  */

object MeasureTypes {
  type ProbMeasure = Measure[Real] with ProbLikeBound[Real]
  type CardMeasure = Measure[InfInt]
}


/**
  * Interval approximation of numeric values
  *
  * @param up    upper bound of the value
  * @param low   lower bound of the value
  * @param exact exact value if exists
  * @param num   the value must be numeric
  * @tparam T the type of the value
  */
class Measure[T] private(val up: T, val low: T, val exact: Option[T])(implicit val num: Numeric[T]) {

  trait ComparatorOp

  object GT extends ComparatorOp

  object LT extends ComparatorOp

  object EQ extends ComparatorOp

  object GE extends ComparatorOp

  object LE extends ComparatorOp

  override lazy val hashCode: Int = FactoryP.sHashCode(up) + 41 * (FactoryP.sHashCode(low) + 41 * FactoryP.sHashCode(exact))

  /**
    * Equality is structural equality
    *
    * @param obj other object
    * @return is equals or not
    */
  override def equals(obj: scala.Any): Boolean = {
    obj match {
      case a: Measure[T] => a.up == up && a.low == low && a.exact == exact
      case _ => false
    }

  }

  /**
    * limit a value of the measure
    *
    * @param value  the initial value
    * @param newUp  the new upper bound
    * @param newLow the new lower bound
    * @return the limited value
    */
  def lim(value: T, newUp: T, newLow: T): T = {
    if (num.gt(value, newUp))
      newUp
    else if (num.lt(value, newLow))
      newLow
    else
      value
  }

  /**
    * Transform a measure to a bounded measure
    *
    * @param newUp  the maximal upper bound
    * @param newLow the minimal lower bound
    * @return bounded measure
    */
  def toBounded(newUp: T, newLow: T): Measure[T] with Bounded[T] = {
    new Measure[T](lim(up, newUp, newLow), lim(low, newUp, newLow), for (e <- exact) yield lim(e, newUp, newLow)) with Bounded[T] {
      val upBound: T = newUp
      val lowBound: T = newLow
    }
  }

  /**
    * Bounded as probability (ie [0, 1])
    *
    * @return a bounded measure
    */
  def toProbability: Measure[T] with ProbLikeBound[T] = {
    val one = num.fromInt(1)
    val zero = num.zero
    val newOverApprox = lim(up, one, zero)
    val newUnderApprox = lim(low, one, zero)
    val newExact = for (e <- exact) yield lim(e, one, zero)
    if (newOverApprox == newUnderApprox)
      new Measure[T](newOverApprox, newUnderApprox, Some(newOverApprox)) with ProbLikeBound[T]
    else
      new Measure[T](newOverApprox, newUnderApprox, newExact) with ProbLikeBound[T]
  }

  /**
    * Addition of measures
    *
    * @param that another measure
    * @return resulting measure
    */
  def +(that: Measure[T]): Measure[T] = {
    {
      for (e1 <- exact; e2 <- that.exact) yield {
        Measure[T](num.plus(e1, e2))
      }
    }.getOrElse(
      new Measure[T](num.plus(up, that.up),
        num.plus(low, that.low),
        None)
    )
  }

  /**
    * Substraction of measures
    *
    * @param that another measure
    * @return resulting measure
    */
  def -(that: Measure[T]): Measure[T] = {
    {
      for (e1 <- exact; e2 <- that.exact) yield {
        Measure[T](num.minus(e1, e2))
      }
    }.getOrElse(
      Measure[T](num.minus(up, that.low),
        num.minus(low, that.up),
        None)
    )
  }

  /**
    * Multiplication of measures
    *
    * @param that another measure
    * @return resulting measure
    */
  def *(that: Measure[T]): Measure[T] = {
    {
      for (e1 <- exact; e2 <- that.exact) yield {
        Measure[T](num.times(e1, e2))
      }
    }.getOrElse(
      Measure[T](
        List(
          num.times(up, that.up),
          num.times(up, that.low),
          num.times(low, that.low),
          num.times(low, that.up)
        ).max,
        List(
          num.times(up, that.up),
          num.times(up, that.low),
          num.times(low, that.low),
          num.times(low, that.up)
        ).min,
        None)
    )
  }

  /**
    * Minimum of measures
    *
    * @param that another measure
    * @return resulting measure
    */
  def min(that: Measure[T]): Measure[T] = {
    Measure(
      num.min(up, that.up),
      num.min(low, that.low),
      for (e1 <- exact; e2 <- that.exact) yield num.min(e1, e2)
    )
  }

  /**
    * Addition of this measure to a value
    *
    * @param that a value
    * @return resulting measure
    */
  def +(that: T): Measure[T] = {
    Measure[T](num.plus(up, that),
      num.plus(low, that),
      for (e1 <- exact) yield num.plus(e1, that))
  }

  /**
    * Subtraction of this measure to a value
    *
    * @param that a value
    * @return resulting measure
    */
  def -(that: T): Measure[T] = {
    Measure[T](num.minus(up, that),
      num.minus(low, that),
      for (e1 <- exact) yield num.minus(e1, that))
  }

  /**
    * Multiplication of this measure to a value
    *
    * @param that a value
    * @return resulting measure
    */
  def *(that: T): Measure[T] = {
    Measure[T](num.times(up, that),
      num.times(low, that),
      for (e1 <- exact) yield num.times(e1, that))
  }

  /**
    * Comparison rules between to measure (ie interval)
    *
    * @param that other measure
    * @return comparison object if comparison possible, none otherwise
    */
  private def tryCompareTo[B >: Measure[T]](that: Measure[T]): Option[ComparatorOp] = exact match {

    case Some(a) => that.exact match {
      case Some(b) => num.compare(a, b) match {
        case 1 => Some(GT)
        case 0 => Some(EQ)
        case -1 => Some(LT)
      }
      case None =>
        if (num.gt(a, that.up))
          Some(GT)
        else if (num.gteq(a, that.up))
          Some(GE)
        else if (num.lt(a, that.low))
          Some(LT)
        else if (num.lteq(a, that.low))
          Some(LE)
        else
          None
    }
    case None => that.exact match {
      case Some(b) =>
        if (num.gt(low, b))
          Some(GT)
        else if (num.gteq(low, b))
          Some(GE)
        else if (num.lt(up, b))
          Some(LT)
        else if (num.lteq(up, b))
          Some(LE)
        else None
      case None =>
        if (num.gt(low, that.up))
          Some(GT)
        else if (num.gteq(low, that.up))
          Some(GE)
        else if (num.lt(up, that.low))
          Some(LT)
        else if (num.lteq(up, that.low))
          Some(LE)
        else None
    }
  }

  def <(that: Measure[T]): Boolean = {
    for (r <- tryCompareTo(that)) yield r == LT
  }.getOrElse(false)

  def <=(that: Measure[T]): Boolean = {
    for (r <- tryCompareTo(that)) yield r == LT || r == EQ || r == LE
  }.getOrElse(false)

  def >(that: Measure[T]): Boolean = {
    for (r <- tryCompareTo(that)) yield r == GT
  }.getOrElse(false)

  def >=(that: Measure[T]): Boolean = {
    for (r <- tryCompareTo(that)) yield r == GT || r == EQ || r == GE
  }.getOrElse(false)

  def <(that: T): Boolean = {
    for (r <- tryCompareTo(Measure(that, that, Some(that)))) yield r == LT
  }.getOrElse(false)

  def <=(that: T): Boolean = {
    for (r <- tryCompareTo(Measure(that, that, Some(that)))) yield r == LT || r == EQ || r == LE
  }.getOrElse(false)

  def >(that: T): Boolean = {
    for (r <- tryCompareTo(Measure(that, that, Some(that)))) yield r == GT
  }.getOrElse(false)

  def >=(that: T): Boolean = {
    for (r <- tryCompareTo(Measure(that, that, Some(that)))) yield r == GT || r == EQ || r == GE
  }.getOrElse(false)

  override def toString: String = {
    "( max:" + num.toDouble(up) + ", min:" + num.toDouble(low) + ", real:" + {
      for (e <- exact) yield num.toDouble(e).toString
    }.getOrElse(exact) + ")"
  }
}

trait Bounded[T] {
  self: Measure[T] =>
  val upBound: T
  val lowBound: T
}

trait ProbLikeBound[T] extends Bounded[T] {
  self: Measure[T] =>
  val upBound: T = num.fromInt(1)
  val lowBound: T = num.zero

  override def +(that: Measure[T]): Measure[T] with ProbLikeBound[T] = {
    {
      for (e1 <- exact; e2 <- that.exact) yield {
        Measure[T](num.plus(e1, e2))
      }
    }.getOrElse(
      Measure[T](num.plus(up, that.up),
        num.plus(low, that.low),
        None)
    ).toProbability
  }

  override def -(that: Measure[T]): Measure[T] with ProbLikeBound[T] = {
    {
      for (e1 <- exact; e2 <- that.exact) yield {
        Measure[T](num.minus(e1, e2))
      }
    }.getOrElse(
      Measure[T](num.minus(up, that.low),
        num.minus(low, that.up),
        None)
    ).toProbability
  }

  override def *(that: Measure[T]): Measure[T] with ProbLikeBound[T] = {
    {
      for (e1 <- exact; e2 <- that.exact) yield {
        Measure[T](num.times(e1, e2))
      }
    }.getOrElse(
      if (num.gteq(low, num.zero) && num.gteq(that.low, num.zero)) {
        Measure[T](num.times(up, that.up),
          num.times(low, that.low),
          None)
      } else {
        that * this //otherwise classic interval mutiplication
      }
    ).toProbability
  }

  override def +(that: T): Measure[T] with ProbLikeBound[T] = {
    Measure[T](num.plus(up, that),
      num.plus(low, that),
      for (e1 <- exact) yield num.plus(e1, that)).toProbability
  }

  override def -(that: T): Measure[T] with ProbLikeBound[T] = {
    Measure[T](num.minus(up, that),
      num.minus(low, that),
      for (e1 <- exact) yield num.minus(e1, that)).toProbability
  }

  override def *(that: T): Measure[T] with ProbLikeBound[T] = {
    Measure[T](num.times(up, that),
      num.times(low, that),
      for (e1 <- exact) yield num.times(e1, that)).toProbability
  }
}

object ProbaMeasure {

  val zero = ProbaMeasure(Real(0.0))

  val one = ProbaMeasure(Real(1.0))

  def apply(overApprox: Real, underApprox: Real, exact: Option[Real])(implicit num: Numeric[Real]): MeasureTypes.ProbMeasure = {
    if (overApprox == underApprox)
      Measure(overApprox, overApprox, Some(overApprox)).toProbability
    else
      Measure(overApprox, underApprox, exact).toProbability
  }

  def apply(value: Real)(implicit num: Numeric[Real]): MeasureTypes.ProbMeasure = {
    Measure(value).toProbability
  }
}

object CardMeasure {

  import InfNumericImplicits._
  import InfNumericTypes._

  val zero = CardMeasure(0)

  val infinity = CardMeasure(POSITIVE_INFINITY)

  def apply(overApprox: InfInt, underApprox: InfInt, exact: Option[InfInt])(implicit num: Numeric[InfInt]): MeasureTypes.CardMeasure = {
    if (overApprox == underApprox)
      Measure(overApprox, overApprox, Some(overApprox))
    else
      Measure(overApprox, underApprox, exact)
  }

  def apply(value: InfInt)(implicit num: Numeric[InfInt]): MeasureTypes.CardMeasure = {
    Measure(value)
  }
}

object Measure {
  def apply[T](overApprox: T, underApprox: T, exact: Option[T])(implicit num: Numeric[T]): Measure[T] = {
    if (overApprox == underApprox)
      new Measure(overApprox, overApprox, Some(overApprox))
    else
      new Measure(overApprox, underApprox, exact)
  }

  def apply[T](value: T)(implicit num: Numeric[T]): Measure[T] = new Measure(value, value, Some(value))
}
