package util.analyzers

import java.io.File
import ast.dsl.InstBoolIdent
import com.microsoft.z3.enumerations.Z3_ast_kind.{Z3_APP_AST, Z3_VAR_AST}
import com.microsoft.z3.enumerations.Z3_decl_kind._
import util.io.FileManager
import util.sat.{FormulaFileWriter, FormulaLike}
import com.microsoft.z3.{Sort => Z3Sort,
  Symbol => Z3Symbol, FuncDecl => Z3FuncDecl,
  Solver => Z3Solver, BoolExpr => Z3BoolExpr,
  Expr => Z3Expr,
  BitVecExpr => Z3BitVecExpr,
  BitVecNum}
import scala.collection.immutable.Set
import scala.sys.process._
import scala.collection.{immutable, mutable}

/**
  * Translates the structure function of a system represented by a z3 expression to a SAT-solver
  * friendly formula in DIMACS or OPB format for prime implicant enumeration using external solvers.
  *
  * @param phi the structure function
  */
class PBMCSEnumerator(phi: Phi) {

  private val z3ToDslCache: immutable.Map[Z3BoolExpr, InstBoolIdent] = phi.events.map(_.swap)
  private val dslToDimacsCache: mutable.Map[InstBoolIdent, Int] = new mutable.OpenHashMap[InstBoolIdent, Int]()
  private val z3ToDimacsCache: mutable.Map[Z3BoolExpr, Int] = new mutable.OpenHashMap[Z3BoolExpr, Int]()
  private val cnfFormula: FormulaLike = new FormulaLike {}

  /** Generates DIMACS and OPB files encoding the prime implicant enumeration problem to solve them using external solvers. */
  def generateFiles(maxOrder: Option[Int], file: File): Unit = {

    this.toCnf(phi)


    // optionally dump resulting instances to file

    // add backtranslation map as header to the formula
    val freeLits = dslToDimacsCache.values.toList
    cnfFormula.addHeaderComment("translation map :")
    for ((k, v) <- dslToDimacsCache) {
      cnfFormula.addHeaderComment(s"${v.toString} -> ${k.toString}")
    }

    maxOrder match {
      case Some(m) =>
        cnfFormula.addAtMostK(freeLits, m)()
        FormulaFileWriter.Opb(cnfFormula, file.getAbsolutePath)
        val cmd :ProcessBuilder = s"${PBEncoderInterface.toolPath} ${file.getAbsolutePath}.opb " #> new File(s"${file.getAbsolutePath}.cnf")
        PBEncoderInterface.startProcess(cmd).exitValue()
      case None =>
        FormulaFileWriter.Cnf(cnfFormula, file.getAbsolutePath)
    }
  }

  private def runSharpCDCL(projScopeFile: File, cnfFile: File, resultFile: File): Unit = {
    // invoke external solver
    import scala.sys.process._
    util.io.WriteToFile(projScopeFile.getAbsolutePath, reThrow = true)(out => {
      dslToDimacsCache.values.foreach(lit => out.write(s"${lit.toString}\n"))
    })
    val command = Seq(
      SharpCDCLInterface.toolPath.toAbsolute.toString(),
      "-countMode=2",
      s"-projection=${projScopeFile.getAbsolutePath}",
      s"-projectToDNF=${resultFile.getAbsolutePath}",
      cnfFile.getAbsolutePath
    )
    SharpCDCLInterface.startProcess(Process(command)).exitValue()
  }

  /**
    * Parses a sharpCDCL model file, possibly containing several models of a same formula.
    * TODO implement real parsing, string splitting + scanning does not scale
    */
  def parseSharpCDCLModelFile(resultFile: File): Iterable[Set[InstBoolIdent]] = {
    // parse result file line per line to return prime implicants
    // warning some model could be non minimal but all MCS are in the set thus we must
    // remove all sut which is a subset of another one
    if (resultFile.exists()) {
      val dimacsToDsl = dslToDimacsCache.map(_.swap)
      val zero = InclusionSet[Set[InstBoolIdent], InstBoolIdent]()
      scala.io.Source.fromFile(resultFile).getLines().foldLeft(zero)((acc, line) => {
        acc + Set(line.split(" ").map(_.toInt).filter(dimacsToDsl.contains).map(dimacsToDsl): _*)
      })
    } else {
      Nil
    }
  }

  def apply(maxOrder: Option[Int], file: File): Iterable[Set[InstBoolIdent]] = {
    generateFiles(maxOrder, file)
    val programFile = FileManager.addFileToTemp("prog.cnf")
    val resFile = FileManager.addFileToTemp("result.txt")
    val modelFile = FileManager.addFileToTemp(s"${file.getName}.cnf")
    runSharpCDCL(programFile, modelFile, resFile)
    parseSharpCDCLModelFile(resFile)
  }

  def run(file: File): File = {
    val progFile = FileManager.addFileToTemp("prog.cnf")
    val resFile = FileManager.addFileToTemp("result.txt")
    val modelFile = FileManager.addFileToTemp(s"${file.getName}.cnf")
    runSharpCDCL(progFile, modelFile, resFile)
    resFile
  }

  def parse(r: File): Iterable[Set[InstBoolIdent]] = parseSharpCDCLModelFile(r)

  /** Translates the structure function to CNF. */
  private def toCnf(phi: Phi): Unit = {

    // translate free variables first to map them to low dimacs indices
    this.z3ToDslCache.keys foreach { v => this.toCnf(v, this.cnfFormula) }

    // translate the expression of Phi to CNF and add resulting literal as as unit clause
    this.cnfFormula.addClause(this.toCnf(phi.expr, this.cnfFormula))()


    addWitnesses(phi, this.cnfFormula)

  }

  private def addWitnesses(phi: Phi, cnfFormula: FormulaLike): Unit = {
    import scala.math._
    val freeLits = dslToDimacsCache.map(p => phi.events(p._1) -> p._2)
    val dimacsTodsl = freeLits.map(_.swap)
    val offsets = freeLits.zipWithIndex.map(p => p._1._1 -> ((cnfFormula.nofVars - freeLits.size) * (p._2 + 1)))
    val initClause = cnfFormula.clauses.toArray
    (1 to (cnfFormula.nofGates * freeLits.size)).foreach(_ => cnfFormula.freshVar)
    freeLits.foreach(p => {
      initClause.init.foreach(c => {
        cnfFormula.addClause(c._1.lits.map(i => {
          if (p._2 == abs(i)) {
            -i
          } else if (dimacsTodsl.contains(abs(i))) {
            i
          } else if (i >= 0) {
            i + offsets(p._1)
          } else {
            i - offsets(p._1)
          }
        }): _*)()
      })
      val lastLit = initClause.last._1.lits.head
      val lastLitTr = {
        if (p._2 == abs(lastLit)) {
          -lastLit
        } else if (dimacsTodsl.contains(abs(lastLit))) {
          lastLit
        } else if (lastLit >= 0) {
          lastLit + offsets(p._1)
        } else {
          lastLit - offsets(p._1)
        }
      }
      cnfFormula.addClause(-lastLitTr :: -p._2 :: Nil: _*)()
    })
  }

  /** Recursively translates a Z3Expr to a CNF formula (memoization of [[toCnfRec]]). */
  private
  def toCnf(node: Z3Expr, cnfFormula: FormulaLike): Int = node match {
    case node: Z3BoolExpr =>
      z3ToDimacsCache.getOrElseUpdate(
        node,
        toCnfRec(node, cnfFormula))
    case _ =>
      throw new Exception(s"formula ${node.toString} is not purely boolean")
  }

  /** Recursively translates the structure function to a CNF formula (does memoization of [[toCnfRec]]). */
  private
  def toCnfRec(node: Z3BoolExpr, cnfFormula: FormulaLike): Int = node.getASTKind match {

    // translate as free variable, update translation cache
    case Z3_VAR_AST =>
      val fresh = cnfFormula.freshVar
      val dsl = this.z3ToDslCache(node)
      this.dslToDimacsCache.update(dsl, fresh)
      fresh

    // functional definition translate using gate variables
    case Z3_APP_AST =>
      val args = node.getArgs

      node.getFuncDecl.getDeclKind match {
        case Z3_OP_NOT =>
          assert(args.length == 1)
          -toCnf(args(0), cnfFormula)

        case Z3_OP_AND =>
          val newArgs = args.toList.map(a => toCnf(a, cnfFormula))
          val fresh = cnfFormula.freshVar
          cnfFormula.defineAnd(fresh, newArgs)()
          fresh

        case Z3_OP_OR =>
          val newArgs = args.toList.map(a => toCnf(a, cnfFormula))
          val fresh = cnfFormula.freshVar
          cnfFormula.defineOr(fresh, newArgs)()
          fresh

        case Z3_OP_IMPLIES =>
          assert(args.length == 2)
          val newArg0 = toCnf(args(0), cnfFormula)
          val newArg1 = toCnf(args(1), cnfFormula)
          val fresh = cnfFormula.freshVar
          cnfFormula.defineImplies(fresh, newArg0, newArg1)()
          fresh

        case Z3_OP_IFF =>
          assert(args.length == 2)
          val newArg0 = toCnf(args(0), cnfFormula)
          val newArg1 = toCnf(args(1), cnfFormula)
          val fresh = cnfFormula.freshVar
          cnfFormula.defineIff(fresh, newArg0, newArg1)()
          fresh

        case Z3_OP_XOR =>
          assert(args.length == 2)
          val newArg0 = toCnf(args(0), cnfFormula)
          val newArg1 = toCnf(args(1), cnfFormula)
          val fresh = cnfFormula.freshVar
          cnfFormula.defineIff(fresh, newArg0, -newArg1)()
          fresh

        case Z3_OP_ITE =>
          assert(args.length == 3)
          val newArg0 = toCnf(args(0), cnfFormula)
          val newArg1 = toCnf(args(1), cnfFormula)
          val newArg2 = toCnf(args(2), cnfFormula)
          val fresh = cnfFormula.freshVar
          cnfFormula.defineIte(fresh, newArg0, newArg1, newArg2)()
          fresh

        case otherKind => throw new Exception(s"in ${node.toString}: unexpected operator Z3_decl_kind = ${otherKind.toString}")
      }
    case otherKind => throw new Exception(s"in ${node.toString}: unexpected AST kind Z3_ast_kind = ${otherKind.toString}")
  }
}
