package util.analyzers

import ast.dsl.InstBoolIdent
import net.sf.javabdd.BDD
import com.microsoft.z3.{Sort => Z3Sort,
  Symbol => Z3Symbol, FuncDecl => Z3FuncDecl,
  Solver => Z3Solver, BoolExpr => Z3BoolExpr,
  Expr => Z3Expr,
  BitVecExpr => Z3BitVecExpr,
  BitVecNum}
/**
  * Trait for BDD based cutsets enumeration on phi
  */
trait PhiCutEnumerationBDD extends PhiCutEnumeration[InstBoolIdent] with BDDAnalyzer {
  self: Phi =>

  val BDDFactory: BDDFactory with Z3BoolExprToBDDTranslator

  /**
    * Truncated computation of the BDD of phi
    *
    * @param max optional bound on cutset cardinality
    * @return the BDD of phi and a mapping of BDD Node to system events
    */
  def _computeBDD(max: Option[Int]): (BDD, Map[Int, InstBoolIdent]) = {
    val eventsList: List[(InstBoolIdent, Z3BoolExpr)] = events.toList.sortBy(p => p._2.getIndex)
    val varMap: List[(InstBoolIdent, Z3BoolExpr, BDD)] = eventsList.map(e => (e._1, e._2, BDDFactory.produceVar()))
    val result = {
      for (m <- max) yield {
        (BDDFactory.boolExprToPartialBdd(expr, varMap.map(t => (t._2, t._3)).toMap, context, m), varMap.map(t => (t._3.`var`(), t._1)).toMap)
      }
    }.getOrElse((BDDFactory.boolExprToBdd(expr, varMap.map(t => (t._2, t._3)).toMap, context), varMap.map(t => (t._3.`var`(), t._1)).toMap))
    result
  }

  /**
    * Compute cutsets on the truncated BDD of [[Phi]]
    *
    * @param max cutset cardinality bound
    * @return cutsets
    */
  def computeCutsUntil(max: Int): Set[Set[InstBoolIdent]] = {
    val (bdd, varMap) = _computeBDD(Some(max))
    val bddCuts = computeCutsBDDUntil(bdd, max)
    bddCuts.map(cut => cut.map(e => varMap(e)))
  }

  /**
    * Compute cutsets on the classic BDD of [[Phi]]
    *
    * @return cutsets
    */
  def computeAllCuts(): Set[Set[InstBoolIdent]] = {
    val (bdd, varMap) = _computeBDD(None)
    val bddCuts = computeCutsBDD(bdd)
    bddCuts.map(cut => cut.map(e => varMap(e)))
  }

  override val phiCutEnumerationProfiler = new Profiler[Option[Int], Iterable[Set[InstBoolIdent]]] {
    override def runWithProfile(i: Option[Int])(iterations: Int = 1): (Iterable[Set[InstBoolIdent]], ExecutionTimes) = {
      val (pair, pTime) = time({
        val (bdd, varMap) = _computeBDD(i)
        i match {
          case Some(m) =>
            (computeCutsBDDUntil(bdd, m), varMap)
          case None =>
            (computeCutsBDD(bdd), varMap)
        }
      }, iterations)
      val (cuts, postTime) = time({
        pair._1.map(cut => cut.map(e => pair._2(e)))
      }, iterations)
      (cuts, ExecutionTimes(0, pTime, postTime))
    }
  }
}
