package util.analyzers

import ast.dsl.InstBoolIdent
import com.microsoft.z3.{Sort => Z3Sort,
  Symbol => Z3Symbol, FuncDecl => Z3FuncDecl,
  Solver => Z3Solver, BoolExpr => Z3BoolExpr,
  Expr => Z3Expr,
  BitVecExpr => Z3BitVecExpr,
  BitVecNum}
import com.microsoft.z3.Status._
/**
  * Trait for SMT resolution based cutsets enumeration
  */
trait PhiCutEnumerationIterativeSMT extends PhiCutEnumeration[InstBoolIdent] {
  self: Phi =>

  private def mkFresh() = context.mkFreshConst("comp", context.mkBoolSort()).asInstanceOf[Z3BoolExpr]

  /**
    * Compute cutsets with SMT problems resolution
    *
    * @param max cutset cardinality bound
    * @return cutsets
    */
  def computeCutsUntil(max: Int): List[Set[InstBoolIdent]] = {

    //initial cutsets
    var result = List[Set[InstBoolIdent]]()

    //instantiate Z3 solver
    val s = context.mkSolver()

    //generate Boolean constants for each system event
    val eventsList: List[(InstBoolIdent, Z3BoolExpr, Z3BoolExpr)] = events.toList.map(e => (e._1, e._2, context.mkBoolConst(e._1.toString())))
    val consts: (List[Z3BoolExpr], List[Z3BoolExpr]) = eventsList.map(t => (t._2, t._3)).unzip

    //activation map of atMost assertions
    val assertions = scala.collection.mutable.Map[Z3BoolExpr, Z3BoolExpr]()

    //substitute phi var by system events and assert it
    s.add(expr.substitute(consts._1.toArray[Z3Expr], consts._2.toArray[Z3Expr]).asInstanceOf[Z3BoolExpr])

    s.check() match {
      case UNSATISFIABLE =>

        //if phi is not satisfiable then it does not exist cutsets
        s.dispose()

      case SATISFIABLE =>

        //if phi is satisfiable, then compute cutsets starting at k=0
        var k = 0
        var over = false

        //while k has not reach the bound or there is not any more cutsets
        while (k <= max && !over) {
          if (s.check(assertions.values.toList: _*) == SATISFIABLE) {

            //generate and add an atMostk assertion
            val atMostN: Z3BoolExpr = context.atMostN(consts._2, k)
            val fresh = mkFresh()
            assertions(atMostN) = context.mkNot(fresh)
            s.add(context.mkOr(atMostN, fresh))

            //while some cutsets can be found
            while (s.check(assertions.values.toList: _*) == SATISFIABLE) {
              val found = eventsList.collect(
                { case (name, _, const) if s.getModel.eval(const, false).isTrue =>
                  (name, const, context.mkBool(s.getModel.eval(const, false).isTrue))
                })
              //save it
              result ::= found.map(t => t._1).toSet

              //block it
              s.add(context.mkNot(context.mkAnd(found.map(t => t._2): _*)))
            }

            //when there is no more cutsets
            if (s.check(assertions.values.toList: _*) == UNSATISFIABLE) {
              //desactivate current atMostK assertion
              assertions(atMostN) = fresh

            } else {
              throw Z3UnknownException(this)
            }
          } else {
            over = true
          }
          //increment cardinality bound
          k += 1
        }
        s.dispose()
      case UNKNOWN => throw Z3UnknownException(this)
    }
    result
  }

  /**
    * Compute cutsets with SMT problems resolution
    *
    * @return cutsets
    */
  def computeAllCuts(): List[Set[InstBoolIdent]] = {

    //initial cutsets
    var result = List[Set[InstBoolIdent]]()

    //instantiate Z3 solver
    val s = context.mkSolver()

    //generate Boolean constants for each system event
    val eventsList: List[(InstBoolIdent, Z3BoolExpr, Z3BoolExpr)] = events.toList.map(e => (e._1, e._2, context.mkBoolConst(e._1.toString())))
    val consts: (List[Z3BoolExpr], List[Z3BoolExpr]) = eventsList.map(t => (t._2, t._3)).unzip

    //every event is necessary to trigger phi
    val isCut = context.mkAnd(eventsList.map(c => context.mkImplies(c._3,
      context.mkNot(expr.substitute(consts._1.toArray[Z3Expr],
        consts._2.map(d => if (c._3 == d) context.mkNot(d) else d).toArray[Z3Expr]) match { case b: Z3BoolExpr => b
      }))): _*)

    //assert phi is triggered and all events are necessary (ie valuation is a cutset)
    s.add(expr.substitute(consts._1.toArray[Z3Expr], consts._2.toArray[Z3Expr]) match { case b: Z3BoolExpr => b })
    s.add(isCut)

    s.check() match {

      //if phi is not satisfiable then it does not exist cutsets
      case UNSATISFIABLE =>
        s.dispose()
        Nil

      //if phi is satisfiable, then compute cutsets
      case SATISFIABLE =>

        //while it exists cutsets
        while (s.check() == SATISFIABLE) {
          val found = eventsList.collect(
            { case (name, _, const) if s.getModel.eval(const, false).isTrue =>
              (name, const, context.mkBool(s.getModel.eval(const, false).isTrue))
            })

          //save it
          result ::= found.map(t => t._1).toSet

          //block it
          s.add(context.mkNot(context.mkAnd(found.map(t => t._2): _*)))

        }
        if (s.check() == UNKNOWN)
          throw Z3UnknownException(this)
        s.dispose()
      case UNKNOWN => throw Z3UnknownException(this)
    }
    result
  }

  override val phiCutEnumerationProfiler = new Profiler[Option[Int], Iterable[Set[InstBoolIdent]]] {
    override def runWithProfile(i: Option[Int])(iterations: Int = 1): (Iterable[Set[InstBoolIdent]], ExecutionTimes) = ???
  }
}
