package util.analyzers

import analyzers._
import ast.dsl.{CompAlternative, CompInstanceIdent}
import util.analyzers.MeasureTypes.{CardMeasure, ProbMeasure}
import util.analyzers.InfNumericImplicits._

/**
  * Trait for additional analysis of the DSE problem solution
  */
trait SolutionAnalyzer extends ConfigurationAnalyzer {
  self: Parsing with DSEProblemSolver with STMDDBuilder with SafetyTheory =>


  private lazy val altIndexMap: Map[(CompInstanceIdent, CompAlternative), Int] = conf.designSpace.flatMap(sub => {
    val allAlts = sub.init +: sub.alternatives
    allAlts.map(alt => (sub.instance, alt) -> allAlts.indexOf(alt))
  }).toMap

  /**
    * Compute the unreliability of the solution with the STMDD
    *
    * @param solution the solution to assess
    * @return the unreliability
    */
  def computeUnrWithMemo(solution: Solution): ProbMeasure = {
    computeUnr((STMDDD, solutionToSelectSeq(solution)))
  }

  /**
    * Transform a solution to a sequence of selection of local indicators in the STMDD
    *
    * @param solution alternative choices for system's component
    * @return sequence of selection of local indicators in the STMDD
    */
  def solutionToSelectSeq(solution: Solution): SelectionSeq = {
    STMDDD.variables.toSeq.map(v => (v.value.path, altIndexMap.get((v.value, solution(v.value)))))
  }

  /**
    * compute the cardinality of a substitution choice
    *
    * @param solution the substitution choice
    * @return the over, under and exact cardinality (if computable)
    */
  def computeCard(solution: Solution): CardMeasure = {
    computeCardRec(STMDDD, solutionToSelectSeq(solution))
  }

  /**
    * Recursively compute the order of a mdd node according to a substitution choice which is
    * card(node)=min_{s \in node.kids} (card(s.tr,choice(node))+card(s.son))
    *
    * @param node   mdd node
    * @param choice substitution chocie
    * @return the order
    */
  def computeCardRec(node: mddFactory.MDD, choice: SelectionSeq): CardMeasure = {
    node match {
      case n: mddFactory.MDDNode =>
        //choice for the node
        val nodeChoiceIndex = getChoiceIndex(choice, n)
        //choice for sons
        val choiceTail = choice.drop(nodeChoiceIndex + 1)

        //computing the minimal order among sons
        n.foldLeft(CardMeasure.infinity)((acc, pair) => {

          val (edge, son) = pair

          //if the current minimum is zero then no need to evaluate son
          if (acc == CardMeasure.zero) {
            CardMeasure.zero
          } else {

            //otherwise use the value the order of the trace in the edge if substitution choice available
            //or the max and min order otherwise
            val nodeOrder = choice(nodeChoiceIndex)._2 match {
              case Some(bit) => CardMeasure(edge.card(bit))
              case None => CardMeasure(edge.maxCard, edge.minCard, None)
            }

            //select the minimum
            acc.min(nodeOrder + computeCardRec(son, choiceTail))
          }
        })
      case _: mddFactory.Terminal =>
        //if the node is a terminal then we reach the failure condition thus order=0
        CardMeasure.zero
    }
  }
}
