package util.analyzers

import analyzers.{Analyser, Parsing, SMTTranslation}
import ast.dsl.{CompSubstitution, DSLSymbol, StdDSLSymbol}
import com.microsoft.z3.BitVecNum
import util.analyzers.analyzerTypes.Real
import com.microsoft.z3.{Sort => Z3Sort,
  Symbol => Z3Symbol, FuncDecl => Z3FuncDecl,
  Solver => Z3Solver, BoolExpr => Z3BoolExpr,
  Expr => Z3Expr,
  BitVecExpr => Z3BitVecExpr,
  BitVecNum}
import com.microsoft.z3.Status._
import scala.collection.mutable

/**
  * Case clas for subsitution analysis with pure SMT exploration (ie without the interpreter)
  *
  * @param sub            the AST node substitution
  * @param confAnalyzer   the current configuration analyser
  * @param componentCache cache containgin already analysed component
  */
class Substitution(val sub: CompSubstitution,
                   val confAnalyzer: Parsing with SMTTranslation,
                   val componentCache: mutable.HashMap[DSLSymbol, Component with RelationBuild
                     with FBDDBuildSMT
                     with FCardinalityBuild])

/**
  * Legality analysis based on SMT problem solving
  */
trait LegalAnalyser extends Analyser {
  self: Substitution =>

  lazy val isLegal: Boolean = _isLegal

  /**
    * Legality analysis of the current substitution
    *
    * @return
    */
  private def _isLegal: Boolean = {

    //gather all possible alternatives
    val allAlt = sub.init +: sub.alternatives

    //create a component analysis for each alternative
    val alternatives = allAlt.map(alt =>
      new Component(confAnalyzer.getCompTranslation(confAnalyzer.conf.namespace.locate(alt.compId))) with TraceBuild)

    //create a component analysis fir the initial component
    val iniComp = confAnalyzer.conf.namespace.locate(sub.init.compId)
    val iniCompTr = new Component(confAnalyzer.getCompTranslation(iniComp)) with TraceBuild

    //check legality for each alternative
    val legal = alternatives.map(alt => {

      //instantiate solver
      val s = iniCompTr.context.mkSolver()

      //create constants for the flow identifiers of the initial component
      val iniConst = iniCompTr.traceFieldVars.map(p => (p._2, iniCompTr.context.mkConst(p._1.toString, p._2.getSort))).toList.sortBy(p => p._1.getIndex)

      //create a is_a_tr predicate for the initial component
      val trIni = iniCompTr.traceFieldExpr.substituteVars(iniConst.map(e => e._2).toArray) match {
        case b: Z3BoolExpr => b
      }

      //create a is_a_tr predicate for the alternative component
      val trAlt = alt.traceFieldExpr.substituteVars(iniConst.map(e => e._2).toArray) match {
        case b: Z3BoolExpr => b
      }

      //try to find a trace of alternative component which is not a trace of initial component
      s.add(iniCompTr.context.mkAnd(iniCompTr.context.mkNot(trIni), trAlt))

      s.check() match {

        //if possible store the trace
        case SATISFIABLE => (SATISFIABLE, iniConst.map(cst =>
          (StdDSLSymbol(Symbol(cst._2.getFuncDecl.getName.toString)), s.getModel.eval(cst._2, true).asInstanceOf[BitVecNum])))

        //if not possible continu the analysis
        case UNSATISFIABLE => (UNSATISFIABLE, Nil)
        case UNKNOWN => throw Z3UnknownException(this)
      }
    })

    //if the problem is not unsat for a given alternative then this one cannot be used for the exploration
    //(ie does not fulfil the inclusion assumption)
    legal.zipWithIndex.collect({
      case ((SATISFIABLE, trace), i) =>
        throw IllegalSubstitution(sub.instance, sub.alternatives(i).compId, trace.mkString(","))
    })
    !legal.exists(p => p._1 == SATISFIABLE)
  }
}

/**
  * Analysis of the alternatives for a given design space
  */
trait AlternativesAnalyser extends Analyser {
  self: Substitution with LegalAnalyser =>

  val BDDFactory: BDDFactory with Z3BoolExprToBDDTranslator

  lazy val characAlternatives: List[Alternative with FUnreliabilityBuild] = _characAlternatives

  /**
    * Compute the trace probability and order for each alternative given in a [[Substitution]] class
    * using SMT problem solving
    *
    * @return the characterised alternatives
    */
  private def _characAlternatives: List[Alternative with FUnreliabilityBuild] = {
    if (!isLegal) {
      Nil
    } else {
      (sub.init +: sub.alternatives).map(alt => {
        val comp = componentCache.getOrElseUpdate(alt.compId.baseId,
          new Component(confAnalyzer.getCompTranslation(confAnalyzer.conf.namespace.locate(alt.compId)))
            with RelationBuild with FBDDBuildSMT with FCardinalityBuild {
            val BDDFactory: BDDFactory with Z3BoolExprToBDDTranslator = self.BDDFactory
          })
        new Alternative(alt, comp, {
          for (t <- confAnalyzer.conf.duration) yield Real(t.repr)
        }.getOrElse(1)) with FUnreliabilityBuild
      })
    }
  }
}
