package util.analyzers

/**
  * Created by rdelmas on 22/06/16.
  */

import ast.dsl.InstBoolIdent
import com.microsoft.z3.enumerations.Z3_ast_kind._
import com.microsoft.z3.enumerations.Z3_decl_kind._
import com.microsoft.z3.{AST => Z3AST, BitVecNum => Z3BitVecNum, BoolExpr => Z3BoolExpr, DatatypeSort => Z3DataTypeSort, Expr => Z3Expr}
import util.sat._

import scala.collection.immutable.{Map => IMap, Set => ISet}
import scala.collection.{concurrent, immutable, mutable}


/** Class for enumrating prime implicants of a structure function.
  *
  * @param phi      The structure function to analyze
  * @param maxOrder Optional maximum cardinality for prime implicants
  * @param fileName Optional file name in which to dump CNF models for pi generation.
  * */
class TREKPIEnumerator(phi: Phi, maxOrder: Option[Int] = None, fileName: Option[String] = None) {

  val z3ToDslCache: immutable.Map[Z3BoolExpr, InstBoolIdent] = phi.events.map(_.swap)
  val dslToPropCache: mutable.Map[InstBoolIdent, Prop.Atom] = new concurrent.TrieMap[InstBoolIdent, Prop.Atom]()
  val z3ToPropCache: mutable.Map[Z3BoolExpr, Prop.Expr] = new concurrent.TrieMap[Z3BoolExpr, Prop.Expr]()


  /** Enumerates prime implicants of the given structure function.
    *
    * @return A pair (z3Pis, dslPis) where z3Pis are the prime implicants translated as Z3BoolExpr
    *         and dslPis are the prime implicants translated as DslLit.
    * */
  def generate(): (List[ISet[Z3BoolExpr]], List[ISet[DslLit]]) = {

    val structureFunction: Prop.Expr = translate(phi)

    val enumerator = PrimeImplicantGenerator(structureFunction, fileName.getOrElse("file"), maxOrder)

    // compute pis
    val primeImplicantsProp: List[ISet[Prop.Expr]] = enumerator.genPrimeImplicants()

    // swap cache
    val propToZ3Cache = z3ToPropCache.map(_.swap)

    // translate back
    val z3Pi: List[ISet[Z3BoolExpr]] = primeImplicantsProp.map(
      pi => pi.map {
        case atom: Prop.Atom => propToZ3Cache(atom)
        case Prop.Not(atom) => phi.context.mkNot(propToZ3Cache(atom))
        case _ => throw new Exception("not a literal")
      }
    )

    // swap cache
    val propToDslCache = dslToPropCache.map(_.swap)

    // translate back
    val dslPi: List[ISet[DslLit]] = primeImplicantsProp.map(
      pi => {
        val newPi: ISet[DslLit] = pi.map {
          case atom: Prop.Atom => PosDslLit(propToDslCache(atom))
          case Prop.Not(atom@Prop.Atom(_)) => NegDslLit(propToDslCache(atom))
          case _ => throw new Exception("not a literal")
        }
        newPi
      }
    )
    // return results
    (z3Pi, dslPi)
  }

  /** Translates the structure function to CNF. */
  private def translate(phi: Phi): Prop.Expr = {

    // translate free variables first to map them to low dimacs indices
    this.z3ToDslCache.keys foreach { v => this.translate(v) }

    // translate the expression of Phi to CNF and add resulting literal as as unit clause
    this.translate(phi.expr)

  }

  /** Recursively translates a Z3Expr to a Prop formula (memoization of [[translate]]). */
  private def translate(node: Z3Expr): Prop.Expr = node match {
    case node: Z3BoolExpr => z3ToPropCache.getOrElseUpdate(node, translateRec(node))
    case _ => throw new Exception(s"formula ${node.toString} is not purely boolean")
  }

  /** Recursively translates the structure function to a CNF formula (does memoization of [[translate]]). */
  private def translateRec(node: Z3BoolExpr): Prop.Expr = node.getASTKind match {

    case Z3_VAR_AST =>
      val fresh = Prop.freshAtom()
      val dsl = this.z3ToDslCache(node)
      this.dslToPropCache.update(dsl, fresh)
      fresh


    case Z3_APP_AST =>
      val args = node.getArgs

      node.getFuncDecl.getDeclKind match {
        case Z3_OP_NOT =>
          assert(args.length == 1)
          Prop.Not(translate(args(0)))

        case Z3_OP_AND =>
          Prop.And(args.toList.map(translate))

        case Z3_OP_OR =>
          Prop.Or(args.toList.map(translate))

        case Z3_OP_IMPLIES =>
          assert(args.length == 2)
          Prop.Implies(translate(args(0)), translate(args(1)))

        case Z3_OP_IFF =>
          assert(args.length == 2)
          Prop.Iff(translate(args(0)), translate(args(1)))

        case Z3_OP_XOR =>
          assert(args.length == 2)
          Prop.Xor(translate(args(0)), translate(args(1)))

        case Z3_OP_ITE =>
          assert(args.length == 3)
          Prop.Ite(translate(args(0)), translate(args(1)), translate(args(2)))

        case otherKind => throw new Exception(s"in ${node.toString}: unexpected operator Z3_decl_kind = ${otherKind.toString}")
      }
    case otherKind => throw new Exception(s"in ${node.toString}: unexpected AST kind Z3_ast_kind = ${otherKind.toString}")
  }

}
