package util.analyzers

import util.io.{FileManager, SystemInfo}

import scala.collection.mutable
import scala.sys.process.{Process, _}
import java.io.File

import analyzers._
import util.io.FileManager

import scala.collection.mutable
import scala.collection.mutable.ListBuffer
import scala.reflect.io.Path
import scala.sys.process._


/**
  * Created by kdelmas on 22/02/17.
  */
case class ToolNotFound(msg: String) extends Exception(msg)

trait ToolInterface {

  val CORRECT_END = 0

  val PATH = "PATH"

  val LIBRARY = "LD_LIBRARY_PATH"

  val exeName: String

  val defaultPath: Path

  val mustContain: String

  val activeProcesses: mutable.ListBuffer[Process]


  var _toolPath: Option[Path] = None

  def toolPath: Path = {
    (for (p <- _toolPath) yield p) orElse
      (for (f <- FileManager.locateFirstFileInEnv(PATH, (f: File) => f.getName == exeName)) yield Path(f.getAbsolutePath)) orElse
      (if (defaultPath.exists) Some(defaultPath.toAbsolute) else None) getOrElse
      (throw ToolNotFound(s"the tool $exeName is not in current path or binlib directory"))
  }

  def toolPathOption: Option[Path] = {
    var result: Option[Path] = None
    try {
      result = Some(toolPath)
    } catch {
      case _: ToolNotFound =>
    }
    result
  }


  def setToolPath(p: Path): Boolean = {
    val path = FileManager.recursiveLocateFirstFile(new File(p.toAbsolute.toString()), (f: File) => f.getName == mustContain)
    if (p.exists && path.isDefined) {
      _toolPath = Some(path.get)
      true
    } else
      false
  }

  def startProcess(pB: ProcessBuilder): Process = {
    val p = pB.run()
    activeProcesses += p
    p
  }

  def destroyAll(): Unit = {
    activeProcesses.foreach(p => p.destroy())
    activeProcesses.clear()
  }

  def destroy(p: Process): Unit = {
    activeProcesses -= p
    p.destroy()
  }

  def addProcess(p: Process): Unit = {
    activeProcesses += p
  }
}


object PBEncoderInterface extends ToolInterface {
  val activeProcesses: ListBuffer[Process] = mutable.ListBuffer[Process]()
  val exeName: String = "pbencoder"
  val defaultPath: Path = "binlib/pbencoder"
  val mustContain: String = exeName
}

object PythonInterface extends ToolInterface {
  val activeProcesses: ListBuffer[Process] = mutable.ListBuffer[Process]()
  val exeName: String = "python"
  val defaultPath: Path = "/usr/bin/python"
  val mustContain: String = exeName
}

object WineInterface extends ToolInterface {
  val activeProcesses: ListBuffer[Process] = mutable.ListBuffer[Process]()
  val exeName: String = "wine"
  val defaultPath: Path = "/usr/bin/wine"
  val mustContain: String = exeName
}

object ClaspInterface extends ToolInterface {
  val activeProcesses: ListBuffer[Process] = mutable.ListBuffer[Process]()
  val exeName: String = "clasp"
  val defaultPath: Path = "binlib/clasp"
  val mustContain: String = exeName
}

object SharpCDCLInterface extends ToolInterface {
  val activeProcesses: ListBuffer[Process] = mutable.ListBuffer[Process]()
  val exeName: String = "sharpCDCL"
  val defaultPath: Path = "binlib/sharpCDCL"
  val mustContain: String = exeName

}

object CygwinInterface extends ToolInterface {
  val activeProcesses: ListBuffer[Process] = mutable.ListBuffer[Process]()
  val exeName: String = "cygwin"
  val defaultPath: Path = "cygwin"
  val mustContain: String = exeName
}

object XSAPInterface extends ToolInterface {
  val activeProcesses: ListBuffer[Process] = mutable.ListBuffer[Process]()
  val exeName: String = "xSAP/scripts"
  val defaultPath: Path = "binlib/xSAP/scripts"
  val mustContain: String = "extend_model.py"
  val mustContainCom: String = "compute_ft.py"

  override def setToolPath(p: Path): Boolean = {
    val extendPath = FileManager.locateFirstFile(new File(p.toAbsolute.toString()), (f: File) => f.getName == mustContain)
    val computePath = FileManager.locateFirstFile(new File(p.toAbsolute.toString()), (f: File) => f.getName == mustContainCom)
    if (p.exists && extendPath.isDefined & computePath.isDefined) {
      _toolPath = Some(p)
      true
    } else
      false
  }
  def killXSAP():Unit={
    val r="[ \t]*([0-9]+).*".r
    try{
      val explorerProcess=SystemInfo.runUnixCmd("ps -A | grep -E xSAP").!!.split("\n")
    explorerProcess.foreach(s=> {
      val r(id) = s
      SystemInfo.runUnixCmd(s"kill $id").run().exitValue()
    })
    } catch {
      case _: Throwable => //TODO handle non-zero return
    }
  }
}


object XFTAInterface extends ToolInterface {
  val activeProcesses: ListBuffer[Process] = mutable.ListBuffer[Process]()
  val exeName: String = "xftar"
  val defaultPath: Path = "binlib/xftar"
  val mustContain: String = exeName
}

object HipHOPSInterface extends ToolInterface {
  val activeProcesses: ListBuffer[Process] = mutable.ListBuffer[Process]()
  val exeName: String = "hipop.exe"
  val defaultPath: Path = "binlib/HipHOPS/hipop.exe"
  val mustContain: String = exeName

  def killInternetExplorer():Unit={
    val r="[ \t]*([0-9]+).*".r
    try{
      val explorerProcess=SystemInfo.runUnixCmd("ps -A | grep -E iexplore").!!.split("\n")
      explorerProcess.foreach(s=> {
        val r(id)=s
        s"kill $id".run().exitValue()
      })
    } catch {
      case _: RuntimeException => //TODO inform that no iexplorer process
    }

  }
}

object Z3JavaSO extends ToolInterface {

  val activeProcesses: ListBuffer[Process] = mutable.ListBuffer[Process]()
  lazy val exeName: String = SystemInfo.userOs match {
    case SystemInfo.Linux=> "libz3java.so"
    case SystemInfo.MAC =>  "libz3java.dylib"
    case SystemInfo.Windows => "libz3java.dll"
    case _ => "libz3java"

  }
  val defaultPath: Path = null
  val mustContain: String = exeName

  override def toolPath: Path = {
    (for (p <- _toolPath) yield p) orElse
      (for (f <- FileManager.locateFirstFileInEnv(LIBRARY, (f: File) => f.getName == exeName)) yield Path(f.getAbsolutePath)) getOrElse
      (throw ToolNotFound(s"the tool $exeName is not in $LIBRARY"))
  }
}



