package util.analyzers

import com.microsoft.z3.enumerations.Z3_ast_kind._
import com.microsoft.z3.enumerations.Z3_decl_kind._
import com.microsoft.z3.{BoolExpr => Z3BoolExpr, Context => Z3Context}
import net.sf.javabdd.BDD
import util.translators.SMTContext

import scala.collection.mutable

/**
  * Created by kdelmas on 08/11/16.
  */

/**
  * Base trait for translating Z3 SMT Boolean expression to BDD
  */
trait Z3BoolExprToBDDTranslator extends Z3ExprTranslators {
  self: BDDFactory =>

  /**
    * Transfrom a Boolean expression to a BDD
    * @param expr the Z3 Boolean expression
    * @param varMap the BDD for Boolean variable (leaves of the Z3 AST)
    * @param context context in which the expression has been built
    * @return the BDD
    */
  def boolExprToBdd(expr: Z3BoolExpr,
                    varMap: Map[Z3BoolExpr, BDD],
                    context: SMTContext): BDD = {
    boolExprToBddRec(expr, varMap, context)(mutable.Map.empty[Z3BoolExpr, BDD])
  }

  private def boolExprToBddRec(expr: Z3BoolExpr,
                               varMap: Map[Z3BoolExpr, BDD],
                               context: Z3Context)(implicit cache: mutable.Map[Z3BoolExpr, BDD]): BDD = {
    expr.getASTKind match {
      case Z3_VAR_AST =>
        cache.getOrElseUpdate(expr, varMap.getOrElse(expr, throw UnknownEvent(expr.getFuncDecl.getName.toString)))
      case Z3_APP_AST =>
        expr.getFuncDecl.getDeclKind match {
          case Z3_OP_AND =>
            cache.getOrElseUpdate(expr, {
              val args = expr.getArgs.map { case b: Z3BoolExpr => b }
              args.foldLeft(one())((bdd, bExpr) => bdd.and(boolExprToBddRec(bExpr, varMap, context)))
            })
          case Z3_OP_OR =>
            cache.getOrElseUpdate(expr, {
              val args = expr.getArgs.map { case b: Z3BoolExpr => b }
              args.foldLeft(zero())((bdd, bExpr) => bdd.or(boolExprToBddRec(bExpr, varMap, context)))
            })
          case Z3_OP_NOT =>
            cache.getOrElseUpdate(expr, {
              val args = expr.getArgs.map { case b: Z3BoolExpr => b }
              boolExprToBddRec(args(0), varMap, context).not()
            })
          case Z3_OP_IMPLIES =>
            cache.getOrElseUpdate(expr, {
              val args = expr.getArgs.map { case b: Z3BoolExpr => b }
              boolExprToBddRec(args(0), varMap, context).imp(boolExprToBddRec(args(1), varMap, context))
            })
          case Z3_OP_IFF =>
            cache.getOrElseUpdate(expr, {
              val args = expr.getArgs.map { case b: Z3BoolExpr => b }
              boolExprToBddRec(args(0), varMap, context).biimp(boolExprToBddRec(args(1), varMap, context))
            })
          case Z3_OP_ITE =>
            cache.getOrElseUpdate(expr, {
              val args = expr.getArgs.map { case b: Z3BoolExpr => b }
              boolExprToBddRec(args(0), varMap, context).ite(boolExprToBddRec(args(1), varMap, context), boolExprToBddRec(args(2), varMap, context))
            })
          case Z3_OP_TRUE =>
            one()
          case Z3_OP_FALSE =>
            zero()
          case _ => throw WrongBoolExpr(expr.toString)
        }
      case _ => throw UnknownZ3AST(expr.toString)
    }
  }

  /**
    * Transform a Boolean expression to a truncated BDD
    * @param expr the Z3 Boolean expression
    * @param varMap the BDD for Boolean variable (leaves of the Z3 AST)
    * @param context context in which the expression has been built
    * @param upTo the maximum number of positive arc in paths of BDD
    * @return the BDD
    */
  def boolExprToPartialBdd(expr: Z3BoolExpr,
                           varMap: Map[Z3BoolExpr, BDD],
                           context: SMTContext,
                           upTo: Int): BDD = {
    boolExprToPartialBddRec(toNNF(expr, varMap.keys.toList, context), varMap, context, upTo)(mutable.Map.empty[(Z3BoolExpr, Int), BDD])
  }

  def boolExprToPartialBddRec(expr: Z3BoolExpr,
                              varMap: Map[Z3BoolExpr, BDD],
                              context: Z3Context,
                              remain: Int)(implicit cache:collection.mutable.Map[(Z3BoolExpr, Int), BDD]): BDD = {
    expr.getASTKind match {
      case Z3_VAR_AST =>
        cache.getOrElseUpdate((expr, remain), {
          if (remain > 0) {
            varMap.getOrElse(expr, throw UnknownEvent(expr.getFuncDecl.getName.toString))
          } else
            zero()
        })
      case Z3_APP_AST =>
        expr.getFuncDecl.getDeclKind match {
          case Z3_OP_AND =>
            cache.getOrElseUpdate((expr, remain), {
              var simplifiedVarMap = varMap
              //skip all variable such that F0==F1 (simplified since subtree will be equals)
              while (expr.substitute(simplifiedVarMap.head._1, context.mkTrue()) == expr.substitute(simplifiedVarMap.head._1, context.mkFalse())) {
                simplifiedVarMap = simplifiedVarMap.tail
              }
              val F0 = expr.substitute(simplifiedVarMap.head._1, context.mkFalse()).simplify().asInstanceOf[Z3BoolExpr]
              val BDDF0 = boolExprToPartialBddRec(F0, simplifiedVarMap.tail, context, remain)
              if (remain > 0) {
                val F1 = expr.substitute(simplifiedVarMap.head._1, context.mkTrue()).simplify().asInstanceOf[Z3BoolExpr]
                val BDDF1 = boolExprToPartialBddRec(F1, simplifiedVarMap.tail, context, remain - 1)
                val result = mkNode(simplifiedVarMap.head._2.`var`(), BDDF1, BDDF0)
                result
              } else {
                val BDDF1 = zero()
                val result = mkNode(simplifiedVarMap.head._2.`var`(), BDDF1, BDDF0)
                result
              }
            })
          case Z3_OP_OR =>
            cache.getOrElseUpdate((expr, remain), {
              var simplifiedVarMap = varMap
              //skip all variable such that F0==F1 (simplified since subtree will be equals)
              while (expr.substitute(simplifiedVarMap.head._1, context.mkTrue()) == expr.substitute(simplifiedVarMap.head._1, context.mkFalse())) {
                simplifiedVarMap = simplifiedVarMap.tail
              }
              val F0 = expr.substitute(simplifiedVarMap.head._1, context.mkFalse()).simplify().asInstanceOf[Z3BoolExpr]
              val BDDF0 = boolExprToPartialBddRec(F0, simplifiedVarMap.tail, context, remain)
              if (remain > 0) {
                val F1 = expr.substitute(simplifiedVarMap.head._1, context.mkTrue()).simplify().asInstanceOf[Z3BoolExpr]
                val BDDF1 = boolExprToPartialBddRec(F1, simplifiedVarMap.tail, context, remain - 1)
                val result = mkNode(simplifiedVarMap.head._2.`var`(), BDDF1, BDDF0)
                result
              } else {
                val BDDF1 = zero()
                val result = mkNode(simplifiedVarMap.head._2.`var`(), BDDF1, BDDF0)
                result
              }
            })
          case Z3_OP_ITE =>
            cache.getOrElseUpdate((expr, remain), {
              var simplifiedVarMap = varMap
              //skip all variable such that F0==F1 (simplified since subtree will be equals)
              while (expr.substitute(simplifiedVarMap.head._1, context.mkTrue()) == expr.substitute(simplifiedVarMap.head._1, context.mkFalse())) {
                simplifiedVarMap = simplifiedVarMap.tail
              }
              val F0 = expr.substitute(simplifiedVarMap.head._1, context.mkFalse()).simplify().asInstanceOf[Z3BoolExpr]
              val BDDF0 = boolExprToPartialBddRec(F0, simplifiedVarMap.tail, context, remain)
              if (remain > 0) {
                val F1 = expr.substitute(simplifiedVarMap.head._1, context.mkTrue()).simplify().asInstanceOf[Z3BoolExpr]
                val BDDF1 = boolExprToPartialBddRec(F1, simplifiedVarMap.tail, context, remain - 1)
                val result = mkNode(simplifiedVarMap.head._2.`var`(), BDDF1, BDDF0)
                result
              } else {
                val BDDF1 = zero()
                val result = mkNode(simplifiedVarMap.head._2.`var`(), BDDF1, BDDF0)
                result
              }
            })
          case Z3_OP_NOT =>
            cache.getOrElseUpdate((expr, remain), {
              val args = expr.getArgs.map { case b: Z3BoolExpr => b }
              val result = boolExprToPartialBddRec(args(0), varMap, context, remain).not()
              result
            })
          case Z3_OP_TRUE =>
            one()
          case Z3_OP_FALSE =>
            zero()
          case _ => throw WrongBoolExpr(expr.toString)
        }
      case _ => throw UnknownZ3AST(expr.toString)
    }
  }
}
