package util.analyzers

import com.microsoft.z3.enumerations.Z3_ast_kind._
import com.microsoft.z3.enumerations.Z3_decl_kind._
import com.microsoft.z3.{AST => Z3AST, BitVecNum => Z3BitVecNum, BoolExpr => Z3BoolExpr, DatatypeSort => Z3DataTypeSort, Expr => Z3Expr}
import util.sat.FormulaLike

import scala.collection._


/**
  * Created by rdelmas on 07/06/16.
  *
  * A translation function from z3 expressions to FormulaLike
  */
object Z3ExprToFormulaLike {

  /** Class storing the translation results. */
  case class Result[A](
                        source: Z3BoolExpr,
                        freeVars: immutable.Map[Z3BoolExpr, A],
                        result: Int,
                        formula: FormulaLike,
                        freeDimacsVars: List[Int],
                        dimacsCache: immutable.Map[Z3BoolExpr, Int]
                      )

  /** Class storing the caches and formula builder used during translation. */
  private case class Context[A](
                                 freeVars: immutable.Map[Z3BoolExpr, A],
                                 targetFormula: FormulaLike,
                                 cache: mutable.Map[Z3BoolExpr, Int]
                               )

  /** Translates the given source expression to a propositional formula using Tseitin rules.
    *
    * @tparam A user defined type
    * @param source   The z3 boolean expression to translate
    * @param freeVars Map that stores the free variables of the source expression and maps them to some representation of type A
    * @param target   Optional FormulaLike container and translation cache (new ones created if not provided)
    **/
  def apply[A](
                source: Z3BoolExpr,
                freeVars: immutable.Map[Z3BoolExpr, A],
                target: Option[(FormulaLike, mutable.Map[Z3BoolExpr, Int])] = None): Result[A] = {

    val (formula, cache) = target.getOrElse((new FormulaLike {}, new concurrent.TrieMap[Z3BoolExpr, Int]()))
    implicit val ctx: Context[A] = Context(freeVars, formula, cache)
    // translate free variables first so they occupy the lowest dimacs indices possible.
    val freeVarsDimacs = ctx.freeVars.keys.map(v => translate(v)).toList
    // translate the source expression
    val result = translate(source)
    // build result instance
    Result(source, freeVars, result, formula, freeVarsDimacs, cache.toMap)

  }

  /** Recursively translates a Z3Expr to a CNF formula (memoization of [[translateRec]]). */
  private def translate[A](node: Z3Expr)(implicit ctx: Context[A]): Int = node match {
    case node: Z3BoolExpr => ctx.cache.getOrElseUpdate(node, this.translateRec(node))
    case _ => throw new Exception(s"formula ${node.toString} is not purely boolean")

  }

  /** Recursively translates the structure function to a CNF formula (does memoization of [[translateRec]]). */
  private def translateRec[A](node: Z3BoolExpr)(implicit ctx: Context[A]): Int = node.getASTKind match {

    // translate as free variable, update translation cache
    case Z3_VAR_AST =>
      ctx.targetFormula.freshVar


    // functional definition translate using gate variables
    case Z3_APP_AST =>

      val args = node.getArgs

      node.getFuncDecl.getDeclKind match {
        case Z3_OP_NOT =>
          assert(args.length == 1)
          -translate(args(0))

        case Z3_OP_AND =>
          val newArgs = args.toList.map(arg => translate(arg))
          val fresh = ctx.targetFormula.freshVar
          ctx.targetFormula.defineAnd(fresh, newArgs)()
          fresh

        case Z3_OP_OR =>
          val newArgs = args.toList.map(arg => translate(arg))
          val fresh = ctx.targetFormula.freshVar
          ctx.targetFormula.defineOr(fresh, newArgs)()
          fresh

        case Z3_OP_IMPLIES =>
          assert(args.length == 2)
          val newArg0 = translate(args(0))
          val newArg1 = translate(args(1))
          val fresh = ctx.targetFormula.freshVar
          ctx.targetFormula.defineImplies(fresh, newArg0, newArg1)()
          fresh

        case Z3_OP_IFF =>
          assert(args.length == 2)
          val newArg0 = translate(args(0))
          val newArg1 = translate(args(1))
          val fresh = ctx.targetFormula.freshVar
          ctx.targetFormula.defineIff(fresh, newArg0, newArg1)()
          fresh

        case Z3_OP_XOR =>
          assert(args.length == 2)
          val newArg0 = translate(args(0))
          val newArg1 = translate(args(1))
          val fresh = ctx.targetFormula.freshVar
          ctx.targetFormula.defineXor(fresh, newArg0, newArg1)()
          fresh

        case Z3_OP_ITE =>
          assert(args.length == 3)
          val newArg0 = translate(args(0))
          val newArg1 = translate(args(1))
          val newArg2 = translate(args(2))
          val fresh = ctx.targetFormula.freshVar
          ctx.targetFormula.defineIte(fresh, newArg0, newArg1, newArg2)()
          fresh

        case otherKind => throw new Exception(s"in ${node.toString}: unexpected operator Z3_decl_kind = ${otherKind.toString}")
      }
    case otherKind => throw new Exception(s"in ${node.toString}: unexpected AST kind Z3_ast_kind = ${otherKind.toString}")
  }
}
