package util.analyzers

import com.microsoft.z3.enumerations.Z3_ast_kind._
import com.microsoft.z3.enumerations.Z3_decl_kind._
import com.microsoft.z3.{BoolExpr => Z3BoolExpr, Context => Z3Context, Expr => Z3Expr, Status => Z3Status}
import util.cardinality.{AtMost, CardinalityEncoder, Sequential}
import util.sat.Prop
import util.translators.SMTContext

import scala.collection.mutable
import scalaz.Memo


/**
  * Created by kdelmas on 04/11/16.
  */
trait Z3ExprTranslators {

  /**
    * Transform a at most n using the sequential encoding
    * @param kids the Boolean expressions addressed by the at most
    * @param n the maximum number of kids which must be true
    * @param SMTContext the Z3 context to create Z3 terms
    * @return Z3 Boolean expression
    */
  def atMostN(kids: List[Z3BoolExpr], n: Int, SMTContext: Z3Context): Z3BoolExpr = {
    if (n >= kids.size)
      SMTContext.mkBool(true)
    else if (n < 0)
      SMTContext.mkBool(false)
    else if (n == 0)
      SMTContext.mkAnd(kids.map(kid => SMTContext.mkNot(kid)): _*)
    else {
      val countEncoder = CardinalityEncoder(kids.size, AtMost, n, Sequential)
      val mkFresh = () => SMTContext.mkFreshConst("countVar", SMTContext.mkBoolSort()).asInstanceOf[Z3BoolExpr]
      val mkAnd = (in: List[Z3BoolExpr]) => SMTContext.mkAnd(in: _*)
      val mkOr = (in: List[Z3BoolExpr]) => SMTContext.mkOr(in: _*)
      val mkNot = (in: Z3BoolExpr) => SMTContext.mkNot(in)
      val translated = countEncoder.translate(kids, mkFresh, mkNot, mkAnd, mkOr)
      translated._2
    }
  }

  /**
    * Negate an expression and apply the de Morgan law on Z3 expression
    * @param expr the expression which is negated
    * @param context the Z3 context to create Z3 terms
    * @return Z3 Boolean expression
    */
  private def applyDeMorganLaw(expr: Z3BoolExpr, context: SMTContext): Z3BoolExpr = {
    expr.getASTKind match {
      case Z3_VAR_AST =>
        context.mkNot(expr)
      case Z3_APP_AST =>
        expr.getFuncDecl.getDeclKind match {
          case Z3_OP_NOT =>
            expr.getArgs.head.asInstanceOf[Z3BoolExpr]
          case Z3_OP_TRUE =>
            context.mkFalse()
          case Z3_OP_FALSE =>
            context.mkTrue()
          case Z3_OP_AND =>
            val args = expr.getArgs.map { case b: Z3BoolExpr => context.mkNot(b) }
            context.mkOr(args: _*)
          case Z3_OP_OR =>
            val args = expr.getArgs.map { case b: Z3BoolExpr => context.mkNot(b) }
            context.mkAnd(args: _*)
          case Z3_OP_ITE =>
            val arg0 = expr.getArgs.apply(0).asInstanceOf[Z3BoolExpr]
            val arg1 = expr.getArgs.apply(1).asInstanceOf[Z3BoolExpr]
            val arg2 = expr.getArgs.apply(2).asInstanceOf[Z3BoolExpr]
            context.mkIte(arg0, context.mkNot(arg1), context.mkNot(arg2))
          case _ => throw WrongBoolExpr("phi")
        }
      case _ => throw UnknownZ3AST(expr.toString)
    }
  }

  /**
    * Push the not operators to the leaf
    * @param expr the Boolean expression
    * @param context the context use to build the expression
    * @return resulting expression
    */
  private def pushNot(expr: Z3BoolExpr, context: SMTContext): Z3BoolExpr = {
    pushNotRec(expr)(mutable.Map.empty[Z3BoolExpr, Z3BoolExpr],context)
  }

  private def pushNotRec(expr: Z3BoolExpr)
                (implicit pushNotCache:mutable.Map[Z3BoolExpr, Z3BoolExpr],
                 context: SMTContext): Z3BoolExpr = expr.getASTKind match{

      case Z3_VAR_AST =>
        expr
      case Z3_APP_AST =>
        expr.getFuncDecl.getDeclKind match {
          case Z3_OP_NOT =>
            expr.getArgs.head.getASTKind match {
              case Z3_VAR_AST => expr
              case Z3_APP_AST =>
                pushNotCache.getOrElseUpdate(expr, {
                  val pushed = applyDeMorganLaw(expr.getArgs.head.asInstanceOf[Z3BoolExpr], context)
                  pushNotRec(pushed)
                })
              case _ => throw UnknownZ3AST(expr.toString)
            }
          case Z3_OP_TRUE =>
            expr
          case Z3_OP_FALSE =>
            expr
          case Z3_OP_AND =>
            pushNotCache.getOrElseUpdate(expr, {
              val args = expr.getArgs.map { case b: Z3BoolExpr => pushNotRec(b) }
              context.mkAnd(args: _*)
            })
          case Z3_OP_ITE =>
            pushNotCache.getOrElseUpdate(expr, {
              val args = expr.getArgs.map { case b: Z3BoolExpr => pushNotRec(b) }
              context.mkIte(args(0), args(1), args(2))
            })
          case Z3_OP_OR =>
            pushNotCache.getOrElseUpdate(expr, {
              val args = expr.getArgs.map { case b: Z3BoolExpr => pushNotRec(b) }
              context.mkOr(args: _*)
            })
          case _ => throw WrongBoolExpr("phi")
        }
      case _ =>
        throw UnknownZ3AST(expr.toString)
    }

  /**
    * Transform the expression in NNF form
    * @param expr the expression to transform
    * @param vars the list of leaves
    * @param context the context use to build the expression
    * @param breakITE break ite into i.t + \neg i.e
    * @return the NNF form
    */
  def toNNF(expr: Z3BoolExpr, vars: List[Z3BoolExpr], context: SMTContext, breakITE: Boolean = false): Z3BoolExpr = {

    val toNNFCache = collection.mutable.Map[Z3Expr, Z3BoolExpr]()

    def toNNFRec(expr: Z3Expr): Z3BoolExpr = {
      expr.getASTKind match {
        case Z3_VAR_AST =>
          toNNFCache.getOrElseUpdate(expr, vars.find(v => v == expr).getOrElse(throw UnknownEvent(expr.getFuncDecl.getName.toString)))
        case Z3_APP_AST =>
          expr.getFuncDecl.getDeclKind match {
            case Z3_OP_AND =>
              toNNFCache.getOrElseUpdate(expr, context.mkAnd(expr.getArgs.map(arg => toNNFRec(arg)): _*))
            case Z3_OP_OR =>
              toNNFCache.getOrElseUpdate(expr, context.mkOr(expr.getArgs.map(arg => toNNFRec(arg)): _*))
            case Z3_OP_NOT =>
              toNNFCache.getOrElseUpdate(expr, context.mkNot(expr.getArgs.map(arg => toNNFRec(arg)).head))
            case Z3_OP_IMPLIES =>
              toNNFCache.getOrElseUpdate(expr,
                context.mkOr(context.mkNot(toNNFRec(expr.getArgs.head)), toNNFRec(expr.getArgs.last)))
            case Z3_OP_IFF =>
              toNNFCache.getOrElseUpdate(expr, {
                val atoB = context.mkOr(context.mkNot(toNNFRec(expr.getArgs.head)), toNNFRec(expr.getArgs.last))
                val btoA = context.mkOr(context.mkNot(toNNFRec(expr.getArgs.last)), toNNFRec(expr.getArgs.head))
                context.mkAnd(atoB, btoA)
              })
            case Z3_OP_ITE =>
              toNNFCache.getOrElseUpdate(expr,
                if (!breakITE) {
                  val arg0 = expr.getArgs.apply(0)
                  val arg1 = expr.getArgs.apply(1)
                  val arg2 = expr.getArgs.apply(2)
                  val i = toNNFCache.getOrElseUpdate(arg0, toNNFRec(arg0))
                  val t = toNNFCache.getOrElseUpdate(arg1, toNNFRec(arg1))
                  val e = toNNFCache.getOrElseUpdate(arg2, toNNFRec(arg2))
                  context.mkIte(i, t, e)
                } else {
                  val i = expr.getArgs.apply(0).asInstanceOf[Z3BoolExpr]
                  val t = expr.getArgs.apply(1).asInstanceOf[Z3BoolExpr]
                  val e = expr.getArgs.apply(2).asInstanceOf[Z3BoolExpr]
                  val r = context.mkOr(context.mkAnd(i, t), context.mkAnd(context.mkNot(i), e))
                  toNNFRec(r)
                })
            case Z3_OP_TRUE =>
              context.mkTrue()
            case Z3_OP_FALSE =>
              context.mkFalse()
            case _ =>
              toNNFCache.getOrElseUpdate(expr, dnfOf(expr, vars, context))
          }
        case _ => throw UnknownZ3AST(expr.toString)
      }
    }

    pushNot(toNNFRec(expr), context)
  }

  /**
    * Transform an expression to its dnfFrom
    * @param expr Boolean expression
    * @param vars the list of leaves terms
    * @param context the context use to build the expression
    * @return the DNF form
    */
  def dnfOf(expr: Z3Expr, vars: List[Z3BoolExpr], context: Z3Context): Z3BoolExpr = {

    //transform leaves terms to uninterpreted constants
    val varToConst = vars.map(v => (v, context.mkBoolConst(s"const${v.getIndex}"))).unzip
    val constToVar = varToConst.zipped.toMap.map(p => p.swap)

    //substitute bound variables by the uninterpreted constants
    val subExpr = expr.substitute(varToConst._1.toArray[Z3Expr], varToConst._2.toArray[Z3Expr]).asInstanceOf[Z3BoolExpr]

    //storage for cubes
    val cubes = mutable.ListBuffer[Z3BoolExpr]()

    //instantiate solver and assert expression
    val s = context.mkSolver()
    s.add(subExpr)

    //while some models exists continue
    while (s.check() == Z3Status.SATISFIABLE) {

      //extract the model and transform it as a cube
      val cube = constToVar.map(p => (p._2, s.getModel.eval(p._1, false))).collect({
        case (v: Z3BoolExpr, m: Z3BoolExpr) if m.isTrue => v
        case (v: Z3BoolExpr, m: Z3BoolExpr) if m.isFalse => context.mkNot(v)
      })
      val model = constToVar.map(p => (p._1, s.getModel.eval(p._1, false))).collect({
        case (v: Z3BoolExpr, m: Z3BoolExpr) if m.isTrue => v
        case (v: Z3BoolExpr, m: Z3BoolExpr) if m.isFalse => context.mkNot(v)
      })

      //add the cube to found cubes
      cubes += context.mkAnd(cube.toList: _*)

      //block the model
      s.add(context.mkNot(context.mkAnd(model.toList: _*)))
    }

    s.dispose()

    //the DNF is the disjunction of cubes
    context.mkOr(cubes: _*)
  }

  private val Z3toPropCache = mutable.OpenHashMap[(Z3BoolExpr, Map[Z3BoolExpr, Prop.Expr]), Prop.Expr]()

  def translate(node: Z3BoolExpr, atoms: Map[Z3BoolExpr, Prop.Expr]): Prop.Expr = {
    Z3toPropCache.clear()
    translateRec((node, atoms))
  }

  /** Recursively translates the structure function to a CNF formula (does memoization of [[translate]]). */
  private val translateRec: ((Z3BoolExpr, Map[Z3BoolExpr, Prop.Expr])) => Prop.Expr = Memo.mutableMapMemo(Z3toPropCache) {
    (p: (Z3BoolExpr, Map[Z3BoolExpr, Prop.Expr])) => {
      val (node, atoms) = p
      node.getASTKind match {
        case Z3_VAR_AST =>
          atoms(node)
        case Z3_APP_AST =>
          val args = node.getArgs
          node.getFuncDecl.getDeclKind match {
            case Z3_OP_NOT =>
              assert(args.length == 1)
              Prop.Not(translateRec((args(0).asInstanceOf[Z3BoolExpr], atoms)))
            case Z3_OP_AND =>
              Prop.And(args.toList.map(a => translateRec((a.asInstanceOf[Z3BoolExpr], atoms))))
            case Z3_OP_OR =>
              Prop.Or(args.toList.map(a => translateRec((a.asInstanceOf[Z3BoolExpr], atoms))))
            case Z3_OP_IMPLIES =>
              assert(args.length == 2)
              Prop.Implies(translateRec((args(0).asInstanceOf[Z3BoolExpr], atoms)), translateRec((args(1).asInstanceOf[Z3BoolExpr], atoms)))
            case Z3_OP_IFF =>
              assert(args.length == 2)
              Prop.Iff(translateRec((args(0).asInstanceOf[Z3BoolExpr], atoms)), translateRec((args(1).asInstanceOf[Z3BoolExpr], atoms)))
            case Z3_OP_XOR =>
              assert(args.length == 2)
              Prop.Xor(translateRec((args(0).asInstanceOf[Z3BoolExpr], atoms)), translateRec((args(1).asInstanceOf[Z3BoolExpr], atoms)))
            case Z3_OP_ITE =>
              assert(args.length == 3)
              Prop.Ite(translateRec((args(0).asInstanceOf[Z3BoolExpr], atoms)), translateRec((args(1).asInstanceOf[Z3BoolExpr], atoms)), translateRec((args(2).asInstanceOf[Z3BoolExpr], atoms)))
            case otherKind => throw new Exception(s"in ${node.toString}: unexpected operator Z3_decl_kind = ${otherKind.toString}")
          }
        case otherKind => throw new Exception(s"in ${node.toString}: unexpected AST kind Z3_ast_kind = ${otherKind.toString}")
      }
    }
  }

}
