package util.analyzers

import analyzers._
import ast.dsl.{FlowConst, InstBoolIdent, InstFlowIdent}
import interpreter.TraceInterpreter
import spire.math.Rational
import ui.{ArchitectureDrawer, MDDDrawer, TraceDrawer}
import com.microsoft.z3.{ BoolExpr => Z3BoolExpr, BitVecNum}

package object analyzerTypes {

  /** concrete implementation of Real numbers used for probabilistic computations must
    * be set to spire Rational for testing and to Double for performances
    */
  type Real = Rational

  /**
    * Map type associating system events to its Z3 variables
    */
  type DSLEventToZ3Map = Map[InstBoolIdent, Z3BoolExpr]

  /**
    * Map type associating BDD variable index (representing events) to their probability
    */
  type ProbabilityMapType = Map[Int, Real]

  /**
    * Representation of a trace given by Z3 ie a model of system's flows
    */
  type Z3TraceType = List[(InstFlowIdent, BitVecNum)]

  /**
    * DSL representation of flow valuation ie set of basic flow constants
    */
  type FlowValue = Set[FlowConst]

  /**
    * implicit object for implicit conversion based on spire rational
    */
  implicit object RationalIsNumeric extends Numeric[Rational] {
    def plus(x: Rational, y: Rational): Rational = x + y

    def minus(x: Rational, y: Rational): Rational = x - y

    def times(x: Rational, y: Rational): Rational = x * y

    def negate(x: Rational): Rational = -x

    def fromInt(x: Int): Rational = Rational(x)

    def toInt(x: Rational): Int = x.intValue()

    def toLong(x: Rational): Long = x.longValue()

    def toFloat(x: Rational): Float = x.floatValue()

    def toDouble(x: Rational): Double = x.doubleValue()

    def compare(x: Rational, y: Rational): Int = x.compare(y)
  }

  /**
    * Type resulting for the mixin of all kind of configuration analysers
    */
  type FullAnalyzer = Parsing
    with SMTTranslation
    with AcceptableCardAnalyzer
    with MonotonyAnalyser
    with CutEnumeratorAll
    with ReliabilityComputer
    with AcceptableReliabilityAnalyzer
    with TraceInterpreter
    with SubstitutionAnalyzerSim
    with STMDDBuilder
    with DSEProblemSolverAll
    with SolutionAnalyzer
    with CardinalityComputer
    with SymbolicFailureRateComputer
    with ArchitectureDrawer
    with TraceDrawer
    with OpenPSAExporter
    with SMVExporter
    with MDDDrawer
    with HipHOPSExporter
    with CircuitGraphExporter

  /**
    * generic Real builder depending on Real implementation
    */
  object Real {
    def apply[A: Numeric](a: A): Real = {
      val ev = implicitly[Numeric[A]]
      ev.toDouble(a)
    }

    def apply(a: String): Real = {
      a.toDouble
    }
  }

}
