/** $Id$
  *
  * @author Rémi Delmas
  *
  */
package util.cardinality

import org.sat4j.core.VecInt
import org.sat4j.specs.ISolver
import org.sat4j.tools.{ClausalCardinalitiesDecorator, DimacsOutputSolver, encoding}

/** Encoding strategy. */
sealed abstract trait Encoding {
  def adapter: encoding.EncodingStrategyAdapter
}

case object Binary extends Encoding {
  def adapter = new encoding.Binary()
}

case object Sequential extends Encoding {
  def adapter = new encoding.Sequential()
}

/* These encodings are not clausal and make the DimacsOutputSolver complain because they are not strictly clausal so we do not use them. */

//case object Binomial  extends Encoding { def adapter = new encoding.Binomial() }
//case object Commander extends Encoding { def adapter = new encoding.Commander() }
//case object Ladder    extends Encoding { def adapter = new encoding.Ladder() }
//case object Product   extends Encoding { def adapter = new encoding.Product() }


/** Relational operators supported in the constraints. */
sealed abstract trait RelOp

case object AtLeast extends RelOp

case object AtMost extends RelOp

case object Exactly extends RelOp

/** Adds a string conversion method to a java.io.ByteArrayOutputStream. */
class MyByteArrayOutputStream() extends java.io.ByteArrayOutputStream() {
  def asString = new String(this.toByteArray, java.nio.charset.StandardCharsets.UTF_8)
}


/** Parses the dimacs format generated by a sat4j DimacsOutputSolver. */
object DimacsParser extends scala.util.parsing.combinator.RegexParsers {
  def apply(input: String): List[List[Int]] = parseAll(dimacs, input) match {
    case Success(result, _) => result
    case failure: NoSuccess => scala.sys.error(failure.toString)
  }

  lazy val zero = """0""".r
  lazy val header = """p""" ~ """cnf""" ~ posInt ~ """X+""".r
  lazy val posInt: Parser[Int] = """[1-9][0-9]*""".r ^^ { case s => s.toInt }
  lazy val negInt: Parser[Int] = ("-" ~> posInt) ^^ { case s => -s }
  lazy val literal: Parser[Int] = (negInt | posInt)
  lazy val clause: Parser[List[Int]] = rep1(literal) <~ zero ^^ { case lits => lits.toList }
  lazy val dimacs: Parser[List[List[Int]]] = header ~> rep(clause)
}

/** This class is used as a template to instanciate cardinality constraints. */
case class CardinalityEncoder(popSize: Int, relop: RelOp, k: Int, encoding: Encoding) {

  private lazy val outStream = new MyByteArrayOutputStream()

  /** The core sat4j solver instance with no solving and dimacs output. */
  private lazy val coreSolver: DimacsOutputSolver = new DimacsOutputSolver(new java.io.PrintWriter(outStream, true))
  coreSolver.newVar(popSize) // reserve indices of our population variables
  coreSolver.setDBSimplificationAllowed(false) // prevent simplifications
  private lazy val popVec: VecInt = new VecInt((1 to popSize).toArray)

  /** The core sat4j decorated with a cardinality encoder. */
  private lazy val cardSolver: ClausalCardinalitiesDecorator[ISolver] = new ClausalCardinalitiesDecorator(coreSolver, encoding.adapter)

  /** Push the template constraint in the solver. */
  relop match {
    case AtLeast => cardSolver.addAtLeast(popVec, k)
    case AtMost => cardSolver.addAtMost(popVec, k)
    case Exactly => cardSolver.addExactly(popVec, k)
  }

  /** The clauses generated from the constraint (serves as a template for the translate method). */
  private lazy val templateClauses = DimacsParser(outStream.asString)

  /** Instanciates the template constraint for the given population.
    *
    * @returns the list of fresh variables created by the instanciation/clausification and the expression of the constraint.
    **/
  def translate[BoolExpr, BoolIdent <: BoolExpr](

                                                  /** The population to be counted. */
                                                  population: List[BoolExpr],

                                                  /** User-provided function to create fresh variables in the user land. */
                                                  mkFresh: () => BoolIdent,

                                                  /** User-provided function to create negations in the user land. */
                                                  mkNot: BoolExpr => BoolExpr,

                                                  /** User-provided function to create conjunctions in the user land. */
                                                  mkAnd: (List[BoolExpr] => BoolExpr),

                                                  /** User-provided function to create disjuctions in the user land. */
                                                  mkOr: (List[BoolExpr] => BoolExpr)

                                                ): (List[BoolIdent], BoolExpr) = {

    assert(population.size == popSize)

    val cache = scala.collection.mutable.Map[Int, BoolExpr]()

    // add a 1 offset because sat4j variables start at 1
    population.zipWithIndex.foreach(x => cache.update(x._2 + 1, x._1))

    var newConsts: List[BoolIdent] = Nil

    def newConst: BoolIdent = {
      val res = mkFresh()
      newConsts ::= res
      res
    }

    def mkClause(l: List[Int]): BoolExpr = mkOr(
      l map {
        i => {
          if (i < 0)
            mkNot(
              cache.getOrElseUpdate(-i, newConst)
            )
          else
            cache.getOrElseUpdate(i, newConst)
        }
      }
    )

    val result = mkAnd(templateClauses.map(mkClause))
    (newConsts, result)
  }

}


