/** $Id$
  *
  * @author $Author$
  * @date $Date$
  **/
package util.graph

/**
  * This package provides dependency graph data structure allowing
  * compute strongly connected components and topological
  * orderings. The nodes of the graph are labelled with instances of an
  * abstract type A.
  */

import java.io._

import scala.collection.mutable.ArrayStack


/** Class representing a dependency graph. */
trait DepGraphU[A] {

  /** Nodes of the graph. */
  final var nodes: ArrayStack[Node] = new ArrayStack()

  /** Class representing a node in a graph, labeled with an instance of class A. */
  final class Node(val data: A) {

    final var uid = -1
    final var index: Option[Int] = None
    final var lowlink: Option[Int] = None
    final var instack: Boolean = false
    final var sorted: Boolean = false

    /** Predecessors of the node (other nodes that must come before this one in the topological ordering). */
    final var predecessors: List[Node] = Nil

    /** Adds a node as dependency to this node. */
    final def addPred(pred: Node): Node = {
      this.predecessors ::= pred
      this
    }

    /** Successor in the topological ordering (None if no order was computed yet). */
    final var successor: Option[Node] = None

    /** Resets the mutable attributes of the node used in traversal algorithms. */
    final def reset(): Unit = {
      successor = None
      index = None
      lowlink = None
      instack = false
      sorted = false
    }

    nodes += this
  }


  /** Iterator on the nodes of the graph (in arbitrary order). */
  final def iterator = this.nodes.iterator

  /** Root node of the topological ordering of the graph */
  final var root: Option[Node] = None

  /** Iterator on the nodes of the graph (in topological order if root is defined). */
  final def inOrder: Iterator[Node] = this.root match {

    case None => {
      throw new Exception("The graph is either empty or not sorted yet.")
    }

    case Some(n) => {

      /* Create a new iterator object */
      new Iterator[Node] {
        var current: Option[Node] = root

        def hasNext = current.isDefined

        def getCurrent: Node = current.get

        def next: Node = {
          if (current.isDefined) {
            val prev = current.get;
            current = prev.successor;
            prev
          } else {
            throw new NoSuchElementException("empty iterator")
          }
        }
      }
    }
  }

  /** Computes the connected components of the graph using the recursive Tarjan algorithm.
    * The method returns a list of list of nodes representing the strongly connected components of the graph.
    * If each scc has only one node in it, then the graph is acyclic.
    * */
  final def sccs: List[List[Node]] = {

    // The list of strongly connected components.
    var _sccs: List[List[Node]] = Nil

    // Stack of traversed nodes.
    val s: ArrayStack[Node] = new ArrayStack()

    // Index of the current node.
    var _index: Int = 0

    // Actions performed when visiting a node during the graph traversal.
    def visit(n: Node): Unit = {
      if (n.index == None) {
        n.index = Some(_index)
        n.lowlink = Some(_index)
        _index += 1
        s += n
        n.instack = true
        for (succ <- n.predecessors) {
          if (succ.index == None) {
            visit(succ)
            n.lowlink = Some(n.lowlink.get min succ.lowlink.get)
          } else if (succ.instack) {
            n.lowlink = Some(n.lowlink.get min succ.index.get)
          }
        }
        if (n.lowlink.get == n.index.get) {
          var nn: Node = null
          var c = List[Node]()
          do {
            nn = s.pop
            nn.instack = false
            c ::= nn
          } while (nn != n)
          _sccs ::= c
        }
      }
    }

    // initialise all nodes
    nodes foreach {
      _.reset
    }

    // traverse all nodes
    nodes foreach {
      visit
    }

    // return the computed components
    _sccs
  }

  /** Recursive DFS topological sorting algorithm.
    *
    * In this order, a node comes after all its predecessors.
    *
    * The graph must be acyclic otherwise the algorithm will compute a wrong ordering.
    *
    * After sucessfull execution, the nodes of the graph can be traversed in topological order using the inOrder method.
    *
    */
  final def topsort(): Iterator[Node] = {

    var res: Option[Node] = None
    var prev: Option[Node] = None

    /* recursive part of the algo */
    def visit(n: Node): Unit = {
      if (!n.instack) {
        n.instack = true
        n.predecessors.foreach(visit)
        // memorize first root
        if (!res.isDefined) {
          res = Some(n)
        }
        // link current node with the previous node
        if (prev.isDefined) {
          prev.get.successor = Some(n)
          prev = Some(n)
        } else {
          prev = Some(n)
        }
      }
    }

    // initialise nodes
    nodes foreach {
      _.instack = false
    }

    // traverse from each node
    nodes foreach {
      visit
    }

    // assign root node in graph
    root = res
    this.inOrder

  }

  /**
    * Prints a dot representation of the graph using  printing functions for nodes and edges provided as arguments
    **/
  final def toDot(filename: String, withOrderEdges: Boolean, nodeLabelFun: Option[A => String], edgeLabelFun: Option[(A, A) => String]): Unit = {

    // node to string function
    def n2s(n: Node): String = {
      "\"%s%d\"".
        format(
          nodeLabelFun match {
            case Some(f) => f(n.data) + "\\n"
            case None => ""
          }, n.uid)
    }

    val out = new BufferedWriter(new FileWriter(filename))
    out.write("digraph Expr{\nrankdir=TB;\n")

    nodes.foreach(
      n => {
        out.write(n2s(n) +
          (root match {
            case Some(r) => if (r == n) "[shape=doublecircle]" else ""
            case None => ""
          }) + ";\n")
        n.predecessors.foreach(d => out.write("%s -> %s;\n".format(n2s(n), n2s(d))))
      }
    )

    if (withOrderEdges) {
      this.inOrder.reduceLeft {
        (n, s) => {
          out.write("%s -> %s [style=dotted];\n".format(n2s(n), n2s(s)))
          s
        }
      }
    }
    out.write("}\n")
    out.close
  }
}
