package util.io

import java.io.{File, FileWriter}

import scala.reflect.io.Path

/**
  * Created by kdelmas on 21/02/17.
  */
object FileManager {


  private val tempDirectory = {
    val temp = new File("temp")
    if (temp.exists())
      temp
    else if (temp.mkdir()) {
      temp
    } else {
      throw new Exception(s"cannot create $temp directory")
    }
  }

  private val testDirectory = {
    val test = new File("testResult")
    if (test.exists())
      test
    else if (test.mkdir()) {
      test
    } else {
      throw new Exception(s"cannot create $test directory")
    }
  }

  private val benchDirectory = {
    val test = new File("bench")
    if (!test.exists())
      throw new Exception("unable to locate bench folder")
    else
      test
  }

  private val benchExpDirectory = {
    val test = new File("bench/exploration")
    if (!test.exists())
      throw new Exception("unable to locate bench folder")
    else
      test
  }

  private val benchCutDirectory = {
    val test = new File("bench/cutsetEnumeration")
    if (!test.exists())
      throw new Exception("unable to locate bench folder")
    else
      test
  }

  def addFileToTemp(s: String): File = {
    new File(tempDirectory, s)
  }

  def addFileToTest(s: String): File = {
    new File(testDirectory, s)
  }

  def cleanTest(): Unit = {
    testDirectory.listFiles().foreach(f => Path(f.getAbsolutePath).deleteRecursively())
  }

  def cleanTemp(): Unit = {
    tempDirectory.listFiles().foreach(f => Path(f.getAbsolutePath).deleteRecursively())
  }

  def locateInBench(name: String) : Option[File] =
    recursiveLocateFirstFile(benchDirectory, (f: File) => f.getName == name)

  def locateInExpBench(name: String) : Option[File] =
    recursiveLocateFirstFile(benchExpDirectory, (f: File) => f.getName == name)

  def locateInCutBench(name: String) : Option[File] =
    recursiveLocateFirstFile(benchCutDirectory, (f: File) => f.getName == name)

  def recursiveLocateFirstFile(dir: File, filter: (File) => Boolean): Option[File] = {
    if (dir.exists()) {
      if (dir.isDirectory) {
        val these = dir.listFiles
        (for (r <- these.find(filter)) yield {
          r
        }) orElse {
          these.filter(_.isDirectory).flatMap(x => recursiveLocateFirstFile(x, filter)).toList match {
            case Nil => None
            case h :: t => Some(h)
          }
        }
      } else {
        Some(dir).filter(filter)
      }
    } else
      None
  }

  def locateFirstFile(dir: File, filter: (File) => Boolean): Option[File] = {
    if (dir.exists()) {
      if (dir.isDirectory) {
        val these = dir.listFiles
        for (r <- these.find(filter)) yield {
          r
        }
      } else {
        Some(dir).filter(filter)
      }
    } else {
      None
    }
  }

  def locateFirstFileInTemp(filter: (File) => Boolean): Option[File] = {
    locateFirstFile(tempDirectory, filter)
  }

  def locateFirstFileInEnv(envVarName: String, filter: (File) => Boolean): Option[File] = {
    val zero: Option[File] = None
    System.getenv(envVarName).split(System.getProperty("path.separator")).foldLeft(zero)((acc, p) => {
      acc match {
        case Some(f) => acc
        case None => locateFirstFile(new File(p), filter)
      }
    })
  }

  def recursiveLocateFirstFileInTemp(filter: (File) => Boolean): Option[File] = {
    recursiveLocateFirstFile(tempDirectory, filter)
  }

  sealed trait Bench {
    val path: Path
    val name: String
    val conf: String
  }

  sealed trait ExplorationBench extends Bench

  sealed trait CutEnumerationBench extends Bench

  object RosaceExp extends ExplorationBench {
    val path: Path = "bench/exploration/KCR/Rosace.kcr"
    val name = "Rosace"
    val conf = "Conf"
    assert(path.exists)
  }

  object HBSExp extends ExplorationBench {
    val path: Path = "bench/exploration/KCR/HBS.kcr"
    val name = "HBS"
    val conf = "Conf"
    assert(path.exists)
  }

  object FuelExp extends ExplorationBench {
    val path: Path = "bench/exploration/KCR/fuel.kcr"
    val name = "Fuel"
    val conf = "Conf"
    assert(path.exists)
  }

  object QuadExp extends ExplorationBench {
    val path: Path = "bench/exploration/KCR/Quadcopter.kcr"
    val name = "Quadcopter"
    val conf = "Conf"
    assert(path.exists)
  }

  object RosaceCut extends CutEnumerationBench {
    val path: Path = "bench/cutsetEnumeration/KCR/Rosace.kcr"
    val name = "Rosace"
    val conf = "Conf"
    assert(path.exists)
  }

  object HBSCut extends CutEnumerationBench {
    val path: Path = "bench/cutsetEnumeration/KCR/HBS.kcr"
    val name = "HBS"
    val conf = "Conf"
    assert(path.exists)
  }

  object FuelCut extends CutEnumerationBench {
    val path: Path = "bench/cutsetEnumeration/KCR/fuel.kcr"
    val name = "Fuel"
    val conf = "Conf"
    assert(path.exists)
  }

  object QuadCut extends CutEnumerationBench {
    val path: Path = "bench/cutsetEnumeration/KCR/Quadcopter.kcr"
    val name = "Quadcopter"
    val conf = "Conf"
    assert(path.exists)
  }

  object GridCut extends CutEnumerationBench {
    val path: Path = "bench/cutsetEnumeration/KCR/grid4x60.kcr"
    val name = "Grid"
    val conf = "Conf"
    assert(path.exists)
  }

  val explorationBench = List(RosaceExp, HBSExp, FuelExp, QuadExp)
  val cutEnumerationBench = List(RosaceCut, HBSCut, FuelCut, QuadCut, GridCut)
}
