/** $Id: LoadDSLFiles.scala 1430 2017-10-17 09:16:44Z kdelmas $
  *
  * @author $Author: kdelmas $
  * @date $Date: 2017-10-17 11:16:44 +0200 (mar. 17 oct. 2017) $
  **/
package util.io

import ast.dsl.{ConfigIdent, ConfigurationDeclaration, PackageDeclaration, RootPackage}
import namespace.dsl._
import parsers.dsl.{ParseFiles, ParserError}

import scalaz._

/** Loads a list of kcr files and returns the root package. If the files contain any error the main program is aborted. */
object LoadDSLFiles {

  def apply(files: List[String], doThrow: Boolean = false): PackageDeclaration = {
    val res = for {
      declMap <- ParseFiles(files)
      prog = RootPackage(declMap.values.toList.flatten)
      namespaceExceptions <- BuildNamespace(prog)
      compCycleExceptions <- CheckCompCycles(prog)
      configCallGraphsExceptions <- BuildConfigCallGraphs(prog)
      typeCheckingExceptions <- CheckTypes(prog)
    } yield {
      (namespaceExceptions ::: compCycleExceptions ::: configCallGraphsExceptions ::: typeCheckingExceptions, prog)
    }

    res match {
      case \/-((warnings, prog)) => {
        warnings foreach {
          println
        }
        prog
      }
      case -\/(error) => {
        if (doThrow) {
          throw new Exception(error.toString)
        } else {
          println(error.toString)
          sys.exit(1)
        }
      }
    }
  }

  def applyOrGetErrors(files: List[String]): (Option[PackageDeclaration], Option[String]) = {
    try {
      val res = for {
        declMap <- ParseFiles(files)
        prog = RootPackage(declMap.values.toList.flatten)
        namespaceExceptions <- BuildNamespace(prog)
        compCycleExceptions <- CheckCompCycles(prog)
        configCallGraphsExceptions <- BuildConfigCallGraphs(prog)
        typeCheckingExceptions <- CheckTypes(prog)
      } yield {
        (namespaceExceptions ::: compCycleExceptions ::: configCallGraphsExceptions ::: typeCheckingExceptions, prog)
      }
      res match {
        case \/-((warnings, prog)) =>
          val textWarnings = if (warnings.nonEmpty) Some(warnings.mkString("\n")) else None
          (Some(prog), textWarnings)
        case -\/(error) =>
          error match {
            case l: Traversable[_] =>
              (None,
                Some(l.collect({
                  case e: ParserError => e.e.getMessage
                  case e => e.toString
                }).mkString("\n")))
            case e => (None, Some(e.toString))
          }
      }
    } catch {
      case e: Exception =>
        (None, Some(e.toString))
    }
  }

  def applyAndInlineOrGetErrors(files: List[String], confId: ConfigIdent): (Option[ConfigurationDeclaration], Option[String]) = {
    try {
      val res = for {
        declMap <- ParseFiles(files)
        prog = RootPackage(declMap.values.toList.flatten)
        namespaceExceptions <- BuildNamespace(prog)
        compCycleExceptions <- CheckCompCycles(prog)
        configCallGraphsExceptions <- BuildConfigCallGraphs(prog)
        typeCheckingExceptions <- CheckTypes(prog)
      } yield {
        (namespaceExceptions ::: compCycleExceptions ::: configCallGraphsExceptions ::: typeCheckingExceptions, prog)
      }
      res match {
        case \/-((warnings, prog)) =>
          val textWarnings = if (warnings.nonEmpty) Some(warnings.mkString("\n")) else None
          val newConf = for (conf <- Some(prog.namespace.locate(confId));
                             inliner <- Some(new inlining.dsl.Inliner(conf));
                             newProg <- inliner.inline()
          ) yield {
            newProg.namespace.locate(ConfigIdent(conf.ident.baseId))
          }
          (newConf, textWarnings)
        case -\/(error) =>
          error match {
            case l: Traversable[_] =>
              (None,
                Some(l.collect({
                  case e: ParserError => e.e.getMessage
                  case e => e.toString
                }).mkString("\n")))
            case e => (None, Some(e.toString))
          }
      }
    } catch {
      case e: Exception =>
        (None, Some(e.toString))
    }
  }
}
