package util.io

import analyzers._
import util.analyzers._


import scala.sys.process.{Process, ProcessBuilder}

/**
  * Created by kevin on 17/10/17.
  */
object SystemInfo {

  sealed trait OS

  object Linux extends OS

  object Windows extends OS

  object MAC extends OS

  object Unknown extends OS

  val userOs: OS = {
    System.getProperty("os.name") match {
      case null => Unknown
      case s if s.toLowerCase.contains("windows") => Windows
      case s if s.toLowerCase.contains("linux") => Linux
      case s if s.toLowerCase.contains("mac") => MAC
      case _ => Unknown
    }
  }

  private val nbTotalCores = Runtime.getRuntime.availableProcessors()

  val usableCores:Int= if(nbTotalCores>=2) nbTotalCores/2 else nbTotalCores

  def runUnixCmd(cmd:String):ProcessBuilder = userOs match {
    case Windows =>
      Process(s"${CygwinInterface.toolPath}/bin/bash.exe -c $cmd")
    case _ =>
      Process(cmd)
  }

  def findAvailableCutComputationMethods(): List[CutComputationMode] = {
    val all = List(HipHOPSCutComputation,
      XSAPCutComputation,
      XFTACutComputation,
      IterativeSATCutComputation,
      PBCutComputation,
      BDDCutComputation
    )
    val available = all.filter(m => {
      m match {
        case HipHOPSCutComputation =>
          HipHOPSInterface.toolPathOption.isDefined && WineInterface.toolPathOption.isDefined
        case XFTACutComputation =>
          XFTAInterface.toolPathOption.isDefined
        case XSAPCutComputation =>
          XSAPInterface.toolPathOption.isDefined
        case IterativeSATCutComputation =>
          ClaspInterface.toolPathOption.isDefined
        case PBCutComputation =>
          SharpCDCLInterface.toolPathOption.isDefined && PBEncoderInterface.toolPathOption.isDefined
        case BDDCutComputation => true
      }
    })
    all.foreach(m => {
      if (!available.contains(m)) {
        m match {
          case IterativeSATCutComputation =>
            println(s"[WARNING] the $m cutset enumeration method is not available, need path to external executable ${ClaspInterface.mustContain}")
          case PBCutComputation =>
            if(SharpCDCLInterface.toolPathOption.isEmpty)
              println(s"[WARNING] the $m cutset enumeration method is not available, need path to external executable ${SharpCDCLInterface.mustContain}")
            if(PBEncoderInterface.toolPathOption.isEmpty)
              println(s"[WARNING] the $m cutset enumeration method is not available, need path to external executable ${PBEncoderInterface.mustContain}")
          case other =>
            println(s"[WARNING] the $other exploration method is not available, path to external tool not provided")
        }
      }
    })
    available
  }

  def findAvailableExpMethods(): List[DesignSpaceExplorationMethod] = {
    val all = List(SMTExploration, GeneticExploration)
    val available = all.filter(m => {
      m match {
        case GeneticExploration =>
          userOs match {
            case Windows => HipHOPSInterface.toolPathOption.isDefined
            case _ => HipHOPSInterface.toolPathOption.isDefined && WineInterface.toolPathOption.isDefined
          }
        case SMTExploration => true
      }
    })
    all.foreach(m => {
      if (!available.contains(m))
        println(s"[WARNING] the $m exploration method is not available, path to external tool not provided")
    })
    available
  }

  def stopToolFor(cutComputationMode: CutComputationMode): Unit = {
    cutComputationMode match {
      case HipHOPSCutComputation =>
        HipHOPSInterface.destroyAll()
      case XSAPCutComputation =>
        XSAPInterface.destroyAll()
        XSAPInterface.killXSAP()
      case XFTACutComputation =>
        XFTAInterface.destroyAll()
      case IterativeSATCutComputation =>
        ClaspInterface.destroyAll()
      case PBCutComputation =>
        PBEncoderInterface.destroyAll()
        SharpCDCLInterface.destroyAll()
      case _ =>
    }
  }

  def stopToolFor(explorationMethod: DesignSpaceExplorationMethod): Unit = {
    explorationMethod match {
      case GeneticExploration =>
        HipHOPSInterface.destroyAll()
        HipHOPSInterface.killInternetExplorer()
      case _ =>
    }
  }
}
