package util.sat

import scala.collection.immutable.{Map => IMap, Set => ISet}
import scala.collection.mutable.{HashSet => MSet, OpenHashMap => MMap}

/**
  * Created by rdelmas on 25/05/16.
  *
  * Abstract syntax trees for various forms of propositional logic and associated translation functions.
  *
  *
  */

/** A generic hashing function for lists of objects. */
object foldHash {
  def apply(args: Any*): Int = {
    args.foldLeft[Int](1)((acc, x) => 41 * acc + x.hashCode)
  }

  def fromIterable(args: Iterable[Any]): Int = {
    args.foldLeft[Int](1)((acc, x) => 41 * acc + x.hashCode)
  }
}

/** Propositional logic. */
object Prop {

  /** Abstract base type for all Prop expressions. */
  sealed trait Expr {
    val kids: List[Expr]
    lazy val atoms: ISet[Atom] = kids.foldLeft(ISet[Atom]())((acc, x) => acc union x.atoms)
    override lazy val hashCode = foldHash.fromIterable(kids)
  }

  /** Counter for number of atoms created in [[Prop]]. */
  private var _nofAtoms: Int = 0

  /** Creates a fresh uniquely identified atom. */
  def freshAtom() = {
    _nofAtoms += 1;
    Atom(_nofAtoms)
  }

  /** Free proposition. */
  case class Atom private(id: Int) extends Expr {
    assert(id > 0)
    val kids = Nil
    override lazy val atoms = Set(this)
    override lazy val hashCode = foldHash(id)
  }

  /** Logical negation. */
  case class Not(kid1: Expr) extends Expr {
    val kids = List(kid1)
  }

  /** Logical conjunction. */
  case class And(kids: List[Expr]) extends Expr

  /** Logical disjunction. */
  case class Or(kids: List[Expr]) extends Expr

  /** Logical implication. */
  case class Implies(kid1: Expr, kid2: Expr) extends Expr {
    val kids = List(kid1, kid2)
  }

  /** Logical equivalence. */
  case class Iff(kid1: Expr, kid2: Expr) extends Expr {
    val kids = List(kid1, kid2)
  }

  /** Logical exlusive disjunction. */
  case class Xor(kid1: Expr, kid2: Expr) extends Expr {
    val kids = List(kid1, kid2)
  }

  /** Logical conditional selection. */
  case class Ite(kid1: Expr, kid2: Expr, kid3: Expr) extends Expr {
    val kids = List(kid1, kid2, kid3)
  }

}


/** Propositional logic solely with negation, conjunction, disjunction and if-then-else. */
object Core {

  /** Abstract base types for all Core expressions. */
  sealed trait Expr {
    val kids: List[Expr]
    lazy val atoms: ISet[Atom] = kids.foldLeft(ISet[Atom]())((acc, x) => acc union x.atoms)
    override lazy val hashCode = foldHash.fromIterable(kids)
  }

  /** Counter for number of atoms created in core. */
  private var _nofAtoms: Int = 0

  /** Creates a fresh uniquely identified atom. */
  def freshAtom() = {
    _nofAtoms += 1;
    Atom(_nofAtoms)
  }

  /** Free proposition. */
  case class Atom private(id: Int) extends Expr {
    assert(id > 0)
    lazy val kids = Nil
    override lazy val atoms = Set(this)
    override lazy val hashCode = foldHash(id)
  }

  /** Logical negation. */
  case class Not(kid1: Expr) extends Expr {
    val kids = List(kid1)
  }

  /** Logical conjunction. */
  case class And(kids: List[Expr]) extends Expr

  /** Logical disjunction. */
  case class Or(kids: List[Expr]) extends Expr

  /** Logical conditional selection. */
  case class Ite(kid1: Expr, kid2: Expr, kid3: Expr) extends Expr {
    val kids = List(kid1, kid2, kid3)
  }

  /** Translator from Prop to Core. */
  class ToCore extends (Prop.Expr => Expr) {
    private val _atomCache: MMap[Prop.Atom, Atom] = new MMap()
    private val _exprCache: MMap[Prop.Expr, Expr] = new MMap()

    /** Returns the translation cache over atoms. */
    def atomCacheFunc: IMap[Prop.Atom, Atom] = IMap() ++ _atomCache

    /** Returns the translation cache over all subexpressions. */
    def exprCacheFunc: IMap[Prop.Expr, Expr] = IMap() ++ _exprCache

    /** Translates the given Prop expression to an equivalent Core expression. */
    def apply(e: Prop.Expr): Expr = _exprCache.getOrElseUpdate(e, applyRec(e))

    /** Recursive traversal and translation of the expression. */
    private def applyRec(e: Prop.Expr): Expr = e match {
      case a@Prop.Atom(id) => {
        val newA = freshAtom(); _atomCache.update(a, newA); newA
      }
      case Prop.Not(k1) => Not(apply(k1))
      case Prop.And(kids) => And(kids map apply)
      case Prop.Or(kids) => Or(kids map apply)
      case Prop.Implies(k1, k2) => apply(Prop.Or(Prop.Not(k1) :: k2 :: Nil))
      case Prop.Iff(k1, k2) => apply(Prop.And(Prop.Implies(k1, k2) :: Prop.Implies(k2, k1) :: Nil))
      case Prop.Xor(k1, k2) => apply(Prop.Iff(k1, Prop.Not(k2)))
      case Prop.Ite(k1, k2, k3) => Ite(apply(k1), apply(k2), apply(k3))
    }
  }

}

/** Propositional logic in negation-normal-form. */
object NNF {

  /** Abstract base type for all NNF expressions. */
  sealed trait Expr {

    /** Height of the expression, by convention literals have height 0. */
    lazy val height: Int = (kids.foldLeft(0)((acc, kid) => math.max(acc, kid.height)) + 1)

    /** List of kids of the expression. */
    val kids: List[NNF.Expr]

    /** Number of kids. */
    lazy val nofKids: Int = kids.size

    override lazy val hashCode: Int = foldHash(kids: _*)

    /** Set of atoms found in the cone of influence of the expression. */
    lazy val atoms: ISet[Atom] = kids.foldLeft(ISet[Atom]())((current, x) => current union x.atoms)

    /** Set of literals found in the cone of influence of the expression. */
    def literals: ISet[Literal] = kids.foldLeft(ISet[Literal]())((current, x) => current union x.literals)

    /** True iff the atom sets of kids are mutually disjoint,
      * ie iff they are syntactically independent,
      * and hence semantically independent.
      * */
    lazy val pairwiseIndepKids: Boolean = {
      val res = kids.foldLeft(true)(
        (soFar, kid) => kids.foldLeft(soFar)(
          (soFar, otherKid) => if (otherKid != kid) soFar & (kid.atoms intersect otherKid.atoms).isEmpty else soFar
        )
      )
      res
    }

    /** True iff the expression has no duplicate kid. */
    lazy val sameKidTwice: Boolean = {
      val res = kids.foldLeft((true, ISet[Expr]())) {
        (acc, kid) => {
          val (soFar, allKids) = acc
          (soFar & !allKids.contains(kid), allKids + kid)
        }
      }
      res._1
    }
  }

  /** Abstract type for literals, ie [[Atom]] or the [[Not]] of an atom. */
  sealed trait Literal extends Expr {
    override lazy val height = 0
  }

  /** Counter for number of atoms created in NNF. */
  private var _nofAtoms: Int = 0

  /** Creates a fresh uniquely identified atom. */
  def freshAtom() = {
    _nofAtoms += 1;
    Atom(_nofAtoms)
  }

  /** An atom, ie a free proposition. */
  case class Atom(id: Int) extends Literal {
    assert(id > 0)
    val kids = Nil
    override lazy val atoms = Set(this)

    override def literals = Set(this)

    override lazy val hashCode = foldHash(id)
  }

  /** Negation (only over an atom since we're in Negation Normal Form). */
  case class Not(kid1: Atom) extends Literal {
    val kids = List(kid1)
  }

  /** Conjunction of expressions. */
  case class And(kids: List[Expr]) extends Expr

  /** Disjunction. */
  case class Or(kids: List[Expr]) extends Expr

  /** Conditional selection. */
  case class Ite(kid1: Expr, kid2: Expr, kid3: Expr) extends Expr {
    val kids = List(kid1, kid2, kid3)
  }

  /** Translator from Core to NNF. */
  class ToNNF extends (Core.Expr => Expr) {
    private val _exprCache: MMap[Core.Expr, Expr] = new MMap()
    private val _atomCache: MMap[Core.Atom, Atom] = new MMap()

    def atomCacheFunc: IMap[Core.Atom, Atom] = IMap() ++ _atomCache

    def exprCacheFunc: IMap[Core.Expr, Expr] = IMap() ++ _exprCache

    def apply(e: Core.Expr): Expr = _exprCache.getOrElseUpdate(e, applyRec(e))

    private def applyRec(e: Core.Expr): Expr = e match {
      case Core.Not(a@Core.Atom(id)) => {
        apply(a); Not(_atomCache(a))
      }
      case Core.Not(Core.Not(k1)) => apply(k1)
      case Core.Not(Core.And(kids)) => apply(Core.Or(kids.map(k => Core.Not(k))))
      case Core.Not(Core.Or(kids)) => apply(Core.And(kids.map(k => Core.Not(k))))
      case Core.Not(Core.Ite(k1, k2, k3)) => apply(Core.Ite(k1, Core.Not(k2), Core.Not(k3)))
      case a@Core.Atom(id) => {
        val newA = freshAtom(); _atomCache.update(a, newA); newA
      }
      case Core.And(kids) => And(kids map apply)
      case Core.Or(kids) => Or(kids map apply)
      case Core.Ite(k1, k2, k3) => Ite(apply(k1), apply(k2), apply(k3))
    }
  }


  /** Builds the dependency graph of an expression (frome leaves to top node). */
  def genDepGraph(e: Expr): util.graph.DepGraphE {type A = Expr} = {
    val depGraph = new util.graph.DepGraphE {
      type A = Expr
    }
    val exprToNode = new MMap[Expr, depGraph.Node]()

    def buildNode(e: Expr): depGraph.Node = {
      exprToNode.getOrElseUpdate(e, buildNodeRec(e))
    }

    def buildNodeRec(e: Expr): depGraph.Node = {
      val n = depGraph.Node(e)
      e.kids.foreach {
        kid => {
          val kidNode = buildNode(kid)
          n.addPred(kidNode)
        }
      }
      n
    }

    buildNode(e)
    depGraph
  }

  /** Builds the inverse dependency graph of an expression (from top node to leaves). */
  def genInverseDepGraph(e: Expr): util.graph.DepGraphE {type A = Expr} = {
    val depGraph = new util.graph.DepGraphE {
      type A = Expr
    }
    val exprToNode = new MMap[Expr, depGraph.Node]()

    def buildNode(e: Expr, parent: Option[depGraph.Node]): depGraph.Node = {
      val node = exprToNode.getOrElseUpdate(e, buildNodeRec(e, parent))
      if (parent.isDefined) node.addPred(parent.get)
      node
    }

    def buildNodeRec(e: Expr, parent: Option[depGraph.Node]): depGraph.Node = {
      val n = depGraph.Node(e)
      e.kids.foreach { kid => buildNode(kid, Some(n)) }
      n
    }

    buildNode(e, None)
    depGraph
  }


  /** @return A map associating to each sub expression of the expression a boolean flag true iff the sub expression is
    *         in a proper subtree tree wrt the root of the expression.
    * */
  def genIsInProperTree(e: Expr, ctx: IMap[Expr, Boolean] = IMap()): (util.graph.DepGraphE {type A = Expr}, IMap[Expr, Boolean]) = {
    // the ctx map will be extended with new bindings
    var res = ctx

    // compute a topological ordering using a depgraph
    val depGraph = genInverseDepGraph(e)

    // propagate the property topologically
    depGraph.topsort() foreach {
      node => {
        val e = node.data
        // a node is in a proper tree iff :
        // - the node has a single parent
        // - the parent is in a proper tree
        val b = node.predecessors match {
          case head :: Nil => res(head.data)
          case _ => false
        }
        res += (e -> b)
      }
    }

    // return augmented context map as result
    (depGraph, res)

  }

  /** TODO fix bug in the way this flag is computed, introduce notion of polarity with monotony.
    *
    * @return A map associating to each sub expression of the expression a boolean flagtrue iff the sub expression has a
    *         montonic influence wrt the root of the expression.
    * */
  def genIsMonotonic(e: Expr, ctx: IMap[Expr, Boolean] = IMap()): (util.graph.DepGraphE {type A = Expr}, IMap[Expr, Boolean]) = {

    // the ctx map will be extended with new bindings
    var res = ctx

    /** True iff e is not the selector of a parent Ite expression. */
    def isMonoWrt(e: Expr, parent: Expr): Boolean = {
      parent match {
        case Not(_) => false
        case Ite(s, _, _) if s == e => false
        case _ => true
      }
    }

    // compute a topological ordering using a depgraph
    val depGraph = genInverseDepGraph(e)

    // propagate the property topologically
    depGraph.topsort() foreach {
      node => {
        val e = node.data
        // a node is monotonic iff for all of its parents :
        // - the parent is in monotonic
        // - the node is in monotonic position wrt the parent
        val b = node.predecessors.foldLeft(true)(
          (soFar, n) => {
            val parent = n.data
            soFar & res(parent) & isMonoWrt(e, parent)
          }
        )
        res += (e -> b)
      }
    }

    // return augmented context map as result
    (depGraph, res)
  }
}

/** Translation function from Prop to NNF.
  * Each instance of the class possesses its own translation cache.
  * */
class PropToNNF extends (Prop.Expr => NNF.Expr) {

  /** Translation function from Prop to Core logic. */
  private val toCore: Core.ToCore = new Core.ToCore

  /** Translation function from Core to NNF logic. */
  private val toNNF: NNF.ToNNF = new NNF.ToNNF

  /** Translates the given express to NNF form. */
  def apply(e: Prop.Expr): NNF.Expr = toNNF(toCore(e))

  /** Returns the current expression translation cache as a function. */
  def atomCacheFunc: PartialFunction[Prop.Atom, NNF.Atom] = toCore.atomCacheFunc andThen toNNF.atomCacheFunc

  /** Returns the current expression translation cache as a function. */
  def exprCacheFunc: PartialFunction[Prop.Expr, NNF.Expr] = toCore.exprCacheFunc andThen toNNF.exprCacheFunc
}

/** Propositional logic in conjunctive normal form. */
object CNF {

  /** Abstract trait for expressions in CNF. */
  sealed trait Expr {
    val kids: List[Expr]
    override val hashCode: Int = foldHash(kids: _*)

    /** Set of atoms found in the cone of influence of the expression. */
    val atoms: ISet[Atom] = kids.view.map(_.atoms).foldLeft(ISet[Atom]())((current, x) => current union x)

    /** Set of literals found in the cone of influence of the expression. */
    def literals: ISet[Literal] = kids.view.map(_.literals).foldLeft(ISet[Literal]())((current, x) => current union x)
  }

  /** Abstract type for literals (atoms or atom negations). */
  sealed trait Literal extends Expr

  /** Counter for number of atoms created in NNF. */
  private var _nofAtoms: Int = 0

  /** Creates a fresh uniquely identified atom. */
  def freshAtom() = {
    _nofAtoms += 1;
    Atom(_nofAtoms)
  }

  /** Free proposition. */
  case class Atom private(id: Int) extends Literal {
    assert(id > 0)
    val kids = Nil
    override val atoms = ISet(this)

    override def literals = ISet(this)

    override val hashCode = foldHash(id)
  }

  /** Logical negation. */
  case class Not(kid1: Atom) extends Literal {
    val kids = List(kid1)
  }

  /** CNF formula. */
  case class Cnf(kids: List[Clause]) extends Expr

  /** Clause. */
  case class Clause(kids: List[Literal]) extends Expr

  /** Translation from NNF to CNF with Tseitin's rules. */
  class ToCNFTseitin extends (NNF.Expr => Cnf) {
    // TODO implement
    def apply(e: NNF.Expr): Cnf = ???
  }

  /** Translation from NNF to CNF with deMorgan's rules. */
  class ToCNFDeMorgan extends (NNF.Expr => Cnf) {
    // TODO implement
    def apply(e: NNF.Expr): Cnf = ???
  }

}
