/**
  * Created by rdelmas on 18/05/16.
  */
package util.sat

import scala.collection.immutable.{HashMap, HashSet, Map, Set}

/** A class implementing the translation algorithm described in the paper
  * "Enumerating Prime Implicants of Propositional Formulae in Conjunctive Normal Form"
  * by Said Jabbour, Joao Marques-Silva, Lakhdar Sais and Yakoub Salhi.
  * */
class CNFPrimeImplicants(source: FormulaLike, target: FormulaLike) {

  /** Translates source clauses to target clauses using the renaming algorithm.
    *
    * @return A backtranslator instance for translating models of target to prime implicants of source.
    **/
  def translate(): CNFPrimeImplicantBackTranslator = {
    // dual-rail translation
    this.renameClauses()
    this.genBijectionConstraints()
    val freeVars = source.inputs.toList.map(v => "x" + this._renamingMap(v).toString).mkString(" ")
    target.addHeaderComment(s"project ${freeVars}")
    val res = new CNFPrimeImplicantBackTranslator(source, target, this._renamingMap.map(e => e.swap))
    res
  }

  /** Translates source and adds a cardinality constraint on the number of positive positive free variables.
    *
    * @return A backtranslator instance for translating models of target to prime implicants of source.
    **/
  def atMostK(maxOrder: Int): CNFPrimeImplicantBackTranslator = {
    val res = this.translate()
    // add constraint on all renamed literals
    //target.addCard( target.Card( this._wasFreeSet.toList, target.AtMost, maxOrder ) )
    // add constraint on literals representing positive variables only
    // TODO cardinality constraint is only added on positive dual-rail literals of the form l_{x} and not l_{~x}
    target.addCard(target.Card(source.inputs.map(v => this._renamingMap(v)).toList, target.AtMost, maxOrder))()
    res
  }

  /** Translates source and adds a minimization criterion over positive free variables.
    *
    * @return A backtranslator instance for translating models of target to prime implicants of source.
    **/
  def minimize(): CNFPrimeImplicantBackTranslator = {
    val res = this.translate()
    val lits = this.source.inputs.map(v => this._renamingMap(v)).toList
    val weigths = lits.map(_ => 1)
    val crit = target.Minimize(lits, weigths)
    target.addCriterion(crit)
    res
  }

  /** Renaming map from source to target variables
    * Entries are of the form:
    *
    * If l is a free (ie. non-gate) literal of the source formula:
    *
    * {{{ l -> x_{l} }}}
    *
    * Where:
    * - l is a free of the source formula (either p or -p)
    * - x_{l} is the renamed literal corresponding to l (either x_{p}  or x_{-p})
    * */
  var _renamingMap: Map[Int, Int] = HashMap.empty

  /** Target literals originating from free variables of the source formula. */
  private var _wasFreeSet: Set[Int] = HashSet.empty

  /** @return An iterator on variables that correspond to variables that were free in source instance. */
  def wasFreeSet: Iterator[Int] = this._wasFreeSet.iterator

  /** True iff the target literal originates from a free variable of the source formaula. */
  private def wasFree(lit: Int): Boolean = this._wasFreeSet.contains(lit)

  /** Index from free variables of the target formula to clauses containing it. */
  private var _clauseIndex: Map[Int, Set[target.Clause]] = HashMap.empty

  /** Applies renaming to all clauses, building index from renamed literals to clauses. */
  final private
  def renameClauses(): Unit = {
    source.clauses.foreach {
      x => {
        val (clause, swOpt) = x
        val newLits = clause.lits.map(renameLit)
        target.addClauseList(newLits)()
        // build clause index
        val newClause = target.Clause(newLits)
        newLits.foreach {
          lit => {
            val prev = this._clauseIndex.getOrElse(lit, HashSet.empty[target.Clause])
            this._clauseIndex = this._clauseIndex.updated(lit, prev + newClause)
          }
        }
      }
    }
  }

  /** Renames a literal of the form p, respectively -p, to a variable of the form x_{p}, respectively x_{-p}. */
  final private
  def renameLit(lit: Int): Int = {
    this._renamingMap.get(lit) match {
      case Some(x) => x
      case None => {
        if (source.isFree(lit)) {
          val x_p = target.freshVar
          val x_not_p = target.freshVar
          val (p, res) = if (lit > 0) (lit, x_p) else (-lit, x_not_p)
          this._renamingMap += (p -> x_p)
          this._renamingMap += (-p -> x_not_p)
          // add a clause so that renamed literals cannot be both true (which would mean that p is true and false simultaneously
          target.addClause(-x_p, -x_not_p)()
          this._wasFreeSet += x_p
          this._wasFreeSet += x_not_p
          res
        } else {
          val x_p = target.freshVar
          val (p, res) = if (lit > 0) (lit, x_p) else (-lit, -x_p)
          this._renamingMap += (p -> x_p)
          this._renamingMap += (-p -> -x_p)
          res
        }
      }
    }
  }

  /** Generates bijection clauses for all literals stemming from free variables. */
  final private
  def genBijectionConstraints(): Unit = this._wasFreeSet.foreach {
    x => {
      val clauses = this._clauseIndex(x)
      this.genBijectionConstraints(x, clauses)
    }
  }

  /** Removes x from the given clause. */
  final private
  def restrict(x: Int, clause: target.Clause): target.Clause = {
    val l = (clause.lits.toSet - x).toList
    target.Clause(l)
  }

  /** Generates additional constraints enforcing the bijection between source's prime implicants and target's models for literal x.
    * Must be called after the clause index has been build (by a prior call to renameClauses).
    * */
  final private
  def genBijectionConstraints(x: Int, clauses: Set[target.Clause]): Unit = {
    // remove x from all clauses that contain it
    val restrictedClauses = clauses.map { clause => restrict(x, clause) }
    // add a gate g = Or(clause.lits) for each clause in the restricted clauses
    val restrictedGates: List[Int] = restrictedClauses.map {
      clause => {
        val g = target.freshVar
        target.defineOr(g, clause.lits)()
        g
      }
    }.toList

    // build a conjunction gate c = And(gi) from the restricted clause gates
    val c = target.freshVar
    target.defineAnd(c, restrictedGates)()
    // add a clause (x => ~c) to the target solver
    target.addClause(-x, -c)()
  }

}

/** A back translator from models of the target formula to prime implicants of the source formula. */
class CNFPrimeImplicantBackTranslator(
                                       /** The source formula. */
                                       val source: FormulaLike,

                                       /** The target formula. */
                                       val target: FormulaLike,

                                       /** The translation map with items of the form {{{(target_lit -> source_lit)}}}. */
                                       val map: Map[Int, Int]) {
  /** Translates the provided models of the target formula to prime implicants of the source formula.
    *
    * @param targetModels a collection of models of the target formsula (each interpreted as a cube).
    * @return a colletcion of prime implicants of the source formula (each interpreted as a cube).
    **/
  def apply(targetModels: Iterable[Set[Int]]): Iterable[Set[Int]] = {
    targetModels.map {
      model => {
        map.flatMap {
          kv => {
            val (target, source) = kv
            if (model.contains(target))
              source :: Nil
            else if (model.contains(-target))
              -source :: Nil
            else
              Nil
          }
        }.toSet
      }
    }
  }
}

