package util.sat

import scala.annotation.tailrec
import scala.collection.immutable

/**
  * Created by rdelmas on 20/06/16.
  */

/** An interface for building circuit-like formulas on top of CNF formulas. */
trait CircuitLike {
  this: CnfLike =>

  /* ============  INPUTS ============== */

  /** Circuit inputs set. */
  private var _inputSet: immutable.Set[Int] = immutable.Set()

  /** Circuit inputs list. */
  private var _inputVec: immutable.Vector[Int] = immutable.Vector()

  /** @return the number of gates of the formula. */
  final def nofInputs: Int = this._inputSet.size

  /** @return An itetator on the gate variables of the formula. */
  final def inputs: Iterator[Int] = this._inputVec.iterator

  /** Declares a literal as an output of the circuit. */
  final def declareInput(lit: Int)(switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    this._inputSet += lit
    this._inputVec :+= lit
    val swOpt = switch.orElse(if (switchable) Some(freshVar) else None)
    this._gates += ((lit, (Input, swOpt)))
    swOpt
  }

  /** True iff the literal was previously declared as a circuit input. */
  final def isInput(lit: Int): Boolean = !this._gates.contains(if (lit > 0) lit else -lit)

  /* =============== OUTPUTS ============ */

  /** Circuit outputs set. */
  private var _outputSet: immutable.Set[Int] = immutable.Set()

  /** Circuit outputs set. */
  private var _outputVec: immutable.Vector[Int] = immutable.Vector()

  /** @return the number of gates of the formula. */
  final def nofOutputs: Int = this._outputSet.size

  /** @return An iterator on the gate variables of the formula. */
  final def outputs: Iterator[Int] = this._outputVec.iterator

  /** Declares a literal as an output of the circuit. */
  final def declareOutput(lit: Int): Unit = {
    this._outputSet += lit
    this._outputVec :+= lit
  }

  /* ============ ASSERTIONS =============== */

  /** Stores assertions of the form (g, sw) where g is a gate literal and sw is a switch literal. */
  private var _assertions: immutable.Vector[(Int, Option[Int])] = immutable.Vector()

  /** @return the number of gates of the formula. */
  final def nofAssertions: Int = this._assertions.size

  /** @return An itetator on the gate variables of the formula. */
  final def assertions: Iterator[(Int, Option[Int])] = this._assertions.iterator

  /** Asserts that a certain literal must be true. */
  final def assert(lit: Int)(switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    val sw = this.addClause(lit)(switch, switchable)
    this._assertions :+= Tuple2(lit, sw)
    sw
  }

  /* =============== GATES ============== */

  /** Abstract type for logical gates used when constructing a CNF formula. */
  sealed trait Gate

  /** Represents an input gate of the circuit. */
  case object Input extends Gate

  /** Negation. */
  final case class Not(lit: Int) extends Gate

  /** Conjunction. */
  final case class And(lits: List[Int]) extends Gate

  /** Disjunction. */
  final case class Or(lits: List[Int]) extends Gate

  /** Implication. */
  final case class Implies(kid1: Int, kid2: Int) extends Gate

  /** Equivalence. */
  final case class Iff(kid1: Int, kid2: Int) extends Gate

  /** Equivalence. */
  final case class Xor(kid1: Int, kid2: Int) extends Gate

  /** If-then-else. */
  final case class Ite(kid1: Int, kid2: Int, kid3: Int) extends Gate

  /** Stores literals defined as circuit gates with optional switch. */
  private var _gates: immutable.Map[Int, (Gate, Option[Int])] = immutable.Map()

  /** @return The number of gates of the formula. */
  final def nofGates: Int = this._gates.size

  /** @return An itetator on gate literal definitions. */
  final def gatesIterator: Iterator[(Int, (Gate, Option[Int]))] = this._gates.iterator

  /** @return A gate Map indexed by literals. */
  final def gates: immutable.Map[Int, (Gate, Option[Int])] = this._gates

  /** @return True iff the literal or its complement is defined as a gate. */
  final def isGate(lit: Int): Boolean = this._gates.get(lit).orElse(this._gates.get(-lit)).isDefined

  /** @return True iff the literal is not a gate. */
  final def isFree(lit: Int): Boolean = !isGate(lit)

  /** Encodes the equation
    * {{{ x = Not(p) }}}
    * as clauses:
    * {{{
    * (~x, ~p)
    * (x, p)
    * }}}
    *
    * @param switch     optional switch literal
    * @param switchable if true and if switch is None, creates a fresh switch literal that allows to switch the gate clauses on and off
    * @return if switch is defined or if switchable is true, Some(fresh) where fresh is a switching literal, None otherwise.
    * */
  final def defineNot(x: Int, p: Int)(switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    val swOpt = mkSwitchLit(switch, switchable)
    this.addClause(-x, -p)(switch = swOpt)
    this.addClause(x, p)(switch = swOpt)
    this._gates = this._gates.updated(x, (Not(p), swOpt))
    swOpt
  }

  /** Encodes the equation
    * {{{ x = And(a1, ... ,an) }}}
    * as clauses:
    * {{{
    * (  x, ~a1, ... , ~an)
    * ( ~x,  a1 )
    * ...
    * ( ~x,  an )
    * }}}
    *
    * @param switch     optional switch literal
    * @param switchable if true and if switch is None, creates a fresh switch literal that allows to switch the gate clauses on and off
    * @return if switch is defined or if switchable is true, Some(fresh) where fresh is a switching literal, None otherwise.
    * */
  final def defineAnd(x: Int, lits: List[Int])(switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    val swOpt = mkSwitchLit(switch, switchable)
    this.addClauseList(x :: lits.map(lit => -lit))(switch = swOpt)
    lits.foreach(lit => this.addClause(-x, lit)(switch = swOpt))
    this._gates = this._gates.updated(x, (And(lits), swOpt))
    swOpt
  }

  /** Encodes the equation
    * {{{ x = Or(a1, ... , an) }}}
    * as clauses:
    * {{{
    * ( ~x,  a1, ... , an)
    * (  x, ~a1)
    * ...
    * (  x, ~an)
    * }}}
    *
    * @param switch     optional switch literal
    * @param switchable if true and if switch is None, creates a fresh switch literal that allows to switch the gate clauses on and off
    * @return if switch is defined or if switchable is true, Some(fresh) where fresh is a switching literal, None otherwise.
    * */
  final def defineOr(x: Int, lits: List[Int])(switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    val swOpt = mkSwitchLit(switch, switchable)
    this.addClauseList(-x :: lits)(switch = swOpt)
    lits.foreach(lit => this.addClause(x, -lit)(switch = swOpt))
    this._gates = this._gates.updated(x, (Or(lits), swOpt))
    swOpt
  }

  /** Encodes the equation
    * {{{ x = Implies(p, q) }}}
    * as clauses:
    * {{{
    * (~x, ~p, q )
    * ( x,  p )
    * ( x, ~q )
    * }}}
    *
    * @param switch     optional switch literal
    * @param switchable if true and if switch is None, creates a fresh switch literal that allows to switch the gate clauses on and off
    * @return if switch is defined or if switchable is true, Some(fresh) where fresh is a switching literal, None otherwise.
    * */
  final def defineImplies(x: Int, p: Int, q: Int)(switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    val swOpt = mkSwitchLit(switch, switchable)
    this.addClause(-x, -p, q)(switch = swOpt)
    this.addClause(x, p)(switch = swOpt)
    this.addClause(x, -q)(switch = swOpt)
    this._gates = this._gates.updated(x, (Implies(p, q), swOpt))
    swOpt
  }

  /** Encodes the equation
    * {{{ x = Iff(p, q) }}}
    * as clauses:
    * {{{
    * ( ~x, ~p,  q )
    * ( ~x,  p, ~q )
    * (  x, ~p, ~q )
    * (  x,  p,  q )
    * }}}
    *
    * @param switch     optional switch literal
    * @param switchable if true and if switch is None, creates a fresh switch literal that allows to switch the gate clauses on and off
    * @return if switch is defined or if switchable is true, Some(fresh) where fresh is a switching literal, None otherwise.
    * */
  final def defineIff(x: Int, p: Int, q: Int)(switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    val swOpt = mkSwitchLit(switch, switchable)
    this.addClause(-x, -p, q)(switch = swOpt)
    this.addClause(-x, p, -q)(switch = swOpt)
    this.addClause(x, -p, -q)(switch = swOpt)
    this.addClause(x, p, q)(switch = swOpt)
    this._gates = this._gates.updated(x, (Iff(p, q), swOpt))
    swOpt
  }

  /** Asserts that a collection of literals are all equal, implemented using a loop of implications.
    *
    * {{{ l1 = l2 = ... = ln := l1 -> l2 & ... & ln -> l1 }}}
    *
    * @param lits       list of literals of size >= 2
    * @param switch     optional switch literal
    * @param switchable if true and if switch is None, creates a fresh switch literal that allows to switch the assertion on and off
    * @return The optional switching literal
    *
    * */
  final def assertIff(lits: List[Int])(switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    lits match {
      case h1 :: h2 :: tail => {
        /* OK */
      }
      case _ => throw new Exception("lits shall contain at least two literals")
    }
    val swOpt = mkSwitchLit(switch, switchable)

    @tailrec
    def loop(lits: List[Int], prev: Option[Int] = None, first: Option[Int] = None): Unit = lits match {
      case h1 :: h2 :: tail if first.isEmpty => {
        addClause(-h1, h2)(switch = swOpt)
        loop(h2 :: tail, prev = Some(h1), first = Some(h1))
      }
      case h1 :: h2 :: tail => {
        addClause(-h1, h2)(switch = swOpt)
        loop(h2 :: tail, prev = Some(h1), first = first)
      }
      case h :: Nil => {
        addClause(-h, first.get)(switch = swOpt)
      }
    }

    loop(lits)
    swOpt
  }

  /** Encodes the equation
    * {{{ x = Xor(p, q) }}}
    * as clauses:
    * {{{
    * ( ~x,  p,  q )
    * ( ~x, ~p, ~q )
    * (  x,  p, ~q )
    * (  x, ~p,  q )
    * }}}
    *
    * @param switch     optional switch literal
    * @param switchable if true and if switch is None, creates a fresh switch literal that allows to switch the gate clauses on and off
    * @return if switch is defined or if switchable is true, Some(fresh) where fresh is a switching literal, None otherwise.
    * */
  final def defineXor(x: Int, p: Int, q: Int)(switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    val swOpt = mkSwitchLit(switch, switchable)
    this.addClause(-x, p, q)(switch = swOpt)
    this.addClause(-x, -p, -q)(switch = swOpt)
    this.addClause(x, p, -q)(switch = swOpt)
    this.addClause(x, -p, q)(switch = swOpt)
    this._gates = this._gates.updated(x, (Xor(p, q), swOpt))
    swOpt
  }

  /** Encodes the equation
    * {{{ x = Ite(s, t, f) }}}
    * as clauses:
    * {{{
    * (  x, ~s, ~t )
    * ( ~x, ~s,  t )
    * (  x,  s, ~f )
    * ( ~x,  s,  f )
    * (  x, ~t, ~f )
    * ( ~x,  f,  t )
    * }}}
    *
    * @param switch     optional switch literal
    * @param switchable if true and if switch is None, creates a fresh switch literal that allows to switch the gate clauses on and off
    * @return if switch is defined or if switchable is true, Some(fresh) where fresh is a switching literal, None otherwise.
    * */
  final def defineIte(x: Int, s: Int, t: Int, f: Int)(switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    val swOpt = mkSwitchLit(switch, switchable)
    this.addClause(x, -s, -t)(switch = swOpt)
    this.addClause(-x, -s, t)(switch = swOpt)
    this.addClause(x, s, -f)(switch = swOpt)
    this.addClause(-x, s, f)(switch = swOpt)
    this.addClause(x, -t, -f)(switch = swOpt)
    this.addClause(-x, t, f)(switch = swOpt)
    this._gates = this._gates.updated(x, (Ite(s, t, f), swOpt))
    swOpt
  }
}
