/**
  * Created by rdelmas on 19/05/16.
  */
package util.sat

import scala.collection.immutable

/** An interface for building formulas in conjunctive normal form. */
trait CnfLike {

  /* ========================= HEADER COMMENTS ================ */

  /** User specified header comments. */
  private var _headerComments: immutable.Vector[String] = immutable.Vector()

  /** Add a header comment in the formula. */
  final def addHeaderComment(comment: String): Unit = this._headerComments :+= comment

  /** @return Iterator over all header comments, in insertion order. */
  final def headerComments: Iterator[String] = this._headerComments.iterator

  /* ======================== VARIABLES ====================== */

  /** Number of allocated vars. */
  private var _nofVars: Int = 0

  /**
    * Creates a fresh variable in the solver.
    *
    * @return the number of variables created in the solver so far, equal to the fresh variable identifier.
    */
  final def freshVar: Int = {
    this._nofVars += 1
    this.freshVarImplem(this._nofVars)
    this._nofVars
  }

  /**
    * Reserve {{{howmany}}} variables to express the problem, reserving dimacs variables identifiers in [1, howmany].
    * Next fresh variables identifiers will start at {{{howmany+1}}}
    *
    * @param howmany number of variables to create
    * @return The total variables used in the solver.
    */
  final def freshVar(howmany: Int): Int = {
    // TODO find proper interaction between implem and our representation of the formula (what if literals are not returned in the same order)
    this._nofVars = howmany
    val resImpl = this.freshVarImplem(this._nofVars, howmany)
    this._nofVars
  }

  /** @return the number of variables created using [[freshVar]] */
  final def nofVars: Int = this._nofVars

  /* ======================== FROZEN VARIABLES ========================= */

  /** Set of frozen literals (that will not be deleted or transformed when processing the formula). */
  private var _frozenVars: immutable.HashSet[Int] = immutable.HashSet()

  /** Freeze the variable underlying the literal. */
  def freeze(lit: Int): this.type = {
    assert(lit != 0)
    _frozenVars += math.abs(lit)
    this
  }

  /** True iff the variable is frozen. */
  def isFrozen(lit: Int): Boolean = _frozenVars.contains(math.abs(lit))

  /* ====================== SWITCHING LITERALS ========================== */

  /** Returns a switching literal: if switch is Some( s ), returns this Some( s ), else if switchable is true, returns a Some( newSwitch ) else returns None. */
  final def mkSwitchLit(switch: Option[Int], switchable: Boolean): Option[Int] = switch match {
    case Some(s) => switch
    case None => if (switchable) Some(freshVar) else None
  }

  /* ======================== CLAUSES ====================== */

  /** A clause. */
  final case class Clause(lits: List[Int])

  /** Clauses of the formula, with optional swith literal. */
  private var _clauses: immutable.Vector[(Clause, Option[Int])] = immutable.Vector()

  /** @return the number of clauses of the formula. */
  final def nofClauses: Int = this._clauses.length

  /** @return an iterator on the clauses currently in the solver. */
  final def clauses: Iterator[(Clause, Option[Int])] = this._clauses.iterator

  /**
    * Adds a list of literals as a clause to the solver.
    *
    * @return if switchable is true, Some(fresh) where fresh is a fresh literal that allows to switch the clause on or off, None otherwise.
    */
  final def addClauseList(lits: List[Int])(switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    val (clause, sw) = toSwitchableClause(lits)(switch, switchable)
    this._clauses :+= Tuple2(clause, sw)
    this.addClauseImplem(clause.lits)
    sw
  }

  /**
    * Adds a collection of literals as a clause to the solver.
    *
    * @return if switchable is true, Some(fresh) where fresh is a fresh literal that allows to switch the clause on or off, None otherwise.
    */
  final def addClause(lits: Int*)(switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    this.addClauseList(lits.toList)(switch, switchable)
  }

  /* Transforms a clause to a switchable clause depending on the switch and switchable arguments). */
  private final def toSwitchableClause(lits: List[Int])(switch: Option[Int] = None, switchable: Boolean = false): (Clause, Option[Int]) = {
    mkSwitchLit(switch, switchable) match {
      case Some(sw) => (Clause(-sw :: lits), Some(sw))
      case None => (Clause(lits), None)
    }
  }

  /* ====================== IMPLEMENTATION METHODS ======================= */

  /** Method to implement to effectively create a fresh var in the underlying representation. */
  def freshVarImplem(superRes: Int, howmany: Int): Int = {
    // TODO find proper interaction between implem and our representation of the formula (what if literals are not returned in the same order)
    superRes
  }

  /** Method to implement to effectively create a fresh var in the underlying representation. */
  def freshVarImplem(superRes: Int): Int = {
    // TODO find proper interaction between implem and our representation of the formula (what if literals are not returned in the same order)
    superRes
  }

  /** Method to use to effectively add a clause to the underlying Solver implementation. */
  def addClauseImplem(lits: List[Int]): Unit = {
    // TODO find proper interaction between implem and our representation of the formula (what if literals are not returned in the same order)
    /* Intentionally empty */
  }

}
