package util.sat

import scala.collection.mutable.{OpenHashMap => MMap}

/**
  * Created by rdelmas on 23/06/16.
  */

/** Implements the dual-rail transformation at clause level.
  * Does not split variables that are frozen in the source formula, and freezes them in the target formula.
  * */
case class DualRailEncoder(source: FormulaLike, target: FormulaLike) {

  /** Translation cache from single var to single var. */
  private val _cache: MMap[Int, Int] = new MMap()

  /** Translation cache from single var to dual-rail vars. */
  private val _func: MMap[Int, (Int, Int)] = new MMap()

  /** Generates dual-rail variables for a list of literals (allows to control which indices ranges will be occupied by dual rail lits).
    * Must be called before [[encode()]] to guarantee intended control over the dual-rail literal indices.
    * Literals not encoded a priori using this method will be enoded on the fly as discovered during the clause-level processing of [[encode()]].
    *
    * @param lits the list of literals to encode
    * @return the list of dual-rail literals resulting from the translation
    **/
  def apply(lits: List[Int]): List[(Int, Int)] = {
    lits.map(lit => {
      apply(lit); bothLits(lit)
    })
  }

  /** Encodes source formula to target formula using dual-rail encoding.
    *
    * Literals not encoded a priori using the [[apply(List[Int])]] will be encoded on the fly as discovered during clause traversal,
    * and hence no guarantees can be given about the index range they will occupy nor on their ordering.
    * */
  def encode(): Unit = {

    // process all variables of the source formula (to maintain relative order of variables).
    (1 to source.nofVars) foreach {
      apply
    }

    // process all clauses
    source.clauses foreach {
      x => {
        val (clause, _) = x
        val newLits: List[Int] = clause.lits.map(apply)
        target.addClauseList(newLits)()
      }
    }
  }

  /** Given a literal, returns a pair (pos, neg)
    * where:
    * - pos is the positive version of the literal.
    * - neg the negative version of the literal.
    * */
  def toPosNeg(lit: Int): (Int, Int) =
    if (lit > 0) (lit, -lit) else (-lit, lit)

  /** Returns ifPos if selector > 0 and isNeg otherwise. */
  def selFromSign(selector: Int, ifPos: Int, ifNeg: Int): Int =
    if (selector > 0) ifPos else ifNeg

  /** Return both dual-rail encoded literals in (l_x, l_not_x). */
  def bothLits(lit: Int): (Int, Int) = {
    val (pos, neg) = toPosNeg(lit)
    (apply(pos), apply(neg))
  }

  /** Returns the current translation cache as an immutable map. */
  def funcMap: Map[Int, (Int, Int)] = scala.collection.immutable.Map() ++ _func

  /** Returns the current translation cache as an immutable map. */
  def cacheMap: Map[Int, Int] = scala.collection.immutable.Map() ++ _cache

  /** Returns the dual rail encoded variable. */
  def apply(lit: Int): Int = {
    _cache.getOrElseUpdate(lit, mkDualRail(lit))
  }

  /** Creates literals and extra clauses in the target formula, returns the encoded litera. */
  private def mkDualRail(lit: Int): Int = {
    val (pos, neg) = toPosNeg(lit)
    val (l_x, l_not_x) =
      if (source.isFrozen(lit)) {
        val x = target.freshVar
        target.freeze(x)
        (x, -x)
      } else {
        val l_x = target.freshVar
        val l_not_x = target.freshVar
        target.addClause(-l_x, -l_not_x)()
        (l_x, l_not_x)
      }
    _cache.update(pos, l_x)
    _cache.update(neg, l_not_x)
    _func.update(pos, (l_x, l_not_x))
    selFromSign(lit, l_x, l_not_x)
  }
}
