package util.sat

/**
  * Created by rdelmas on 26/06/16.
  *
  * Unit-tests of the dual rail enoding wrt the Ternary semantics of operators.
  *
  */
object DualRailEncodingTests extends App {

  runTests()

  import TernaryLogic._
  import UnitProp._


  def cartesianProd[A](s: List[List[A]], prefix: List[A] = Nil, result: List[List[A]] = Nil): List[List[A]] = {
    s match {
      case Nil => prefix.reverse :: result
      case head :: tail => head.flatMap(e => cartesianProd(tail, e :: prefix, result))
    }
  }

  def cartesianProdMap[A, B](s: List[List[A]], f: List[A] => B, prefix: List[A] = Nil, result: List[B] = Nil): List[B] = {
    s match {
      case Nil => f(prefix.reverse) :: result
      case head :: tail => head.flatMap(e => cartesianProdMap(tail, f, e :: prefix, result))
    }
  }

  case class TableEntry(
                         kids: List[(((Int, Int), String), TernaryVal)],
                         o: ((Int, Int), String),
                         oExpected: TernaryVal,
                         oObtained: TernaryVal,
                         formulaStatus: FStatus,
                         l_o_value: Option[Boolean],
                         l_not_o_value: Option[Boolean]) {

    def litValueToString(value: Option[Boolean]): String = value match {
      case Some(v) => if (v) "1" else "0"
      case None => "?"
    }

    val kidsSpaceString = kids.map(k => "--").mkString("-+-")

    def printHeader(): Unit = {
      val kidsHeaderString = kids.map(k => k._1._2).mkString(" | ")
      println("| " + kidsHeaderString + """ | $o$ | $f(o)$ | $f$ status | $f(l_{o})$ | $f(l_{\lnot{o}})$ |""")
      println("|-" + kidsSpaceString + "-+-----+--------+------------+------------+-------------------|")
    }

    def print() = {
      val kidsValueString = kids.map(k => k._2).mkString(" | ")
      println(s"| ${kidsValueString} | ${oExpected} |  ${oObtained}   | ${formulaStatus} | ${litValueToString(l_o_value)} | ${litValueToString(l_not_o_value)} |")
      println("|-" + kidsSpaceString + "-+------+--------------+------------+----------------------------+--------------------------------|")
    }
  }

  def testEncoding(propF: Prop.Expr, atomFunc: PartialFunction[Prop.Atom, (Int, Int)], kids: List[((Lit, Lit), String)], output: ((Lit, Lit), String), evalOp: (List[TernaryVal]) => TernaryVal, f: FormulaLike): List[TableEntry] = {

    val allVals = List(TernaryTrue, TernaryFalse, TernaryUndef)
    val domains = kids.map(kid => allVals.map(value => (kid, value)))
    val baseFormula = convert(f)

    println(s"*** CNF Encoding")
    println()
    println("Atoms to dual rail literals mapping:")
    propF.atoms foreach {
      atom => {
        println(" - " + atom.toString + """ $\rightarrow$ """ + atomFunc(atom).toString)
      }
    }
    println()
    println("Clausal form:")
    println()
    println(baseFormula.map(clause => clause.mkString("""$\{ """, " , ", """ \}$""")).mkString("""$\{$ """, " , ", """ $\}$"""))

    def runTestCase(kidsWithVals: List[(((Int, Int), String), TernaryVal)]): TableEntry = {
      val kidVals: List[TernaryVal] = kidsWithVals map { k => k._2 }
      val (l_o, l_not_o) = output._1
      val oExpected = evalOp(kidVals)
      val baseFormulaWithUnits: Formula = kidsWithVals.foldLeft(baseFormula)(
        (acc, kid) => {
          val ((kidLits, kidName), kidVal) = kid
          val (pos, neg) = kidLits
          val units = kidVal.toFormula(pos, neg)
          units ::: acc
        })
      val (status, model) = propagate(baseFormulaWithUnits)
      val l_o_value = model.get(l_o)
      val l_not_o_value = model.get(l_not_o)
      val oObtained = TernaryVal.fromDualRail(l_o_value, l_not_o_value)
      TableEntry(kidsWithVals, output, oExpected, oObtained, status, l_o_value, l_not_o_value)
    }

    // run the test for each element of the procuct
    cartesianProdMap(domains, (x: List[(((Int, Int), String), TernaryVal)]) => runTestCase(x))
  }


  def testOp(
              f: Prop.Expr,
              evalOp: (List[TernaryVal]) => TernaryVal,
              drF: FormulaLike,
              atomFunc: PartialFunction[Prop.Atom, (Int, Int)],
              exprFunc: PartialFunction[Prop.Expr, (Int, Int)]
            ): Unit = {
    val kids = f.atoms.toList
    val inNames = kids.map(a => "$a_{" + a.id.toString + "}$")
    val inVars = kids.map(a => atomFunc(a))
    val ins = inVars zip inNames
    val outName = "o"
    val outVar = exprFunc(f)
    val out = (outVar, outName)
    println()
    println(s"** Results for o = ${f.toString}")

    val resultTable = testEncoding(f, atomFunc, kids = ins, output = out, evalOp = evalOp, f = drF)

    println()
    println(s"*** Unit propagation results")
    println()

    if (!resultTable.isEmpty) {
      resultTable.head.printHeader()
      resultTable.foreach(_.print())
    } else {
      println("experiment failed")
    }
  }

  def runTests(): Unit = {

    import Prop._

    {
      val notProp = Not(Atom(1))
      val (notCircuit, notAtoms, notExprs) = PropToDualRailEncoder(notProp)
      testOp(notProp, evalNot, notCircuit, notAtoms, notExprs)
    }

    {
      val notProp = Not(Atom(1))
      val (notCircuit, notAtoms, notExprs) = PropToTernaryEncoder(notProp)
      testOp(notProp, evalNot, notCircuit, notAtoms, notExprs)
    }


    val nofKids = 3
    val kids = (1 to nofKids).view.map(Atom(_)).toList

    {
      val andProp = And(kids)
      val (andCircuit, andAtoms, andExprs) = PropToDualRailEncoder(andProp)
      testOp(andProp, evalAnd, andCircuit, andAtoms, andExprs)
    }

    {
      val andProp = And(kids)
      val (andCircuit, andAtoms, andExprs) = PropToTernaryEncoder(andProp)
      testOp(andProp, evalAnd, andCircuit, andAtoms, andExprs)
    }

    {
      val orProp = Or(kids)
      val (orCircuit, orAtoms, orExprs) = PropToDualRailEncoder(orProp)
      testOp(orProp, evalOr, orCircuit, orAtoms, orExprs)
    }

    {
      val orProp = Or(kids)
      val (orCircuit, orAtoms, orExprs) = PropToTernaryEncoder(orProp)
      testOp(orProp, evalOr, orCircuit, orAtoms, orExprs)
    }
    {
      val iteProp = Ite(kids(0), kids(1), kids(2))
      val (iteCircuit, iteAtoms, iteExprs) = PropToDualRailEncoder(iteProp)
      testOp(iteProp, evalIte, iteCircuit, iteAtoms, iteExprs)
    }
    {
      val iteProp = Ite(kids(0), kids(1), kids(2))
      val (iteCircuit, iteAtoms, iteExprs) = PropToTernaryEncoder(iteProp)
      testOp(iteProp, evalIte, iteCircuit, iteAtoms, iteExprs)
    }
  }

}
