/** @author rdelmas */
package util.sat

import java.io.{BufferedWriter, FileWriter, IOException}

/** A object for printing FormulaLike objects to dimacs, OPB, circuit and graphviz files. */
object FormulaFileWriter {

  object Cnf {

    /**
      * Writes the current solver state to a file using the dimacs format. Only safe to call when the solver does not contain cardinality constraints.
      *
      * @param fileName the name of the file
      */
    final def apply(formula: FormulaLike, fileName: String, addPrimerHeader: Boolean = false): Unit = {
      val fName = fileName + ".cnf"
      implicit var out: BufferedWriter = null
      try {
        out = new BufferedWriter(new FileWriter(fName))

        if (formula.nofPBCst > 0) {
          println(s"[warning] cardinality constraints will be printed as comments in '${fileName}.cnf'.")
        }

        if (addPrimerHeader) {
          writePrimerHeader(formula.nofVars, formula.nofClauses, formula.nofInputs)
        } else {
          writeHeader(formula.nofVars, formula.nofClauses)
        }
        formula.headerComments.foreach {
          this.writeComment
        }
        writeCards(formula.cards.map(_._1))
        writeClauses(formula.clauses.map(_._1))
      } catch {
        case e: IOException => {
          println(s"Could not create file ${fName} for writing: ${e.toString}")
          sys.exit(1)
        }
        case e: Throwable => throw e
      } finally {
        if (out != null) {
          out.close()
        }
      }
    }

    /** Writes a comment line to out. */
    private def writeComment(comment: String)(implicit out: BufferedWriter): Unit = {
      out.write("c ")
      out.write(comment)
      out.write("\n")
    }

    /** Writes a header line to out using given nofVars and nofClauses. */
    private def writeHeader(nofVars: Int, nofClauses: Int)(implicit out: BufferedWriter): Unit = {
      out.write(s"p cnf ${nofVars} ${nofClauses}\n")
    }

    /** Writes a header line to out using given nofVars and nofClauses for primer soft. */
    private def writePrimerHeader(nofVars: Int, nofClauses: Int, nofInputs: Int)(implicit out: BufferedWriter): Unit = {
      out.write(s"c n orig vars $nofInputs\n")
      out.write(s"p cnf ${nofVars} ${nofClauses}\n")
    }

    /** Writes a clause in dimacs format to the out writer. */
    private def writeClause(clause: CnfLike#Clause)(implicit out: BufferedWriter): Unit = {
      clause.lits foreach { lit => {
        out.write(lit.toString);
        out.write(" ")
      }
      }
      out.write("0\n")
    }

    /** Writes a clause data base (must be in dimacs format, ie a zero-separated list of clauses represented as tuples of signed integers) to out. */
    private def writeClauses(clauses: Iterator[CnfLike#Clause])(implicit out: BufferedWriter): Unit = {
      clauses foreach {
        this.writeClause
      }
    }

    /** Writes a cardinality constraint in cnf as a comment. */
    private def writeCard(c: FormulaLike#PseudoBoolCst)(implicit out: BufferedWriter): Unit =
      c match {
        case cst: FormulaLike#ClassicPBCst =>
          out.write("c ")
          cst.lits.foreach { lit => {
            out.write(s" ${cst.weights(lit)}*");
            out.write(lit.toString);
            out.write(" ")
          }
          }
          out.write(cst.relOp.repr)
          out.write(" ")
          out.write(cst.k.toString)
          out.write("\n")
        case cst: FormulaLike#Card =>
          cst.toAtLeast foreach {
            newC => {
              out.write("c ")
              newC.lits.foreach { lit => {
                out.write(lit.toString);
                out.write(" ")
              }
              }
              out.write(newC.relOp.repr)
              out.write(" ")
              out.write(newC.k.toString)
              out.write("\n")
            }
          }
      }

    /** Writes a collection of cardinality constraints to out as comments. */
    private def writeCards(cards: Iterator[FormulaLike#PseudoBoolCst])(implicit out: BufferedWriter): Unit = {
      cards foreach {
        this.writeCard
      }
    }
  }

  /** File printing methods for opb models. */
  object Opb {

    /**
      * Writes the current solver state instance to a file using the OPB format.
      *
      * @param fileName the name of the file
      */
    def apply(formula: FormulaLike, fileName: String): Unit = {
      val fName = fileName + ".opb"
      implicit var out: BufferedWriter = null
      try {
        out = new BufferedWriter(new FileWriter(fName))
        writeHeader(formula.nofVars, formula.nofClauses + formula.nofPBCst)
        formula.headerComments foreach {
          this.writeComment
        }
        writeCriteria(formula.criteria)
        writeCards(formula.cards.map(_._1))
        writeClauses(formula.clauses.map(_._1))
      } catch {
        case e: IOException => {
          {
            println(s"couldn not create file ${fName} for writing: ${e.toString}")
            sys.exit(1)
          }
        }
        case e: Throwable => throw e
      } finally {
        if (out != null) {
          out.close()
        }
      }
    }

    /** Writes a comment line to out. */
    private def writeComment(comment: String)(implicit out: BufferedWriter): Unit = {
      out.write("* ")
      out.write(comment)
      out.write("\n")
    }

    /** Writes a header line to out for given number of variables and constraints. */
    private def writeHeader(nofVars: Int, nofConstraints: Int)(implicit out: BufferedWriter): Unit = {
      out.write(s"* #variable= ${nofVars} #constraint= ${nofConstraints}\n")
    }

    /** Writes a literal ( lit != 0 ) to out.  */
    private def writeLit(lit: Int)(implicit out: BufferedWriter): Unit = {
      if (lit > 0) {
        out.write("x")
        out.write(lit.toString)
      } else /* if ( lit < 0 ) */ {
        out.write("~x")
        out.write((-lit).toString)
      }
    }

    /** Writes a weigth value to out. */
    private def writeWeight(weight: Int)(implicit out: BufferedWriter): Unit = {
      if (weight > 0) {
        out.write("+")
        out.write(weight.toString)
      } else {
        out.write("-")
        out.write((-weight).toString)
      }
    }

    /** Writes a collection of criteria to out. */
    private def writeCriteria(crits: Iterator[FormulaLike#Criterion])(implicit out: BufferedWriter): Unit = {
      crits foreach {
        this.writeCriterion
      }
    }

    /** Writes an optimization criterion to out. */
    private def writeCriterion(crit: FormulaLike#Criterion)(implicit out: BufferedWriter): Unit = {
      val minCrit = crit.toMinimize
      out.write("min: ")
      (crit.weights zip crit.lits) foreach {
        e => {
          val (w, l) = e
          out.write(" ")
          this.writeWeight(w)
          out.write(" ")
          this.writeLit(l)
        }
      }
      out.write(";\n")
    }

    /** Writes a clause to out in OPB format to out. */
    private def writeClause(clause: FormulaLike#Clause)(implicit out: BufferedWriter): Unit = {
      clause.lits foreach {
        lit => {
          this.writeWeight(1)
          out.write(" ")
          this.writeLit(lit)
          out.write(" ")
        }
      }
      out.write(">= 1;\n")
    }

    /** Writes a clause data base in OPB format to out. */
    private def writeClauses(clauses: Iterator[FormulaLike#Clause])(implicit out: BufferedWriter): Unit = {
      clauses foreach {
        this.writeClause
      }
    }

    private def writeCard(c: FormulaLike#PseudoBoolCst)(implicit out: BufferedWriter): Unit =
      c match {
        case cst: FormulaLike#ClassicPBCst =>
          cst.lits foreach {
            lit => {
              writeWeight(cst.weights(lit))
              out.write(" ")
              writeLit(lit)
              out.write(" ")
            }
          }
          out.write(cst.relOp.repr)
          out.write(" ")
          out.write(cst.k.toString)
          out.write(";\n")
        case cst: FormulaLike#Card =>
          cst.toAtLeast foreach {
            newC => {
              newC.lits foreach {
                lit => {
                  writeWeight(1)
                  out.write(" ")
                  writeLit(lit)
                  out.write(" ")
                }
              }
              out.write(newC.relOp.repr)
              out.write(" ")
              out.write(newC.k.toString)
              out.write(";\n")
            }
          }
      }

    private def writeCards(cards: Iterator[FormulaLike#PseudoBoolCst])(implicit out: BufferedWriter): Unit = {
      cards foreach {
        this.writeCard
      }
    }
  }

  object Circuit {
    /**
      * Writes the current solver state instance to a file using the OPB format.
      *
      * @param fileName the name of the file
      */
    def apply(formula: FormulaLike, fileName: String): Unit = {
      val fName = fileName + ".circuit"
      implicit var out: BufferedWriter = null

      try {
        out = new BufferedWriter(new FileWriter(fName))
        writeHeader(
          formula.nofInputs,
          formula.nofOutputs,
          formula.nofGates,
          formula.nofCriteria,
          formula.nofPBCst,
          formula.nofAssertions)
        writeComment("created by FormulaFileWriter")
        formula.headerComments.foreach { c => writeComment(c) }
        writeComment("optimization criteria (lexicographic)")
        writeCriteria(formula)
        writeComment("circuit inputs")
        writeInputs(formula)
        writeComment("circuit outputs")
        writeOutputs(formula)
        writeComment("circuit gates")
        writeGates(formula)
        writeComment("asserted gates")
        writeAssertions(formula)
        writeComment("cardinality constraints")
        writeCards(formula)
      } catch {
        case e: IOException => {
          println(s"couldn not create file ${fName} for writing: ${e.toString}")
          sys.exit(1)
        }
        case e: Throwable => throw e
      } finally {
        if (out != null) {
          out.close()
        }
      }
    }

    private def writeHeader(
                             nofInputs: Int,
                             nofOutputs: Int,
                             nofGates: Int,
                             nofCrits: Int,
                             nofCards: Int,
                             nofAssertions: Int
                           )(implicit out: BufferedWriter): Unit = {
      out.write(s"* circuit #inputs ${nofInputs} #outputs ${nofOutputs} #gates ${nofGates} #criteria ${nofCrits} #assertions ${nofAssertions} #cards ${nofCards} \n")
    }

    private def writeComment(comment: String)(implicit out: BufferedWriter): Unit = {
      out.write("* ")
      out.write(comment)
      out.write("\n")
    }

    private def writeLit(lit: Int)(implicit out: BufferedWriter): Unit = {
      if (lit > 0) {
        out.write("x")
        out.write(lit.toString)
      } else {
        out.write("~x")
        out.write((-lit).toString)
      }
    }

    private def writeInputs(f: FormulaLike)(implicit out: BufferedWriter): Unit = f.inputs foreach {
      v => {
        out.write(s"input: ")
        this.writeLit(v)
        out.write("\n")
      }
    }

    private def writeOutputs(f: FormulaLike)(implicit out: BufferedWriter): Unit = f.outputs foreach {
      v => {
        out.write(s"output: ")
        this.writeLit(v)
        out.write("\n")
      }
    }

    private def writeAssertions(f: FormulaLike)(implicit out: BufferedWriter): Unit = f.assertions foreach {
      x => {
        val (g, sw) = x
        out.write(s"assert: ")
        this.writeLit(g)
        if (sw.isDefined) {
          out.write(s" switched by ")
          this.writeLit(sw.get)
        }
        out.write("\n")
      }
    }

    private def writeSwitch(switch: Option[Int])(implicit out: BufferedWriter): Unit = {
      if (switch.isDefined) {
        out.write(" switched by ")
        this.writeLit(switch.get)
      }
    }

    private def writeGates(f: FormulaLike)(implicit out: BufferedWriter): Unit = f.gates foreach {
      x => {
        val (v, (g, sw)) = x
        out.write(s"gate: ")
        this.writeLit(v)
        out.write(" = ")

        g match {
          case f.Input => {
            out.write("input ")
            this.writeLit(v)
            this.writeSwitch(sw)
          }
          case f.Not(lit) => {
            out.write("not ")
            this.writeLit(lit)
            this.writeSwitch(sw)
          }
          case f.And(lits) => {
            out.write("and")
            lits foreach {
              l => {
                out.write(" ")
                this.writeLit(l)
              }
            }
            this.writeSwitch(sw)
          }
          case f.Or(lits) => {
            out.write("or")
            lits foreach {
              l => {
                out.write(" ")
                this.writeLit(l)
              }
            }
            this.writeSwitch(sw)
          }
          case f.Implies(p, q) => {
            out.write("implies")
            out.write(" ")
            this.writeLit(p)
            out.write(" ")
            this.writeLit(q)
            this.writeSwitch(sw)
          }
          case f.Xor(p, q) => {
            out.write("xor")
            out.write(" ")
            this.writeLit(p)
            out.write(" ")
            this.writeLit(q)
            this.writeSwitch(sw)
          }
          case f.Iff(p, q) => {
            out.write("iff")
            out.write(" ")
            this.writeLit(p)
            out.write(" ")
            this.writeLit(q)
            this.writeSwitch(sw)
          }
          case f.Ite(c, t, e) => {
            out.write("ite")
            out.write(" ")
            this.writeLit(c)
            out.write(" ")
            this.writeLit(t)
            out.write(" ")
            this.writeLit(e)
            this.writeSwitch(sw)
          }
        }
        out.write("\n")
      }
    }

    private def writeCards(f: FormulaLike)(implicit out: BufferedWriter): Unit = f.cards foreach {
      x => {
        val (c, swOpt) = x
        out.write("assert card: ")
        c.lits foreach {
          l => {
            out.write(" + ")
            this.writeLit(l)
          }
        }
        out.write(" ")
        out.write(c.relOp.repr)
        out.write(" ")
        out.write(c.k.toString)
        out.write("\n")
      }
    }

    private def writeCriteria(f: FormulaLike)(implicit out: BufferedWriter): Unit = f.criteria foreach {
      crit =>
        crit match {
          case f.Maximize(lits, weights, k) => {
            out.write("max: ")
            out.write(k.toString)
            (weights zip lits) foreach {
              x => {
                val (w, l) = x
                out.write(" + ")
                out.write(w.toString)
                out.write(" * ")
                this.writeLit(l)
              }
            }
            out.write("\n")
          }
          case f.Minimize(lits, weights, k) => {
            out.write("min: ")
            out.write(k.toString)
            (weights zip lits) foreach {
              x => {
                val (w, l) = x
                out.write(" + ")
                out.write(w.toString)
                out.write(" * ")
                this.writeLit(l)
              }
            }
            out.write("\n")
          }
        }
    }
  }

}
