package util.sat

/**
  * Created by rdelmas on 21/06/16.
  */

import scala.collection.immutable.{Vector, Map => IMap, Set => ISet}
import scala.collection.mutable.{HashSet => MSet, OpenHashMap => MMap}

/** This class takes a boolean formula as input, and outputs several files which which can be used for prime implicant enumeration. */
case class PrimeImplicantGenerator(
                                    formula: Prop.Expr,
                                    filePrefix: String,
                                    maxOrder: Option[Int] = None,
                                    encodeForPrimer: Boolean = false,
                                    properTreeOpt: Boolean = false,
                                    monotonyOpt: Boolean = false
                                  ) {

  /** Check that NNF Atoms occupy the [1, nofAtoms] dimacs range in the circuit encoder. */
  private def checkPrimerIndices(nnf: NNF.Expr, f: FormulaLike, encoder: NNFToCircuitEncoder) = {
    val (min, max) = nnf.atoms.foldLeft((f.nofVars, 0))(
      (acc, atom) => {
        val (min, max) = acc
        val i = encoder(atom)
        val newMin = math.min(min, i)
        val newMax = math.max(max, i)
        (newMin, newMax)
      })
    assert(min == 1 & max == nnf.atoms.size)
  }

  /** Generates the list of projection scope literals that must be used in the dual rail formula. */
  private def genProjetctionLits(nnf: NNF.Expr, primaryEncoder: NNFToCircuitEncoder, dualRailEncoder: DualRailEncoder): List[Int] = {
    nnf.atoms.flatMap(
      a => {
        val (pos, neg) = dualRailEncoder.bothLits(primaryEncoder(a))
        pos :: neg :: Nil
      }).toList
  }

  /** Generates the projection scope header string for the Clingo solver. */
  private def genClingoProjectionString(nnf: NNF.Expr, primaryEncoder: NNFToCircuitEncoder, dualRailEncoder: DualRailEncoder): String = {
    genProjetctionLits(nnf, primaryEncoder, dualRailEncoder).map(i => s"x${i}").mkString("project ", " ", "")
  }

  /** Runs the clingo solver in clasp mode with model projection, returns prime implicants. */
  private def runClingo(opbFile: String, resultFile: String): List[ISet[Int]] = {
    // invoque external solver
    import scala.sys.process._
    val command = s"binlib/claspProj.sh ${opbFile} ${resultFile}"
    println("[verbose] " + command)
    command.!
    //val command s"clingo --mode=clasp -n0 -q0 --parse-ext --out-hide -t 4 --config=handy ${opbFile} > ${resultFile}"
    //println("[verbose] " + command)
    //val result  = s"clingo --mode=clasp -n0 -q0 --parse-ext --out-hide -t 4 --config=handy " + opbFile #> resultFile !
    parseClingoModel(resultFile)
  }

  /**
    * Parses a clingo model file, containing potentially several models of a formula.
    * TODO implement real parsing, string scanning + splitting does not scale
    */
  def parseClingoModel(resultFile: String): List[ISet[Int]] = {
    val zero: (List[ISet[Int]]) = Nil
    scala.io.Source.fromFile(resultFile).getLines().foldLeft(zero)(
      (acc, line) => {
        if (line.startsWith("c Answer")) {
          // new model
          ISet[Int]() :: acc
        } else if (line.startsWith("v")) {
          // it is a value line, append to current head set
          val values = line.drop(2).split(" ").map(s => s.replace("x", "").toInt)
          acc.head ++ values :: acc.tail
        } else {
          // do nothing on other type of lines
          acc
        }
      })
  }

  /** Generates the file that defines the projection scope for the sharpCDCL solver. */
  private def genSharpCDCLScopeFile(nnf: NNF.Expr, primaryEncoder: NNFToCircuitEncoder, dualRailEncoder: DualRailEncoder, projScopeFile: String): Unit = {
    val projectionLits = genProjetctionLits(nnf, primaryEncoder, dualRailEncoder)
    util.io.WriteToFile(projScopeFile, true)(out => {
      projectionLits.foreach(lit => out.write(s"${lit.toString}\n"))
    })
  }

  /** Runs the sharpCDCL solver on the given file with projection. */
  private def runSharpCDCL(projScopeFile: String, cnfFile: String, resultFile: String): List[ISet[Int]] = {
    // invoque external solver
    import scala.sys.process._
    val command = s"binlib/sharpCDCL.sh ${projScopeFile} ${resultFile} ${cnfFile}"
    println("[verbose] " + command)
    command.!
    parseSharpCDCLModelFile(resultFile)
  }

  /**
    * Parses a sharpCDCL model file, possibly containing several models of a same formula.
    * TODO implement real parsing, string splitting + scanning does not scale
    */
  def parseSharpCDCLModelFile(resultFile: String): List[ISet[Int]] = {
    // parse result file line per line to return prime implicants
    val zero: List[ISet[Int]] = Nil
    scala.io.Source.fromFile(resultFile).getLines().foldLeft(zero)((acc, line) => {
      ISet(line.split(" ").map(_.toInt): _*) :: acc
    })
  }

  def genPrimeImplicants(): List[ISet[Prop.Expr]] = {

    /** Translator to negation normal form. */
    val toNNF = new PropToNNF

    /** Formula in negation normal form. */
    val nnf = toNNF(formula)

    /** Container for generated formula */
    val circuit = new FormulaLike {}

    /** Container for generated formula */
    val primerPos = new FormulaLike {}

    /** Container for generated formula */
    val primerNeg = new FormulaLike {}

    /** Translator used to translate NNF formula to circuit. */
    val primaryEncoder = NNFToCircuitEncoder(nnf, circuit, switchableGates = true)

    /** Translator used to translate NNF formula to circuit. */
    val primerPosEncoder = NNFToCircuitEncoder(nnf, primerPos, switchableGates = false)

    /** Translator used to translate NNF formula to circuit. */
    val primerNegEncoder = NNFToCircuitEncoder(nnf, primerNeg, switchableGates = false)

    /** Witness circuit translator. */
    val witnessEncoder = WitnessEncoder(nnf, primaryEncoder, circuit)

    /** Dual rail encoded formula. */
    val dualRail: FormulaLike = new FormulaLike {}

    /** The encoder for the dual rail translation. */
    val dualRailEncoder = new DualRailEncoder(circuit, dualRail)

    /** Switch constraints encoder */
    val switchEncoder = SwitchEncoder(nnf, circuit, dualRail, dualRailEncoder)

    /** Bijection constraints encoder. */
    val bijectionEncoder = BijectionEncoder(nnf, circuit, primaryEncoder, witnessEncoder, dualRail, dualRailEncoder)

    /** Monotonicity optims encoder. */
    val monotonyOptimEncoder = MonotonyOptimEncoder(nnf, circuit, primaryEncoder, witnessEncoder, dualRail, dualRailEncoder)

    /** Kids Independence optims encoder. */
    val properTreeOptimEncoder = ProperTreeOptimEncoder(nnf, circuit, primaryEncoder, witnessEncoder, dualRail, dualRailEncoder)

    /** gate independent from all parent's other kids optimisation. */
    val indepFromAllParentsOtherKids = IndepFromAllParentsOtherKids(nnf, circuit, primaryEncoder, witnessEncoder, dualRail, dualRailEncoder)

    // ==================== ENCODE ========================= //

    if (encodeForPrimer) {
      // Encode in format compatible with the primer solver for reference
      primerPosEncoder.encode()
      primerPos.addClause(primerPosEncoder.exprCacheFunc(nnf))()
      checkPrimerIndices(nnf, primerPos, primerPosEncoder)
      val fp = filePrefix + "_primer_p"
      FormulaFileWriter.Cnf(primerPos, fp, true)
      println(s"[verbose] formula written to file '${fp}.cnf'.")

      primerNegEncoder.encode()
      primerNeg.addClause(-primerNegEncoder.exprCacheFunc(nnf))()
      checkPrimerIndices(nnf, primerPos, primerPosEncoder)
      val fn = filePrefix + "_primer_n"
      FormulaFileWriter.Cnf(primerNeg, fn, true)
      println(s"[verbose] formula written to file '${fn}.cnf'.")
    }

    // encode primary circuit
    primaryEncoder.encode()

    // encode witness circuits
    witnessEncoder.encode()

    // apply dual-rail transformation to primary + witness formula
    dualRailEncoder.encode()

    // encode switch constraints in the dual rail formula
    switchEncoder.encode()

    // encode PI to models bijection constraints in the dual rail formula.
    bijectionEncoder.encode()

    // apply the "proper subtree" optimization to dual-rail formula.
    if (properTreeOpt) {
      properTreeOptimEncoder.encode()
    }

    // apply the "monotony" optimizaton to the dual-rail formula.
    if (monotonyOpt) {
      monotonyOptimEncoder.encode()
    }

    //====================== SOLVE ============================== //

    val drLits = genProjetctionLits(nnf, primaryEncoder, dualRailEncoder)

    val dualRailFile = filePrefix + "_drb"

    // TODO combine sharpCDCL with pbencoder to handle cardinality and pseudo boolean constraints

    val drModels = maxOrder match {

      // if cardinality constraints are used, use clingo with projection
      case Some(k) => {
        {
          dualRail.addAtMostK(drLits, k)(switch = None, switchable = false)
          dualRail.addHeaderComment(genClingoProjectionString(nnf, primaryEncoder, dualRailEncoder))
          FormulaFileWriter.Opb(dualRail, dualRailFile)
          println(s"[verbose] formula written to file '${dualRailFile}.opb'.")
          runClingo(dualRailFile + ".opb", dualRailFile + ".models")
          // TODO ??? dualRail.addCriterion(dualRail.Minimize(drLits, drLits.map(l => 1), 0)) ???
        }
      }

      // else use sharpCDCL with projection (much faster)
      case None => {
        {
          FormulaFileWriter.Cnf(dualRail, dualRailFile)
          println(s"[verbose] formula written to file '${dualRailFile}.cnf'.")
          val scopeFile = dualRailFile + ".vars"
          genSharpCDCLScopeFile(nnf, primaryEncoder, dualRailEncoder, scopeFile)
          runSharpCDCL(scopeFile, dualRailFile + ".cnf", dualRailFile + ".models")
        }
      }
    }

    // ================== TRANSLATE BACK MODELS TO PRIME IMPLICANTS ========================= //

    // conversion from prop to nnf atoms
    val propToNNF = toNNF.atomCacheFunc

    // conversion from nnf atoms to circuit variables
    val nnfToCircuit = primaryEncoder.atomCacheFunc

    // conversion from cricuit variables to dual rail variables pairs
    val circuitToDr = dualRailEncoder.funcMap

    // translate back to Prop formula prime implicants
    val propPrimeImplicants = drModels.foldLeft(Nil: List[ISet[Prop.Expr]]) {
      (acc, pi) => {
        formula.atoms.foldLeft(ISet[Prop.Expr]()) {
          (acc, atom) => {
            val (pos_lit, neg_lit) = circuitToDr(nnfToCircuit(propToNNF(atom)))
            if (pi.contains(pos_lit)) acc + atom else if (pi.contains(neg_lit)) acc + Prop.Not(atom) else acc
          }
        } :: acc
      }
    }

    // translate back to NNF prime implicants
    val nnfPrimeImplicants = drModels.foldLeft(Nil: List[ISet[NNF.Literal]]) {
      (acc, pi) => {
        nnf.atoms.foldLeft(ISet[NNF.Literal]()) {
          (acc, atom) => {
            val circVar = nnfToCircuit(atom)
            val (pos_lit, neg_lit) = circuitToDr(circVar)
            if (pi.contains(pos_lit)) acc + atom else if (pi.contains(neg_lit)) acc + NNF.Not(atom) else acc
          }
        } :: acc
      }
    }


    if (encodeForPrimer) {
      // translate back to primer circuit implicants

      val primerPrimeImplicants = nnfPrimeImplicants.map(
        pi => pi.map(
          lit => lit match {
            case NNF.Not(atom) => -primerPosEncoder(atom)
            case atom: NNF.Atom => primerPosEncoder(atom)
          }))

      val primerPIFile = filePrefix + "_primer_implicants.dnf"

      util.io.WriteToFile(primerPIFile, true) {
        out => {
          primerPrimeImplicants.foreach { pi => out.write(pi.mkString("", " ", " 0\n")) }
        }
      }
      println(s"[verbose] prime implicants for primer model written to '${primerPIFile}'")

      val primerBCFile = filePrefix + "_primer_blocking_clauses.cnf"

      util.io.WriteToFile(primerBCFile, true) {
        out => {
          primerPrimeImplicants.foreach { pi => out.write(pi.map(x => -x).mkString("", " ", " 0\n")) }
        }
      }
      println(s"[verbose] blocking clauses for primer model written to '${primerBCFile}'")
    }

    // return results
    propPrimeImplicants

    // TODO saturate results with consesus rule.
  }
}

/**
  * Adds new constraints to the dual-rail encoding for an optimization that is applied when a given gate g is such that:
  * -  the COI of g does not intersect the COI all of its parent's other kids
  *
  * Under such conditions, the gate is part of a prime implicant iff when different from "?" some clause 'c' in the set of all clauses defining its parents gates is such that c - {g}, is false.
  *
  * Applied only on gates of the primary circuit.
  */
// TODO find proper criterion for applying the optimization
case class IndepFromAllParentsOtherKids(nnf: NNF.Expr, circuit: FormulaLike, primaryEncoder: NNFToCircuitEncoder, witnessEncoder: WitnessEncoder, drFormula: FormulaLike, drEncoder: DualRailEncoder) {

  lazy val toCircuitLit = primaryEncoder.exprCacheFunc

  lazy val toDrLits = primaryEncoder.exprCacheFunc andThen drEncoder.funcMap

  /** Generates additional clauses encoding the optimization. */
  def mkOptim(gateLit: Int, clauses: MSet[ISet[Int]]): Unit = {
    val restrictedClausesWithNeg = clauses.map {
      clause => {
        val x = drFormula.freshVar
        drFormula.defineOr(x, (clause - gateLit).toList)()
        -x
      }
    }.toList
    val someClauseFalse = drFormula.freshVar
    drFormula.defineOr(someClauseFalse, restrictedClausesWithNeg)()
    val implication = drFormula.freshVar
    drFormula.defineImplies(implication, gateLit, someClauseFalse)()
    drFormula.addClause(implication)()
  }

  /** Encodes the optimization formulae. */
  def encode(): Unit = {

    // build index from literals to clauses in which they appear
    val clauseIndex: MMap[Int, MSet[ISet[Int]]] = new MMap()
    drFormula.clauses foreach {
      x => {
        val (clause, swOpt) = x
        val cl = ISet() ++ clause.lits
        clause.lits.foreach(lit => {
          val prevSet = clauseIndex.getOrElseUpdate(math.abs(lit), new MSet())
          prevSet.add(cl)
        })
      }
    }

    // traverse the nnf formula graph in topological order and apply optimisation
    val depGraph = NNF.genDepGraph(nnf)
    var nofOpt = 0
    // iterate from leaves to root
    depGraph.topsort() foreach {
      node => {
        val g = node.data
        val (posG, negG) = toDrLits(g)
        val posGCl = clauseIndex(posG)
        val negGCl = clauseIndex(negG)
        val parents = node.predecessors.map(_.data)
        val allOtherKidsIncludingG = node.predecessors.flatMap(_.data.kids)

        val canOptimizeG = false

        // TODO find proper criterion to enable optim
        //        val canOptimizeG = allOtherKidsIncludingG.foldLeft(true) {
        //          (soFar, kid) => {
        //            allOtherKidsIncludingG.foldLeft(soFar) {
        //              (soFar, otherKid) => {
        //                soFar && ( kid == otherKid && kid == g || kid.atoms.intersect(otherKid.atoms).isEmpty )
        //              }
        //            }
        //          }
        //        }

        //        // node can be optimized iff it g is COI independent of all its parents other kids
        //        val canOptimizeG = parents.foldLeft(true)(
        //          (soFar, p) => soFar && p.kids.foldLeft(soFar)(
        //            (soFar, kid) => soFar && ( kid == g || kid.atoms.intersect(g.atoms).isEmpty )
        //          )
        //        )

        if (canOptimizeG) {
          nofOpt += 1
          parents foreach {
            parent => {
              val (posP, negP) = toDrLits(parent)
              // fetch all clauses in which both the parent and g appear
              val posPCl = clauseIndex(posP)
              val negPCl = clauseIndex(negP)
              // set of clauses in which both l_g and either l_parent or l_not_parent appear, with l_g removed from each clause
              val pClauses = posPCl.intersect(posGCl).union(negPCl.intersect(posGCl))
              mkOptim(posP, pClauses)
              // set of clauses in which both l_not_g and either l_parent or l_not_parent appear, with l_not_g removed from each clause
              val nClauses = posPCl.intersect(negGCl).union(negPCl.intersect(negGCl))
              mkOptim(negP, nClauses)
            }
          }
        }
      }
    }
    if (nofOpt > 0) {
      println(s"[verbose] indep from all parent's other kids: ${nofOpt} nodes optimized.")
    }
  }
}

/* Adds the following constraints for each gate of the primary circuit:
  *  If a gate is in a monotonic position wrt the root of the formula, only values 1 and ? are allowed.
  *  // TODO disabled until bug fixed...
  *  */
case class MonotonyOptimEncoder(nnf: NNF.Expr, circuit: FormulaLike, primaryEncoder: NNFToCircuitEncoder, witnessEncoder: WitnessEncoder, drFormula: FormulaLike, drEncoder: DualRailEncoder) {

  /** Encodes the optimization formulae. */
  def encode(): Unit = {

    val toCircuitLit = primaryEncoder.exprCacheFunc orElse witnessEncoder.exprCacheFunc
    val (depGraph, monotonic) = NNF.genIsMonotonic(nnf)
    var nofOpt = 0
    monotonic.foreach {
      x => {
        val (e, isMono) = x
        if (isMono) {
          //println( "monotonic: " + e.toString )
          nofOpt += 1
          e match {
            case e: NNF.Literal => {
              val (_, negLit) = drEncoder.bothLits(toCircuitLit(e))
              drFormula.addClause(-negLit)()
            }
            case e: NNF.Or => {
              val (_, negLit) = drEncoder.bothLits(toCircuitLit(e))
              drFormula.addClause(-negLit)()
            }
            case e: NNF.And => {
              val (_, negLit) = drEncoder.bothLits(toCircuitLit(e))
              drFormula.addClause(-negLit)()
            }
            case e: NNF.Ite => {
              val (_, negLit) = drEncoder.bothLits(toCircuitLit(e))
              drFormula.addClause(-negLit)()
            }
          }
        } else {
          println("not monotonic: " + e.toString)
        }
      }
    }
    if (nofOpt > 0) {
      println(s"[verbose] monotony: ${nofOpt} nodes optimized.")
    }
  }
}

/**
  * Implements the optimisation applicable to gates for which all kids are independent and have a unique parent,
  * ie applicable only to parts of the nnf formula that are genuine trees from the root and down to input literals.
  *
  * This optim consists in forcing the maximum number of undefined kids for each possible output value of the gates
  *
  *
  * For And(kids):
  * - assert (o = 0) => nofKid(?) >= g.nofKids-1
  * - assert (o = ?) => nofKids(?) >= g.nofKids
  *
  * For Or(kids):
  * - assert (o = 1) => nofKid(?) >= g.nofKids-1
  * - assert (o = ?) => nofKids(?) >= g.nofKids
  *
  * For Ite:
  * - assert (o = 0) => optFalse,
  *
  * with:
  * - define optFalse := And(optFalseFalse, optFalseTrue, optFalseUndef)
  *
  * with:
  * - define optFalseFalse := (s = 0) => (t = ? and f = 0)
  * - define optFalseTrue  := (s = 1) => (t = 0 and f = ?)
  * - define optFalseUndef := (s = ?) => (t = ? and f = ?)
  *
  * And
  * - assert (o = 1) => optTrue
  *
  * with:
  * - define optTrue := And( optTrueFalse, optTrueTrue, optTrueUndef )
  *
  * with:
  * - define optTrueFalse := (s = 0) => (t = ? and f = 1)
  * - define optTrueTrue  := (s = 1) => (t = 1 and f = ?)
  * - define optTrueUndef := (s = ?) => (t = 1 and f = 1)
  *
  * And
  * - assert (o = ?) => optUndef
  *
  * with:
  * - define optUndef := (t = ? and f = ?)
  *
  */
case class ProperTreeOptimEncoder(nnf: NNF.Expr, circuit: FormulaLike, primaryEncoder: NNFToCircuitEncoder, witnessEncoder: WitnessEncoder, drFormula: FormulaLike, drEncoder: DualRailEncoder) {

  /** Encodes the optimization formulae. */
  def encode(): Unit = {

    import NNF._
    val toCircuitLit = primaryEncoder.exprCacheFunc orElse witnessEncoder.exprCacheFunc

    def toDualRailLits(e: Expr) = drEncoder.bothLits(toCircuitLit(e))

    def getDrSwitch(e: Expr) = drEncoder(circuit.gates(toCircuitLit(e))._2.get)

    // traverse main formula
    val (depGraph, isInProperTree) = genIsInProperTree(nnf)
    var nofOpt = 0
    depGraph.topsort() foreach {
      genConstraints
    }

    def genConstraints(n: depGraph.Node): Unit = {
      val e = n.data
      if (isInProperTree(e) && e.pairwiseIndepKids) {
        nofOpt += 1
        e match {
          case l: Literal => {
            // do nothing
          }
          case And(kids) => {
            val swE = getDrSwitch(e)
            val (posE, negE) = toDualRailLits(e)
            val kidSwitches = kids.map(kid => getDrSwitch(kid))
            val kidLits = kids.map(kid => toDualRailLits(kid))
            drFormula.addAtLeastK(kidSwitches.map(sw => -sw), e.nofKids - 1)(switch = Some(negE))
            drFormula.addAtLeastK(kidSwitches.map(sw => -sw), e.nofKids)(switch = Some(-swE))
          }
          case Or(kids) => {
            val swE = getDrSwitch(e)
            val (posE, negE) = toDualRailLits(e)
            val kidSwitches = kids.map(kid => getDrSwitch(kid))
            val kidLits = kids.map(kid => toDualRailLits(kid))
            drFormula.addAtLeastK(kidSwitches.map(sw => -sw), e.nofKids - 1)(switch = Some(posE))
            drFormula.addAtLeastK(kidSwitches.map(sw => -sw), e.nofKids)(switch = Some(-swE))
          }
          case Ite(s, t, f) => {

            val swE = getDrSwitch(e)
            val (posE, negE) = toDualRailLits(e)

            val swS = getDrSwitch(s)
            val (posS, negS) = toDualRailLits(s)

            val swT = getDrSwitch(t)
            val (posT, negT) = toDualRailLits(t)

            val swF = getDrSwitch(f)
            val (posF, negF) = toDualRailLits(f)

            val optFalse = drFormula.freshVar
            val optTrue = drFormula.freshVar
            val optUndef = drFormula.freshVar

            /* assert (o = 0) => optFalse */
            drFormula.addClause(-negE, optFalse)()

            val optFalseFalse = drFormula.freshVar
            val optFalseTrue = drFormula.freshVar
            val optFalseUndef = drFormula.freshVar

            /* define optFalse := And( optFalseFalse, optFalseTrue, optFalseUndef) */
            drFormula.defineAnd(optFalse, List(optFalseFalse, optFalseTrue, optFalseUndef))()

            /* define optFalseFalse := (s = 0) => (t = ? and f = 0) */
            {
              val conj = drFormula.freshVar
              drFormula.defineAnd(conj, List(-swT, negF))()
              drFormula.defineImplies(optFalseFalse, negS, conj)()
            }

            /* define optFalseTrue := (s = 1) => (t = 0 and f = ?) */
            {
              val conj = drFormula.freshVar
              drFormula.defineAnd(conj, List(negT, -swF))()
              drFormula.defineImplies(optFalseTrue, posS, conj)()
            }

            /* define optFalseUndef := (s = ?) => (t = 0 and f = 0) */
            {
              val conj = drFormula.freshVar
              drFormula.defineAnd(conj, List(negT, negF))()
              drFormula.defineImplies(optFalseUndef, -swS, conj)()
            }

            /* assert (o = 1) => optTrue */
            drFormula.addClause(-negE, optTrue)()

            val optTrueFalse = drFormula.freshVar
            val optTrueTrue = drFormula.freshVar
            val optTrueUndef = drFormula.freshVar

            /* define optTrue := And(optTrueFalse, optTrueTrue, optTrueUndef) */
            drFormula.defineAnd(optTrue, List(optTrueFalse, optTrueTrue, optTrueUndef))()

            /* define optTrueFalse := (s = 0) => (t = ? and f = 1) */
            {
              val conj = drFormula.freshVar
              drFormula.defineAnd(conj, List(-swT, posF))()
              drFormula.defineImplies(optTrueFalse, negS, conj)()
            }

            /* define optTrueTrue := (s = 1) => (t = 1 and f = ?) */
            {
              val conj = drFormula.freshVar
              drFormula.defineAnd(conj, List(posT, -swF))()
              drFormula.defineImplies(optTrueTrue, posS, conj)()
            }

            /* define optTrueUndef := (s = ?) => (t = 1 and f = 1) */
            {
              val conj = drFormula.freshVar
              drFormula.defineAnd(conj, List(posT, posF))()
              drFormula.defineImplies(optTrueUndef, -swS, conj)()
            }

            /* assert (o = ?) => optUndef */
            drFormula.addClause(-swE, optUndef)()

            /* define optUndef := (t = ? and f = ?) */
            drFormula.defineAnd(optUndef, List(-swT, -swF))()
          }
        }
      }
    }

    if (nofOpt > 0) {
      println(s"[verbose] indep kids: ${nofOpt} nodes optimized.")
    }

  }
}

/**
  * Adds constraints to the dual rail formula that enforce a bijection between models of the dual rail formula and prime implicants of the original formula:
  * - the input atom of each witness circuit is equal to the inverse of the corresponding input when is switch is on, and to the input otherwise.
  * - the positive dual-rail literal of the witness circuit must be false (by adding a unit clause).
  */
case class BijectionEncoder(nnf: NNF.Expr, circuit: FormulaLike, primaryEncoder: NNFToCircuitEncoder, witnessEncoder: WitnessEncoder, drFormula: FormulaLike, drEncoder: DualRailEncoder) {
  def encode(): Unit = {
    // assert that primary output is true
    drFormula.addClause(drEncoder.bothLits(primaryEncoder.output.get)._1)()
    witnessEncoder.encoders.values.foreach {
      encoder => {
        // add clauses defining witness inputs in terms of primary inputs
        // witnessAtom = if (switchAtom) invert(Atom) else Atom
        val primA = primaryEncoder(encoder.keyAtom)
        val (_, Some(sw)) = circuit.gates(primA)
        val swPrimA = drEncoder(sw)
        val (posPrimA, negPrimA) = drEncoder.bothLits(primA)
        val (posWitA, negWitA) = drEncoder.bothLits(encoder.input.get)
        val (posWitOut, negWitOut) = drEncoder.bothLits(encoder.output.get)

        // make it undef when witness selected
        drFormula.addClause(-posWitA)(switch = Some(swPrimA))
        drFormula.addClause(-negWitA)(switch = Some(swPrimA))

        // make it equal to primary when witness not selected
        val posWitEqPrim = drFormula.freshVar
        drFormula.defineIff(posWitEqPrim, posWitA, posPrimA)()
        drFormula.addClause(posWitEqPrim)(switch = Some(-swPrimA))
        val negWitEqPrim = drFormula.freshVar
        drFormula.defineIff(posWitEqPrim, posWitA, posPrimA)()
        drFormula.addClause(negWitEqPrim)(switch = Some(-swPrimA))

        // add clause asking that witness output pos literal is false
        drFormula.addClause(-swPrimA, -posWitOut)()
      }
    }
  }
}

/** Encodes switches definitions in the dual rail formula to enforce proper ternary valued models. */
case class SwitchEncoder(nnf: NNF.Expr, circuit: FormulaLike, drFormula: FormulaLike, drEncoder: DualRailEncoder) {
  def encode(): Unit = {
    // iterate over all gates of the circuit, and generate corresponding gate switch constraints in the dual rail formula
    circuit.gatesIterator.foreach {
      x => {
        val (gate, (gateType, Some(sw))) = x
        val (posGate, negGate) = drEncoder.bothLits(gate)
        val gateSw = drEncoder(sw)
        gateType match {
          case circuit.Input => {
            drFormula.defineOr(gateSw, posGate :: negGate :: Nil)()
          }
          case circuit.Not(kid) => {
            val (posKid, negKid) = drEncoder.bothLits(kid)
            drFormula.defineOr(gateSw, posKid :: negKid :: Nil)()
            drFormula.addClause(gateSw, -posGate)()
            drFormula.addClause(gateSw, -negGate)()
          }
          case circuit.And(kids) => {
            val (posKids, negKids) = kids.map(k => drEncoder.bothLits(k)).unzip
            val allKidsTrue = drFormula.freshVar
            drFormula.defineAnd(allKidsTrue, posKids)()
            drFormula.defineOr(gateSw, allKidsTrue :: negKids)()
            drFormula.addClause(gateSw, -posGate)()
            drFormula.addClause(gateSw, -negGate)()
          }
          case circuit.Or(kids) => {
            val (posKids, negKids) = kids.map(k => drEncoder.bothLits(k)).unzip
            val allKidsFalse = drFormula.freshVar
            drFormula.defineAnd(allKidsFalse, negKids)()
            drFormula.defineOr(gateSw, allKidsFalse :: posKids)()
            drFormula.addClause(gateSw, -posGate)()
            drFormula.addClause(gateSw, -negGate)()
          }
          case circuit.Ite(s, t, f) => {
            val (l_s, l_not_s) = drEncoder.bothLits(s)
            val (_, Some(sws)) = circuit.gates(s)
            val sw_s = drEncoder(sws)
            val (l_t, l_not_t) = drEncoder.bothLits(t)
            val (_, Some(swt)) = circuit.gates(t)
            val sw_t = drEncoder(swt)
            val (l_f, l_not_f) = drEncoder.bothLits(f)
            val (_, Some(swf)) = circuit.gates(f)
            val sw_f = drEncoder(swf)
            val case1 = drFormula.freshVar
            drFormula.defineAnd(case1, l_s :: sw_t :: Nil)()
            val case2 = drFormula.freshVar
            drFormula.defineAnd(case2, l_not_s :: sw_f :: Nil)()
            val iff1 = drFormula.freshVar
            drFormula.defineIff(iff1, l_t, l_f)()
            val iff2 = drFormula.freshVar
            drFormula.defineIff(iff2, l_not_t, l_not_f)()
            val case3 = drFormula.freshVar
            drFormula.defineAnd(case3, -sw_s :: sw_t :: sw_f :: iff1 :: iff2 :: Nil)()
            drFormula.defineOr(gateSw, case1 :: case2 :: case3 :: Nil)()
            drFormula.addClause(gateSw, -posGate)()
            drFormula.addClause(gateSw, -negGate)()
          }
          case _ => throw new Exception("impossible")
        }
      }
    }
  }
}

/**
  * Translator for a primary circuit. If the switchableGates parameter is false, then atoms of the nnf formula are guaranteed to use the index range [1, nnf.atoms.size].
  *
  * @param nnf             the nnf expression to translate
  * @param circuit         the target circuit in which to encode the formula
  * @param switchableGates if set to true, all gates created in the circuit will be switchable
  *
  */
case class NNFToCircuitEncoder(nnf: NNF.Expr, circuit: FormulaLike, switchableGates: Boolean) {

  /** Inputs of the primary circuit. */
  var inputs: Vector[Int] = Vector()

  /** Output of the primary circuit. */
  var output: Option[Int] = None

  /** Memoization map from nnf atoms to circuit variables. */
  private val _atomCache: MMap[NNF.Atom, Int] = new MMap()

  /** Memoization map from nnf expressions to circuit variables. */
  private val _exprCache: MMap[NNF.Expr, Int] = new MMap()

  /** @return The current atom translation cache as an immutable function. */
  def atomCacheFunc: PartialFunction[NNF.Expr, Int] = scala.collection.immutable.Map() ++ _atomCache

  /** @return The current expression translation cache as an immutable function. */
  def exprCacheFunc: PartialFunction[NNF.Expr, Int] = scala.collection.immutable.Map() ++ _exprCache

  /** @return An immutable map from nnf Atoms to circuit variables. */
  def atoms: Map[NNF.Atom, Int] = Map() ++ _atomCache

  /**
    * Translate nnf formula to primary circuit, asserts root gate is true and and declares it as output.
    * Encodes nnf atoms first to that their corresponding variables and optional switches
    * use the lowest possible indices of the target formula.
    */
  def encode(): Unit = {

    // encode atoms first so that free literals are in the lowest indices range
    nnf.atoms.foreach(apply)

    // encode whole formula
    val o = apply(nnf)
    circuit.declareOutput(o)
    output = Some(o)
  }

  def apply(e: NNF.Expr): Int = {
    _exprCache.getOrElseUpdate(e, applyRec(e))
  }

  /** Translates the primary nnf formula to a boolean circuit in the result formula. */
  private def applyRec(e: NNF.Expr): Int = e match {
    case a@NNF.Atom(id) => {
      val x = circuit.freshVar
      inputs :+= x
      _atomCache.update(a, x)
      val swOpt = circuit.declareInput(x)(switchable = switchableGates)
      if (swOpt.isDefined) {
        circuit.freeze(swOpt.get)
      }
      x
    }
    case NNF.Not(atom) => {
      val newAtom = apply(atom)
      val x = circuit.freshVar
      val swOpt = circuit.defineNot(x, newAtom)(switchable = switchableGates)
      if (swOpt.isDefined) {
        circuit.freeze(swOpt.get)
      }
      x
    }
    case NNF.And(kids) => {
      val newKids = kids.map(apply)
      val x = circuit.freshVar
      val swOpt = circuit.defineAnd(x, newKids)(switchable = switchableGates)
      if (swOpt.isDefined) {
        circuit.freeze(swOpt.get)
      }
      x
    }
    case NNF.Or(kids) => {
      val newKids = kids.map(apply)
      val x = circuit.freshVar
      val swOpt = circuit.defineOr(x, newKids)(switchable = switchableGates)
      if (swOpt.isDefined) {
        circuit.freeze(swOpt.get)
      }
      x
    }
    case NNF.Ite(kid1, kid2, kid3) => {
      val newKid1 = apply(kid1)
      val newKid2 = apply(kid2)
      val newKid3 = apply(kid3)
      val x = circuit.freshVar
      val swOpt = circuit.defineIte(x, newKid1, newKid2, newKid3)(switchable = switchableGates)
      if (swOpt.isDefined) {
        circuit.freeze(swOpt.get)
      }
      x
    }
  }
}

/** Witness circuit translation function. */
case class WitnessEncoder(nnf: NNF.Expr, primaryEncoder: NNFToCircuitEncoder, circuit: FormulaLike) extends {

  /** Witness translators map. */
  val encoders: MMap[NNF.Atom, NNFAtomWitnessEncoder] = new MMap()

  // Initialise containers.
  for (keyAtom <- nnf.atoms) {
    encoders.update(keyAtom, NNFAtomWitnessEncoder(nnf, keyAtom, primaryEncoder, circuit))
  }

  /** Returns the current expression translation cache as an immutable function. */
  def atomCacheFunc: PartialFunction[NNF.Expr, Int] = {
    encoders.values.foldLeft(primaryEncoder.atomCacheFunc) {
      (acc, encoder) => {
        acc orElse encoder.atomCacheFunc
      }
    }
  }

  /** Returns the current expression translation cache as an immutable function. */
  def exprCacheFunc: PartialFunction[NNF.Expr, Int] = {
    encoders.values.foldLeft(primaryEncoder.exprCacheFunc) {
      (acc, encoder) => {
        acc orElse encoder.exprCacheFunc
      }
    }
  }

  /** Generates a witness circuit from the nnf formula for each atom of the nnf. */
  def encode(): Unit = {
    for (encoder <- encoders.values) {
      encoder.encode()
    }
  }
}

/** Translates NNF expressions to witness circuit for keyAtom. */
case class NNFAtomWitnessEncoder(nnf: NNF.Expr, keyAtom: NNF.Atom, primaryEncoder: NNFToCircuitEncoder, circuit: FormulaLike) extends (NNF.Expr => Int) {

  /** New free input of the witness circuit corresponding to the key atom. */
  var input: Option[Int] = None

  /** Outputs of the witness circuit. */
  var output: Option[Int] = None

  /** Memoization map from nnf expressions to circuit variables. */
  private val _exprCache: MMap[NNF.Expr, Int] = new MMap()

  /** Memoization map from nnf atims to circuit variables. */
  private val _atomCache: MMap[NNF.Atom, Int] = new MMap()

  /** Returns the current atom translation cache as an immutable function. */
  def atomCacheFunc: PartialFunction[NNF.Expr, Int] = scala.collection.immutable.Map() ++ _atomCache

  /** Returns the current expression translation cache as an immutable function. */
  def exprCacheFunc: PartialFunction[NNF.Expr, Int] = scala.collection.immutable.Map() ++ _exprCache

  /** @return An immutable map from nnf Atoms to circuit variables. */
  def atoms: Map[NNF.Atom, Int] = Map() ++ _atomCache

  /** Translate nnf formula to witness circuit, declares resulting gate as output. */
  def encode(): Unit = {
    // generate a witness
    val o = apply(nnf)
    output = Some(o)
    circuit.declareOutput(o)
  }

  /** Translates the given NNF formula as a witness circuit. */
  def apply(e: NNF.Expr): Int = _exprCache.getOrElseUpdate(e, applyRec(e))

  /** Translates the NNF formula as a witness circuit. */
  private def applyRec(e: NNF.Expr): Int = e match {
    case a@NNF.Atom(id) => {
      if (a == keyAtom) {
        val x = circuit.freshVar
        input = Some(x)
        val swOpt = circuit.declareInput(x)(switchable = true)
        val sw = swOpt.get
        circuit.freeze(sw)
        x
      } else {
        primaryEncoder(a)
      }
    }
    case NNF.Not(atom) => {
      val oldAtom = primaryEncoder(atom)
      val newAtom = apply(atom)
      if (newAtom != oldAtom) {
        val x = circuit.freshVar
        val swOpt = circuit.defineNot(x, newAtom)(switchable = true)
        val sw = swOpt.get
        circuit.freeze(sw)
        x
      } else {
        primaryEncoder(e)
      }
    }
    case NNF.And(kids) => {
      val oldKids = kids.map(k => primaryEncoder(k))
      val newKids = kids.map(k => apply(k))
      if (newKids != oldKids) {
        val x = circuit.freshVar
        val swOpt = circuit.defineAnd(x, newKids)(switchable = true)
        val sw = swOpt.get
        circuit.freeze(sw)
        x
      } else {
        primaryEncoder(e)
      }
    }
    case NNF.Or(kids) => {
      val oldKids = kids.map(k => primaryEncoder(k))
      val newKids = kids.map(k => apply(k))
      if (newKids != oldKids) {
        val x = circuit.freshVar
        val swOpt = circuit.defineOr(x, newKids)(switchable = true)
        val sw = swOpt.get
        circuit.freeze(sw)
        x
      } else {
        primaryEncoder(e)
      }
    }
    case NNF.Ite(kid1, kid2, kid3) => {
      val oldKid1 = primaryEncoder(kid1)
      val oldKid2 = primaryEncoder(kid2)
      val oldKid3 = primaryEncoder(kid3)
      val newKid1 = apply(kid1)
      val newKid2 = apply(kid2)
      val newKid3 = apply(kid3)
      if (newKid1 != oldKid1 || newKid2 != oldKid2 || newKid3 != oldKid3) {
        val x = circuit.freshVar
        val swOpt = circuit.defineIte(x, newKid1, newKid2, newKid3)(switchable = true)
        val sw = swOpt.get
        circuit.freeze(sw)
        x
      } else {
        primaryEncoder(e)
      }
    }
  }
}
