package util.sat

/**
  * Created by rdelmas on 17/05/16.
  */

import scala.util.Try

trait ProblemLike {

  /** @return {{{Success(m)}}} with m the current model of the problem, must be called after [[isSatisfiable]] or [[isSatisfiable( assumptions )]], {{{Failure( e )}}} if no current model is available. */
  def model: Try[List[Int]]

  /** Generates a prime implicant of the problem if some exists.
    *
    * @return {{{Success(pi)}}} of the problem if any exists, {{{Failure(e)}}} otherwise.
    * */
  def primeImplicant: Try[List[Int]]

  /** Checks if the given literal is part of the prime implicant computed using the previous call to [[primeImplicant]].
    *
    * @return {{{Success(b)}}} where b indicates if lit is in the current prime implicant, {{{Failure(e)}}} if no current prime implicant is available.
    * */
  def primeImplicant(lit: Int): Try[Boolean]

  /** Checks the problem for satisfiability.
    *
    * @return true iff the problem is satisfiable.
    * */
  def isSatisfiable: Boolean

  /** Checks the problem for satisfiability.
    *
    * @param assumptions   a set of literals to assert during the check.
    * @param globalTimeout if true, timeout counter will not be reset between call.
    * @return true iff the problem is satisfiable under the given assumptions.
    * */
  def isSatisfiable(assumptions: List[Int], globalTimeout: Boolean): Boolean

  /** Checks the problem for satisfiability.
    *
    * @param globalTimeout if true, the timeout counter will not reset between each call.
    * @return true iff the problem is satisfiable.
    * */
  def isSatisfiable(globalTimeout: Boolean): Boolean

  /** Checks the problem for satisfiability.
    *
    * @param assumptions set of literals to assert during the check.
    * @return true iff the problem is satisfiable under the given assumptions.
    * */
  def isSatisfiable(assumptions: List[Int]): Boolean

  /** Checks the problem for satisfiability
    *
    * @return {{{Success(model)}}} if any exists, {{{Failure(e)}}} otherwise.
    */
  def findModel: Try[List[Int]]

  /** Checks the problem for satisfiability
    *
    * @param assumptions a set of literals to assert during the check.
    * @return {{{Succes(m)}}} if any exists, {{{Failure(e)}}} otherwise.
    */
  def findModel(assumptions: List[Int]): Try[List[Int]]

}
