package util.sat

import scala.collection.immutable

/**
  * Created by rdelmas on 20/06/16.
  */

/** An interface for building pseudo-boolean problems on top of CNF formulas. */
trait PseudoBooleanLike {
  this: CnfLike with CircuitLike =>

  /** Abstract type for relational operators used in cardinality constraints. */
  sealed trait RelOp {
    val repr: String
  }

  case object AtMost extends RelOp {
    val repr = "<="
  }

  case object AtLeast extends RelOp {
    val repr = ">="
  }

  case object Exactly extends RelOp {
    val repr = "="
  }

  sealed trait PseudoBoolCst {
    val lits: List[Int]
    val weights: Map[Int, Int]
    val relOp: RelOp
    val k: Int
  }

  /** A classic PB constraint f the form {{{ sum(weights(lit)*lit) relop k }}}. */
  final case class ClassicPBCst(lits: List[Int], relOp: RelOp, k: Int, weights: Map[Int, Int]) extends PseudoBoolCst

  /** A cardinality constraint of the form {{{ sum(lits) relop k }}}. */
  final case class Card(lits: List[Int], relOp: RelOp, k: Int) extends PseudoBoolCst {

    val weights = Map[Int, Int]().withDefaultValue(1)
    // Sanity check
    if (k < 0) throw new Exception(s"in cardinality constraint ${this.toString} k is negative")

    // Sanity check
    relOp match {
      case AtLeast | Exactly => if (lits.size < k) throw new Exception(s"cardinality constraint ${this.toString} is trivially false since |lits| < k")
      case _ => ()
    }

    /** Converts any cardinality constraint to an equivalent list (understood as conjunction) of AtLeast constraints. */
    def toAtLeast: List[Card] = relOp match {
      case AtLeast => this :: Nil
      case AtMost => Card(lits.map { lit => -lit }, AtLeast, lits.length - k) :: Nil
      case Exactly => Card(lits, AtLeast, k) :: Card(lits, AtMost, k).toAtLeast
    }
  }

  /** Cardinality constraints of the formula with optional switch literal. */
  private var _constraints: immutable.Vector[(PseudoBoolCst, Option[Int])] = immutable.Vector()

  /** @return an iterator on the cardinality constraints currently in the solver. */
  final def cards: Iterator[(PseudoBoolCst, Option[Int])] = this._constraints.iterator

  /** @return the number of cardinality constraints of the formula. */
  final def nofPBCst: Int = this._constraints.length

  /** Rewrites a cardinality constraint {{{ sum(lits) relop k }}}  to a switchable constraint depending on 'switch and switchable arguments'. */
  private
  final def toSwitchableCard(c: Card)(switch: Option[Int] = None, switchable: Boolean = false): (Card, Option[Int]) = {
    val swOpt = mkSwitchLit(switch, switchable)
    swOpt match {
      case Some(sw) => {

        val newLits = c.relOp match {
          /** rewrite
            * {{{ l1 + ... + ln <= k }}}
            * to
            * {{{ f1 + ... + fn <= k }}}
            * with forall i
            * {{{ fi = sw and li }}}
            */
          case AtMost => c.lits map {
            lit => {
              val x = freshVar
              defineAnd(x, sw :: lit :: Nil)()
              x
            }
          }

          /** rewrite
            * {{{ l1 + ... + ln >= k }}}
            * to
            * {{{ f1 + ... + fn >= k }}}
            * with forall i
            * {{{ fi = ~sw or li }}}
            */
          case AtLeast => c.lits map {
            lit => {
              val x = freshVar
              defineOr(x, -sw :: lit :: Nil)()
              x
            }
          }

          /** rewrite
            * {{{l1 + ... + ln >= k}}}
            * to
            * {{{f1 + ... + fn + ~sw + ... (k times) ... + ~sw >= k}}}
            * with forall i
            * {{{ fi = sw and li }}}
            */
          case Exactly => {
            var newLits = c.lits map {
              lit => {
                val x = freshVar
                defineAnd(x, sw :: lit :: Nil)()
                x
              }
            }
            val i = c.k
            while (i > 0) {
              newLits ::= -sw
            }
            newLits
          }
        }

        val newC = Card(newLits, c.relOp, c.k)
        (newC, Some(sw))
      }

      case None => (c, None)
    }
  }

  /** Adds a cardinality constraint with specified relop and constant to the solver.
    *
    * @param switch     optional switch literal
    * @param switchable if true and if switch is None, creates a fresh switch literal that allows to switch the constraint on and off
    * @return if switch is defined or if switchable is true, Some(fresh) where fresh is a fresh literal that allows to switch the clause on or off, None otherwise.
    */
  final def addCard(c: Card)(switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    val swOpt = mkSwitchLit(switch, switchable)
    c.toAtLeast.foreach {
      card => {
        val (newCard, _) = this.toSwitchableCard(card)(switch = swOpt)
        this._constraints :+= Tuple2(newCard, swOpt)
        this.addCardImplem(newCard)
      }
    }
    swOpt
  }

  final def addPbCst(c: ClassicPBCst)(switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    val swOpt = mkSwitchLit(switch, switchable)
    this._constraints :+= Tuple2(c, swOpt)
    swOpt
  }

  /** Adds a cardinality constraint {{{ sum(lits) <= k }}}.
    *
    * @param switchable optionally add a masking literal that allows to switch the clause on and off
    * @return if switchable is true, Some(fresh) where fresh is a fresh literal that allows to switch the clause on or off, None otherwise.
    **/
  final def addAtMostK(lits: List[Int], k: Int)(switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    this.addCard(Card(lits, AtMost, k))(switch, switchable)
  }

  /** Adds a cardinality constraint {{{ sum(lits) >= k }}}.
    *
    * @param switchable optionally add a masking literal that allows to switch the clause on and off
    * @return if switchable is true, Some(fresh) where fresh is a fresh literal that allows to switch the clause on or off, None otherwise.
    **/
  final def addAtLeastK(lits: List[Int], k: Int)(implicit switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    this.addCard(Card(lits, AtLeast, k))(switch, switchable)
  }

  /** Adds a cardinality constraint {{{ sum(lits) <= k }}}.
    *
    * @param switchable optionally add a masking literal that allows to switch the clause on and off
    * @return if switchable is true, Some(fresh) where fresh is a fresh literal that allows to switch the clause on or off, None otherwise.
    **/
  final def addExactlyK(lits: List[Int], k: Int)(implicit switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    this.addCard(Card(lits, Exactly, k))(switch, switchable)
  }

  /**
    *
    * @param lits
    * @param k
    * @param switch
    * @param switchable
    * @return
    */
  final def addAtMostKGate(lits: List[Int], k: Int)(implicit switch: Option[Int] = None, switchable: Boolean = false): Option[Int] = {
    val x = freshVar
    this.addPbCst(ClassicPBCst(lits :+ -x, Exactly, k, Map(x -> k).withDefaultValue(1)))(switch, switchable)
    this.addPbCst(ClassicPBCst(lits.map(-_) :+ x, Exactly, lits.size - k + 1, Map(x -> {
      lits.size - k + 1
    }).withDefaultValue(1)))(switch, switchable)
    Some(x)
  }

  /* ============================== IMPLEMENTATION METHODS ======================= */

  /** Method to use to effectively add a cardinality constraint to the underlying Solver implementation. */
  def addCardImplem(c: Card): Unit = {
    // TODO find proper interaction between implem and our representation of the formula (what if literals are not returned in the same order)
    /* Intentionally empty */
  }


}
