/** @author rdelmas */
package util.sat

import scala.util.Try

/** A builder for dimacs models, returning a result of type R. */
trait SolverLike {

  /** @return {{{Success(model)}}} if any model is currently available (must follow a successfull [[ProblemLike.isSatisfiable]], {{{Failure(exception)}}} otherwise . */
  def getModel: Try[List[Int]]

  /** Discards the current model (meant to be used when iterating on models of a formula). */
  def discardModel(): Unit

  /** Lifts the negation of the current model to a clause (to be used as a blocking clause when enumerating models), optionally projecting the model over the given set of literals.
    *
    * @param projectOnLits A list of literals on which to restrict the clause
    * @return {{{Success(clause)}}} if successfull, {{{Failure(exception)}}} otherwise.
    **/
  def liftCurrentModelToClause(projectOnLits: Option[Set[Int]] = None): Try[List[Int]]

  /** Lifts the current model to a cube, optionally projecting the model over the given set of literals.
    *
    * @param projectOnLits A list of literals on which to restrict the cube
    * @return {{{Success(cube)}}} if successfull, {{{Failure(exception)}}} otherwise.
    **/
  def liftCurrentModelToCube(projectOnLits: Option[Set[Int]] = None): Try[List[Int]]

  /** Sets a timeout expressed in seconds.
    *
    * @param duration timeout duration in seconds.
    **/
  def setTimeout(duration: Int): Unit

  /** Sets a timeout expressed as a number of conflicts.
    *
    * @param nofConflicts timeout duration in number of conflicts reached.
    */
  def setTimeoutOnConflicts(nofConflicts: Int): Unit

  /** Sets a timeout expressed in milliseconds.
    *
    * @param duration timeout duration in milliseconds.
    **/
  def setTimeoutMs(duration: Long): Unit

  /** @return the currently set timeout value in seconds. */
  def getTimeout: Int

  /** @return the currently set timeout value in milliseconds. */
  def getTimeoutMs: Long

  /** Force timeout to expire. */
  def expireTimeout(): Unit

  /** Resets the solver to a clean state. */
  def reset(): Unit

  /** Enable/disable clause DB simplification. */
  def setDBSimplificationAllowed(value: Boolean): Unit

  /** True iff DB simplification is allowed. */
  def isDBSimplficationAllowed: Boolean

  /** Removes all learnt clauses fromt the solver. */
  def clearLearntClauses(): Unit

  /** Adds a blocking clause to the solver. */
  def addBlockingClause(c: List[Int])(implicit switchable: Boolean = false): Unit

  /** Returns a clause explaning the unsat state of the constraints (must follow a call to [[ProblemLike.isSatisfiable]] that returned false.
    *
    * @return {{{Success(clause)}}} on success, {{{Failure( exception )}}} otherwise.
    **/
  def unsatExplanation: Try[List[Int]]

  /** Enable/disable the "keep solver hot" feature. */
  def setKeepSolverHot(value: Boolean): Unit

  /** True iff the solver is kept hot between check-sats. */
  def isSolverKeptHot: Boolean

}


