package util.sat

/**
  * Created by rdelmas on 26/06/16.
  *
  * Defines ternary semantics (true, false, X) for propositional operators and links with standard boolean semantics.
  */
object TernaryLogic {

  import UnitProp._

  /** Companion object allowing to create ternary values from dual rail encoded literals or from standard propositions. */
  object TernaryVal {

    /** @param posLitVal The value of l_x
      * @param negLitVal The value of l_not_x
      * @return The ternary value encoded by the (l_x, l_not_x) value pair.
      **/
    def fromDualRail(posLitVal: Option[Boolean], negLitVal: Option[Boolean]): TernaryVal = {
      (posLitVal, negLitVal) match {
        case (Some(true), Some(true)) => TernaryConflict
        case (Some(true), Some(false)) => TernaryTrue
        case (Some(true), None) => TernaryTrue
        case (Some(false), Some(true)) => TernaryFalse
        case (None, Some(true)) => TernaryFalse
        case (Some(false), None) => TernaryUndef
        case (Some(false), Some(false)) => TernaryUndef
        case (None, Some(false)) => TernaryUndef
        case (None, None) => TernaryUndef
      }
    }

    /** @param posLitVal The value of l_x
      * @param negLitVal The value of l_not_x
      * @return The ternary value encoded by the (l_x, l_not_x) value pair.
      * */
    def fromDualRail(posLitVal: Boolean, negLitVal: Boolean): TernaryVal = {
      (posLitVal, negLitVal) match {
        case (true, false) => TernaryTrue
        case (false, true) => TernaryFalse
        case (false, false) => TernaryUndef
        case (true, true) => TernaryConflict
      }
    }

    /** @param litVal The value of the literal.
      * @return The corresponding ternary value.
      **/
    def fromNormal(litVal: Boolean): TernaryVal = {
      if (litVal) TernaryTrue else TernaryFalse

    }
  }

  /** Base type for all ternary values. */
  sealed abstract trait TernaryVal {
    def toFormula(posLit: Lit, negLit: Lit): Formula
  }

  /** The {{{True}}} value. */
  case object TernaryTrue extends TernaryVal {
    override def toString = "1"

    def toFormula(posLit: Lit, negLit: Lit): Formula = List(List(posLit), List(-negLit))
  }

  /** The {{{False}}} value. */
  case object TernaryFalse extends TernaryVal {
    override def toString = "0"

    def toFormula(posLit: Lit, negLit: Lit): Formula = List(List(-posLit), List(negLit))
  }

  /** The {{{Undef}}} value. */
  case object TernaryUndef extends TernaryVal {
    override def toString = "?"

    def toFormula(posLit: Lit, negLit: Lit): Formula = List(List(-posLit), List(-negLit))
  }

  /** The {{{Conflict}}} value. */
  case object TernaryConflict extends TernaryVal {
    override def toString = "C"

    override def toFormula(posLit: Lit, negLit: Lit): Formula = List(List(posLit), List(negLit))
  }

  /** Performs ternary evaluation under a given expression given a context map.
    *
    * @param e   Expression to evaluate
    * @param ctx The evaluation context, assigns ternary values to Atoms by their integer identifier.
    * @return The evaluation of e in ctx.
    * */
  def eval(e: Core.Expr)(implicit ctx: Map[Int, TernaryVal]): TernaryVal = {
    import Core._
    e match {
      case And(kids) => evalAnd(kids.map(eval))
      case Or(kids) => evalOr(kids.map(eval))
      case Ite(k1, k2, k3) => evalIte(List(k1, k2, k3).map(eval))
      case Not(kid) => evalNot(List(eval(kid)))
      case Atom(uid) => evalAtom(uid, ctx)
    }
  }

  /** Performs ternary evaluation under a given expression given a context map.
    *
    * @param e   Expression to evaluate
    * @param ctx The evaluation context, assigns ternary values to Atoms by their integer identifier.
    * @return The evaluation of e in ctx.
    * */
  def eval(e: NNF.Expr)(implicit ctx: Map[Int, TernaryVal]): TernaryVal = {
    import NNF._
    e match {
      case And(kids) => evalAnd(kids.map(eval))
      case Or(kids) => evalOr(kids.map(eval))
      case Ite(k1, k2, k3) => evalIte(List(k1, k2, k3).map(eval))
      case Not(kid) => evalNot(List(eval(kid)))
      case Atom(id) => evalAtom(id, ctx)
    }
  }

  /** Evaluates an atom based on its unique identifier by a lookup into the given ctx map. */
  def evalAtom(id: Int, ctx: Map[Int, TernaryVal]): TernaryVal = {
    ctx(id)
  }

  /** Evaluates a negation based on its kids values. */
  def evalNot(kid: List[TernaryVal]): TernaryVal = kid match {
    case kid :: Nil => {
      kid match {
        case TernaryUndef => TernaryUndef
        case TernaryTrue => TernaryFalse
        case TernaryFalse => TernaryTrue
        case TernaryConflict => TernaryConflict
      }
    }
    case _ => throw new Exception("wrong number of kids for not")
  }

  /** Evaluates a conjunction based on its kids values. */
  def evalAnd(kids: List[TernaryVal]): TernaryVal = kids.foldLeft(TernaryTrue: TernaryVal) {
    (soFar, kid) =>
      soFar match {
        case TernaryUndef => kid match {
          case TernaryUndef | TernaryTrue => TernaryUndef
          case TernaryFalse => TernaryFalse
          case TernaryConflict => TernaryConflict
        }
        case TernaryTrue => kid match {
          case TernaryUndef => TernaryUndef
          case TernaryTrue => TernaryTrue
          case TernaryFalse => TernaryFalse
          case TernaryConflict => TernaryConflict
        }
        case TernaryFalse => kid match {
          case TernaryUndef | TernaryTrue | TernaryFalse => TernaryFalse
          case TernaryConflict => TernaryConflict
        }
        case TernaryConflict => TernaryConflict
      }
  }

  /** Evaluates a disjunction based on its kids values. */
  def evalOr(kids: List[TernaryVal]): TernaryVal = kids.foldLeft(TernaryFalse: TernaryVal) {
    (soFar, kid) =>
      soFar match {
        case TernaryUndef => kid match {
          case TernaryUndef | TernaryFalse => TernaryUndef
          case TernaryTrue => TernaryTrue
          case TernaryConflict => TernaryConflict
        }
        case TernaryTrue => kid match {
          case TernaryUndef | TernaryTrue | TernaryFalse => TernaryTrue
          case TernaryConflict => TernaryConflict
        }
        case TernaryFalse => kid match {
          case TernaryUndef => TernaryUndef
          case TernaryTrue => TernaryTrue
          case TernaryFalse => TernaryFalse
          case TernaryConflict => TernaryConflict
        }
        case TernaryConflict => TernaryConflict
      }
  }


  /** Evaluates an if-then-else based on its kids values. */
  def evalIte(kids: List[TernaryVal]): TernaryVal = {
    kids match {
      case selector :: ifTrue :: ifFalse :: Nil => {
        if (selector == TernaryConflict || ifTrue == TernaryConflict || ifFalse == TernaryConflict) {
          TernaryConflict
        } else {
          selector match {
            case TernaryTrue => ifTrue
            case TernaryFalse => ifFalse
            case TernaryUndef => if (ifTrue == ifFalse) ifTrue else TernaryUndef
            case _ => throw new Exception("impossible")
          }
        }
      }
      case _ => throw new Exception("wrong number of kids for ite")
    }
  }

}
