package util.sat

/**
  * Created by rdelmas on 25/06/16.
  */

import scala.collection.immutable.Map

/** Very naive implementation of a unit propagation routine
  * only meant to test propagation properties of
  * clausal encodings for small problems/operators.
  * Performance is not an issue, correctness is however required.
  *
  * */
object UnitProp {

  type Lit = Int
  type Clause = List[Lit]
  type Formula = List[Clause]
  type Model = Map[Lit, Boolean]

  def convert(f: FormulaLike): Formula = f.clauses.map(x => x._1.lits.map(x => x).toList).toList

  /** Clause satisfiability status. */
  sealed trait CStatus

  case object CConflicting extends CStatus {
    override def toString = "CONFLICT"
  }

  case object CSat extends CStatus {
    override def toString = "SAT"
  }

  case object CUnknown extends CStatus {
    override def toString = "UNDEF"
  }

  case class CUnit(lit: Lit) extends CStatus {
    override def toString = s"UNIT(${lit})"
  }

  /** Formula satisfiability status. */
  sealed trait FStatus

  case object FUnsat extends FStatus {
    override def toString = "UNSAT"
  }

  case object FSat extends FStatus {
    override def toString = "SAT"
  }

  case object FUnknown extends FStatus {
    override def toString = "UNDEF"
  }

  /** Applies the negation to value depending on the sign of lit. */
  def applySign(lit: Lit, value: Boolean) = if (lit > 0) value else !value

  /** Returns the value of the literal as found in model. */
  def litValue(lit: Lit, model: Model): Option[Boolean] = {
    for (value <- model.get(math.abs(lit)))
      yield applySign(lit, value)
  }

  /** Clause level propagation. */
  def propagate(clause: Clause, model: Model): CStatus = {
    val zero: (CStatus, List[Lit]) = (CUnknown, Nil)
    val (status, unassigned) = clause.foldLeft(zero) {
      (acc, lit) => {
        val (status, unassigned) = acc
        status match {
          case CSat => (CSat, Nil)
          case CUnknown => litValue(lit, model) match {
            case Some(value) if value => (CSat, Nil)
            case Some(value) if !value => (CUnknown, unassigned)
            case None => (CUnknown, lit :: unassigned)
          }
          case _ => throw new Exception("impossible")
        }
      }
    }
    status match {
      case CSat => CSat
      case CUnknown => unassigned match {
        // still unknown and no unassigned lit, conflict
        case Nil => CConflicting
        // sill unknown and single unassigned lit, propagate unit
        case lit :: Nil => CUnit(lit)
        // more than one litteral unassigned, nothing can be deduced
        case _ => CUnknown
      }
      case _ => throw new Exception("impossible")
    }
  }

  /** Formula level single pass of propagation to detect status and units. */
  def propagateSP(formula: Formula, model: Model, fstatus: FStatus = FSat, units: List[CUnit] = Nil): (FStatus, List[CUnit]) = {
    formula match {
      case Nil => (fstatus, units)
      case clause :: tail => {
        propagate(clause, model) match {
          case CSat => propagateSP(tail, model, fstatus, units)
          case CUnknown => propagateSP(tail, model, FUnknown, units)
          case CConflicting => propagateSP(tail, model, FUnsat, units)
          case u: CUnit => propagateSP(tail, model, FUnknown, u :: units)
        }
      }
    }
  }

  /** Formula level propagation until fixed point.
    *
    * @param formula the formula to run propagation on
    * @return A pair (result, model), the model can be inconsistent with formula if status is FUnsat
    **/
  def propagate(formula: Formula): (FStatus, Model) = {

    /** Extends a model with from a list of unit clauses. */
    def unitsToModel(units: List[CUnit], model: Model): Model = {
      model ++ units.map(u => if (u.lit > 0) (u.lit -> true) else (-u.lit -> false))
    }

    // initial model is empty and status unknown
    var model: Model = Map()
    var status: FStatus = FUnknown

    // initialis propagation loop by collecting any single literal clause and lifting into CUnit:s
    var units = formula.collect { case lit :: Nil => CUnit(lit) }

    // propagate while new CUnits are found
    while (!units.isEmpty) {
      model = unitsToModel(units, model)
      units = Nil
      propagateSP(formula, model) match {
        case (FSat, _) => status = FSat
        case (FUnsat, _) => status = FUnsat
        case (FUnknown, newUnits) => {
          status = FUnknown
          units = newUnits
        }
      }
    }
    (status, model)
  }
}
