/** $Id$
  *
  * @author $Author$
  * @date $Date$
  **/
package util.stack.immutable

import util.stack._

import scala.collection.GenTraversableOnce
import scala.collection.immutable.Set


/** A simple mock proxy that implements the assertion stack using a singly linked list. */
class StackSolverProxyMock[BoolExpr](val name: String = "no_name_given") extends StackSolverProxy[BoolExpr] {
  /** The stack of assertions. */
  private var _stack: List[List[BoolExpr]] = Nil :: Nil

  def add(a: BoolExpr): Unit = {
    _stack = (a :: _stack.head) :: _stack.tail
  }

  def add(as: List[BoolExpr]): Unit = {
    _stack = (as ::: _stack.head) :: _stack.tail
  }

  def push(): Unit = {
    _stack = Nil :: _stack
  }

  def pop(n: Int = 1): Unit = {
    _stack = _stack.drop(n)
  }

  def reset(): Unit = {
    _stack = Nil :: Nil
  }

  override def toString: String =
    s"""StackSolverProxyMock(${name}\n${_stack.zipWithIndex.map(x => s"Level ${x._2}:\n" + x._1.map(_.toString).mkString(", ")).mkString("\n")}\n)"""

  /** Returns true iff the given keep stacks contains the same elements than. */
  def sameContents(that: immutable.AssertionStack[BoolExpr]): Boolean = {
    val zero: Set[BoolExpr] = Set.empty
    val keepStackContents: Set[BoolExpr] = that.biFoldLeft(zero, zero)({ (acc, e) => acc + e._1 }, { (acc, e) => acc union e })
    val myContents: Set[BoolExpr] = this._stack.foldLeft(zero)(
      (acc, e) => {
        val res = e.foldLeft(zero)({ (acc, e) => acc + e })
        acc union res
      }
    )
    keepStackContents == myContents
  }
}

/** A proxy that prints the dialog with its KeepStack on stdout. */
class PrintingSolverProxyMock[BoolExpr](val name: String = "no_name_given") extends StackSolverProxy[BoolExpr] {
  def add(a: BoolExpr): Unit = println(s"${name}.add( ${a.toString} )")

  def add(as: List[BoolExpr]): Unit = println(s"${name}.add( ${as.map(_.toString).mkString(", ")} )")

  def push(): Unit = println(s"${name}.push()")

  def pop(n: Int = 1): Unit = println(s"${name}.pop(${n})")

  def reset(): Unit = println(s"${name}.reset(n)")

  override def toString: String = s"PrintingSolverProxy(${name})"
}

/** A proxy that does nothing. */
class UnitSolverProxyMock[BoolExpr](val name: String = "no_name_given") extends StackSolverProxy[BoolExpr] {
  def add(a: BoolExpr): Unit = ()

  def add(as: List[BoolExpr]): Unit = ()

  def push(): Unit = ()

  def pop(n: Int = 1): Unit = ()

  def reset(): Unit = ()

  override def toString: String = s"UnitSolverProxyMock(${name})"
}

/** An immutable stack of lists of items of type BoolExpr, each item paired with a boolean value indicating if we want the item to stay in the stack when popping the stack.
  * Embeds a proxy to a statefull SMT solver, and enforces the item keeping mechanism on the proxy.
  */
sealed trait AssertionStack[BoolExpr] {
  outer =>

  /** The statefull solver proxy on which this stack operates. */
  val solverProxy: StackSolverProxy[BoolExpr]

  /** Number of levels of the stack. */
  val size: Int

  /** The top level of the stack. */
  val top: List[(BoolExpr, Boolean)]

  /** The bottom of the stack. */
  val bottom: List[List[(BoolExpr, Boolean)]]

  /** Adds a new empty level on the stack. */
  def push: AssertionStack[BoolExpr] = {

    // perform side effects on the proxy
    solverProxy.push()

    new AssertionStack[BoolExpr] {
      val solverProxy: StackSolverProxy[BoolExpr] = outer.solverProxy
      val size: Int = outer.size + 1
      val top: List[(BoolExpr, Boolean)] = Nil
      val bottom: List[List[(BoolExpr, Boolean)]] = outer.top :: outer.bottom
    }
  }

  /** Pops n levels of the stack, throw exception if n larger than size-1. */
  def pop(n: Int): AssertionStack[BoolExpr] = {
    assert(n < outer.size)
    var i = n
    var toKeep: List[(BoolExpr, Boolean)] = Nil
    var stack = outer.top :: outer.bottom
    while (i > 0) {
      toKeep :::= stack.head.collect { case x if x._2 => x }
      stack = stack.tail
      i -= 1
    }
    val newTop = stack.head reverse_::: toKeep

    // perform side effects on the proxy
    solverProxy.pop(n)
    solverProxy.add(toKeep.map(x => x._1))

    new AssertionStack[BoolExpr] {
      val solverProxy: StackSolverProxy[BoolExpr] = outer.solverProxy
      val size: Int = outer.size - n
      val top: List[(BoolExpr, Boolean)] = newTop
      val bottom: List[List[(BoolExpr, Boolean)]] = stack.tail
    }
  }

  /** Pops the whole stack down to size 1, all things to keep will end up in the first level. */
  def reset: AssertionStack[BoolExpr] = this.pop(this.size - 1)

  /** Prepends an item to the current top level. */
  def ::(a: BoolExpr, keep: Boolean = false): AssertionStack[BoolExpr] = {
    // perform side effects on the proxy
    solverProxy.add(a)
    new AssertionStack[BoolExpr] {
      val solverProxy: StackSolverProxy[BoolExpr] = outer.solverProxy
      val size: Int = outer.size
      val top: List[(BoolExpr, Boolean)] = (a, keep) :: outer.top
      val bottom: List[List[(BoolExpr, Boolean)]] = outer.bottom
    }
  }

  /** Prepends a list of items to the current top level. */
  def :::(as: List[(BoolExpr, Boolean)]): AssertionStack[BoolExpr] = {
    // perform side effects on the proxy
    solverProxy.add(as.map(_._1))
    new AssertionStack[BoolExpr] {
      val solverProxy: StackSolverProxy[BoolExpr] = outer.solverProxy
      val size: Int = outer.size
      val top: List[(BoolExpr, Boolean)] = as ::: outer.top
      val bottom: List[List[(BoolExpr, Boolean)]] = outer.bottom
    }
  }

  /** Horizontally map:s a function, level per level, returning list of results (one result per level). */
  def mapH[B](f: ((BoolExpr, Boolean)) => B): List[List[B]] =
    outer.top.map(f) :: outer.bottom.map(l => l.map(f))

  /** Horizontally flatMap:s a function, level per level, returning list of results (one result per level). */
  def flatMapH[B](f: ((BoolExpr, Boolean)) => GenTraversableOnce[B]): List[List[B]] =
    outer.top.flatMap(f).toList :: outer.bottom.map(l => l.flatMap(f).toList)

  /** Horizontally foldLeft:s each level of the stack with a given function, returns a list of result (one result per level). */
  def foldLeftH[B](zero: B)(f: (B, (BoolExpr, Boolean)) => B): List[B] =
    outer.top.foldLeft(zero)(f) :: outer.bottom.map(l => l.foldLeft(zero)(f))

  /** Folds each level of the stack with the given zero1 and f1 yielding a list containing one result per level, which is then folded using zero2 and f2. */
  def biFoldLeft[B1, B2](zero1: B1, zero2: B2)(f1: (B1, (BoolExpr, Boolean)) => B1, f2: (B2, B1) => B2): B2 =
    this.foldLeftH(zero1)(f1).foldLeft(zero2)(f2)

  /** True iff all elements of all levels of the stack satisfy predicate p. */
  def forall(p: (BoolExpr, Boolean) => Boolean): Boolean =
    biFoldLeft(true, true)({ (acc, x) => acc && p(x._1, x._2) }, { (acc, x) => acc && x })

  /** True iff some element of some level of the stack satisfies predicate p. */
  def exists(p: (BoolExpr, Boolean) => Boolean): Boolean =
    biFoldLeft(false, false)({ (acc, x) => acc || p(x._1, x._2) }, { (acc, x) => acc || x })

  /** Returns true iff the stack contains some element for which the keep flag is true. */
  def containsToKeep: Boolean = this.exists((e, keep) => keep)

  /** Returns true iff the stack contains some element for which the keep flag is false. */
  def containsNotKeep: Boolean = this.exists((e, keep) => !keep)

  /** Returns true the index of the first stack level from the top where the predicate is true. */
  def levelWhere(p: List[(BoolExpr, Boolean)] => Boolean): Int = {
    if (p(this.top))
      0
    else
      this.bottom.zipWithIndex.find(x => p(x._1)) match {
        case Some((_, index: Int)) => index + 1
        case None => -1
      }
  }

  private def findFirst(ls: List[List[(BoolExpr, Boolean)]], p: List[(BoolExpr, Boolean)] => Boolean): Option[(List[(BoolExpr, Boolean)], Int)] =
    ls.zipWithIndex.foldLeft(None: Option[(List[(BoolExpr, Boolean)], Int)])(
      (acc, x) => {
        val (l, i) = x
        acc match {
          case res@Some(_) => res
          case None => if (p(l)) Some((l, i)) else None
        }
      }
    )

  /** Returns the first level from the top which satisfies the given predicate together with the integer representing its depth from the top of the stack, None if not found. */
  def findLevelFromTop(p: List[(BoolExpr, Boolean)] => Boolean): Option[(List[(BoolExpr, Boolean)], Int)] =
    this.findFirst((this.top :: this.bottom), p)

  /** Returns the first level from the bottom of the stack which satisfies the given predicate together with the integer representing its depth from the bottom of the stack, None if not found. */
  def findLevelFromBottom(p: List[(BoolExpr, Boolean)] => Boolean): Option[(List[(BoolExpr, Boolean)], Int)] =
    this.findFirst((this.top :: this.bottom).reverse, p)

  override def toString = {
    s"AssertionStack(\n${(top :: bottom).map(_.map(_.toString).mkString(", ")).mkString("\n-------------\n")}\n)"
  }

}


/** Companion object (allows to create instances of the trait given a particular proxy object, defaults to the UnitSolverProxy[BoolExpr]. */
object AssertionStack {
  /** Returns a new AssertionStack[BoolExpr] with an empty level at the bottom. */
  def apply[BoolExpr](sp: StackSolverProxy[BoolExpr] = new UnitSolverProxyMock[BoolExpr]()) = new AssertionStack[BoolExpr] {
    val solverProxy: StackSolverProxy[BoolExpr] = sp
    val size: Int = 1
    val top: List[(BoolExpr, Boolean)] = Nil
    val bottom: List[List[(BoolExpr, Boolean)]] = Nil
  }
}


