package util.translators

import com.microsoft.z3.{BitVecExpr, AST => Z3AST, BoolExpr => Z3BoolExpr, Context => Z3Context, Expr => Z3Expr, IntExpr => Z3IntExpr, Symbol => Z3Symbol, _}
import translators.translatorsTypes.PathIdentType

import scala.collection.mutable

/**
  * Created by kdelmas on 04/11/16.
  */
sealed trait Key

case class FlowKey(path: PathIdentType) extends Key

case class EventKey(path: PathIdentType) extends Key

case class TypeKey(path: PathIdentType) extends Key

case class CompKey(path: PathIdentType) extends Key

case class Tuple(name: String, sorts: List[BitVecSort])

object CacheImplicits {

  implicit object isEventCache extends EventKey(Nil)

  implicit object isFlowCache extends FlowKey(Nil)

  implicit object isTypeCache extends TypeKey(Nil)

  implicit object isCompCache extends CompKey(Nil)

}

/**
  * Immutable cache containing translation of DSL nodes to Z3 SMT expressions
  * Allow to store all kinds of translation and preserve type information
  */
trait ImmutableCache {

  protected val _cacheEvt: mutable.OpenHashMap[EventKey, Z3BoolExpr]

  protected val _cacheFlow: mutable.OpenHashMap[FlowKey, BitVecExpr]

  protected val _cacheTuple: mutable.OpenHashMap[TypeKey, Tuple]

  protected val _cacheComp: mutable.OpenHashMap[CompKey, DatatypeExpr]

  def collect[C](f: PartialFunction[(EventKey, Z3BoolExpr), C])(implicit tag: EventKey): mutable.Iterable[C] = _cacheEvt.collect(f)

  def collect[C](f: PartialFunction[(FlowKey, BitVecExpr), C])(implicit tag: FlowKey): mutable.Iterable[C] = _cacheFlow.collect(f)

  def collect[C](f: PartialFunction[(TypeKey, Tuple), C])(implicit tag: TypeKey): mutable.Iterable[C] = _cacheTuple.collect(f)

  def collect[C](f: PartialFunction[(CompKey, DatatypeExpr), C])(implicit tag: CompKey): mutable.Iterable[C] = _cacheComp.collect(f)

  def filter(p: ((EventKey, Z3BoolExpr)) => Boolean)(implicit tag: EventKey): mutable.OpenHashMap[EventKey, Z3BoolExpr] = _cacheEvt.filter(p)

  def filter(p: ((FlowKey, BitVecExpr)) => Boolean)(implicit tag: FlowKey): mutable.OpenHashMap[FlowKey, BitVecExpr] = _cacheFlow.filter(p)

  def filter(p: ((TypeKey, Tuple)) => Boolean)(implicit tag: TypeKey): mutable.OpenHashMap[TypeKey, Tuple] = _cacheTuple.filter(p)

  def filter(p: ((CompKey, DatatypeExpr)) => Boolean)(implicit tag: CompKey): mutable.OpenHashMap[CompKey, DatatypeExpr] = _cacheComp.filter(p)

  def apply(key: EventKey) = _cacheEvt(key)

  def apply(key: FlowKey) = _cacheFlow(key)

  def apply(key: TypeKey) = _cacheTuple(key)

  def apply(key: CompKey) = _cacheComp(key)

  def get(key: EventKey): Option[Z3BoolExpr] = _cacheEvt.get(key)

  def get(key: FlowKey): Option[BitVecExpr] = _cacheFlow.get(key)

  def get(key: TypeKey): Option[Tuple] = _cacheTuple.get(key)

  def get(key: CompKey): Option[DatatypeExpr] = _cacheComp.get(key)
}

/**
  * Extension of immutable cache with mutable methods
  */
class MutableCache extends ImmutableCache {

  protected val _cacheEvt: mutable.OpenHashMap[EventKey, Z3BoolExpr] = mutable.OpenHashMap.empty

  protected val _cacheFlow: mutable.OpenHashMap[FlowKey, BitVecExpr] = mutable.OpenHashMap.empty

  protected val _cacheTuple: mutable.OpenHashMap[TypeKey, Tuple] = mutable.OpenHashMap.empty

  protected val _cacheComp: mutable.OpenHashMap[CompKey, DatatypeExpr] = mutable.OpenHashMap.empty

  def update(pair: (EventKey, Z3BoolExpr))(implicit tag: EventKey): Unit = _cacheEvt.update(pair._1, pair._2)

  def update(pair: (FlowKey, BitVecExpr))(implicit tag: FlowKey): Unit = _cacheFlow.update(pair._1, pair._2)

  def update(pair: (TypeKey, Tuple))(implicit tag: TypeKey): Unit = _cacheTuple.update(pair._1, pair._2)

  def update(pair: (CompKey, DatatypeExpr))(implicit tag: CompKey): Unit = _cacheComp.update(pair._1, pair._2)

  def getOrElseUpdate(key: EventKey, default: Z3BoolExpr): Z3BoolExpr = _cacheEvt.getOrElseUpdate(key, default)

  def getOrElseUpdate(key: FlowKey, default: BitVecExpr): BitVecExpr = _cacheFlow.getOrElseUpdate(key, default)

  def getOrElseUpdate(key: TypeKey, default: Tuple): Tuple = _cacheTuple.getOrElseUpdate(key, default)

  def getOrElseUpdate(key: CompKey, default: DatatypeExpr): DatatypeExpr = _cacheComp.getOrElseUpdate(key, default)

  def toImmutable: ImmutableCache = {
    val myCacheEvt = _cacheEvt
    val myCacheFlow = _cacheFlow
    val myCacheComp = _cacheComp
    val myCacheTuple = _cacheTuple
    new ImmutableCache {
      protected val _cacheFlow: mutable.OpenHashMap[FlowKey, BitVecExpr] = myCacheFlow
      protected val _cacheEvt: mutable.OpenHashMap[EventKey, Z3BoolExpr] = myCacheEvt
      protected val _cacheTuple: mutable.OpenHashMap[TypeKey, Tuple] = myCacheTuple
      protected val _cacheComp: mutable.OpenHashMap[CompKey, DatatypeExpr] = myCacheComp
    }
  }
}
