package util.translators

import util.analyzers.SubstitutionVar
import com.microsoft.z3.{BitVecExpr, AST => Z3AST, BoolExpr => Z3BoolExpr, Context => Z3Context, Expr => Z3Expr, IntExpr => Z3IntExpr, Status => Z3Status, Symbol => Z3Symbol, _}
import translators.translatorsTypes._
import util.cardinality.{AtLeast, AtMost, CardinalityEncoder, Sequential}

import scala.collection.mutable

/**
  * Created by kdelmas on 04/11/16.
  */

/**
  * Trait for utility operations on Z3 SMT terms
  */
trait SMTContext extends Z3Context {


  /**
    * Create an uninterpreted Boolean constant
    * @param pathIdent the identifier of the constant
    * @return Z3 Boolean constant
    */
  def mkBoolConst(pathIdent: PathIdentType): Z3BoolExpr = {
    mkConst(pathIdent.mkString("."), mkBoolSort()).asInstanceOf[Z3BoolExpr]
  }

  /**
    * Create Z3 bit vector constant
    * @param pathIdent the name of the constant
    * @param sort the bit vector sort of the constant
    * @return the bitvector constant
    */
  def mkBitVecConst(pathIdent: PathIdentType, sort: BitVecSort): BitVecExpr = {
    mkConst(pathIdent.mkString("."), sort).asInstanceOf[BitVecExpr]
  }

  /**
    * Create a Boolean bounded variable
    * @param index the De Bruijn index
    * @return bounded variable
    */
  def mkBoolBound(index: Int): Z3BoolExpr = {
    mkBound(index, mkBoolSort()).asInstanceOf[Z3BoolExpr]
  }

  /**
    * Create a bit vector bounded variable
    * @param index the De Bruijn index
    * @param sort the bit vector sort
    * @return bounded variable
    */
  def mkBitVecBound(index: Int, sort: BitVecSort): BitVecExpr = {
    mkBound(index, sort).asInstanceOf[BitVecExpr]
  }

  /**
    * Create a predicate declaration
    * @param name the name of the predicate
    * @param domain its domain
    * @return the function declaration
    */
  def mkBoolFunDecl(name: String, domain: List[Sort]): FuncDecl = {
    mkFuncDecl(name, domain.toArray, mkBoolSort())
  }

  /**
    * Create an if then else structure
    * @param i the if expression
    * @param t the then expression
    * @param e the else expression
    * @tparam T the type of the else and then expression
    * @return the if-then-else term
    */
  def mkIte[T <: Z3Expr](i: Z3BoolExpr, t: T, e: T): T = mkITE(i, t, e).asInstanceOf[T]

  /**
    * Substitute the variable of Z3 SMT expression by some other variables
    * @param in the initial expression
    * @param by the map from initial to new variables
    * @tparam T the type of the expression
    * @tparam U the type of the variables
    * @return the expression where variables have been substituted
    */
  def substituteIn[T <: Z3Expr, U <: Z3Expr](in: T, by: Map[U, U]): T = {
    val vars = by.unzip
    in.substitute(vars._1.toArray[Z3Expr], vars._2.toArray[Z3Expr]).asInstanceOf[T]
  }

  /**
    * Return the index of a given element in the enumeration sort of Z3 contained in a substitution variable
    * @param const the substitution constant
    * @param value the value to find
    * @return the index of the value
    */
  def getValuePos(const: SubstitutionVar, value: Z3Expr): Option[Int] = {
    val index = const.sort.getConsts.indexOf(value)
    if (index != -1) Some(index) else None
  }

  /**
    * Find the ith element of the enumeration sort of Z3 contained in a substitution variable
    * @param const the substitution constant
    * @param index the value to find
    * @return the index of the value
    */
  def getValueAtPos(const: SubstitutionVar, index: Int): Option[Z3Expr] = {
    if (index >= 0 && index < const.sort.getConsts.length)
      Some(const.sort.getConsts.apply(index))
    else
      None
  }

  /**
    * Transform a set of
    * @param cube the set of substitution variable / value couples
    * @return the conjunction of the "sub = val" equalities
    */
  @deprecated("Should use the Choose predicate to force value of a substitution predicate","KCR analyzer 1.0")
  def toCube(cube: Traversable[(SubstitutionVar, Z3Expr)]): Z3BoolExpr = {
    cube.headOption match {
      case Some(_) => mkAnd(cube.map(p => mkEq(p._1.z3Var, p._2)).toSeq: _*)
      case None => mkTrue()
    }
  }

  /**
    * Express at most n with Sequential encoding
    * @param kids the Boolean expression addressed by the at most
    * @param n the bound
    * @return the equivalent Boolean expression
    */
  def atMostN(kids: List[Z3BoolExpr], n: Int): Z3BoolExpr = {
    if (n >= kids.size)
      mkBool(true)
    else if (n < 0)
      mkBool(false)
    else if (n == 0)
      mkAnd(kids.map(kid => mkNot(kid)): _*)
    else {
      val countEncoder = CardinalityEncoder(kids.size, AtMost, n, Sequential)
      val mkFresh = () => mkFreshConst("countVar", mkBoolSort()).asInstanceOf[Z3BoolExpr]
      val myMkAnd = (in: List[Z3BoolExpr]) => mkAnd(in: _*)
      val myMkOr = (in: List[Z3BoolExpr]) => mkOr(in: _*)
      val myMkNot = (in: Z3BoolExpr) => mkNot(in)
      val translated = countEncoder.translate(kids, mkFresh, myMkNot, myMkAnd, myMkOr)
      translated._2
    }
  }

  /**
    * BDD encoding of the at least n
    * @param kids the Boolean expression addressed by the at least
    * @param k the bound
    * @return the equivalent BDD expression expressed as if-then-else
    */
  def atLeastNBDD(kids: List[Z3BoolExpr], k: Int): Z3BoolExpr = {
    if (k > kids.size)
      mkBool(false)
    else if (k <= 0)
      mkBool(true)
    else if (k == 1)
      mkOr(kids: _*)
    else if (k == kids.size)
      mkAnd(kids: _*)
    else {
      val nodes = mutable.OpenHashMap.empty[(Int, Int), Z3BoolExpr]
      val n = kids.size
      //the at least BDD ia k x (n-k+1) diamond of the following form (-> means low son and | high son)
      //e_1 -> ... e_{n-k+1} -> 0
      // |            |
      //e_2 -> ... e_{n-k+2} -> 0
      // ...
      //e_k -> ... e_{n} ->0
      // |   ...    |
      // 1  ....    1
      for (j <- (0 to n - k).reverse) {
        for (i <- (0 until k).reverse) {
          val t = if (i != k - 1) nodes((i + 1, j)) else mkTrue()
          val e = if (j != n - k) nodes((i, j + 1)) else mkFalse()
          nodes((i, j)) = mkIte(kids(i + j), t, e)
        }
      }
      nodes((0, 0))
    }
  }
}


object SMTContext {
  def apply(): SMTContext = new SMTContext {}
}

